Package: scan
Type: Package
Title: Single-Case Data Analyses for Single and Multiple Baseline
        Designs
Version: 0.68.1
Date: 2026-04-27
Authors@R: c(
  person("Juergen", "Wilbert", email = "juergen.wilbert@uni-muenster.de", 
    role = c("cre", "aut"), comment = c(ORCID = "0000-0002-8392-2873")), 
  person("Timo", "Lueke", email = "timo.lueke@uni-graz.at", role = "aut", 
    comment = c(ORCID = "0000-0002-2603-7341")))
Depends: R (>= 4.1.0)
Imports: stats, nlme, MCMCglmm, utils, methods, graphics, car, gt,
        knitr, kableExtra, readxl, magrittr, miniUI, rstudioapi, rlang,
        cli
Suggests: testthat (>= 3.0.0), shiny, yaml, scplot
Description: A collection of procedures for analysing, visualising, 
  and managing single-case data. Multi-phase and multi-baseline designs are 
  supported. Analysing methods include regression models 
  (multilevel, multivariate, bayesian), between case standardised mean difference, 
  overlap indices ('PND', 'PEM', 'PAND', 'NAP', 'PET', 'tau-u', 'IRD', 'baseline corrected tau', 
  'CDC'), and randomization tests. Data preparation functions support outlier 
  detection, handling missing values, scaling, and custom transformations. 
  An export function helps to generate html, word, and latex tables in a 
  publication friendly style. A shiny app allows to use scan in a graphical 
  user interface.
  More details can be found in the online book 'Analyzing single-case data with 
  R and scan', Juergen Wilbert (2026)
  <https://jazznbass.github.io/scan-Book/>.
License: GPL (>= 3)
URL: https://github.com/jazznbass/scan/,
        https://jazznbass.github.io/scan-Book/,
        https://jazznbass.github.io/scan/
BugReports: https://github.com/jazznbass/scan/issues
LazyData: true
Encoding: UTF-8
RoxygenNote: 7.3.3
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-27 14:18:20 UTC; wilbert
Author: Juergen Wilbert [cre, aut] (ORCID:
    <https://orcid.org/0000-0002-8392-2873>),
  Timo Lueke [aut] (ORCID: <https://orcid.org/0000-0002-2603-7341>)
Maintainer: Juergen Wilbert <juergen.wilbert@uni-muenster.de>
Repository: CRAN
Date/Publication: 2026-04-28 05:10:02 UTC
Built: R 4.4.3; ; 2026-05-01 15:32:55 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("$",scdf)
S3method("[",scdf)
S3method(anova,sc_hplm)
S3method(anova,sc_mplm)
S3method(anova,sc_plm)
S3method(as.data.frame,scdf)
S3method(c,scdf)
S3method(coef,sc_hplm)
S3method(coef,sc_plm)
S3method(export,sc_ac)
S3method(export,sc_bcsmd)
S3method(export,sc_bctau)
S3method(export,sc_bplm)
S3method(export,sc_cdc)
S3method(export,sc_desc)
S3method(export,sc_hplm)
S3method(export,sc_ird)
S3method(export,sc_mplm)
S3method(export,sc_nap)
S3method(export,sc_outlier)
S3method(export,sc_overlap)
S3method(export,sc_pand)
S3method(export,sc_pem)
S3method(export,sc_pet)
S3method(export,sc_plm)
S3method(export,sc_pnd)
S3method(export,sc_power)
S3method(export,sc_rand)
S3method(export,sc_smd)
S3method(export,sc_tauu)
S3method(export,sc_trend)
S3method(export,scdf)
S3method(export,scdf_summary)
S3method(fetch,sc_bplm)
S3method(fetch,sc_hplm)
S3method(fetch,sc_mplm)
S3method(fetch,sc_plm)
S3method(na.omit,scdf)
S3method(plot,scdf)
S3method(print,sc_ac)
S3method(print,sc_bcsmd)
S3method(print,sc_bctau)
S3method(print,sc_bplm)
S3method(print,sc_cdc)
S3method(print,sc_desc)
S3method(print,sc_design)
S3method(print,sc_hplm)
S3method(print,sc_ird)
S3method(print,sc_mplm)
S3method(print,sc_nap)
S3method(print,sc_outlier)
S3method(print,sc_overlap)
S3method(print,sc_pand)
S3method(print,sc_pem)
S3method(print,sc_pet)
S3method(print,sc_plm)
S3method(print,sc_pnd)
S3method(print,sc_power)
S3method(print,sc_rand)
S3method(print,sc_rci)
S3method(print,sc_smd)
S3method(print,sc_tauu)
S3method(print,sc_trend)
S3method(print,scdf)
S3method(subset,scdf)
S3method(summary,scdf)
S3method(transform,scdf)
export("%>%")
export("scdf_attr<-")
export(across_cases)
export(add_dummy_variables)
export(add_l2)
export(all_cases)
export(as_scdf)
export(autocorr)
export(batch_apply)
export(between_smd)
export(bplm)
export(cdc)
export(center_at)
export(combine)
export(convert)
export(corrected_tau)
export(describe)
export(design)
export(estimate_design)
export(export)
export(fetch)
export(fill_missing)
export(fillmissingSC)
export(first_of)
export(hplm)
export(import_scdf)
export(ird)
export(is.scdf)
export(local_regression)
export(moving_mean)
export(moving_median)
export(mplm)
export(nap)
export(outlier)
export(overlap)
export(overlapSC)
export(pand)
export(pem)
export(pet)
export(plm)
export(plotSC)
export(plot_rand)
export(pnd)
export(power_test)
export(power_testSC)
export(rSC)
export(rand.test)
export(randSC)
export(rand_test)
export(random_scdf)
export(ranks)
export(rci)
export(rciSC)
export(readSC)
export(readSC.excel)
export(read_scdf)
export(rescale)
export(rowwise)
export(sample_names)
export(scdf)
export(scdf_attr)
export(select_cases)
export(select_phases)
export(set_dvar)
export(set_mvar)
export(set_na_at)
export(set_pvar)
export(set_vars)
export(shift)
export(shinyscan)
export(smd)
export(smooth_cases)
export(standardize)
export(style_plot)
export(tauUSC)
export(tau_u)
export(trend)
export(truncate_phase)
export(writeSC)
export(write_scdf)
import(MCMCglmm)
import(graphics)
import(gt)
import(kableExtra)
import(knitr)
import(methods)
import(nlme)
import(stats)
importFrom(car,Anova)
importFrom(cli,col_blue)
importFrom(magrittr,"%>%")
importFrom(utils,capture.output)
importFrom(utils,combn)
importFrom(utils,install.packages)
importFrom(utils,read.csv)

# scan - Single-Case Data Analyses for Single and Multiple Baseline Designs

<!-- badges: start -->
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/scan?color=blue)](https://CRAN.R-project.org/package=scan)
[![Downloads](https://cranlogs.r-pkg.org/badges/grand-total/scan?color=orange)](https://CRAN.R-project.org/package=scan)
<!-- badges: end -->

A collection of procedures for analysing, visualising, 
and managing single-case data. Multiphase and multibaseline designs are 
supported.

Analysing methods include 

- regression models (multilevel, multivariate, bayesian), 
- between-case standardized-mean difference, 
- overlap indices (including 'PND', 'PEM', 'PAND', 'NAP', 'PET', 'IRD', 'CDC'), 
- Tau-U methods (including meta analysis; baseline corrected tau), and 
- randomization tests. 

Data preparation functions support outlier detection, 
handling missing values, scaling, and custom transformations.   
An export function helps to generate html, word, and latex tables in a 
publication friendly style. A shiny app allows to use scan in a graphical 
userinterface.

More details can be found in the online book 'Analyzing single-case data with 
R and scan', Juergen Wilbert (2026)  
<https://jazznbass.github.io/scan-Book/>.

## Installation

You can install the released version of scan from CRAN with:

``` r
install.packages("scan")
```

Or the development version from GitHub with:

``` r
# install.packages("pak")
pak::pak("jazznbass/scan")
```

server <- function(input, output, session) {

  session$onFlushed(function() {
    if (!is.null(getOption("scan.shinyscan.initial"))) {
      updateNavbarPage(session, "scan", selected = "Stats")
    }
  }, once = TRUE)
  
  observeEvent(input$darkmode, {
    session$setCurrentTheme(
      if (isTRUE(input$darkmode)) res$theme_dark else res$theme_light
    )
  }, ignoreInit = TRUE)
  
  ## set_dvar ----
  dvar_choices_update <- reactive({
    scdf <- transformed()
    req(inherits(scdf, "scdf"))
    choices <- lapply(scdf, function(x) names(x)) |> unlist() |> unique()
    id <- which(!choices %in% scdf_attr(scdf)[c("var.mt", "var.phase")] |> unlist())
    updateSelectInput(
      session, 
      inputId = "setdvar", 
      choices = choices,
      selected = scdf_attr(scdf)["var.values"]
    )
  })
  
  # Updates when a nav tab is selected ------
  
  observeEvent(input$scan, {
    
    if (input$scan == "Transform") {
      dvar_choices_update()
    }
    
    if (input$scan == "Plot") {
      choices <- lapply(transformed(), function(x) names(x)) |> unlist() |> unique()
      id <- which(!choices %in% scdf_attr(transformed())[c("var.values", "var.mt", "var.phase")] |> unlist())
      updateSelectInput(
        session, 
        inputId = "scplot_add", 
        choices = c("None", choices[id])
      )
      choices <- correct_casenames(transformed(), string = TRUE)
      choices <- setNames(1:length(choices), choices)
      updateSelectInput(
        session, 
        inputId = "scplot_select_case", 
        choices = c("all", choices)
      )
    }
    if (input$scan == "Stats") {
      choices <- correct_casenames(transformed(), string = TRUE)
      choices <- setNames(1:length(choices), choices)
      updateSelectInput(
        session, 
        inputId = "stats_select_case", 
        choices = c("all", choices)
      )
    }
  })
  
  # +++ Tab scdf +++ ----
  
  ## Startup message ----
  
  output$scdf_messages <- renderPrint({
    if (is.null(getOption("scan.shinyscan.initial"))) {
      cat(res$msg$startup)
    }
  })

  ## Render scdf ----
  my_scdf <- reactiveVal(getOption("scan.shinyscan.initial", NULL))
  import_file <- reactiveVal()
  
  scdf_render <- reactive({
    
    if (!input$scdf_output_format){
      output$scdf_output <- renderPrint({
        if (!inherits(my_scdf(), "scdf")) validate(res$msg$no_case)
        if (identical(length(my_scdf()), 0)) validate(res$msg$no_case)
        do.call("summary", list(my_scdf()))
      })
    } else if (input$scdf_output_format) {
      output$scdf_output <- renderPrint({
        if (!inherits(my_scdf(), "scdf")) validate(res$msg$no_case)
        if (identical(length(my_scdf()), 0)) validate(res$msg$no_case)
        do.call("convert", list(
          my_scdf(), inline = as.logical(input$scdf_syntax_phase_structure)
        ))
      })
    }
    
    choices <- c(res$new_case, correct_casenames(my_scdf(), string = TRUE))
    updateSelectInput(session, "new_select_case", choices = choices)
  })

  observeEvent(input$new_clear_fields, {
    updateTextAreaInput(session, "new_values", value = "")
    updateTextInput(session, "new_mt", value = "")
    updateTextAreaInput(session, "new_variables", value = "")
    updateTextInput(session, "new_casename", value = "")
  }) 
  
  observeEvent(input$scdf_output_format, scdf_render()) 
  
  output$load_messages <- renderPrint({
    if (is.null(getOption("scan.shinyscan.initial"))) {
      cat(res$msg$startup)
    }
  })
  
  ## summary render ----
  render_summary <- reactive({
    tmp <- capture.output(out <- summary(import_file()) |> export())
    if (input$scan_export_engine == "gt")
      out <- gt::as_raw_html(out)
    HTML(out)
  })
  
  ## input example ----
  observeEvent(input$scdf_example, {
    if (input$scdf_example != "(none)") {
      import_file(paste0("scan::", input$scdf_example) |> str2lang() |> eval())
      scdf_render()
      msg <- paste("Example", input$scdf_example, 
                 "selected. Press 'Import scdf' to import it.")
      output$load_messages <- renderPrint(cat(msg))
      #showNotification(msg, type = "message")

      output$load_output_html <- renderUI({
        render_summary()
      })
     
      output$scdf_messages <- renderPrint(cat(""))
    } else {
      import_file(NULL)
    }
  })

  ## upload (load) ------
  
  load_file <- function () {
    
    if (length(input$upload$datapath) == 0) return()
    ext <- tolower(tools::file_ext(input$upload$datapath))
    
    if (ext == "rds") {
      readRDS(input$upload$datapath) |> import_file()
    } 
    if (ext == "r") {
      new <- readLines(input$upload$datapath)
      new <- paste0(new, collapse = "\n")
      .tmp <- new.env()
      eval(parse(text = new), envir = .tmp) |> import_file()
    } 
    if (ext %in% c("csv", "txt")) {
      sep <- input$scdf_csv
      dec <- if (sep == ";") "," else "."
      utils::read.table(
        input$upload$datapath, 
        header = TRUE, 
        stringsAsFactors = FALSE, 
        sep = sep,
        dec = dec,
        quote = "\"", 
        fill = TRUE, 
        comment.char = ""
      ) |> import_file()
    }
    if (ext %in% c("xlsx", "xls")) {
      readxl::read_excel(
        input$upload$datapath
      ) |>  
        as.data.frame() |> 
        import_file()
    }
    
    msg <- "File selected. Press 'Import scdf' to import it"
    output$load_messages <- renderText(msg)
    showNotification(msg, type = "message")
    
    if (inherits(import_file(), "data.frame")) {
      choices <- names(import_file())
      updateSelectInput(
        session, "scdf_load_cvar", choices = choices, 
        selected = guess_col(choices, c("case", "id", "subject", "name"))
      )
      updateSelectInput(
        session, "scdf_load_pvar", choices = choices, 
        selected = guess_col(choices, c("phase", "condition", "segment"))
      )
      updateSelectInput(
        session, "scdf_load_mvar", choices = choices, 
        selected = guess_col(choices, c("measurement", "mt", "time", "t", "day", "session"))
      )
      updateSelectInput(
        session, "scdf_load_dvar", choices = choices, 
        selected = guess_col(choices, c("values", "score", "y", "outcome"))
      )
      
      output$load_output_html <- renderUI({
        if (input$scan_export_engine == "gt") {
          import_file() |> gt::gt() |> gt::as_raw_html() |> HTML()
        } else {
          import_file() |> knitr::kable(format = "html") |> HTML()
        }
      })
    }
    
    if (inherits(import_file(), "scdf")) {
      output$load_output_html <- renderUI({
        render_summary()
      })
    }
  }
  
  observeEvent(input$scdf_csv, load_file())
  
  observeEvent(input$upload, load_file())

  observeEvent(input$scdf_import, {
    
    if (!inherits(import_file(), "scdf") && 
        !inherits(import_file(), "data.frame")) {
      msg <- "Sorry,\n the file you tried to load is not valid."
      output$load_messages <- renderPrint(msg)
      showNotification(msg, type = "message")
      return(NULL)
    } 
    
    if (inherits(import_file(), "data.frame")) {
      tryCatch(
        read_scdf(
          import_file(),
          cvar = input$scdf_load_cvar,
          pvar = input$scdf_load_pvar,
          dvar = input$scdf_load_dvar,
          mvar = input$scdf_load_mvar,
          na =  eval(str2lang(paste0("c(", input$scdf_load_na, ")")))
        ) |> import_file(),
        error = function(e) {
          msg <- "Sorry,\n the file you tried to load is not a valid scdf file."
          #showNotification(msg, type = "error")
          output$load_messages <- renderPrint({msg; "\n\n"; e})
        }
      )
    }
    
    my_scdf(import_file())
    scdf_render()
    msg <- "loaded file successfully"
    output$load_messages <- renderPrint(cat(msg))
    showNotification(msg, type = "message")
    
    # update dvar selection choices in transform tab
    dvar_choices_update()
    
    output$load_output_html <- renderUI({
      render_summary()
    })
    output$scdf_messages <- renderPrint(cat(""))
 
  })
  
  
  ## download (save) ----
  output$scdf_save <- downloadHandler(
    filename = function() {
      scdf <- my_scdf()
      out <- paste(
        input$scdf_save_prefix,
        sprintf("%02d", length(scdf)),
        paste0(unique(scdf[[1]]$phase), collapse = ""),
        format(Sys.time(), format = "%y%m%d-%H%M%S"),
        sep = "-"
      )
      paste0(out, input$scdf_save_format)
    },
    
    content = function(file) {
      scdf <- my_scdf()
      
      if (!inherits(scdf, "scdf")) {
        output$scdf_messages <- renderPrint(cat(res$error_msg$scdf_save))
        showNotification(res$error_msg$scdf_save, type = "message")
      } else {
        msg <- "Saved file"
        output$scdf_messages <- renderPrint(cat(msg))
        res$error_msg$scdf_save
      }
      
      if (input$scdf_save_format == ".rds") 
        saveRDS(scdf, file)
      if (input$scdf_save_format == ".R") 
        convert(scdf, file = file)
      if (input$scdf_save_format == ".csv") 
        write_scdf(scdf, filename = file)

    }
  )
  
  ## select case ----
  
  observeEvent(input$new_select_case, {
    case <- input$new_select_case

    if (!identical(case, res$new_case)) {
      casenames <- correct_casenames(my_scdf(), string = TRUE)
      id <- match(case, c(res$new_case, casenames))

      scdf <- my_scdf()[id - 1]
      dv <- scan:::dv(scdf)
      mt <- scan:::mt(scdf)
      phase <- scan:::phase(scdf)
     
      # values string
      x <- split(
        scdf[[1]][[dv]], 
        scdf[[1]][[phase]]
      )
      x <- mapply(
        function(x, n) {
          paste0(n, " = ", paste0(x, collapse = ", "), collapse = "")
        },
        x, names(x)
      )
      values_string <- paste0(x,collapse = "\n")
      updateTextAreaInput(session, "new_values", value = values_string)
      
      # mt string
      imt <- scdf[[1]][[mt]]
      if (identical(as.integer(imt), seq_len(length(imt)))) {
        mt_string <- ""
      } else {
        mt_string <- paste0(imt, collapse = ", ") 
      }
      updateTextInput(session, "new_mt", value = mt_string)
      
      # additional var string
      names_add_var <- names(scdf[[1]])[which(!names(scdf[[1]]) %in% c(dv, phase, mt))]
      if (length(names_add_var) > 0) {
        add_var_str <- lapply(scdf[[1]][, names_add_var, drop = FALSE], function(x) paste0(x, collapse = ", "))
        add_var_str <- paste0(names(add_var_str), " = ", add_var_str, collapse = "\n")
        updateTextAreaInput(session, "new_variables", value = add_var_str)
      }
      
      # name str
      name_str <- names(my_scdf())[id - 1]
      if(!is.null(name_str) && !identical(name_str, "") && !is.na(name_str))
        updateTextInput(session, "new_casename", value = name_str)
    }
  })
  
  ## save case --------
  observeEvent(input$new_save_case, {
    tryCatch({
      values <- paste0("c(", trim(input$new_values), ")")
      dvar <- "values"
      if (inherits(my_scdf(), "scdf")) {
        dvar <- scdf_attr(my_scdf(), "var.values")
      }
      
      call <- paste0(dvar, " = ", values)
      call <- c(call, paste0("dvar = ", deparse(dvar)))
      if (input$new_mt != "") call <- c(call, paste0("mt = c(", input$new_mt, ")"))
      
      if (input$new_variables != "") {
        variables <- input$new_variables |>
          strsplit("\n")  |>
          unlist() |>
          lapply(function(y) strsplit(y, "=")) |>
          unlist(recursive = FALSE) |>
          lapply(function(y) paste0(y[1], " =c(", y[2], ")")) |>
          unlist()
        call <- c(call, variables)
      }
      
      if (input$new_casename != "") {
        call <- c(call, paste0("name = ", deparse(input$new_casename)))
      } else {
        call <- c(call, paste0("name = ", deparse(sample_names())))
      }
      
      call <- paste0(call, collapse = ",")
      call <- paste0("scdf(", call, ")")
      
      new <- call |> str2lang() |> eval()
      
      scdf <- my_scdf()
      casenames <- correct_casenames(scdf, string = TRUE)
      
      case <- input$new_select_case
      
      position <- 0
      if (length(which(casenames %in% case) > 0)) {
        position <- which(casenames %in% case)
      }
     
      if (position == 0) {
        if (length(scdf) > 0) new <- c(scdf, new)
        msg <- "Appended case"
      } else {
        scdf[[position]] <- new[[1]]
        if (is.null(names(new))) names(new) <- ""
        names(scdf)[position] <- names(new)
        new <- scdf
        msg <- paste("Replaced case", case)
      }
      
      output$scdf_messages <- renderPrint(cat(msg))
      showNotification(msg, type = "message")
      
      my_scdf(new)
      scdf_render()
    },
    error = function(e) {
      msg <- paste0(res$error_msg$invalid_case, "\n\n", e)
      output$scdf_messages <- renderPrint(cat(msg))
      #showNotification(msg, type = "error")
    }
      
    )
  })
  
  ## remove cases --------
  observeEvent(input$new_remove_case, {
    
    casenames <- correct_casenames(my_scdf(), string = TRUE)
    active_case <- input$new_select_case
    
    position <- 0
    if (length(which(casenames %in% active_case) > 0)) {
      position <- which(casenames %in% active_case)
    }
    
    if (position > 0) {
      my_scdf(my_scdf()[-position])
      msg <- paste("removed case", casenames[position])
    } else {
      msg <- paste("No case removed. please selct case first.")
    }
    
    output$scdf_messages <- renderPrint(cat(msg))
    showNotification(msg, type = "message")
    
    scdf_render()
  })

  ## clear cases --------
  observeEvent(input$new_clear_cases, {
    my_scdf(NULL)
    
    msg <- "Cleared cases"
    output$scdf_messages <- renderPrint(cat(msg))
    showNotification(msg, type = "message")

    scdf_render()
  })

  # +++ Tab Transform +++ ----

  ### render ----
  transformed <- reactive({
    out <- my_scdf()
    syntax = "scdf"
    
    # select_cases 
    if (input$select_cases != "") {
      call <- str2lang(paste0("select_cases(out, ", input$select_cases,")"))
      out <- eval(call)
      syntax <- c(syntax, paste0("select_cases(", input$select_cases,")"))
    }
    
    # select_phases 
    if (input$select_phasesA != "" || input$select_phasesB != "") {
      out <- paste0(
          "select_phases(out, A = c(", input$select_phasesA, "), B = c(",
          input$select_phasesB, "))"
        ) |>
        str2lang() |> 
        eval()
  
      syntax <- c(
        syntax, 
        paste0(
          "select_phases(A = c(", input$select_phasesA, "), B = c(",
          input$select_phasesB, "))"
        )
      )
    }

    # subset - filter
    if (input$subset != "") {
      args <-  list(str2lang(input$subset))
      out <- do.call("subset", c(list(out),args))
      syntax <- c(syntax, paste0("subset(",  input$subset, ")"))
    }

    # transform
    if (input$transform != "") {
      arg <- paste0("transform(out,", trim(input$transform),")")
      out <- str2lang(arg) |> eval()
      syntax <- c(
        syntax, paste0("transform(", gsub("\n", ", ", trim(input$transform)),")")
      )
    }

    # set_dvar
    if (input$setdvar != "") {
      args <-  list(input$setdvar)
      out <- do.call("set_dvar", c(list(out),args))
      syntax <- c(syntax, paste0("set_dvar(",  deparse(input$setdvar), ")"))
    }

    if (length(syntax) > 1) {
      syntax <- syntax[-1]
      syntax <- paste0(
        "scdf", res$pipe_br, " ",
        paste0(syntax, collapse = paste0(res$pipe_br, " "))
      )
    }

    output$transform_syntax <- renderPrint(cat(syntax))

    out
  })

  ### save ----
  output$transformed_save <- downloadHandler(
    filename = function() {
      scdf <- transformed()
      out <- paste(
        input$transform_save_prefix,
        sprintf("%02d", length(scdf)),
        paste0(unique(scdf[[1]]$phase), collapse = ""),
        format(Sys.time(), format = "%y%m%d-%H%M%S"),
        sep = "-"
      )
      paste0(out, input$transform_save_format)
    },
    content = function(file) {
      if (input$transform_save_format == ".rds") 
        saveRDS(transformed(), file)
      if (input$transform_save_format == ".R") 
        convert(transformed(), file = file)
      if (input$transform_save_format == ".csv") 
        write_scdf(transformed(), filename = file)
    }
  )
  
  ### output ------
  
  output$transform_scdf <- renderPrint({
    if(!inherits(my_scdf(), "scdf")) validate(res$msg$no_case)
    print(transformed(), rows = 100)
  })

  output$transform_html <- renderUI({
    if(!inherits(my_scdf(), "scdf")) validate(res$msg$no_case)
    
    options(scan.export.engine = input$scan_export_engine)
    
    if (getOption("scan.export.engine") == "gt")
      export(transformed(), caption = "") |> gt::as_raw_html() |> HTML()
    else
      export(transformed(), caption = "") |> HTML()
    
  })
  
  # +++ Tab Stats ++++ -----
  
  stats_class <- reactiveVal()
  stats_args_tab_initialized <- reactiveVal(FALSE)
  
  ### Calculate ---- 
  calculate_stats <- reactive({
    if (!inherits(my_scdf(), "scdf")) validate(res$msg$no_case)
    scdf <- transformed()
    
    if (input$stats_select_case != "all") {
      scdf <- str2lang(
        paste0("select_cases(scdf, ", input$stats_select_case, ")")
      ) |> eval()
    }
    
    call <- get_stats_call()
    call <- paste0("scan::", call)
    
    tryCatch(
      out <- str2lang(call) |> eval(),
      error = function(e) {
        msg <- paste0("Sorry, could not proceed with the calculation:\n\n", e)
        #showNotification(msg, type = "error")
        validate(msg)
      }
    )
    stats_class(class(out))
    update_print()
    out
  })

  update_print <-  reactive({
     stats_class <- stats_class()
     if (is.null(stats_class)) return(NULL)
     out_func <- if (input$stats_out) "export" else "print" 
     call <- paste0("formals(scan:::", out_func, ".", stats_class, ")")
     tryCatch(
       formals <- str2lang(call) |> eval(),
       error = function(e) {
         #showNotification(res$error_msg$html_output, type = "error")
         validate(res$error_msg$html_output)
         return(NULL)
       }
     )
     filter <- lapply(formals, function(x) !is.symbol(x)) |> unlist()
     formals <- formals[filter]
     formals <- lapply(formals, function(x) {
       if (is.character(x)) x <- paste0(paste0("\"", x, "\""))
       x
     })
     id <- which(!names(formals) %in% c("select", "caption", "footnote", "filename"))
     formals <- formals[id]
     if (length(formals) == 0) {
       placeholder <- ""
     } else {
       placeholder <- paste0(names(formals), " = ", formals, collapse = ", ")
     }
     updateTextInput(
       session, inputId = "stats_print_arguments", value = placeholder
     )
    
  })
 
  ### Output ----
  output$stats_html <- renderUI({
    results <- calculate_stats()
    print_args <- input$stats_print_arguments
    options(scan.export.engine = input$scan_export_engine)
    if (!identical(print_args, "")) {
      print_args <- paste0(", ", print_args)
      call<- paste0("export(results, ", print_args, ")")
    } else {
      call <- paste0("export(results)")
    }
    
    tryCatch(
      if (getOption("scan.export.engine") == "gt") {
        str2lang(call) |> eval() |> gt::as_raw_html() |> HTML()
      } else {
        str2lang(call) |> eval() |> print() |> as.character() |> HTML()
      },
      error = function(e) {
        validate(res$error_msg$html_output)
      }
    )    

  })

  output$stats_text <- renderPrint({
    results <- calculate_stats()
   
    print_args <- input$stats_print_arguments
    if (print_args != "") {
      print_args <- paste0(", ", print_args)
      call<- paste0("print(results, ", print_args, ")")
    } else call <- "print(results)"
    
    str2lang(call) |> eval()     
  })

  observeEvent(input$stats_tabs, {
    if (identical(input$stats_tabs, "Arguments")) {
      stats_args_tab_initialized(TRUE)
    }
  })
  
  ### observe stats_func ----
  
  observeEvent(input$stats_func, {
    ## show description
    if(input$stats_description) {
      hf <- help(input$stats_func, package = "scan")
      res <- utils:::.getHelpFile(hf)
      desc <- Filter(function(x) attr(x, "Rd_tag") == "\\description", res)
      if (length(desc) > 0) {
        desc <- paste0(unlist(desc)[-1], collapse = "")
        desc <- gsub("\n", "<br>", trimws(desc))
        output$stats_description <- renderUI({
          tags$p(HTML(desc), style = "font-size:18px; color:darkblue;")
        })
      }
    }
    updateTextAreaInput(session, "stats_print_arguments", value = "")
    stats_args_tab_initialized(FALSE)
    output$stats_arguments <- renderUI(render_stats_args_input())
  })
  
  output$stats_syntax <- renderPrint({
    cat(get_stats_call())
  })


  ### get stat func arguments ------
  stat_arg_names <- reactive({
    .formals <- formals(eval(str2lang(paste0("scan::", input$stats_func))))
    args <- names(.formals)
    values <- .formals

    # remove some arguments
    id <- which(!args %in% c(
      "dvar", "pvar", "mvar", "phases",
      "data", "scdf", "data.l2", "offset", "lag.max",
      "graph", "output", "...")
    )
    
    # add dvar for mplm
    if (input$stats_func == "mplm") {
      id <- c(which(args == "dvar") , id)
      values[which(args == "dvar")] <- scan:::dv(transformed())
    }
    
    list(names = args[id], values = values[id])
  })

  ### create argument inputs ----
  render_stats_args_input <- reactive({
    args <- stat_arg_names()
    
    out <- vector("list", length(args$names))
    if (length(out) > 0) {
      for (i in 1:length(out)) {
        
        value <- args$values[[i]]
        if (is.character(value)) value <- deparse(value)
        
        if (!is.name(value) && !inherits(value, "call") && isTRUE(is.na(value))) 
          value <- substitute(value) |> deparse()
        if (is.null(value)) value <- substitute(value) |> deparse()
        if (!is.numeric(value) && !is.logical(value) && !is.character(value) &&
            !is.call(value)) {
          value <- substitute(value) |> deparse()
        }
        if (is.call(value)) {
          if (is.character(eval(value))) {
            value <- eval(value)
          } else {
            value <- substitute(value) |> deparse()
          }
        }
        outvalue <- value
        
        if (length(value) > 1) {
          choices <- setNames(quoted(value), value)
          selected <- names(choices)[1]
          out[[i]] <- selectInput(
            args$names[i], args$names[i],
            choices = choices,
            selected = selected
          )
        } else if (is.numeric(value)) {
          out[[i]] <- numericInput(
            args$names[i], args$names[i], value = outvalue
          )
        } else if (is.logical(value)) {
          #choices <- c("FALSE" = FALSE, "TRUE" = TRUE)
          out[[i]] <- checkboxInput( 
            args$names[i], 
            tags$span(args$names[i], class = "chklabel-big"),
            value = outvalue
          )
        } else {
          out[[i]] <- textInput(args$names[i], args$names[i], value = outvalue)
        }
      }
      return(out)
    } else {
      return(NULL)
    }
  })

  ### get stats call ----
  get_stats_call <- reactive({
    func <- input$stats_func
    # returns function without args if tab has not been initialised

    if (!stats_args_tab_initialized()) {
      return(paste0(input$stats_func, "(scdf)"))
    }
   
    full_args <- stat_arg_names()
    args <- full_args
    values <- sapply(args$names, function(name) input[[name]])
    args <- args$names
    options(scan.rename.predictors = input$rename_predictors)
    
    if (!input$stats_default) {
      id_default <- mapply(
        function(origin, new) {
          
          if (typeof(origin) == "language") origin <- eval(origin)
          if (typeof(origin) == "symbol") origin <- deparse(origin)
          if (identical(new, origin)) return(TRUE)
          if (identical(new, as.character(origin[1]))) return(TRUE)
          if (identical(new, deparse(origin))) return(TRUE)
          
          if (substr(deparse(origin[1]), 1,3) == "NA_") origin <- NA
          if (identical(new, deparse(origin[1]))) return(TRUE)
          
          return(FALSE)
        },
        origin = full_args$values,
        new = values
      ) |> unlist()
      values[id_default] <- ""
    }
    
    id <- which(values != "")
    
    args <- args[id]
    values <- values[id]
    
    str_scdf <- "scdf"
    call <- paste0(
      input$stats_func, "(", str_scdf,
      if (length(args > 0)) {
        paste0(", ",paste0(args, " = ", values, collapse = ", "))
      } else {
        ""
      },
      ")"
    )

    call
  })

  ### Save ------

  output$stats_save <- downloadHandler(
    
    filename = function() {
      scdf <- transformed()
      out <- paste(
        input$prefix_output_stats, input$stats_func,
        sprintf("%02d", length(scdf)),
        paste0(unique(scdf[[1]]$phase), collapse = ""),
        format(Sys.time(), format = "%y%m%d-%H%M%S"),
        sep = "-"
      )
      ext <- if (!input$stats_out) "txt" else "html"
      if (input$setting_output_docx && input$stats_out) ext <- "docx" 
      out <- paste0(out, ".", ext)
      out
    },
    content = function(file) {
      
      if (!input$stats_out) {
        results <- calculate_stats()
        print_args <- input$stats_print_arguments
        if (print_args != "") {
          print_args <- paste0(", ", print_args)
          call<- paste0("print(results, ", print_args, ")")
        } else call <- "print(results)"
        
        call <- paste0("capture.output(", call, ")")
        writeLines(str2lang(call) |> eval(), con = file)
      }
      
      if (input$stats_out) {
        results <- calculate_stats()
        print_args <- input$stats_print_arguments
        if (print_args != "") {
          print_args <- paste0(", ", print_args)
          call<- paste0(
            "export(results, ", print_args, ", filename = '", file, "')" 
          )
        } else {
          call <- paste0("export(results, filename = '", file, "')")
        }
        
        out <- if (getOption("scan.export.engine") == "gt")
          str2lang(call) |> eval() |> gt::as_raw_html()
        else
          str2lang(call) |> eval()
      }
    }
  )
  
  # +++ Tab Power test +++ -----
  
  output$pt_results <- renderPrint(cat(res$placeholder$pt))
  
  ## Render -----
  render_power_test <- reactive({
    syntax <- paste0(
      "design(\n  n = ", input$design_n, ", ",
      "phase_design = list(", input$design_phase, "), \n  ",
      "trend = ", input$design_trend, ", ",
      "level = list(", input$design_level, "), ",
      "slope = list(", input$design_slope, "), \n  ",
      "start_value = ", input$design_start, ", ",
      #"s = ", input$design_s, ", ",
      "rtt = ", input$design_rtt, ", ",
      "distribution = ", deparse(input$design_distribution),
      "\n)"
    )  
    
    ci <- input$pt_ci
    
    pt_method <- deparse(as.list(input$pt_method))
    if (input$pt_method_user != "") {
      
      if (length(input$pt_method) >0) {
        pt_method <- paste0("\"", input$pt_method, "\"",collapse = ",")
        pt_method <- paste0(
          "list(user = function(scdf) {", input$pt_method_user,"}, ",
          pt_method, ")"
        )
      } else {
        pt_method <- paste0("list(user = function(scdf) {", input$pt_method_user,"})")
      }
    }
    
    syntax <- paste0(
      syntax, res$pipe,
      "  power_test(method = ", pt_method, ", ",
      "effect = ", deparse(input$pt_effect), ", ",
      "n_sim = ", input$pt_n, 
      if (!identical(ci, "")) paste0(", ci = ", ci),
      ")"
    )
    
    syntax
  })
  
  ## Output ----
  output$pt_syntax <- renderPrint({
    cat(render_power_test())
  })
  
  ## Plot ----
  
  observeEvent(input$desigh_plot_button, {
    call <- paste0(
      "design(\n  n = ", input$design_n, ", ",
      "phase_design = list(", input$design_phase, "), \n  ",
      "trend = ", input$design_trend, ", ",
      "level = list(", input$design_level, "), ",
      "slope = list(", input$design_slope, "), \n  ",
      "start_value = ", input$design_start, ", ",
      #"s = ", input$design_s, ", ",
      "rtt = ", input$design_rtt, ", ",
      "distribution = ", deparse(input$design_distribution),
      "\n)", res$pipe,
      "random_scdf()", res$pipe,
      "scplot()"
    )
    
    output$plot_design <- renderPlot(res = 100, {
      str2lang(call) |> eval()
    })
  })
  
  ## Analyse ----
  
  observeEvent(input$pt_compute, {
    
    phase_structure <- eval(str2lang(
      paste0("list(", input$design_phase, ")")
    ))
    if (length(phase_structure) > 2) {
      output$pt_results <- renderPrint({
        cat("Sorry, power-tests are only possible for designs with two phases.")
      })
    } else {
      call <- render_power_test()
      output$pt_results <- renderPrint({cat("Calculating ...")})
      res <- tryCatch(
        str2lang(call) |> eval(),
        error = function(e) {
          msg <- paste0("Sorry, could not proceed calculation:\n\n", e)
          #showNotification(msg, type = "error")
          msg
        }
          
      )
      output$pt_results <- renderPrint({
        if (inherits(res, "character")) cat(res) else res
      })
    }
    
  })
  
  ## quit app -----
  observeEvent(input$navpage, {
    if (input$navpage == "Quit") {
      session$sendCustomMessage(
        type = "closeWindow", 
        list(message = "window.close();")
      )
      stopApp()
    }
  })
  
  # +++ Tab Plot +++ -----
  
  create_scplot_call <- reactive({
    themes <- c(input$scplot_theme_1, input$scplot_theme_2, input$scplot_theme_3)
    themes <- themes[which(themes != "None")]
    
    call <- c(
      "transformed()",
      if (input$scplot_select_case != "all") {
        paste0("select_cases(", input$scplot_select_case, ")")
      },
      "scplot()",
      if (!is.na(input$scplot_ymin) || !is.na(input$scplot_ymax)) {
        paste0("set_yaxis(limits = c(", input$scplot_ymin, ",", input$scplot_ymax, "))")
      },
      if (!is.na(input$scplot_xinc)) {
        paste0("set_xaxis(increment = ", input$scplot_xinc, ")")
      },
      if (input$scplot_stats_mean_a) 'add_statline("mean", phase = 1)',
      if (input$scplot_stats_median_a) 'add_statline("median", phase = 1)',
      if (input$scplot_stats_max_a) 'add_statline("max", phase = 1)',
      if (input$scplot_stats_min_a) 'add_statline("min", phase = 1)',
      if (input$scplot_stats_trend) 'add_statline("trend")',
      if (input$scplot_stats_mean) 'add_statline("mean")',
      if (input$scplot_stats_median) 'add_statline("median")',
      if (input$scplot_stats_moving) 'add_statline("moving mean")',
      if (input$scplot_stats_loess) 'add_statline("loess", span = 0.4)',
      if (input$scplot_stats_trend_a) 'add_statline("trendA")',
      if (input$scplot_add != "None") paste0('set_dataline("', input$scplot_add, '")'),
      
      paste0("set_theme(", paste0("'", themes, "'", collapse = ", "), ")"),
      if (input$scplot_text_size > 6) paste0('set_base_text(size = ', input$scplot_text_size, ')'),
      if (input$scplot_legend_position != "none") {
        #paste0('add_legend(position = ',  input$scplot_legend_position , ')')
        glue::glue('add_legend(position = "{input$scplot_legend_position}")')
      },
      if (!input$scplot_ylabel == "") paste0('set_ylabel(label = ', deparse(input$scplot_ylabel), ')'),
      if (!input$scplot_xlabel == "") paste0('set_xlabel(label = ', deparse(input$scplot_xlabel), ')'),
      if (!input$scplot_footer) 'add_caption(label = NULL)',
      if (!input$scplot_phasenames == "") {
        pn <- input$scplot_phasenames |> strsplit(",") |> unlist()
        pn <- paste0("'", pn, "'", collapse = ", ")
        paste0('set_phasenames(labels = c(', pn, '))')
      },
      if (!input$scplot_casenames == "") {
        pn <- input$scplot_casenames |> strsplit(",") |> unlist()
        pn <- paste0("'", pn, "'", collapse = ", ")
        paste0('set_casenames(labels = c(', pn, '))')
      }
    )
    call <- paste0(call, collapse = " |>\n\t")
    call
  })
  
  
  
  
  output$scplot_syntax <- renderPrint({

    #call <- create_scplot_call()
    #cat(call)
  })
  
  ## Render plot ----

  output$scplot_plot <- renderPlot(res = 120, {
    if (!inherits(my_scdf(), "scdf")) validate(res$msg$no_case)
    req(inherits(my_scdf(), "scdf"))
    
    call <- paste0(create_scplot_call(), " |> print()")
    out <- NULL
    tryCatch(
      out <- str2lang(call) |> eval(),
      error = function(x) {
        out <- paste0(res$error_msg$plot, "\n\n", x)
        
      }
    )
    out
  })
  
  ## Save Render plot new----
  output$saveplot_2 <- downloadHandler(
    filename = function() {
      scdf <- transformed()
      out <- paste(
        input$prefix_output_plot,
        sprintf("%02d", length(scdf)),
        paste0(unique(scdf[[1]]$phase), collapse = ""),
        format(Sys.time(), format = "%y%m%d-%H%M%S"),
        sep = "-"
      )
      paste0(out, ".png")
    },
    content = function(file) {
      ggplot2::ggsave(
        file, 
        paste0(create_scplot_call(), " |> print()") |> str2lang() |> eval(),
        width = input$width, height = input$height,
        dpi = input$dpi, units = "px",  device = "png"
      )
    }
  )
  
  
}

errors <- c()

if (length(errors) > 0) stop(errors)

suppressPackageStartupMessages({
library(scan)
library(scplot)
library(shiny)
})

options(
  scan.export.kable_styling = list(
    bootstrap_options = c("striped", "condensed"), 
    full_width = FALSE,
    position = "left"
  )
)

res <- list()

# choices ------

res$choices <- list()
examples <- data(package = "scan")$results[, 3]
filter <- startsWith(examples,"Grosche2014") | 
          endsWith(examples,"Leidig2018_l2") | 
          startsWith(examples,"exampleAB_50.l2")
examples <- examples[!filter]

res$choices$examples <- c("(none)", examples)

res$choices$scplot_examples <- c(
"(empty selection)" = "",
"Trend lines" = 'add_statline("trend")',
"Baseline trend" = 'add_statline("trendA")',
"Max A" = 'add_statline("max", phase = "A")',
"Means" = 'add_statline("mean")',
"Medians" = 'add_statline("median")',
"Moving average" = 'add_statline("movingMean")',
"Smoothed line" = 'add_statline("loess", span = 0.4)'
)

res$choices$scplot_templates_design <- c(
  "",
  "Legend" = 'add_legend()',
  "Add title" = 'add_title("A new plot", color = "darkblue", size = 1.3)',
  "Add caption" = 'add_caption("Note. What a nice plot!", face = "italic", colour = "darkred")',
  "Set axis labels" = 'set_ylabel("Score", color = "darkred", angle = 0)
set_xlabel("Session", color = "darkred")',
  "Set phase names" = 'set_phasenames(labels = c("Baseline", "Intervention", "Extended", "Follow-up"), color = "darkblue", size = 0.9, face = "italic")',
  "Set case names" = 'set_casenames(position = "strip", background = list(fill = "lightblue"))',
  "Resize size" = 'set_base_text(size = 19)',
  "Background" = 'set_background(fill = "grey90", color = "black", size = 2)'

)

res$choices$scplot_templates_annotate <- c(
  "",
  "Marks" = 'add_marks(case = "all", position = \"values < mean(values)\", shape = 16, size = 2)',
  "Text" = 'add_text("Hallo", case = 1, x = 5, y = 20)',
  "Arrow" = 'add_arrow(case = 1, 2, 70, 6, 55, color = "darkred")'
)  

themes <- names(scplot:::.scplot_themes)

for(i in seq_along(themes)) {
  res$choices$scplot_templates_design[[paste0("Theme ", themes[i])]] <-
    paste0('set_theme("', themes[i], '")')
}

res$choices$fn_stats <- c(
  "Descriptives" = "describe",
  "Standardized mean differences" = "smd",
  "Overlap indices" = "overlap",
  "Data trends" = "trend",
  "Auto correlations" = "autocorr",
  "Piecewise regression" = "plm",
  "Hierarchical piecewise regression" = "hplm",
  "Conservative Dual-Criterion" = "cdc",
  "Tau U" = "tau_u",
  "Baseline corrected tau" = "corrected_tau",
  "Percentage of non-overlapping data" =  "pnd",
  "Percent exceeding the median" = "pem",
  "Percent exceeding the trend" = "pet",
  "Percentage of all non-overlapping data" = "pand",
  "Improvement rate difference" = "ird",
  "Nonoverlap of all Pairs" = "nap",
  "Randomization test" = "rand_test",
  "Outlier analysis" = "outlier"
)

.name <- function(x, at, label) {
  names(x)[which(x == at)] <- label
  x
}


res$choices$pt_method <- names(scan:::.opt$mc_fun)  |> 
  setNames(names(scan:::.opt$mc_fun)) |> 
  .name("plm_level", "Regression (level effect)") |> 
  .name("plm_slope", "Regression (slope effect)") |> 
  .name("plm_poisson_level", "Regression (level effect; frequencies)") |> 
  .name("plm_poisson_slope", "Regression (slope effect; frequencies)") |> 
  .name("hplm_level", "Multilevel-regression (level effect)") |> 
  .name("hplm_slope", "Multilevel-regression (slope effect)") |> 
  .name("tauU", "Tau-U (trend A)") |>
  .name("tauU_slope", "Tau-U (trend A and B)") |>
  .name("tauU_meta", "Metaanalysis Tau-U (trend A)") |>
  .name("tauU_slope_meta", "Metaanalysis Tau-U (trend A and B)") |>
  .name("base_tau", "Baseline corrected tau") |>
  .name("rand", "Randomization test") |> 
  .name("rand_decrease", "Randomization test (decreasing)")




# placeholder ----

res$placeholder$values <- "Enter values here to create a new case. E.g. \nA = 1,2,3,4,3 \nB = 7,6,7,8,7,6"

res$placeholder$transform <- 'e.g.
values = scale(values)
values = local_regression(values)
values2 = values - max(values[phase=="A"])
across_cases(values2 = scale(values)
'

res$placeholder$plot_arguments <- '(choose one or more of the templates below and experiment with the syntax here.)
'

res$placeholder$mt <- "(optional, e.g. 1,2,4,6,7,8,9,12,13)"

res$placeholder$variables <-
"(optional, e.g., depression = 1,4,3,5,6,5,7
separate multiple variables with linebreaks)"

res$placeholder$pt <- "Power calculation may take some time. Click 'Run' to start calculation."

# div ------

res$div$settings <- paste0(
  "background-color:#f0f0f0; ",
  "border: 1px solid black; border-radius: 5px; ",
  "padding-left: 10px; padding-right: 10px; ",
  "padding-top: 0px; padding-bottom: 0px; "
)

res$div$pt <- paste0(
  "background-color:#f0f0f0; ",
  "border: 1px solid black; border-radius: 5px; ",
  "padding-left: 5px; padding-right: 5px; ",
  "padding-top: 0px; padding-bottom: 0px; "
)

# error ----

res$error_msg$invalid_case <- "Sorry!
The last case you tried to add didn't have a valid case definition."

res$error_msg$plot <- "Sorry!
The plot arguments are not valid."

res$error_msg$scdf_save <- "Sorry!
The last file you saved is corrupt. Did you forget to add a case before saving?"

# msg ----

res$msg$startup <-
"Welcome to 'shiny scan'!

You can:

1. create a new case (fill in 'values' and click 'Add')
2. load a dataset (click 'Load file' to import an rds, csv, or excel file)
3. choose an example scdf (choose from 'Load example')

'exampleABC' is a good place to start.

The basic procedure is:

1. Choose or create an scdf (Single Case Data Frame) in the 'Data' tab.
2. Optionally refine the scdf in the 'Transform' tab (e.g. select cases, recombine phases)
3. Analyse the data in the 'Stats' tab.
4. Create a plot in the 'Plot' tab.

Look at the 'Help' tab for more information.

Have fun!
"

res$msg$no_case_scdf <-
"No case has been defined yet.
You can:
1. create a new case (fill in 'values' and click 'Add')
2. load a dataset (click 'Load file' to import an rds, csv, or excel file)
3. choose an example scdf (from 'Load example')
"

res$msg$no_case <-
"There is no case defined yet.
Please define a case on the 'Data' tab first.
"

res$help_page <- structure(
  "<h4 id=\"welcome-to-shiny-scan\">Welcome to <em><strong>shiny scan</strong></em>!</h4>
  <p><em>Shiny-scan</em> is a graphical surface for <em>scan</em> (Single-Case Data Analysis). <em>scan</em> is an R package.</p>
  <p>The basic procedure is:</p>\n<ol>
  <li>Choose/ create a single case file in the <strong>Data</strong> tab.</li>
  <li>Optionally refine the case in the <strong>Transform tab</strong> (select cases, recombine phases, etc.)</li>
  <li>Analyse the data in the <strong>Stats tab</strong>.</li>
  <li>Create a plot in the <strong>Plot tab</strong>.</li>\n</ol>
  <p>Analysis and plots are based on the scdf after any changes from the <strong>Transform tab</strong>.</p>
  <p>Here are helpful links:</p>
  <p><a href=\"https://jazznbass.github.io/scan/\">Help pages for scan</a></p>
  <p><a href=\"https://jazznbass.github.io/scan-Book/\">Online book for single case analysis with scan</a></p>
  <p><a href=\"https://jazznbass.github.io/scplot/\">Help pages for scplot</a></p>
  <p>Have fun!</p>",
  html = TRUE, class = c("html", "character")
)

### little help-functions

trim <- function(x) {
  gsub("\n", ", ", trimws(x))
  x <- gsub("\n", ", ", trimws(x))
  gsub(",,", ",", trimws(x))
}

quoted <- function(x) {
  out <- paste0("\"", x, "\"")
  out[is.na(x)] <- NA
  out
}

n2br <- function(x) gsub("\n", "<br>", x)

source("resources.R")

# scdf ------
tab_scdf <-   tabPanel(
  "Data",
  sidebarLayout(
    sidebarPanel(
      textAreaInput(
        "values", "Values", placeholder = res$placeholder$values
      ),
      textInput("mt", "Measurement times", placeholder = res$placeholder$mt),
      textAreaInput(
        "variables", "Additional variables",
        placeholder = res$placeholder$variables
      ),
      textInput("casename", "Case name", placeholder = "(optional)"),
      actionButton("add_case", "Add case"),
      actionButton("remove_case", "Remove case"),
      actionButton("remove_all", "Clear all cases"),
      br(),
      div(style="display:inline-block;",
          radioButtons(
            "remove_which", "Position", choices = c("last" = "last", "at:" = "at"), inline = TRUE
          ),
      ),
      div(style="display:inline-block; padding-left: 0px",
          numericInput("remove_at", "", min = 1,value = 1,width = "75px"),
      ),
      br(),
      selectInput(
        "scdf_example", "Load example", choices = res$choices$examples,
      ),
      fileInput(
        "upload", NULL, accept = c(".csv", ".rds", ".xlsx", ".xls", ".R", ".r"),
        buttonLabel = "Load file"
      ),
      downloadButton("scdf_save", "Save scdf"),
    ),

    mainPanel(
      verbatimTextOutput("scdf_messages"),
      verbatimTextOutput("scdf_output")
    )
  )
)

# Transform -----
tab_transform <- tabPanel(
  "Transform",
  sidebarLayout(
    sidebarPanel(
      textInput(
        "select_cases", "Select cases",
        placeholder = "e.g.: 1, Anja, 3:5"
      ),
      div(style="display:inline-block; vertical-align: top",
        textInput("select_phasesA", "Combine phases to A", placeholder = "(e.g.: 1)")
      ),
      div(style="display:inline-block; vertical-align: top; padding-left: 30px;",
        textInput("select_phasesB", "Combine phases to B", placeholder = "(e.g.: 2,3)")
      ),
      
      textInput(
        "subset", "Filter measurments",
        placeholder = 'e.g.: mt > mean(values[phase == "A"])'
      ),
      textAreaInput(
        "transform", "Transform variables", rows = 5,
        placeholder = res$placeholder$transform
      ),
      textInput(
        "setdvar", "Set dependent variable", placeholder = "e.g.: depression"
      ),
      downloadButton("transformed_save", "Save transformed scdf"),
    ),
    mainPanel(
      verbatimTextOutput("transform_syntax"),
      conditionalPanel(
        'input.transform_output_format == "Text"', verbatimTextOutput("transform_scdf")
      ),
      conditionalPanel(
        'input.transform_output_format == "Html"', htmlOutput("transform_html")
      )
    )
  )
)


# Stats -----
tab_stats <- tabPanel(
  "Stats",
  sidebarLayout(
    sidebarPanel(
      selectInput(
        "func",
        "Statistic",
        choices = res$choices$fn_stats
      ),
      uiOutput("stats_arguments")
    ),
    mainPanel(
      div(
        style="display:inline-block; vertical-align: top",
        radioButtons(
          "stats_out", "Output format", c("Text", "Html"), "Text",
          inline = TRUE
        )
      ),
      div(
        style="display:inline-block; vertical-align: top; padding-left: 30px;",
        textInput(
          "stats_print_arguments", "Output arguments",
          placeholder = "e.g.: flip = TRUE; digits = 2; meta = FALSE"
        )
      ),
      div(
        style="display:inline-block; padding-left: 30px;",
        downloadButton("stats_save", "Save output")
      ),
      hr(),
      verbatimTextOutput("stats_syntax"),
      conditionalPanel(
        'input.stats_out == "Text"', verbatimTextOutput("stats_text")
      ),
      conditionalPanel(
        'input.stats_out == "Html"', tableOutput("stats_html")
      )
    )
  )
)


# Plot -----
tab_plot <- tabPanel(
  "Plot",
  sidebarLayout(
    sidebarPanel(
      textAreaInput(
        "plot_arguments", "Arguments", value = "",rows = 5,
        placeholder = res$placeholder$plot_arguments
      ),
      selectInput("scplot_examples", "Stats templates",
          choices = names(res$choices$scplot_examples)
      ),
      selectInput("scplot_templates_design", "Design templates",
                    choices = names(res$choices$scplot_templates_design)
      ),
      selectInput("scplot_templates_annotate", "Annotate templates",
                  choices = names(res$choices$scplot_templates_annotate)
      ),
      downloadButton("saveplot", "Save plot", inline = FALSE)
    ),
    mainPanel(
      verbatimTextOutput("plot_syntax"),
      plotOutput("plot_scdf", width = 1000,height = 800,)
    )
  )
)

# Power-test -----
tab_power_test <- tabPanel(
  "Power-test",
  fluidRow(
    column(2, div(style = res$div$pt,
      h3("Case design"),
      br(),
      textInput("design_n", "n", value = 1),
      textInput("design_phase", "Phase design", value = "A = 5, B = 15"),
      textInput("design_trend", "Trend", value = "0.02"),
      textInput("design_slope", "Slope", value = "0"),
      textInput("design_level", "Level", value = "1"),
      textInput("design_start", "Start value", value = 50),
      #textInput("design_s", "Standard deviation", value = 10),
      textInput("design_rtt", "Reliabiliy", value = 0.8),
      selectInput("design_distribution", "Distribution", 
                 choices = c("normal", "poisson", "binomial")
      ),
    )),
    column(2, div(style = res$div$pt,
      h3("Analysis"),
      br(),
      checkboxGroupInput(
        "pt_method", "Method(s)", 
        choices = res$choices$pt_method,selected = "plm_level"
      ),
      selectInput(
       "pt_effect", "Null effect for", choices = c("level", "slope")
      ),
      numericInput(
        "pt_n", "Number of simulations", min = 30, max = 10000, value = 100
      ),
      textInput("pt_ci", "Confidence intervall", placeholder = "e.g.: 0.95"),   
    )),
    column(4, div(style = res$div$pt,
      br(),
      verbatimTextOutput("pt_syntax"),
      actionButton("pt_compute", "Run"),
      hr(),
      verbatimTextOutput("pt_results", placeholder = TRUE)
    )),
    column(4, div(style = res$div$pt,
      actionButton("desigh_plot_button", "Create plot example"),
      plotOutput("plot_design", width = 600, height = 800)
    )),
  ),
 
)

# settings -----
tab_settings <- tabPanel(
  "Settings",
  fluidRow(
    column(2, div(
      style = res$div$settings, 
      h3("Data"),
      radioButtons(
        "scdf_output_format", "Show output as", 
        choices = c("Summary", "Syntax"), 
        inline = TRUE
      ),
      radioButtons(
        "scdf_syntax_phase_structure", "Syntax phase structure", 
        choices = c("phase_design" = FALSE, "inline" = TRUE), 
        inline = TRUE
      ),
      textInput("scdf_save_prefix", "Prefix save filename", value = "scdf"),
      radioButtons(
       "scdf_save_format", "Save format", 
       choices = c("R object" = ".rds", "R syntax" = ".R", "csv" = ".csv"), 
       inline = TRUE),
      textInput("scdf_load_na", "Missing values in import file", value = '"", "NA"'),

    )),
    column(2, div(
      style = res$div$settings, 
      h3("Transform"),
      radioButtons(
        "transform_output_format", "Show output as", c("Text", "Html"), inline = TRUE
      ),
      textInput(
        "transform_save_prefix", 
        "Prefix save filename", 
        value = "scdf-transformed"
      ),
      radioButtons(
        "transform_save_format", "Save format", 
        choices = c("R object" = ".rds", "R syntax" = ".R", "csv" = ".csv"), 
        inline = TRUE)
    )),
    column(2, div(
      style = res$div$settings, 
      h3("Stats"),
      radioButtons(
        "stats_default", "Show defaults", choices = c("No", "Yes"),
        inline = TRUE
      ),
      textInput("prefix_output_stats", "Prefix save filename", value = "scan-stat")
    )),
    column(2, div(
      style = res$div$settings, 
      h3("Plot"),
      #numericInput("plot_display_res", "Display resolution", value = 120, min = 10, max = 4000),
      textInput("prefix_output_plot", "Prefix save filename", value = "scplot"),
      numericInput("width", "Export width", value = 800, min = 100, max = 2000),
      numericInput("height", "Export height", value = 600, min = 100, max = 2000),
      numericInput("dpi", " Export dpi", value = 100, min = 50, max = 600)
    ))
  )
)

# Help -----

tab_help <- tabPanel(
  "Help",
  res$help_page
)


# About -----

tab_about <- tabPanel(
  "About",
  h4("Running:"),
  h4(paste0(
    "scan ",
    utils::packageVersion("scan")," (",utils::packageDate('scan'), ")"
  )),
  h4(paste0(
    "scplot ",
    utils::packageVersion("scplot")," (",utils::packageDate('scplot'), ")"
  )),
  hr(),
  h4("Please cite as:"),
  h4({x<-citation("scan"); class(x)<-"list"; attributes(x[[1]])$textVersion}),
  hr(),
  h4("(c) Jürgen Wilbert, 2023")
)

# ui ------

ui <- navbarPage(
  id = "navpage",
  title = "Shiny-Scan",
  theme = "cerulean.min.css",
  tab_scdf,
  tab_transform,
  tab_stats,
  tab_plot,
  tab_power_test,
  tab_settings,
  tab_help,
  tab_about,
  tabPanel(title = "Quit")
)

source("resources.R")

left_width <- 400

# ---------- Data: New ----------

tab_scdf <- nav_panel(
  "New / Edit",
  layout_sidebar(
    sidebar = sidebar(
      title = NULL,
      open  = "always", width = left_width,
      selectInput("new_select_case", "Select case",
                  choices = res$new_case, selected = res$new_case),
      textAreaInput("new_values", "Values (dependent variable)",
                    placeholder = res$placeholder$values, rows = 3),
      textInput("new_mt", "Measurement times", placeholder = res$placeholder$mt),
      textAreaInput("new_variables", "Additional variables",
                    placeholder = res$placeholder$variables, rows = 3),
      textInput("new_casename", "Case name", 
                placeholder = res$placeholder$casename),
      
      div(class = "d-grid gap-2",
          actionButton("new_save_case",   "Save case",    class = "btn-primary"),
          actionButton("new_remove_case", "Remove case",  class = "btn-light"),
          actionButton("new_clear_cases", "Clear all cases",
                       class = "btn-warning"),
          actionButton("new_clear_fields", "Clear input fields",
                       class = "btn-dark")),
      
      tags$hr(),
      downloadButton("scdf_save", "Save scdf", class = "btn-success")
    ),
    
    # main content
    mainPanel(
      verbatimTextOutput("scdf_messages"),
      verbatimTextOutput("scdf_output")
    )
  )
)

# ---------- Data: Load ----------
card_load_example <- card(
  card_body(
    selectInput("scdf_example", "Choose example", 
                choices = res$choices$examples, size = 15,
                selectize = FALSE),
    tags$hr(),
    actionButton("scdf_import", "Import example", class = "btn-primary")
  )
)
  
card_load_import <- card(
  card_body(
    fileInput("upload", NULL,
              accept = c(".csv", ".rds", ".xlsx", ".xls", ".R", ".r", ".txt"),
              buttonLabel = "Choose file"),
    selectInput("scdf_load_cvar", "Case variable", choices = "case"),
    selectInput("scdf_load_pvar", "Phase variable", choices = "phase"),
    selectInput("scdf_load_dvar", "Dependent variable", choices = "values"),
    selectInput("scdf_load_mvar", "Measurement time variable", choices = "mt"),
    textInput("scdf_load_na", "Missing values", value = '"", "NA"'),
    selectInput("scdf_csv", "Separators (only for .csv and .txt files)", 
                choices = res$choices$separators),
    tags$hr(),
    actionButton("scdf_import",   "Import scdf",    class = "btn-primary")
  )
)

card_load_output <- card(
  card_body(
    verbatimTextOutput("load_messages"),
    verbatimTextOutput("load_output"),
    htmlOutput("load_output_html")
  )
)
  
nav_panel_load <- nav_panel("Load",
  layout_columns(
    col_widths = c(3, 9),
    navset_tab(
      nav_panel("Import", card_load_import),
      nav_panel("Examples", card_load_example)
    ),
    card_load_output
))

# ---------- Transform ----------
tab_transform <- layout_sidebar(
  sidebar = sidebar(
    title = NULL, # title = "Transform",
    open  = "always", width = left_width,
    textInput("select_cases", "Select cases", placeholder = "e.g.: 1, Anja, 3:5"),
    div(class = "d-flex gap-3",
        textInput("select_phasesA", "Combine phases to A", placeholder = "(e.g.: 1)"),
        textInput("select_phasesB", "Combine phases to B", placeholder = "(e.g.: 2,3)")
    ),
    textInput("subset", "Filter measurements", 
              placeholder = 'e.g.: mt > mean(values[phase == "A"])'),
    textAreaInput("transform", "Transform variables", rows = 5, 
                  placeholder = res$placeholder$transform),
    selectInput("setdvar", "Set dependent variable", choices = ""),
    downloadButton("transformed_save", "Save transformed scdf", class = "btn-success")
  ),
  mainPanel(
    input_switch("transform_output_format", "HTML", value = FALSE),
    verbatimTextOutput("transform_syntax"),
    conditionalPanel(
      '!input.transform_output_format', verbatimTextOutput("transform_scdf")
    ),
    conditionalPanel( 'input.transform_output_format', htmlOutput("transform_html"))
  )
)

# ---------- Stats ----------

card_stats_main <- card(
  card_body(
    h5("Select case"),
    selectInput("stats_select_case", NULL, choices = "all"),
    h5("Select analysis"),
    selectInput("stats_func", NULL, choices = res$choices$fn_stats,
                selectize = FALSE, size = length(res$choices$fn_stats))
  )
)

card_stats_args <- card(
  card_body(
    uiOutput("stats_arguments")
  )
)

card_stats_output <- card(
  card_body(
    tags$head(tags$style(HTML("
      .toolbar { display:flex; gap:16px; align-items:center; }
      .toolbar .stretch { flex:1; }"
    ))),
    div(class = "toolbar",
        div(class = "stretch", textAreaInput(
          "stats_print_arguments", label = "Output arguments",
          rows = 1, width = "100%", placeholder = res$placeholder$stats_out_args
        )),
        downloadButton("stats_save", label = "Save", class = "btn-success")
    ),
    div(class = "toolbar",
        input_switch("stats_out", "HTML", TRUE),
        input_switch("stats_description","Method description", FALSE)
    ), 
    conditionalPanel('input.stats_description', htmlOutput("stats_description")),
    verbatimTextOutput("stats_syntax"),
    conditionalPanel('!input.stats_out', verbatimTextOutput("stats_text")),
    conditionalPanel('input.stats_out', tableOutput("stats_html"))
  )
)

tab_stats <- layout_columns(
  col_widths = c(3, 9),
  navset_tab(
    id = "stats_tabs",
    nav_panel("Main", card_stats_main),
    nav_panel("Arguments", card_stats_args)
  ),
  card_stats_output
)


# ---------- Plot -------

card_plot_theme <- card(
  card_body(
    h5("Select case"),
    selectInput("scplot_select_case", NULL, choices = "all"),
    
    h5("Themes"),
    selectInput("scplot_theme_1", "Complete", 
                choices = res$scplot_themes_complete, selected = "default"),
    selectInput("scplot_theme_2", "Add element", 
                choices = res$scplot_themes_element),
    selectInput("scplot_theme_3", "Add element", 
                choices = res$scplot_themes_element),
    h5("Legend position"),
    selectInput("scplot_legend_position", NULL,
                choices = res$choices$legend_position, 
                selected = "none"),
    h5("Textsize"),
    sliderInput("scplot_text_size", NULL, min = 6, max = 25, value = 6)
  )
)

card_plot_axis <- card(
  card_body(
    h5("Axis"),
    numericInput("scplot_xinc", "x increment", value = NA),
    layout_columns(
      col_widths = c(6, 6),
      numericInput("scplot_ymin", "y min", value = NA),
      numericInput("scplot_ymax", "y max", value = NA)
    ),
    textInput("scplot_xlabel", "x label"),
    textInput("scplot_ylabel", "y label"),
    textInput("scplot_phasenames", "Phasenames"),
    textInput("scplot_casenames", "Casenames"),
    input_switch("scplot_footer", "Footnote", FALSE),
    
  )
)

card_plot_lines <- card(
  card_body(
    h5("Baseline"),
    input_switch("scplot_stats_trend_a", "Trend", FALSE),
    input_switch("scplot_stats_mean_a", "Mean", FALSE),
    input_switch("scplot_stats_median_a", "Median", FALSE),
    input_switch("scplot_stats_max_a", "Max", FALSE),
    input_switch("scplot_stats_min_a", "Min", FALSE),
    
    h5("Phases"),
    input_switch("scplot_stats_mean", "Mean", FALSE),
    input_switch("scplot_stats_median", "Median", FALSE),
    input_switch("scplot_stats_trend", "Slope", FALSE),
    
    h5("Curves"),
    input_switch("scplot_stats_moving", "Moving mean", FALSE),
    input_switch("scplot_stats_loess", "Smoothed line", FALSE),
    h5("Add variable"),
    selectInput("scplot_add", NULL, choices = "None")
  )
)

card_plot_save <- card(
  card_body(
    layout_columns(
      col_widths = c(4, 4, 4),
      numericInput("width",  "Width",  value = 1600, min = 100, max = 4000),
      numericInput("height", "Height", value = 1200, min = 100, max = 4000),
      numericInput("dpi",    "Dpi",    value = 200,  min = 50,  max = 1200),
    ),
    downloadButton("saveplot_2", "Save plot", class = "btn-success")
  )
)

card_plot_output <- card(
  card_body(
    #verbatimTextOutput("scplot_syntax"),
    plotOutput("scplot_plot")
  ), height = "100vh"
)


tab_plot <- layout_columns(
  col_widths = c(3, 9),
  navset_tab(
    nav_panel("Theme", card_plot_theme),
    nav_panel("Axis", card_plot_axis),
    nav_panel("Lines", card_plot_lines),
    nav_panel("Save", card_plot_save)
  ),
  card_plot_output
)


# ---------- Power-test ----------

card_design <- card(
  card_body(
    textInput("design_n", "n", value = 1),
    textInput("design_phase", "Phase design", value = "A = 5, B = 15"),
    textInput("design_trend", "Trend", value = "0.02"),
    textInput("design_slope", "Slope", value = "0"),
    textInput("design_level", "Level", value = "1"),
    textInput("design_start", "Start value", value = 50),
    textInput("design_rtt", "Reliabiliy", value = 0.8),
    selectInput("design_distribution", "Distribution", 
                choices = c("normal", "poisson", "binomial"))
  )
)

card_analysis <- card(
  card_body(
    checkboxGroupInput("pt_method", "Method(s)", 
                       choices = res$choices$pt_method, selected = "plm_level"),
    textInput("pt_method_user", "User method", value = ""),
    selectInput("pt_effect", "Null effect for", choices = c("level", "slope")),
    numericInput("pt_n", "Number of simulations", min = 30, max = 10000, value = 100),
    textInput("pt_ci", "Confidence interval", placeholder = "e.g.: 0.95")
  )
)

card_run <- card(
  card_body(
    verbatimTextOutput("pt_syntax"),
    input_task_button("pt_compute", "Run"),
    hr(),
    verbatimTextOutput("pt_results", placeholder = TRUE)
  ), height = "50vh"
)

card_plot <- card(
  card_body(
    actionButton("desigh_plot_button", "Create plot example"),
    plotOutput("plot_design")#, width = 800, height = 800)
  ), height = "50vh"
)


 
tab_power_test <- layout_columns(
  col_widths = c(2, 3, 7),
  heights_equal = "row",
  card_design,
  card_analysis,
  div(class = "d-flex flex-column gap-3 h-100",
    card_run,
    card_plot
  )
)

#tab_power_test <- navset_tab(nav_panel("Power test", tab_power_test))

# ---------- Settings ----------
tab_settings <- layout_columns(
  col_widths = c(2,2,2,2),
  card(
    card_header("Data"),
    card_body(
      input_switch("scdf_output_format", "Show scdf syntax", value = FALSE),
      input_switch("scdf_syntax_phase_structure",
                   tags$span("Phase structure inline", class = "chklabel-big"),
                   value = FALSE),
      textInput("scdf_save_prefix", "Prefix save filename", value = "scdf"),
      radioButtons("scdf_save_format", "Save format",
                   choices = c("R object" = ".rds", "R syntax" = ".R", "csv" = ".csv"),
                   inline = TRUE)
    )
  ),
  card(
    card_header("Transform"),
    card_body(
      textInput("transform_save_prefix", "Prefix save filename",
                value = "scdf-transformed"),
      radioButtons("transform_save_format", "Save format",
                   choices = c("R object" = ".rds", "R syntax" = ".R", "csv" = ".csv"),
                   inline = TRUE)
    )
  ),
  card(
    card_header("Stats"),
    card_body(
      input_switch("stats_default",
                   tags$span("Show defaults", class = "chklabel-big"),
                   value = FALSE),
      radioButtons("rename_predictors", "Rename predictors",
                   choices = c("full", "concise", "no"), inline = TRUE),
      input_switch("setting_output_docx", "Save html as docx", value = FALSE),
      textInput("prefix_output_stats", "Prefix save filename", value = "scan-stat")
    )
  ),
  card(
    card_header("Plot"),
    card_body(
      textInput("prefix_output_plot", "Prefix save filename", value = "scplot"),
      #numericInput("width",  "Export width",  value = 1600, min = 100, max = 4000),
      #numericInput("height", "Export height", value = 1200, min = 100, max = 4000),
      #numericInput("dpi",    "Export dpi",    value = 200,  min = 50,  max = 1200)
    )
  ),
  card(
    card_header("General"),
    card_body(
      radioButtons("scan_export_engine", "Html engine",
                   choices = c("gt", "kable"), inline = TRUE)
    )
  )
)

# ---------- Help ----------
navbar_help <- nav_menu(
  title = "Help",
  nav_panel("Info", res$help_page),
  nav_panel(
    "About",
    h4("Running:"),
    h4(paste0("scan ",   utils::packageVersion("scan"),   " (", 
              utils::packageDate('scan'),   ")")),
    h4(paste0("scplot ", utils::packageVersion("scplot"), " (", 
              utils::packageDate('scplot'), ")")),
    hr(),
    h4("Please cite as:"),
    h4({
      x <- citation("scan")
      class(x) <- "list"
      attributes(x[[1]])$textVersion
    }),
    hr(),
    h4("(c) Jürgen Wilbert, 2026")
  ),
  nav_panel(title = "Quit")
)

# ---------- UI ----------
ui <- tagList(
  tags$head(tags$link(rel = "icon", type = "image/png", href = "logo.png")),
  page_navbar(
    title = "scan",
    id = "scan",
    theme = res$theme_light,
    navbar_options = navbar_options(class = "bg-primary", theme = "dark"),
    #nav_menu("Data", nav_panel_load, tab_scdf),
    nav_panel("Data", nav_panel_load),
    nav_panel("New/Edit", tab_scdf),
    nav_panel("Transform",  tab_transform),
    nav_panel("Stats",      tab_stats),
    nav_panel("Plot",       tab_plot),
    nav_panel("Power",      tab_power_test),
    nav_panel("Settings",   tab_settings),
    navbar_help,
    nav_spacer(),
    
    nav_item(input_switch("darkmode", "Dark mode", value = FALSE))
  )
)

server <- function(input, output, session) {

  # scdf ----
  
  ## startup message ----
  
  output$scdf_messages <- renderPrint(cat(res$msg$startup))

  ## Render ----
  my_scdf <- reactiveVal()
  scdf_render <- reactive({
    
    if (!inherits(my_scdf(), "scdf")) validate(res$msg$no_case)
    
    if (input$scdf_output_format == "Summary"){
      output$scdf_output <- renderPrint({
        do.call("summary", list(my_scdf()))
      })
    } else if (input$scdf_output_format == "Syntax") {
      output$scdf_output <- renderPrint({
        req(inherits(my_scdf(), "scdf"))
        do.call("convert", list(
          my_scdf(), inline = as.logical(input$scdf_syntax_phase_structure)
        ))
      })
    }

  })

  observeEvent(input$scdf_output_format, scdf_render()) 
  
  ## input example ----
  observeEvent(input$scdf_example, {
    if (input$scdf_example != "(none)") {
      my_scdf(paste0("scan::", input$scdf_example) |> str2lang() |> eval())
      scdf_render()
      output$scdf_messages <- renderPrint(
        cat(paste0("loaded example ", input$scdf_example))
      )
    } else {
      my_scdf(NULL)
    }
  })

  ## upload (load) ------
  observeEvent(input$upload, {
    ext <- tools::file_ext(input$upload$datapath)
    if (ext == "rds") {
      new <- readRDS(input$upload$datapath)
    } else if (ext %in% c("r", "R")) {
      new <- readLines(input$upload$datapath)
      new <- paste0(new, collapse = "\n")
      .tmp <- new.env()
      eval(parse(text = new), envir = .tmp)
      new <- .tmp$study
    } else {
      na <- eval(str2lang(paste0("c(", input$scdf_load_na, ")")))
      new <- read_scdf(input$upload$datapath, na = na)
    }

    if (!inherits(new, "scdf")) {
      output$scdf_messages <- renderText(
        "Sorry,\n the file you tried to load is not a valid scdf file.")
    } else {
      my_scdf(new)
      scdf_render()
      output$scdf_messages <- renderPrint(cat(paste0("loaded file successfully")))
    }

  })

  ## download (save) ----
  output$scdf_save <- downloadHandler(
    filename = function() {
      scdf <- my_scdf()
      out <- paste(
        input$scdf_save_prefix,
        sprintf("%02d", length(scdf)),
        paste0(unique(scdf[[1]]$phase), collapse = ""),
        format(Sys.time(), format = "%y%m%d-%H%M%S"),
        sep = "-"
      )
      paste0(out, input$scdf_save_format)
    },
    
    content = function(file) {
      scdf <- my_scdf()
      
      if (!inherits(scdf, "scdf")) {
        output$scdf_messages <- renderPrint(cat(res$error_msg$scdf_save))
      } else {
        output$scdf_messages <- renderPrint(cat("Saved file"))
      }
      
      if (input$scdf_save_format == ".rds") 
        saveRDS(scdf, file)
      if (input$scdf_save_format == ".R") 
        convert(scdf, file = file)
      if (input$scdf_save_format == ".csv") 
        write_scdf(scdf, filename = file)

    }
  )

  ## new cases --------
  observeEvent(input$add_case, {
    tryCatch({
      values <- paste0("c(", trim(input$values), ")")
      dvar <- "values"
      if (inherits(my_scdf(), "scdf")) {
        dvar <- scdf_attr(my_scdf(), "var.values")
      }

      call <- paste0(dvar, " = ", values)
      call <- c(call, paste0("dvar = ", deparse(dvar)))
      if (input$mt != "") call <- c(call, paste0("mt = c(", input$mt, ")"))
      
      if (input$variables != "") {
        variables <- input$variables |>
          strsplit("\n")  |>
          unlist() |>
          lapply(function(y) strsplit(y, "=")) |>
          unlist(recursive = FALSE) |>
          lapply(function(y) paste0(y[1], " =c(", y[2], ")")) |>
          unlist()
        call <- c(call, variables)
      }

      if (input$casename != "") {
        call <- c(call, paste0("name = ", deparse(input$casename)))
      } else {
        call <- c(call, paste0("name = \"case\""))
      }
      call <- paste0(call, collapse = ",")
      call <- paste0("scdf(", call, ")")

      new <- call |> str2lang() |> eval()
      if (input$remove_which == "last") {
        if (length(my_scdf()) > 0) new <- c(my_scdf(), new)
        my_scdf(new)
        output$scdf_messages <- renderPrint(cat("Appended case"))
        scdf_render()
      } 
      
      if (input$remove_which == "at") {
        at <- input$remove_at
        if (length(my_scdf()) >= at - 1) {
          if (at == 1) {
            new <- c(new, my_scdf())
          } else if (at == length(my_scdf()) + 1) {
            new <- c(my_scdf(), new)
          } else {
            new <- c(my_scdf()[1:(at-1)], new, my_scdf()[at:(length(my_scdf()))])
          }
          my_scdf(new)
          output$scdf_messages <- renderPrint(cat("Added case at position", input$remove_at))
          scdf_render() 
        }
      }  

    },
    error = function(e)
      output$scdf_messages <- renderText(
        paste0(res$error_msg$invalid_case, "\n\n", e)
      )
    )
  })

  ## remove cases --------
  observeEvent(input$remove_case, {
    if (input$remove_which == "last") {
      if (length(my_scdf()) > 1) {
        my_scdf(my_scdf()[-length(my_scdf())])
      } else (my_scdf(NULL))
    }
    
    if (input$remove_which == "at") {
      at <- input$remove_at
      if (length(my_scdf()) >= at)
        my_scdf(my_scdf()[-input$remove_at])
    }
    
    output$scdf_messages <- renderPrint(cat("removed case"))
    scdf_render()
  })

  ## remove all cases --------
  observeEvent(input$remove_all, {
    my_scdf(NULL)
    output$scdf_messages <- renderPrint(cat("Cleared cases"))
    scdf_render()
    
  })

  # Transform ----

  ## render ----
  transformed <- reactive({
    out <- my_scdf()
    syntax = "scdf"
    if (input$select_cases != "") {
      call <- str2lang(paste0("select_cases(out, ", input$select_cases,")"))
      out <- eval(call)
      syntax <- c(syntax, paste0("select_cases(",input$select_cases,")"))
    }

    if (input$select_phasesA != "" || input$select_phasesB != "") {
 
      out <- paste0(
          "select_phases(out, A = c(", input$select_phasesA, "), B = c(",
          input$select_phasesB, "))"
        ) |>
        str2lang() |> 
        eval()
      
      syntax <- c(
        syntax, 
        paste0(
          "select_phases(A = c(", input$select_phasesA, "), B = c(",
          input$select_phasesB, "))"
        )
      )
    }

    if (input$subset != "") {
      args <-  list(str2lang(input$subset))
      out <- do.call("subset", c(list(out),args))
      syntax <- c(syntax, paste0("subset(",  input$subset, ")"))
    }

    if (input$transform != "") {
      arg <- paste0("transform(out,", trim(input$transform),")")
      out <- str2lang(arg) |> eval()
      syntax <- c(
        syntax, paste0("transform(", gsub("\n", ", ", trim(input$transform)),")")
      )
    }

    if (input$setdvar != "") {
      args <-  list(input$setdvar)
      out <- do.call("set_dvar", c(list(out),args))
      syntax <- c(syntax, paste0("set_dvar(",  deparse(input$setdvar), ")"))
    }

    if (length(syntax)>1) {
      syntax <- syntax[-1]
      syntax <- paste0(
        "scdf %>%\n  ",
        paste0(syntax, collapse = " %>%\n  ")
      )

    }

    output$transform_syntax <- renderPrint(cat(syntax))

    out
  })

  ## save ----
  output$transformed_save <- downloadHandler(
    filename = function() {
      scdf <- transformed()
      out <- paste(
        input$transform_save_prefix,
        sprintf("%02d", length(scdf)),
        paste0(unique(scdf[[1]]$phase), collapse = ""),
        format(Sys.time(), format = "%y%m%d-%H%M%S"),
        sep = "-"
      )
      paste0(out, input$transform_save_format)
    },
    content = function(file) {
      if (input$transform_save_format == ".rds") 
        saveRDS(transformed(), file)
      if (input$transform_save_format == ".R") 
        convert(transformed(), file = file)
      if (input$transform_save_format == ".csv") 
        write_scdf(transformed(), filename = file)
    }
  )
  
  ## output ------
  
  output$transform_scdf <- renderPrint({
    if(!inherits(my_scdf(), "scdf")) validate(res$msg$no_case)
    print(transformed(), rows = 100)
  })

  output$transform_html <- renderUI({
    if(!inherits(my_scdf(), "scdf")) validate(res$msg$no_case)
    export(transformed(), caption = "") |> HTML()
  })
  
  # Stats -----

  ## Calculate ---- 
  calculate_stats <- reactive({
    if (!inherits(my_scdf(), "scdf")) validate(res$msg$no_case)
    scdf <- transformed()
    call <- get_stats_call()
    tryCatch(
      str2lang(call) |> eval(),
      error = function(e)
        validate(paste0("Sorry, could not proceed calculation:\n\n", e))
    )
  })

  ## Output ----
  output$stats_html <- renderUI({
    results <- calculate_stats()
    print_args <- input$stats_print_arguments
    if (print_args != "") {
      print_args <- paste0(", ", print_args)
      call<- paste0("export(results, ", print_args, ")")
    } else call <- "export(results)"
    tryCatch(
      str2lang(call) |> eval() |> HTML(),
      error = function(e)
        validate("Sorry, no html export for this function available yet.")
    )
  })

  output$stats_text <- renderPrint({
    results <- calculate_stats()

    print_args <- input$stats_print_arguments
    if (print_args != "") {
      print_args <- paste0(", ", print_args)
      call<- paste0("print(results, ", print_args, ")")
    } else call <- "print(results)"

    str2lang(call) |> eval()
  })

  output$stats_syntax <- renderPrint({
    cat(get_stats_call())
  })


  ## Arguments ------
  stat_arg_names <- reactive({
    args <- names(formals(input$func))
    values <- formals(input$func)

    id <- which(!args %in% c(
      "dvar", "pvar", "mvar", "phases", "meta_method",
      "data", "scdf", "data.l2", "offset", "lag.max",
      "graph", "output", "...")
    )
    args <- args[id]
    values <- values[id]
    list(names = args, values = values)
  })

  output$stats_arguments <- renderUI({
    args <- stat_arg_names()

    out <- vector("list", length(args$names))
    if (length(out) > 0) {
      for (i in 1:length(out)) {

        value <- args$values[[i]]
        if (is.character(value)) value <- deparse(value)
        if (isTRUE(is.na(value))) value <- substitute(value) |> deparse()
        if (is.null(value)) value <- substitute(value) |> deparse()
        if (!is.numeric(value) && !is.logical(value) && !is.character(value) &&
            !is.call(value)) {
          value <- substitute(value) |> deparse()
        }
        if (is.call(value)) {
          if (is.character(eval(value))) {
            value <- eval(value)
          } else {
            value <- substitute(value) |> deparse()
          }
        }
        if (input$stats_default == "Yes") outvalue <- value else outvalue = NULL

        if (length(value) > 1) {
          choices <- setNames(quoted(value), value)
          if (input$stats_default == "No")
            choices <- c("(default)" = "", choices)
          selected <- names(choices)[1]
          out[[i]] <- selectInput(
            args$names[i], args$names[i],
            choices = choices,
            selected = selected
          )
        } else if (is.numeric(value)) {
          out[[i]] <- numericInput(
            args$names[i], args$names[i], value = outvalue
          )
        } else if (is.logical(value)) {
          choices <- c("FALSE", "TRUE")
          if (input$stats_default == "No")
            choices <- c("(default)" = "", choices)
          out[[i]] <- radioButtons(
            args$names[i], args$names[i],
            choices = choices,
            inline = TRUE, selected = outvalue
          )
        } else {
          out[[i]] <- textInput(args$names[i], args$names[i], value = outvalue)
        }
      }
      return(out)
    }
  })

  get_stats_call <- reactive({
    args <- stat_arg_names()
    values <- sapply(args$names, function(name) input[[name]])
    args <- args$names

    id <- which(values != "")

    args <- args[id]
    values <- values[id]

    call <- paste0(
      input$func, "(scdf",
      if (length(args > 0)) {
        paste0(", ",paste0(args, " = ", values, collapse = ", "))
      } else {
        ""
      },
      ")"
    )
    call
  })

  ## Save ------
  
  output$stats_save <- downloadHandler(
    
    filename = function() {
      scdf <- transformed()
      out <- paste(
        input$prefix_output_stats, input$func,
        sprintf("%02d", length(scdf)),
        paste0(unique(scdf[[1]]$phase), collapse = ""),
        format(Sys.time(), format = "%y%m%d-%H%M%S"),
        sep = "-"
      )
      
      if (input$stats_out == "Html") out <- paste0(out, ".html")
      if (input$stats_out == "Text") out <- paste0(out, ".txt")
      out
    },
    content = function(file) {
      
      if (input$stats_out == "Text") {
        results <- calculate_stats()
        print_args <- input$stats_print_arguments
        if (print_args != "") {
          print_args <- paste0(", ", print_args)
          call<- paste0("print(results, ", print_args, ")")
        } else call <- "print(results)"
        call <- paste0("capture.output(", call, ")")
        writeLines(str2lang(call) |> eval(), con = file)
      }
      
      if (input$stats_out == "Html") {
        results <- calculate_stats()
        print_args <- input$stats_print_arguments
        if (print_args != "") {
          print_args <- paste0(", ", print_args)
          call<- paste0("export(results, ", print_args, ")")
        } else {
          call <- "export(results)"
        }
        out <- str2lang(call) |> eval()
        kableExtra::save_kable(out, file)
        #writeLines(out, con = file)
      }
      
    }
  )
  
  # Plot -----

  ## Render ----
  render_plot <- reactive({
    req(inherits(my_scdf(), "scdf"))
    call <- paste0("scplot(transformed())")
    if (trimws(input$plot_arguments) != "") {
      plot_args <- trimws(input$plot_arguments)
      plot_args <- gsub("\n+", "\n", plot_args)
      call <- paste0(
        call, "%>% ", gsub("\n", " %>% ", plot_args)
      )
    }
    call <- paste0("print(",call,")")
    tryCatch(
      str2lang(call) |> eval(),
      error = function(x)
        output$plot_syntax <- renderText(paste0(res$error_msg$plot, "\n\n", x))
    )
  })

  observeEvent(input$scplot_templates_design, {
    new_value <- unname(
      res$choices$scplot_templates_design[input$scplot_templates_design]
    )
    old_value <- input$plot_arguments
    if (old_value == "") {
      value <- new_value
    } else {
      value <- paste0(input$plot_arguments, "\n", new_value)
    }
  updateTextAreaInput(inputId = "plot_arguments", value = value)
  })

  observeEvent(input$scplot_templates_annotate, {
    new_value <- unname(
      res$choices$scplot_templates_annotate[input$scplot_templates_annotate]
    )
    old_value <- input$plot_arguments
    if (old_value == "") {
      value <- new_value
    } else {
      value <- paste0(input$plot_arguments, "\n", new_value)
    }
    updateTextAreaInput(inputId = "plot_arguments", value = value)
  })
  
  observeEvent(input$scplot_examples, {
    if ("(empty selection)" == input$scplot_examples) {
      value <- ""
    } else {
      new_value <- unname(res$choices$scplot_examples[input$scplot_examples])
      old_value <- input$plot_arguments
      if (old_value == "") {
        value <- new_value
      } else {
        value <- paste0(input$plot_arguments, "\n", new_value)
      }
    }
    updateTextAreaInput(inputId = "plot_arguments", value = value)
  })

  observeEvent(input$plot_arguments, render_plot_syntax())

  ## Output ----
 
  render_plot_syntax <- reactive({
    call <- paste0("scplot(scdf)")
    if (trimws(input$plot_arguments) != "") {
      call <- paste0(
        call, "%>%\n  ", gsub("\n", " %>%\n  ", trimws(input$plot_arguments))
      )
    }
    output$plot_syntax <- renderPrint({
      cat(call)
    })
  })
  
  output$plot_scdf <- renderPlot(res = 120,{
    render_plot()
  })

  ## Save ----
  output$saveplot <- downloadHandler(
    filename = function() {
      scdf <- transformed()
      out <- paste(
        input$prefix_output_plot,
        sprintf("%02d", length(scdf)),
        paste0(unique(scdf[[1]]$phase), collapse = ""),
        format(Sys.time(), format = "%y%m%d-%H%M%S"),
        sep = "-"
      )
      paste0(out, ".png")
    },
    content = function(file) {
      ggplot2::ggsave(
        file, render_plot(), width = input$width, height = input$height,
        dpi = input$dpi, units = "px",  device = "png"
      )
    }
  )

  # Power test -----
  
  output$pt_results <- renderPrint(cat(res$placeholder$pt))
  
  ## Render -----
  render_power_test <- reactive({
    syntax <- paste0(
      "design(\n  n = ", input$design_n, ", ",
      "phase_design = list(", input$design_phase, "), \n  ",
      "trend = ", input$design_trend, ", ",
      "level = list(", input$design_level, "), ",
      "slope = list(", input$design_slope, "), \n  ",
      "start_value = ", input$design_start, ", ",
      #"s = ", input$design_s, ", ",
      "rtt = ", input$design_rtt, ", ",
      "distribution = ", deparse(input$design_distribution),
      "\n)"
    )  
    
    ci <- input$pt_ci

    syntax <- paste0(
      syntax, " %>% \n",
      "  power_test(method = ", deparse(input$pt_method), ", ",
      "effect = ", deparse(input$pt_effect), ", ",
      "n_sim = ", input$pt_n, ", ",
      "ci = ", ci,
      ")"
    )
    
    syntax
  })
  
  ## Output ----
  output$pt_syntax <- renderPrint({
    cat(render_power_test())
  })
  
  ## Plot ----
  
  observeEvent(input$desigh_plot_button, {
    call <- paste0(
      "design(\n  n = ", input$design_n, ", ",
      "phase_design = list(", input$design_phase, "), \n  ",
      "trend = ", input$design_trend, ", ",
      "level = list(", input$design_level, "), ",
      "slope = list(", input$design_slope, "), \n  ",
      "start_value = ", input$design_start, ", ",
      #"s = ", input$design_s, ", ",
      "rtt = ", input$design_rtt, ", ",
      "distribution = ", deparse(input$design_distribution),
      "\n) %>% ",
      "random_scdf() %>% ",
      "scplot()"
    )
    
    output$plot_design <- renderPlot(res = 100, {
      str2lang(call) |> eval()
    })
    #tryCatch(
    #  str2lang(call) |> eval(),
    #  error = function(x)
    #    output <- renderText(paste0(res$error_msg$plot, "\n\n", x))
    #)

  })
  
  ## Analyse ----
  
  observeEvent(input$pt_compute, {
    
    phase_structure <- eval(str2lang(
      paste0("list(", input$design_phase, ")")
    ))
    if (length(phase_structure) > 2) {
      output$pt_results <- renderPrint({
        cat("Sorry, power-tests are only possible for designs with two phases.")
      })
    } else {
      call <- render_power_test()
      output$pt_results <- renderPrint({cat("Calculating ...")})
      res <- tryCatch(
        str2lang(call) |> eval(),
        error = function(e)
          paste0("Sorry, could not proceed calculation:\n\n", e)
      )
      output$pt_results <- renderPrint({
        if (inherits(res, "character")) cat(res) else res
      })
    }
    
  })
  
  # quit app -----
  observeEvent(input$navpage, {
    if (input$navpage == "Quit") {
      stopApp()
    }
  })
}

#errors <- c()
#
#if (length(errors) > 0) stop(errors)

suppressPackageStartupMessages({
library(scan)
library(scplot)
library(shiny)
library(bslib)
})

options(
  scan.export.kable_styling = list(
    bootstrap_options = c("striped", "condensed"), 
    full_width = FALSE,
    position = "left"
  ),
  scan.export.engine = "gt",
  scan.export.title.prefix = NULL
)

res <- list()

res$pipe <- " |> "
res$pipe_br <- paste0(res$pipe, "\n")

res$new_case <- "-new case-"

# choices ------

res$choices <- list()

## examples ----

examples <- data(package = "scan")$results[, 3]
filter <- startsWith(examples,"Grosche2014") | 
          endsWith(examples,"Leidig2018_l2") | 
          startsWith(examples,"exampleAB_50.l2")
examples <- examples[!filter] |> sort()

res$choices$examples <- c("(none)", examples)

## themes ----

themes <- lapply(scplot:::.scplot_themes, function(x) x$theme_type) |> unlist()

if (!is.null(themes)) {
  complete <- names(themes)[which(themes == "complete")]
  element <- names(themes)[which(!themes == "complete")]
} else {
  complete <- names(scplot:::.scplot_themes)
  element <- names(scplot:::.scplot_themes)
}


res$scplot_themes_complete <- complete
res$scplot_themes_element <- c("None", element)

## stats ----
res$choices$fn_stats <- c(
  "Descriptives" = "describe",
  "Standardized mean differences" = "smd",
  "Between-Case Standardized Mean Difference" = "between_smd",
  "Overlap indices" = "overlap",
  "Data trends" = "trend",
  "Auto correlations" = "autocorr",
  "Piecewise regression" = "plm",
  "Multivariate piecewise regression" = "mplm",
  "Hierarchical piecewise regression" = "hplm",
  "Bayesian piecewise regression" = "bplm",
  "Conservative Dual-Criterion" = "cdc",
  "Tau U" = "tau_u",
  "Baseline corrected tau" = "corrected_tau",
  "Percentage of non-overlapping data" =  "pnd",
  "Percent exceeding the median" = "pem",
  "Percent exceeding the trend" = "pet",
  "Percentage of all non-overlapping data" = "pand",
  "Improvement rate difference" = "ird",
  "Nonoverlap of all Pairs" = "nap",
  "Randomization test" = "rand_test",
  "Outlier analysis" = "outlier"
)



## power analysis ----

.name <- function(x, at, label) {
  names(x)[which(x == at)] <- label
  x
}

res$choices$pt_method <- names(scan:::.opt$mc_fun)  |> 
  setNames(names(scan:::.opt$mc_fun)) |> 
  .name("plm_level", "Regression (level effect)") |> 
  .name("plm_slope", "Regression (slope effect)") |> 
  .name("plm_poisson_level", "Regression (level effect; frequencies)") |> 
  .name("plm_poisson_slope", "Regression (slope effect; frequencies)") |> 
  .name("hplm_level", "Multilevel-regression (level effect)") |> 
  .name("hplm_slope", "Multilevel-regression (slope effect)") |> 
  .name("tauU", "Tau-U (trend A)") |>
  .name("tauU_slope", "Tau-U (trend A and B)") |>
  .name("tauU_meta", "Metaanalysis Tau-U (trend A)") |>
  .name("tauU_slope_meta", "Metaanalysis Tau-U (trend A and B)") |>
  .name("base_tau", "Baseline corrected tau") |>
  .name("rand", "Randomization test") |> 
  .name("rand_decrease", "Randomization test (decreasing)") |> 
  .name("rand_slope", "Randomization test slope") |>
  .name("rand_slope_decrease", "Randomization test slope (decreasing)")

## legend position ----
res$choices$legend_position <- c(
  "none" = "none",
  "top" = "top",
  "right" = "right",
  "bottom" = "bottom",
  "left" = "left"
)

res$choices$separators <- c("comma" = ",", "semicolon" = ";", "tab" = "\t", 
                            "space" = " ")

# placeholder ----

res$placeholder$transform <- 'e.g.
values = scale(values)
values = local_regression(values)
values2 = values - max(values[phase=="A"])
across_cases(values2 = scale(values)
'

res$placeholder$values <- "Enter values here to create a new case. E.g. \nA = 1,2,3,4,3 \nB = 7,6,7,8,7,6"
res$placeholder$mt <- "(optional, e.g. 1,2,4,6,7,8,9,12,13)"
res$placeholder$variables <-
  "(optional, e.g., depression = 1,4,3,5,6,5,7
separate multiple variables with linebreaks)"
res$placeholder$casename <- "(optional. Random if left empty)"

res$placeholder$stats_out_args <- "" #"e.g.: decimals = 3; meta = FALSE"

res$placeholder$pt <- "Power calculation may take some time. Click 'Run' to start calculation."

# div ------

res$div$settings <- paste0(
  "background-color:#f0f0f0; ",
  "border: 1px solid black; border-radius: 5px; ",
  "padding-left: 10px; padding-right: 10px; ",
  "padding-top: 0px; padding-bottom: 0px; "
)

res$div$pt <- paste0(
  "background-color:#f0f0f0; ",
  "border: 1px solid black; border-radius: 5px; ",
  "padding-left: 5px; padding-right: 5px; ",
  "padding-top: 0px; padding-bottom: 0px; "
)

# error ----

res$error_msg$invalid_case <- "Sorry!
The case you just tried to add didn't have a valid case definition."

res$error_msg$plot <- "Sorry!
The plot arguments are not valid."

res$error_msg$scdf_save <- "Sorry!
The last file you saved is corrupt. Did you forget to add a case before saving?"

res$error_msg$html_output <- "Sorry!
HTML export is not available. Please switch to text output instead."

# msg ----

res$msg$startup <-
"Welcome to 'shiny scan'!

You can:

1. load a dataset: click 'Choose file' to import an rds, csv, or excel file)
2. choose an example scdf from 'Choose example')
3. create a new case (Data -> New -> fill in 'values' and click 'save case')

'exampleABC' is a good place to start.

The basic procedure is:

1. Choose or create an scdf (Single Case Data Frame) in the 'Data' tab.
2. Optionally refine the scdf in the 'Transform' tab (e.g. select cases, recombine phases)
3. Analyse the data in the 'Stats' tab.
4. Create a plot in the 'Plot' tab.

Look at the 'Help' tab for more information.

Have fun!
"

res$msg$no_case_scdf <-
"No case has been defined yet.
You can:
1. create a new case (fill in 'values' and click 'save case')
2. load a dataset (Data -> Load -> click 'Open file' to import an rds, csv, or excel file)
3. choose an example scdf (Data -> Load -> 'Choose example')
"

res$msg$no_case <-
"There is no case defined yet.
Please define a case on the 'Data' tab first.
"

res$msg$load_page <- "Please choose an example scdf or open a file in rds, xlsx, csv, or R-code format."

res$help_page <- structure(
  "<h4 id=\"welcome-to-shiny-scan\">Welcome to <em><strong>shiny scan</strong></em>!</h4>
  <p><em>Shiny-scan</em> is a graphical surface for <em>scan</em> (Single-Case Data Analysis). <em>scan</em> is an R package.</p>
  <p>The basic procedure is:</p>\n<ol>
  <li>Choose/ create a single case file in the <strong>Data</strong> tab.</li>
  <li>Optionally refine the case in the <strong>Transform tab</strong> (select cases, recombine phases, etc.)</li>
  <li>Analyse the data in the <strong>Stats tab</strong>.</li>
  <li>Create a plot in the <strong>Plot tab</strong>.</li>\n</ol>
  <p>Analysis and plots are based on the scdf after any changes from the <strong>Transform tab</strong>.</p>
  <p>Here are helpful links:</p>
  <p><a href=\"https://jazznbass.github.io/scan-Book/ch_shinyscan.html\">A short introduction to shiny scan</a></p>
  <p><a href=\"https://jazznbass.github.io/scan-Book/\">Online book for single case analysis with scan</a></p>
  <p><a href=\"https://jazznbass.github.io/scan/\">Technnical help pages for scan</a></p>
  <p><a href=\"https://jazznbass.github.io/scplot/\">Technnical help pages for scplot</a></p>
  <p>Have fun!</p>",
  html = TRUE, class = c("html", "character")
)

## ---- define themes (global scope) ----

theme <- getOption("scan.shiny.theme")
if (!theme %in% bslib::bootswatch_themes()) {
  message(
    theme, " is not a valid theme. Use one of these: ", 
    paste0(bslib::bootswatch_themes(), collapse = ", "))
  theme <- "cerulean"
}

res$theme_light <- bs_theme(version = 5, bootswatch = theme)


res$theme_dark <- res$theme_light |>
  bs_theme_update(
    # High-contrast dark surface + readable text
    bg = "#0f172a", fg = "#e5e7eb",
    # Cards/tables to match dark surface
    "card-bg"            = "#111827",
    "card-border-color"  = "#1f2937",
    "table-bg"           = "#0f172a",
    "table-border-color" = "#1f2937"
    # (Optional) tweak accent colors; by default we keep Cerulean's primary blue
    # "primary" = "#2FA4E7"
  )



### little help-functions

trim <- function(x) {
  gsub("\n", ", ", trimws(x))
  x <- gsub("\n", ", ", trimws(x))
  gsub(",,", ",", trimws(x))
}

quoted <- function(x) {
  out <- paste0("\"", x, "\"")
  out[is.na(x)] <- NA
  out
}

correct_casenames <- function(x, string = FALSE) {
  
  if (is.null(x) || identical(length(x), 0)) return(NULL)
    
  new_names <- names(x)
  if (is.null(new_names)) {
    new_names <- paste0("[case #", 1:length(x), "]")
  } else {
    nonames <- which(is.na(new_names) | new_names == "")
    new_names[nonames] <- paste0("[case #", nonames, "]")
  }
  
  while(anyDuplicated(new_names) != 0) {
    id <- anyDuplicated(new_names) 
    new_names[id] <- paste0(new_names[id], "'")
  }
  
  if (string) {
    return(new_names)
  }
  
  names(x) <- new_names
  x
}

guess_col <- function(cols, hints) {
  hit <- which(tolower(cols) %in% hints)
  if (length(hit)) cols[hit[1]] else cols[1]
}

n2br <- function(x) gsub("\n", "<br>", x)

render_summary <- function(scdf) {
  out <-  summary(scdf) |> export()
  if (getOption("scan.export.engine") == "gt")
    out <- out |> gt::as_raw_html(out)
  HTML(out)
}

#' @describeIn bplm Print results
#' @inheritParams print.sc
#' @order 2
#' @param x An object returned by [bplm()]
#' @export
print.sc_bplm <- function(x, digits = 3, ...) {
  cat("Bayesian Piecewise Linear Regression\n\n")
  cat("Contrast model: ", 
      x$model$interaction.method, " (", 
      paste0(names(x$contrast), ": ",x$contrast, collapse = ", "), 
      ")\n", sep = "")
  
  if (x$N > 1) cat(x$N, "Cases\n\n")
  
  out <- .output_bplm(x)
  
  cat("Deviance Information Criterion:", out$dic, "\n\n")
  
  cat("B-structure - Fixed effects (", out$formula$fixed, ")\n\n", sep = "")
  print(round_numeric(out$fixed, digits))
  
  if (!is.null(out$random)) {
    cat("\nG-Structure - Random effects (", out$formula$random, ")\n\n", sep = "")
    print(round_numeric(out$random, digits), row.names = FALSE, ...)
  }
  
  if (!is.null(out$correlation)) {
    cat("\nCorrelation\n")
    print(round_numeric(out$correlation, digits), row.names = FALSE, ...)
  }
  
  cat("\nR-Structure - Residuals\n\n")
  print(round(out$residuals, digits))
  
  
}

#' @describeIn bplm Export results as html table (see [export()])
#' @order 3
#' @inheritParams export
#' @export
export.sc_bplm <- function(object, 
                           caption = NA, 
                           footnote = NA, 
                           filename = NA,
                           round = 2,
                           nice = TRUE,
                           ...) {
  
  if (is.na(caption)) {
    if (object$N == 1) {
      caption <- paste0(
        "Bayesian Piecewise Linear Regression predicting variable '", 
        attr(object, opt("dv")),  "'"
      )
    } else {
      caption <- paste0(
        "Bayesian Multilevel Piecewise Linear Regression predicting variable '", 
        attr(object, opt("dv")),  "'"
      )
    }
  }
  
  footnote <- c(
    str_contrasts(object$model$interaction.method, object$contrast),
    paste0("N = ", object$N, " cases")
  )

  results <- .output_bplm(object)
  out <- results$fixed
  dat_g <- results$random
 
  dat_r <- cbind(
    data.frame(" " = "Residuals", check.names = FALSE),
    t(round(results$residuals, round))
  )
  
  nrow_g <- nrow(dat_g)
  nrow_b <- nrow(out)
  nrow_r <- nrow(dat_r)
  
  
  if (nice) {
    out$p <- .nice_p(out$p)
    if (!is.null(dat_g$p)) dat_g$p <- .nice_p(dat_g$p)
  }
  
  # B-structure -----
  out[, ] <- lapply(out[, ], function(x)
    if (inherits(x, "numeric")) as.character(round(x, round)) else x
  )
  out <- cbind(
    Predictors = rownames(out), 
    out, 
    stringsAsFactors = FALSE
  )
  rownames(out) <- NULL

  # add G-structure -----
  
  if (!is.null(results$random)) {
    names(dat_g)[1] <- " "
    dat_g[, ] <- lapply(dat_g, function(x)
      if (inherits(x, "numeric")) as.character(round(x, round)) else x
    )
    
    tmp_row <- (nrow_b + 1):(nrow_b + nrow_g + 1)
    out[tmp_row, ] <- ""
    
    out[tmp_row, 1:ncol(dat_r)] <- rbind(
      colnames(dat_g), 
      dat_g, 
      stringsAsFactors = FALSE
    )
  }
 
  # add R-structure -----

  tmp_row <- (nrow(out) + 1):(nrow(out) + nrow_r + 1)
  out[tmp_row, ] <- ""
  
  out[tmp_row, 1:ncol(dat_r)] <- rbind(
    colnames(dat_r), 
    dat_r, 
    stringsAsFactors = FALSE
  )
  
  # model ---
  
  new_row <- (nrow(out) + 1)
  out[new_row, ] <- ""
  
  out[new_row, 1:2] <- c(
    "DIC", as.character(round(results$dic, 1))
  )

  row_group <- if (!is.null(results$random)) {
    list(
      "Fixed effects (B-Structure)" = 1:nrow_b,
      "Random effects (G-Structure)" = (nrow_b + 1) : (nrow_b + nrow_g + 1),
      "Residuals (R-Structure)" = (nrow_b + nrow_g + 2) : (nrow_b + nrow_g + nrow_r + 2),
      "Model" = (nrow_b + nrow_g + nrow_r + 2) : nrow(out)
    )
  } else {
    list(
      "Fixed effects (B-Structure)" = 1:nrow_b,
      "Residuals (R-Structure)" = (nrow_b + 1) : (nrow_b + nrow_r + 1),
      "Model" = (nrow_b + nrow_r + 1) : nrow(out)
    )
  }
  
  
  table <- .create_table(
    out,
    caption = caption,
    footnote = footnote,
    row_group = row_group
  )

  if (getOption("scan.export.engine") == "kable") {
    if (!is.null(results$random)) {
      table <- table |>
        #pack_rows("Fixed effects", 1, nrow_out, indent = FALSE) |>
        pack_rows("\nRandom effects (G-Structure)", nrow_b + 1, nrow_b + nrow_g  + 1, indent = FALSE) |>
        pack_rows("\nResiduals (R-Structure)", nrow_b + nrow_g + 2, nrow_b + nrow_g + nrow_r + 2, indent = FALSE) |>
        pack_rows("\nModel", nrow(out), nrow(out), indent = FALSE) |>
        #row_spec(nrow_out + nrow(dat_g) + 1, hline_after = TRUE) |>
        row_spec(nrow_b, hline_after = TRUE) |> 
        row_spec(nrow_b + nrow_g + 1, hline_after = TRUE)
    }
    
    if (is.null(results$random)) {
      table <- table |>
        #pack_rows("\nRandom effects (G-Structure)", nrow_b + 1, nrow_b + nrow_g  + 1, indent = FALSE) |>
        pack_rows("\nResiduals (R-Structure)", nrow_b + 1, nrow_b + nrow_r + 1, indent = FALSE) |>
        pack_rows("\nModel", nrow(out), nrow(out), indent = FALSE) |>
        #row_spec(nrow_out + nrow(dat_g) + 1, hline_after = TRUE) |>
        row_spec(nrow_b, hline_after = TRUE) 
    }
    
  }
  
  if (!is.na(filename)) .save_export(table, filename)
  
  table
  
}


.output_bplm <- function(x) {
  
  out <- list()
  
  model_summary <- summary(x$mcmcglmm)
  out$dic <- model_summary$DIC
  
  # B-structure -----
  
  out$formula$fixed <- deparse(model_summary$fixed.formula)
  
  md <- as.data.frame(model_summary$solutions)
  colnames(md) <- c("B", "lower 95% CI", "upper 95% CI", "sample size", "p")
  row.names(md) <- rename_predictors(row.names(md), x)
  out$fixed <- md
  
  # G-structure -----
  
  if (!is.null(x$model$random)) {
    
    out$formula$random <- deparse(model_summary$random.formula)
    
    #vcv <- x$mcmc$VCV[, grep("case", colnames(x$mcmc$VCV))]
    #G_structure <- apply(vcv, 2, function(x) {
    #  c(mean = mean(x), lower = quantile(x, 0.025), upper = quantile(x, 0.975))
    #})
    #G_structure <- as.data.frame(t(G_structure))
    #G_matrix <- posterior.mode(vcv) |> as.data.frame()
    G_structure <- model_summary$Gcovariances |> as.data.frame()
    row.names(G_structure) <- rename_predictors(row.names(G_structure), x)
    row.names(G_structure) <- gsub(".case", "", row.names(G_structure))
    row.names(G_structure) <- gsub("case", "Intercept", row.names(G_structure))
    G_structure <- cbind(Parameter = row.names(G_structure), G_structure)
    row.names(G_structure) <- NULL
    
    G_structure <- G_structure[, -5]
    
    if (nrow(G_structure) > 1) {
      matching_elements <- grepl("^(.*):\\1$", G_structure$Parameter)
      G_structure_variance <- G_structure[matching_elements, ]
      G_structure_covariance <- G_structure[!matching_elements, ]
      
      G_structure_variance$Parameter <- sub(":.*", "", G_structure_variance$Parameter)
      
      G_structure_covariance$Parameter <- sapply(
        strsplit(G_structure_covariance$Parameter, ":", fixed = TRUE), 
        function(y) paste(sort(y), collapse = ":")
      )
      
      G_structure_covariance <- 
        G_structure_covariance[duplicated(G_structure_covariance$Parameter), ]
      
      names(G_structure_variance) <- c("Parameter", "SD", "lower 95% CI", "upper 95% CI") 
      names(G_structure_covariance) <- c("Parameter", "Correlation", "lower 95% CI", "upper 95% CI") 
      
      #cat("Standard deviation\n")
      G_structure_variance[, 2:4] <- sqrt(G_structure_variance[, 2:4])
      
      out$random <- G_structure_variance
      
      cor_vars <- strsplit(G_structure_covariance$Parameter, ":", fixed = TRUE)
      
      std_covar <- sapply(cor_vars, function(x) {
        var1 <- which(G_structure_variance$Parameter == x[1])
        var2 <- which(G_structure_variance$Parameter == x[2])
        G_structure_variance$SD[var1] * G_structure_variance$SD[var2]
      })
      
      for (i in 1:nrow(G_structure_covariance)) {
        G_structure_covariance[i, 2:4] <- G_structure_covariance[i,2:4] / std_covar[i]  
      }
      
      out$correlation <- G_structure_covariance
      
    } else {
      names(G_structure) <- c("Parameter", "SD", "lower 95% CI", "upper 95% CI") 
      G_structure[, 2:4] <- sqrt(G_structure[, 2:4])
      #cat("Standard deviation\n")
      
      out$random <- G_structure
      
    }
    
  }
  
  R_structure <- model_summary$Rcovariances[, -4]
  names(R_structure) <- c("SD", "lower 95% CI", "upper 95% CI") 
  out$residuals <- sqrt(R_structure)
  
  out
}

#' Percent exceeding the trend (PET)
#'
#' The `pet` function returns the percentage of Phase B data points that exceed
#' the prediction based on the Phase A trend. A binomial test against a 50/50
#' distribution is calculated. It also calculates the percentage of Phase B data
#' points that exceed the upper (or lower) 95 percent confidence interval of the
#' predicted progression.
#' 
#' The PET is calculated by first fitting a linear model to the Phase A data
#' to estimate the trend. Then, for each data point in Phase B, it is
#' determined whether it exceeds the predicted value from the Phase A trend.
#' The PET is the percentage of Phase B data points that exceed this predicted
#' value. Additionally, a binomial test is performed to assess whether the
#' observed PET is significantly greater than what would be expected by chance
#' (i.e., 50%).
#'
#' @inheritParams .inheritParams
#' @param ci Width of the confidence interval. Default is `ci = 0.95`.
#' @return 
#'  |  |  |
#'  | --- | --- |
#'  | `PET` | Percent exceeding the trend. |
#'  | `ci` | Width of confidence interval. |
#'  | `decreasing` | Logical argument from function call (see Arguments above). | 
#' @author Juergen Wilbert
#' @family overlap functions
#' @examples
#'
#' ## Calculate the PET and use a 99%-CI for the additional calculation
#' # create random example data
#' design <- design(n = 5, slope = 0.2)
#' dat <- random_scdf(design, seed = 23)
#' pet(dat, ci = .99)
#'
#' @export
pet <- function(data, 
                dvar, pvar, mvar, 
                ci = 0.95, 
                decreasing = FALSE, 
                phases = c(1, 2)) {
  
  # set attributes to arguments else set to defaults of scdf
  if (missing(dvar)) dvar <- dv(data) else dv(data) <- dvar
  if (missing(pvar)) pvar <- phase(data) else phase(data) <- pvar
  if (missing(mvar)) mvar <- mt(data) else mt(data) <- mvar
  
  data <- .prepare_scdf(data, na.rm = TRUE)
  data <- recombine_phases(data, phases = phases)$data
  
  N <- length(data)
  
  if (ci != 0) se_factor <- qnorm(ci) else se_factor <- 0 
  
  pet    <- rep(NA, N)
  pet_ci <- rep(NA, N)
  p      <- rep(NA, N)
  
  for(i in 1:N) {
    formula <- as.formula(paste0(dvar, "~", mvar))
    model <- lm(
      formula, 
      data = data[[i]][data[[i]][, pvar] == "A",], 
      na.action = na.omit
    )
    B <- data[[i]][data[[i]][, pvar] == "B",]
    res <- predict(model, B, se.fit = TRUE)
    nB <- nrow(B)
    if(!decreasing) {
      pet_ci[i] <- mean(B[, dvar] > (res$fit + res$se.fit * se_factor)) * 100
      pet[i]    <- mean(B[, dvar] > res$fit)*100
      p[i]      <- binom.test(
        sum(B[, dvar] > res$fit), nB, alternative = "greater"
      )$p.value
    } else {
      pet_ci[i] <- mean(B[, dvar] < (res$fit - res$se.fit * se_factor)) * 100
      pet[i] <- mean(B[, dvar] < res$fit) * 100
      p[i] <- binom.test(
        sum(B[, dvar] < res$fit), nB, alternative = "greater"
      )$p.value
    }
  }

  pet <- data.frame(
    Case = revise_names(data),
    PET = pet, 
    "PET CI" = pet_ci, 
    binom.p = p, 
    check.names = FALSE
  )
  
  out <- list(
    PET = pet,
    ci = ci,
    decreasing = decreasing
  )
  class(out) <- c("sc_pet")
  attributes(out)[opts("phase", "dv")] <- list(pvar, dvar)
  out
}

#' @describeIn tau_u Print results
#' @order 2
#' @inheritParams print.sc
#' @param x Object returned from [tau_u()].
#' @param complete Print all parameters.
#' @param select Character vector with name of variables to be included. When
#'   the vector is named, variables are renamed appropriately.
#' @param nice_p If TRUE, p-values are printed in publication friendly form.
#' @export
#' 
print.sc_tauu <- function(x, 
                          complete = FALSE, 
                          digits = "auto", 
                          select = c(
                            "Tau", "CI lower", "CI upper", "SD_S", "Z", "p"
                          ), 
                          nice_p = TRUE,
                          ...) {
  
  if (digits == "auto") digits <- 2
  
  cat("Tau-U\n")
  cat("Method:", x$method, "\n")
  cat("Applied Kendall's Tau-", x$tau_method, "\n", sep = "")
 
  if (!is.na(x$ci)) {
    cat(x$ci * 100, "% CIs for tau are reported.\n", sep = "")
    cat("CI method: ", x$ci_method, "\n\n", sep = "")
  } else cat("\n")
  
  out <- x$table
  
  if (length(out) > 1 && x$meta_analyses) {
    cat("Tau-U meta analyses:\n")
    
    cat("Weight method: ", x$meta_weight_method, "\n", sep = "")
    if (!is.na(x$ci)) cat(x$ci * 100, "% CIs are reported.\n", sep = "")
    cat("\n")
    print(x$Overall_tau_u, row.names = FALSE, digits = digits)
    cat("\n")
  }
  
  if (!complete) {
    select_vars <- select
    select_rows <- match(
      c(
        "A vs. B", 
        "A vs. B - Trend A",
        "A vs. B + Trend B", 
        "A vs. B + Trend B - Trend A"
      ), row.names(x$table[[1]])
    )
    
    out <- lapply(x$table, function(x) {
      x <- round(x[select_rows, select_vars], digits)
      if (nice_p) x$p <- .nice_p(x$p)
      if (!is.null(names(select))) names(x) <- names(select)
      x
    })
    
  } else {
    out <- lapply(out, function(x) {
      x <- round(x, digits)
      if (nice_p) x$p <- .nice_p(x$p) else x$p <- round(x$p, digits)
      x
    })
  }
  
  for(i in seq_along(out)) {
    cat("Case:", names(out)[i], "\n")
    print(out[[i]], ...)
    cat("\n")
  }
  
}

#' Between-Case Standardized Mean Difference
#'
#' Calculates a standardized mean difference from a multilevel model as
#' described in Pustejovsky et al. (2014)
#'
#' The BC-SMD is calculate as `BC-SMD = Phase difference / sqrt(residual +
#' random_intercept)`. This is most closely related to Cohen's *d*. If you want
#' to have the most exact estimation based on the between case variance, you
#' have to exclude the residual variance by setting the argument
#' `include_residuals = FALSE` you get `BC-SMD = Phase difference /
#' sqrt(random_intercept)`. The 'base' model only includes the phase level as a
#' predictor like originally proposed by Hedges et al. Whereas the 'Full plm'
#' model includes the trend and the phase slope as additional predictors.

#' @param data Either an scdf or an object returned from the [hplm()] or
#'   [bplm()] function.
#' @param method Either `"REML"` or "`MCMglmm"`. This indicated which
#'   statistical method is applied to calculate the model.
#' @param ci A numeric between 0 and 1 setting the width of the confidence
#'   interval (when method is REML) or the credible interval (when method is
#'   MCMCglmm). The default is `0.95` for a 95-percent interval.
#' @param include_residuals Logical. See details.  
#' @param ... Further arguments passed to the [hplm()] or [bplm()]function.
#' @return An object of class sc_bcsmd. It is a list containing the following
#'  elements:
#'  - `models`: A list of data frames containing the BC-SMD results for each
#'  model calculated.
#'  - `ci`: The width of the confidence/credible interval.
#'  - `method`: The method used for model estimation.
#' @author Juergen Wilbert
#' @references Pustejovsky, J. E., Hedges, L. V., & Shadish, W. R. (2014).
#'   Design-Comparable Effect Sizes in Multiple Baseline Designs: A General
#'   Modeling Framework. Journal of Educational and Behavioral Statistics,
#'   39(5), 368–393. https://doi.org/10.3102/1076998614547577
#' @examples
#' ## Create a example scdf:
#' des <- design(
#'   n = 150,
#'   phase_design = list(A1 = 10, B1 = 10, A2 = 10, B2 = 10, C = 10),
#'   level = list(B1 = 1, A2 = 0, B2 = 1, C = 1),
#'   rtt = 0.7,
#'   random_start_value = TRUE
#' )
#' study <- random_scdf(des)
#'
#' ## Standard BC-SMD return:
#' between_smd(study)
#'
#' ## Specify the model and provide an hplm object:
#' model <- hplm(study, contrast_level = "preceding", slope = FALSE,  trend = FALSE)
#' between_smd(model)
#'
#' ## excluding the residuals gives a more accurate estimation:
#' between_smd(model, include_residuals = FALSE)
#'
#' @order 1
#' @export
between_smd <- function(data,
                        method = c("REML", "MCMCglmm"),
                        ci = 0.95,
                        include_residuals = TRUE,
                        ...) {
  
  
  
  check_args(
    by_call(method),
    within(ci, 0, 1),
    is_logical(include_residuals)
  )
  
  models <- list()
  if (inherits(data, "sc_hplm")) {
    if (data$N == 1) {
      abort("Between-Case Standardized Mean Difference can not be computed ",
           "with one case.")
    }
    method <- "REML"
    models[["Provided"]] <- data
  } else if (inherits(data, "sc_bplm")) {
    if (data$N == 1) {
      abort("Between-Case Standardized Mean Difference can not be computed ",
           "with one case.")
    }
    method <- "MCMCglmm"
    models[["Provided"]] <- data
  } else if (inherits(data, "scdf")) {
    if (length(data) == 1) {
      abort("Between-Case Standardized Mean Difference can not be computed ",
           "with one case.")
    }
    
    if (method == "REML") {
      models[["Base model"]] <- hplm(
        data, 
        slope = FALSE, 
        trend = FALSE, 
        level = TRUE,
        method = "REML",
        ...
      ) 
      models[["Full plm model"]] <- hplm(
        data, 
        slope = TRUE, 
        trend = TRUE, 
        level = TRUE,
        method = "REML",
        ...
      ) 
    }
    if (method == "MCMCglmm") {
      models[["Base"]] <- bplm(
        data,
        slope = FALSE,
        trend = FALSE,
        level = TRUE,
        ...
      )
      models[["Full plm"]] <- bplm(
        data,
        slope = TRUE,
        trend = TRUE,
        level = TRUE,
        ...
      )
    }
  } 
  
  extract_bc_smd <- function(model, ci) {
    
    model <- summary(model$hplm)
    ids <- which(startsWith(rownames(model$tTable), "phase"))
    beta <- model$tTable[ids, "Value"]
    se_beta <- model$tTable[ids, "Std.Error"]
    var_within <- model$sigma^2
    var_between <- as.numeric(VarCorr(model)["(Intercept)", "Variance"])
    
    if (include_residuals) variance <- var_within + var_between 
    if (!include_residuals) variance <- var_between 
    
    bc_smd <- beta / sqrt(variance)
    se_bc_smd <- se_beta / sqrt(variance)
    z <- qnorm((1 - (1 - ci)/2))
    ci_lower <- bc_smd - z * se_bc_smd
    ci_upper <- bc_smd + z * se_bc_smd

    out <- list(
      "Effect" = rownames(model$tTable)[ids],
      "BC-SMD" = bc_smd,
      se = se_bc_smd,
      ci_lower = ci_lower,
      ci_upper = ci_upper
    )
   
    out
    #if (!include_residuals) beta / sqrt(var_within + var_between)
    
    #beta <- model$hplm$coefficients$fixed[ids]
    #resid <- as.numeric(nlme::VarCorr(model$hplm)["Residual", "Variance"])
    #inter <- as.numeric(nlme::VarCorr(model$hplm)["(Intercept)", "Variance"])
    #if (include_residuals) beta / sqrt(sum(resid, inter)) else beta / sqrt(inter)
   
  }
  
  extract_bc_smd_bayes <- function(model, ci) {
    model <- model$mcmcglmm
    
    ids <- which(startsWith(colnames(model$Sol), "phase"))
    # Extract posterior samples
    betas <- model$Sol[, ids, drop = FALSE]
    var_resid_samples <- model$VCV[, "units"]
    var_random_intercept_samples <- model$VCV[, "case"]
    
    if (include_residuals) variance <- var_resid_samples + var_random_intercept_samples 
    if (!include_residuals) variance <- var_random_intercept_samples 
    
    # Compute BC-SMD per iteration
    bc_smds <- betas / sqrt(variance)
    
    # Summarize posterior
    posterior_mean <- mean(bc_smds)
    posterior_ci <- quantile(bc_smds, probs = c((1 - ci) / 2, 1 - ((1 - ci) / 2)))
    
    out <- list(
      "Effect" = colnames(model$Sol[, ids, drop = FALSE]),
      "BC-SMD" = posterior_mean,
      "Sample size" = nrow(model$Sol),
      ci_lower = posterior_ci[1],
      ci_upper = posterior_ci[2]
    )
    return(out)
    ########
    # model <- summary(model$mcmcglmm)
    # ids <- which(startsWith(row.names(model$solutions), "phase"))
    # a <- setNames(model$solutions[ids, 1], rownames(model$solutions)[ids])
    # resid <- model$Rcovariances["units", "post.mean"]
    # inter <- model$Gcovariances["case", "post.mean"]
    # #if (include_residuals) a / sqrt(sum(resid, inter)) else a / sqrt(inter)
    
  }

  if (method == "REML") {
    out <- lapply(models, function(x) { 
      extract_bc_smd(x, ci) |> 
        as.data.frame(check.names = FALSE)
    })
  }
  
  if (method == "MCMCglmm") {
    out <- lapply(models, function(x) { 
      extract_bc_smd_bayes(x, ci) |> 
        as.data.frame(check.names = FALSE)
    })
  }
  
  out <- list(
    models = out,
    ci = ci,
    method = method
  )
 
  class(out) <- "sc_bcsmd"
  nm <- opts("phase", "mt", "dv")
  attributes(out)[nm] <- models[[1]][nm]
  out
}

test_that("between_smd errors for a single case", {
  expect_error(
    between_smd(exampleAB$Johanna, method = "REML"),
    "with one case"
  )
})

test_that("between_smd returns sc_bcsmd with REML models and expected columns", {
  res <- between_smd(exampleAB[1:3], method = "REML", ci = 0.90, include_residuals = TRUE)
  res_no_resid <- between_smd(exampleAB[1:3], method = "REML", ci = 0.90, include_residuals = FALSE)
  
  # BC-SMD should increase when residual variance is excluded
  base_with <- res$models[["Base model"]][["BC-SMD"]]
  base_without <- res_no_resid$models[["Base model"]][["BC-SMD"]]
  expect_gt(mean(base_without), mean(base_with))
})

#' Print an scdf
#'
#' @param x An scdf object
#' @param cases Number of cases to be printed. "fit" fits the number to the current screen width.
#' @param rows Number of rows to be printed. 
#' @param cols Columns to be printed. "Main" only prints the dependent, measurement-time and phase variable.
#' @param long Logical. If TRUE cases are printed in one by a time.
#' @param digits Number of digits.
#' @param ... Further arguments passed to the print function. 
#' @details Print options for scdf objects could be set globally:
#' option(scan.print.cases = "all"), 
#' option(scan.print.rows = 10), 
#' option(scan.print.cols = "main"),
#' option(scan.print.long = TRUE),
#' option(scan.print.digits = 0),
#' option(scan.print.scdf.name = FALSE)
#' @export
print.scdf <- function(x, 
                       cases  = getOption("scan.print.cases"), 
                       rows   = getOption("scan.print.rows"), 
                       cols   = getOption("scan.print.cols"), 
                       long   = getOption("scan.print.long"), 
                       digits = getOption("scan.print.digits"), ...) {
  
  v_bar <- getOption("scan.print.bar")
  
  row.names <- FALSE
  N <- length(x)

  names(x) <- revise_names(x)

  if (identical(cases, "all")) cases <- N
  if (is.numeric(cases) && cases > N) cases <- N
  if (N == 1) cat("#A single-case data frame with one case\n\n")
  if (N > 1)  
    cat("#A single-case data frame with", number_word(N), "cases\n\n")
  
  if (identical(cols, "main")) {
    cols <- c(attr(x, opt("dv")), attr(x, opt("phase")), attr(x, opt("mt")))
  }
  
  if (!identical(cols, "all")) for(i in 1:N) x[[i]] <- x[[i]][, cols]
  
  if(getOption("scan.print.scdf.name")) {
    for(i in 1:N) {
      names(x[[i]])[1] <- paste0(names(x)[i], ": ", names(x[[i]])[1])
    }
  }
  
  if(identical(cases, "fit")) {
    tmp_func <- function(case) nchar(paste0(names(case), collapse = " ")) + 3
    tmp_width <- cumsum(lapply(x, tmp_func))
    cases <- length(which(getOption("width") >= tmp_width))
    if(cases == 0) cases <- 1
  }
  
  max_row <- max(unlist(lapply(x, nrow)))
  
  if(!long) {
    
    for(i in 1:cases) {
      
      n_row <- nrow(x[[i]])
      # round and change to character
      for(j in names(x[[i]])) {
        if(is.numeric(digits) && is.numeric(x[[i]][, j])) {
          x[[i]][, j] <- round(x[[i]][, j], digits)
        }
        #x[[i]][, j] <- as.character(x[[i]][, j])
        x[[i]][[j]] <- as.character(x[[i]][[j]])
      }
      if (n_row < max_row) x[[i]][(n_row + 1):max_row, names(x[[i]])] <- ""
    }
  }
  
  if (rows == "all") long <- TRUE
  if (!long) {
    if (max_row < rows) rows <- max_row
    out <- lapply(x[1:cases], function(x) x[1:rows, ])
    #if (cases > 1) out <- lapply(out, function(x) {x$"|" <- "|"; x})
    if (cases > 1) out <- lapply(out, function(x) {x[[v_bar]] <- v_bar; x})

    names <- lapply(out, names)
    out <- as.data.frame(out)
    names(out) <- unlist(names[1:cases])
    print(out, row.names = row.names, ...)
  }
  
  if (long) {
    for(case in 1:N) {
      print(x[[case]], row.names = row.names, digits = digits, ...)
      cat("\n")
    }
  }
  
  if (max_row > rows) 
    cat("# ... up to", number_word(max_row - rows), "more rows\n")
  if ((N - cases)  > 1) cat("# ", number_word(N - cases), "more cases\n")
  if ((N - cases) == 1) cat("# One more case\n")
  
}

#' @describeIn tau_u Export results as html table
#' @order 3
#' @inheritParams export
#' @param meta If TRUE, the results of the meta analysis will be exported. If
#'   FALSE, each single-case is exported.
#' @param select Character vector with name of variables to be included. When
#'   the vector is named, variables are renamed appropriately.
#' @export
export.sc_tauu <- function(object, 
                           caption = NA, 
                           footnote = NA, 
                           filename = NA,
                           select = "auto",
                           meta = FALSE,
                           round = 3,
                           decimals = 3,
                           ...) {
  
  if (is.na(caption)) {
    if (meta) 
      caption <- c("Overall Tau-U") 
    else 
      caption <- "Tau-U analyses"
    #, .phases_string(A, B))
  }

  if (is.na(footnote)) {
    footnote <- paste(
      "Method is '", object$method, 
      "'. Analyses based on Kendall's Tau ", object$tau_method, ". ",
      object$ci * 100, "% CIs for tau are reported",
      collapse = ""
    )
  }
  
  if (meta) {
    out <- object$Overall_tau_u
    row_group <- NULL
  }
  
  if (!meta && getOption("scan.export.engine") == "kable") {
    tables <- object$table
    names_models <- c(" ", row.names(tables[[1]]))
    
    n_cases <- length(tables)
    out <- rbind(tables[[1]])
    out <- rbind(NA, out)
    
    for(i in 2:n_cases) out <- rbind(out, NA, tables[[i]])
  
    row.names(out) <- NULL
    Model <- rep(names_models, length = nrow(out)) 
    Case <- lapply(names(tables), function(x) c(x, rep(" ", 6))) |> unlist()
    out <- cbind(Case, Model, out)
    
    if (identical(select, "auto")) {
        select <- c("Case", "Model", "Tau", "CI lower", "CI upper", "Z", "p")
    }
    
  }  
  
  if (identical(select, "auto") && meta) {
      select <- c(
        "Model", "Tau U" = "Tau_U", "se", "CI lower", "CI upper", "z", "p"
      )
  }
  
  if (!meta && getOption("scan.export.engine") == "gt") {
    tables <- object$table
    out <- do.call(rbind, tables)
    out <- cbind(Model = rep(rownames(tables[[1]]), length(tables)), out)
    rownames(out) <- NULL
    
    if (identical(select, "auto")) {
        select <- c("Model", "Tau", "CI lower", "CI upper", "Z", "p")
    }
    
    row_group <- vector("list", length(tables))
    names(row_group) <- names(tables)
    
    for (i in 1:length(tables)) {
      .start <- 1 + (i - 1) * nrow(tables[[1]])
      row_group[[i]] <- .start : (.start + nrow(tables[[1]]) - 1)
    }
  }
  
  out$p <- .nice_p(out$p)
  out <- .select(out, select)
  
  opts <- options()
  
  tmp <- getOption("scan.export.kable")
  tmp$align <- c("l", "l", rep("c", ncol(out) - 2)) 
  options(scan.export.kable = tmp)
  options(knitr.kable.NA = "")
  
  table <- .create_table(
    out, 
    caption = caption,
    footnote = footnote,
    row_group = row_group,
    decimals = decimals,
    ...
  )
  
  options(opts)
  # finish ------------------------------------------------------------------
  
  if (!is.na(filename)) .save_export(table, filename)
  table
}

#' Estimate single-case design
#'
#' This functions takes an scdf and extracts design parameters. The resulting
#' object can be used to randomly create new scdf files with the same underlying
#' parameters. This is useful for Monte-Carlo studies and bootstrapping
#' procedures.
#' 
#' The function uses the plm function to estimate parameters for each
#' single-case. If more than two single-cases are included in the scdf,
#' the between case variance depicting the overall performance (s) is
#' estimated unless s is provided. The reliability of the measurements (rtt)
#' is estimated for each case unless rtt is provided. If overall_rtt is set
#' to TRUE, rtt estimations will be based on all cases and identical for each
#' case. If overall_effects is set to TRUE, trend, level, and slope
#' effect estimations will be identical for each case.
#' 
#' The resulting design object can be used as input for the random_scdf
#' function to create new random scdf files based on the estimated parameters.
#' This allows to create bootstrap samples or Monte-Carlo datasets based on
#' the characteristics of an existing dataset.
#'   
#' @inheritParams .inheritParams
#' @param s The standard deviation depicting the between case variance of the
#'   overall performance. If more than two single-cases are included in the
#'   scdf, the variance is estimated if s is set to NULL. If s is provided, this
#'   value is used.
#' @param rtt The reliability of the measurements. The reliability is estimated
#'   when rtt = NULL. If rtt is provided, this value is used for all single-cases.
#' @param overall_rtt Ignored when `rtt` is set. If TRUE, rtt estimations will
#'   be based on all cases and identical for each case. If FALSE rtt is
#'   estimated for each case separately. Default is TRUE.
#' @param overall_effects If TRUE, trend, level, and slope effect estimations
#'   will be identical for each case. If FALSE, effects are estimated for each
#'   case separately. Default is FALSE.
#' @param ... Further arguments passed to the plm function used for parameter
#'   estimation. For example, the model argument can be used to select
#'   different piecewise regression models. See plm for details.
#' @return A list of parameters for each single-case. Parameters include name,
#'   length, and starting measurement time of each phase, trend, level, and
#'   slope effects for each phase, start value, standard deviation, and
#'   reliability for each case. This list can be used as input for the
#'   random_scdf function to create new random scdf files based on the estimated
#'   parameters.
#' @author Juergen Wilbert
#' @examples
#' # create a random scdf with predefined parameters
#' set.seed(1234)
#' design <- design(
#'   n = 10, trend = -0.02,
#'   level = list(0, 1), rtt = 0.8,
#'   s = 1
#' )
#' scdf<- random_scdf(design)
#'
#' # Estimate the parameters based on the scdf and create a new random scdf
#' # based on these estimations
#' design_est <- estimate_design(scdf, rtt = 0.8)
#' scdf_est <- random_scdf(design_est)
#'
#' # Analyze both datasets with an hplm model. See how similar the estimations
#' # are:
#' hplm(scdf, slope = FALSE)
#' hplm(scdf_est, slope = FALSE)
#'
#' # Also similar results for pand and randomization tests:
#' pand(scdf)
#' pand(scdf_est)
#' rand_test(scdf)
#' rand_test(scdf_est)
#' @export

estimate_design <- function(data, dvar, pvar, mvar, 
                            s = NULL, 
                            rtt = NULL, 
                            overall_effects = FALSE, 
                            overall_rtt = TRUE,
                            model = "JW", 
                            ...) {

  # set attributes to arguments else set to defaults of scdf
  if (missing(dvar)) dvar <- dv(data)
  if (missing(pvar)) pvar <- phase(data)
  if (missing(mvar)) mvar <- mt(data)
  dv(data) <- dvar
  phase(data) <- pvar
  mt(data) <- mvar
  
  data <- .prepare_scdf(data)
  N <- length(data)
  case_names <- names(data)

  if (is.null(case_names)) case_names <- revise_names(data)

  cases <- lapply(data, function(x) {
    df <- as.list(.phasestructure(x, pvar))
    names(df)[1:2] <- c("length", "phase")
    df
  })

  error <- c()
  fitted <- c()

  for (i in 1:N) {
    plm_model <- plm(data[i], model = model, ...)$full
    res <- coef(plm_model)
    var_fitted <- var(plm_model$fitted.values)
    var_residuals <- var(plm_model$residuals)
    n_phases <- length(cases[[i]]$phase)
    cases[[i]]$start_value <- res[1]
    cases[[i]]$trend <- res[2]
    cases[[i]]$level <- c(0, res[3:(1 + n_phases)])
    cases[[i]]$slope <- c(0, res[(2 + n_phases):(2 + 2 * (n_phases - 1))])
    
    if (is.null(rtt)) {
      cases[[i]]$rtt <- var_fitted / (var_fitted + var_residuals)
    } else {
      cases[[i]]$rtt <- rtt
    }
    
    cases[[i]]$missing_prop <- mean(is.na(data[[i]][[dvar]]))
    
    #missing_prop <- c()
    # for (y in 1:length(cases[[i]]$phase)) {
    #   missing_prop <- c(
    #     missing_prop, 
    #     sum(is.na(data[[i]][cases[[i]]$start[y]:cases[[i]]$stop[y], dvar])) / cases[[i]]$length[y]
    #   )
    # }
    #cases[[i]]$error <- var_residuals
    #cases[[i]]$fitted <- var_fitted
    
    error <- c(error, plm_model$residuals)
    fitted <- c(fitted, plm_model$fitted.values)
  }

  if (N > 2 && is.null(s)) 
    s <- sd(sapply(cases, function(x) x$start_value[1]), na.rm = TRUE)
  
  if (overall_effects) {
    level <- rowMeans(sapply(cases, function(x) x$level))
    trend <- mean(sapply(cases, function(x) x$trend))
    slope <- rowMeans(sapply(cases, function(x) x$slope))
    for (i in 1:N) {
      cases[[i]]$trend <- trend
      cases[[i]]$level <- level
      cases[[i]]$slope <- slope
    }
  }
  
  vars <- c(
    "phase", "length", "rtt", "missing_prop", "extreme_prop", 
    "extreme_low", "extreme_high", "trend", "level", "slope", "start_value", 
    "s", "start", "stop"
  )

  for (i in 1:N) {
    cases[[i]]$level <- cases[[i]]$level / s
    cases[[i]]$slope <- cases[[i]]$slope / s
    cases[[i]]$trend <- cases[[i]]$trend / s
    cases[[i]]$s <- s
    if (overall_rtt) cases[[i]]$rtt <- var(fitted) / (var(fitted) + var(error))
    cases[[i]]$extreme_prop <- 0
    cases[[i]]$extreme_low <- -4
    cases[[i]]$extreme_high <- -3
    cases[[i]] <- cases[[i]][vars]
  }

  out <- list(
    cases = cases,
    distribution = "normal"
  )

  class(out) <- c("sc_design")
  attributes(out)[opts("phase", "mt", "dv")] <- list(pvar, mvar, dvar)
  out
}

test_that("estimate_design", {
  scdf <- exampleABC
  result <- estimate_design(scdf)
  expect_s3_class(result, "sc_design")
  expected_vars <- c(
    'phase', 'length', 'rtt', 'missing_prop', 'extreme_prop', 
    'extreme_low', 'extreme_high', 'trend', 'level', 'slope', 'start_value', 
    's', 'start', 'stop'
  )
  for (i in seq_along(result$cases)) {
    expect_equal(sort(names(result$cases[[i]])), sort(expected_vars))
  }
  
  value <- round(sum(unlist(lapply(result$cases, function(x)  unlist(x[-1])))), 4)
  expect_equal(value, 594.4769)  
})

#' @rdname print.sc
#' @export
print.sc_nap <- function(x, 
                         digits = "auto", 
                         nice = TRUE, 
                         complete = FALSE, 
                         ...) {
  
  if (digits == "auto") digits <- 2
  cat("Nonoverlap of All Pairs\n\n")
  
  out <- as.data.frame(x$nap)
  if (!complete) out <- out[, -(4:7)]
  if (nice) out$p <- .nice_p(unlist(out$p))
  print(out, digits = digits, row.names = FALSE)
  
}

#' Combine single-case data frames into one scdf
#' 
#' Combines several single-case data frames (scdf) into one scdf object.
#'
#' @param ... scdf objects to be combined.
#' @param dvar Character string. Name of the dependent variable. Defaults to the
#'   dependent variable of the first case provided.
#' @param pvar Character string. Name of the phase variable. Defaults to the
#'   phase variable of the first case provided.
#' @param mvar Character string. Name of the measurement-time variable. Defaults
#'   to the measurement-time variable of the first case provided.
#' @param author author of the data.
#' @param info additional information on the scdf file.
#' @return A scdf. If not set differently, the attributes of this scdf are
#'   copied from the first scdf provided (i.e the first argument of the
#'   function).
#' @author Juergen Wilbert
#' @keywords data manipulation
#' @export
combine <- function(..., 
                    dvar = NULL, 
                    pvar = NULL, 
                    mvar = NULL, 
                    info = NULL, 
                    author = NULL) {
  scdfs <- list(...)
  
  source_attr <- attributes(scdfs[[1]])
  
  case_names <- unlist(lapply(scdfs, names))
   
  data <- unlist(scdfs, recursive = FALSE)
  
  attributes(data) <- source_attr
  
  if (!is.null(dvar)) dv(data) <- dvar
  if (!is.null(mvar)) mt(data) <- mvar
  if (!is.null(pvar)) phase(data) <- pvar
  if (!is.null(info)) scdf_attr(data, "info") <- info
  if (!is.null(author)) scdf_attr(data, "author") <- author
  
  names(data) <- case_names
  if (!is.null(names(scdfs))) {
    .names <- names(scdfs)[which(names(scdfs) != "")]
    names(data)[which(names(scdfs) != "")] <- .names
    
  }
  
  # check class scdf validity
  if (opt("rigorous_class_check")) {
    results <- .check_scdf(data)
    if (!isTRUE(results)) {
      if(length(results$warnings) > 0) {
        warn(results$warnings)
      }
      if(length(results$errors) > 0) {
        abort(results$errors)
      }
    } 
  }
  
  data
}

##' @rdname combine
##' @export
c.scdf <- function(...) {
  combine(...)
}

#' Select an scdf case by name
#' 
#' Selects a single case from a scdf object by its name.
#'
#' @param x A scdf object.
#' @param i A case name from x. If i is not a name of x, a warning is issued.
#' @return A scdf object containing only the selected case.
#' @author Juergen Wilbert
#' @rdname Subsetting
#' @keywords internal
#' @export
`$.scdf`<- function(x, i) {
  if (is.character(i) && !(i %in% names(x))) {
    warn("Unknown case: '", i, "'.")
  }
  out <- x[i]
  attr(out, opt("scdf")) <- attr(x, opt("scdf"))
  class(out) <- c("scdf", "list")
  out
}

##' @rdname Subsetting
##' @export
`[.scdf`<- function(x, i) {
  class(x) <- "list"
  out <- x[i]
  attr(out, opt("scdf")) <- attr(x, opt("scdf"))
  class(out) <- c("scdf", "list")
  out
}



#' Test for scdf objects
#' 
#' Tests for objects of type "scdf".
#'
#' @param x An object to be tested.
#' @return Returns TRUE or FALSE depending on whether its argument is of scdf type or not.
#' @author Juergen Wilbert
#' @export
is.scdf <- function(x) inherits(x, "scdf")

#' Remove missing values from scdf
#' 
#' Removes any row with a missing value from each single-case data frame in an scdf.
#'
#' @param object A scdf.
#' @param ... not implemented yet.
#' @return A scdf object.
#' @author Juergen Wilbert
#' @export
na.omit.scdf <- function(object, ...) {
  for (i in 1:length(object))
    object[[i]] <- na.omit(object[[i]])
  object
}

.create_fixed_formula <- function(dvar, mvar, 
                                  slope, level, trend, 
                                  var_phase, 
                                  var_inter,
                                  intercept = TRUE) {
  parameters <- c()
  
  if (intercept) parameters <- "1"
  if (trend)  parameters <- c(parameters, mvar)
  if (level) parameters <- c(parameters, var_phase)
  if (slope) parameters <- c(parameters, var_inter)

  out <- paste0(dvar, " ~ ", paste0(parameters, collapse = " + "))
  
  as.formula(out, env = parent.frame())
}

.create_random_formula <- function(mvar, 
                                   slope, 
                                   level, 
                                   trend, 
                                   var_phase, 
                                   var_inter,
                                   intercept = TRUE,
                                   syntax = "lm") {
  
  parameters <- c()
  
  if (intercept) parameters <- "1"
  if (trend) parameters <- c(parameters, mvar)
  if (level) parameters <- c(parameters, var_phase)
  if (slope) parameters <- c(parameters, var_inter)

  if (syntax == "lm") {
    out <- paste0("~ ", paste0(parameters, collapse = " + "), "|case")
  }
  
  if (syntax == "mcmc") {
    out <- paste0("~ us(", paste0(parameters, collapse = " + "), "):case")
  }
  
  as.formula(out, env = parent.frame())
}

rename_predictors <- function(rn, x) {
  
  style <- getOption("scan.rename.predictors")
  if (identical(style, FALSE)) return(rn)
  if (identical(style, "no")) return(rn)
  
  str_mt <- attr(x, opt("mt"))
  str_phase <- attr(x, opt("phase"))
  str_slope <- getOption("scan.string.dummy.slope")
  str_phase <- getOption("scan.string.dummy.phase")
  
  if (style == "full") {
    rn[which(rn == str_mt)] <- paste0("Trend (", str_mt, ")")
    rn <- gsub("(Intercept)", "Intercept", rn, fixed = TRUE)
    rn <- gsub(
      paste0(str_phase, "(\\w+)"), 
      paste0("Level phase \\1 (", str_phase, "\\1)"), 
      rn
    )
    rn <- gsub(
      paste0(str_slope, "(\\w+)"), 
      paste0("Slope phase \\1 (", str_slope, "\\1)"), 
      rn
    )
  } else if (style == "concise"){
    rn[which(rn == str_mt)] <- "Trend"
    rn <- gsub("(Intercept)", "Intercept", rn, fixed = TRUE)
    rn <- gsub(paste0(str_phase, "(\\w+)"), "Level \\1", rn)
    rn <- gsub(paste0(str_slope, "(\\w+)"), "Slope \\1", rn)
  } else {
    abort("Ill defined scan.rename.predictors option.", 
         "Must be one of 'concise', 'full' or, 'no'.")
  }
  
  rn
}

#' as_scdf
#'
#' Converts a data frame to an scdf object.
#'
#' @param object A data.frame
#' @param cvar Sets the "case" variable. Defaults to `case`.
#' @param pvar Sets the "phase" variable. Defaults to `phase`.
#' @param dvar Sets the "values" variable. Defaults to `values`.
#' @param mvar Sets the variable name of the "mt" variable. Defaults to `mt`.
#' @param phase_names A character vector with phase names. Defaults to the phase
#'   names provided in the phase variable.
#' @param sort_cases If set TRUE, the resulting list is sorted by label names
#'   (alphabetically increasing).
#' @return An scdf.
#' @family data manipulation functions
#' @export

as_scdf <- function(object,
                    cvar = "case",
                    pvar = "phase",
                    dvar = "values",
                    mvar = "mt",
                    phase_names = NULL,
                    sort_cases = FALSE) {

  original_attr <- scdf_attr(object)
  if (!is.null(original_attr)) {
    pvar <- phase(object)
    dvar <- dv(object)
    mvar <- mt(object)
    notify("Found scdf attributes and replaced function arguments.")
  }
  
  ## check file -----
  
  check_args(
    is_true(dvar %in% names(object), "Variable '", dvar, "' is missing."),
    is_true(mvar %in% names(object), "Variable '", mvar, "' is missing."),
    is_true(pvar %in% names(object), "Variable '", pvar, "' is missing."),
    not(any(is.na(object[[cvar]])), "Variable '", cvar, "' has a missing value."),
    not(any(is.na(object[[pvar]])), "Variable '", pvar, "' has a missing value.")
  )
  
  if (!cvar %in% names(object)) {
    notify("Casename variable not found. Assuming one case.")
    object[[cvar]] <- "unnamed"
  }


  if (!sort_cases) {
    object[[cvar]] <- factor(object[[cvar]], levels = unique(object[[cvar]]))
  } else {
    object[[cvar]] <- factor(object[[cvar]])
  }

  object[[pvar]] <- factor(object[[pvar]], levels = unique(object[[pvar]]))
  if (!is.null(phase_names)) levels(object[[pvar]]) <- phase_names

  case_names <- levels(object[[cvar]])
  object <- split(object, object[[cvar]])
  object <- lapply(object, function(x) {
    x[, -which(names(x) == cvar)]
  })
  names(object) <- case_names

  class(object) <- c("scdf", "list")
  
  if (!is.null(original_attr)) {
    scdf_attr(object) <- original_attr
  } else {
    phase(object) <- pvar
    dv(object) <- dvar
    mt(object) <- mvar
  }

  object

}

test_that("main", {
  
  df <- as.data.frame(exampleABC)

  scdf <- suppressMessages(as_scdf(df))
  expect_equal(scan:::check_scdf(scdf), TRUE)
  expect_equal(attributes(scdf)$scdf, attributes(exampleABC)$scdf)
  
  df <- as.data.frame(exampleABC$Rosalind)
  df$case <- NULL
  scdf <- suppressMessages(as_scdf(df))
  expect_equal(scan:::check_scdf(scdf), TRUE)
  
  
  df <- as.data.frame(exampleAB_score)
  scdf <- suppressMessages(as_scdf(df))
  expect_equal(scan:::check_scdf(scdf), TRUE)
  expect_equal(names(df)[-1], names(scdf[[1]]))
  
})

# test_that("warnings", {
#   
#   scdf <- exampleABC
#   l2 <- data.frame(
#     case = c("Marie", "Heinz", "Lise"),
#     gender = c(0,0,2),
#     age = c(23,35,12)
#   )
#   
#   expect_warning(scdf |> add_l2(l2))
#   
#   l2 <- data.frame(
#     name = c("Marie", "Rosalind", "Lise"),
#     gender = c(0,0,2),
#     age = c(23,35,12)
#   )
#   
#   expect_warning(scdf$Marie |> add_l2(l2[1,]))
#   
#})

#' Single-case data generator
#'
#' The \code{random_scdf} function generates random single-case data frames for
#' monte-carlo studies and demonstration purposes. \code{design} is used to set
#' up a design matrix with all parameters needed for the \code{random_scdf}
#' function.
#'
#' The generated data can be normally distributed, Poisson-distributed, or
#' binomially distributed. The default is normally distributed data.
#'
#' @param design A design matrix which is created by \code{design} and specifies
#'   all parameters. If \code{design} is \code{NULL} (default), the design
#'   parameters have to be specified via the \code{...} argument. If a numeric
#'   value is provided instead of a design matrix, it is interpreted as the
#'   number of cases \code{n}.
#' @param round Rounds the scores to the defined decimal. To round to the second
#'   decimal, set \code{round = 2}. Default is \code{NA} (no rounding).
#' @param random_names Is \code{FALSE} by default. If set \code{random_names =
#'   TRUE} cases are assigned random first names. If set \code{"neutral", "male"
#'   or "female"} only gender neutral, male, or female names are chosen. The
#'   names are drawn from the 2,000 most popular names for newborns in 2012 in
#'   the U.S. (1,000 male and 1,000 female names).
#' @param seed A seed number for the random generator. If \code{NULL} (default),
#'   no seed
#' @param ... arguments that are directly passed to the \code{design} function
#'   for a more concise coding.
#' @return A single-case data frame. See \code{\link{scdf}} to learn about this
#'   format.
#' @family mc fucntions
#' @author Juergen Wibert
#' @keywords datagen
#' @export
#' @examples
#'
#' ## Create random single-case data and inspect it
#' design <- design(
#'   n = 3, rtt = 0.75, slope = 0.1, extreme_prop = 0.1,
#'   missing_prop = 0.1
#' )
#' dat <- random_scdf(design, round = 1, random_names = TRUE, seed = 123)
#' describe(dat)
#'
#' ## And now have a look at poisson-distributed data
#' design <- design(
#'   n = 3, B_start = c(6, 10, 14), mt = c(12, 20, 22), start_value = 10,
#'   distribution = "poisson", level = -5, missing_prop = 0.1
#' )
#' dat <- random_scdf(design, seed = 1234)
#' pand(dat, decreasing = TRUE)

random_scdf <- function(design = NULL, 
                        round = NA, 
                        random_names = FALSE, 
                        seed = NULL, 
                        ...) {
  
  if (!is.null(seed)) set.seed(seed)
  if (is.numeric(design)) {
    warn("The first argument is expected to be a design matrix created by ", 
            "design(). If you want to set n, please name the first ",
            "argument with n = ...")
    n <- design
    design <- NULL
  }
  if (is.null(design)) design <- design(...)

  n <- length(design$cases)

  out <- vector("list", n)
  
  for (i in 1:n) {
    
    mt <- sum(design$cases[[i]]$length, na.rm = TRUE)
    start_value <- design$cases[[i]]$start_value[1]
    s <- design$cases[[i]]$s[1]
    
    .rtt <- design$cases[[i]]$rtt[1]
    error <- sqrt(((1 - .rtt) / .rtt) * s^2)
    
    trend <- design$cases[[i]]$trend[1]
    level <- design$cases[[i]]$level
    slope <- design$cases[[i]]$slope
    length <- design$cases[[i]]$length
    missing_prop <- design$cases[[i]]$missing_prop
    extreme_prop <- design$cases[[i]]$extreme_prop
    extreme_low <- design$cases[[i]]$extreme_low
    extreme_high <- design$cases[[i]]$extreme_high
    
    if (design$distribution %in% c("normal", "gaussian")) {
      trend <- trend * s
      slope <- slope * s
      level <- level * s
      extreme_low <- extreme_low * s
      extreme_high <- extreme_high * s
    }
    
    start_values <- c(start_value, rep(0, mt - 1))
    trend_values <- c(0, rep(trend, mt - 1))
    slope_values <- c()
    level_values <- c()
    
    for (j in 1:length(length)) {
      slope_values <- c(slope_values, rep(slope[j], length[j]))
      level_values <- c(level_values, level[j], rep(0, length[j] - 1))
    }
    
    true_values <- start_values + trend_values + slope_values + level_values
    true_values <- cumsum(true_values)
    
    if (design$distribution %in% c("normal", "gaussian")) {
      error_values <- rnorm(mt, mean = 0, sd = error)
      measured_values <- true_values + error_values
    }

    if (design$distribution %in% c("poisson")) {
      true_values[true_values < 0] <- 0
      measured_values <- rpois(length(true_values), lambda = true_values)
    }

    if (design$distribution == "binomial") {
      true_values[true_values < 0] <- 0
      true_values[true_values > 1] <- 1
      measured_values <- rbinom(
          n = length(true_values), 
          size = design$n_trials,
          prob = true_values
      )
    }
    
    if (extreme_prop > 0) {
      .ids <- which(runif(mt) <= extreme_prop)
      .error <- runif(length(.ids), min = extreme_low, max = extreme_high)
      measured_values[.ids] <- measured_values[.ids] + .error
    }

    if (missing_prop > 0) {
      .ids <- sample(1:mt, missing_prop * mt)
      measured_values[.ids] <- NA
    }

    if (!is.na(round)) {
      measured_values <- round(measured_values, round)
    }

    if (design$distribution %in% c("binomial", "poisson")) {
      measured_values[measured_values < 0] <- 0
    }
    
    # fast df assignment
    df <- list(
      phase = rep(design$cases[[i]]$phase, length), 
      values = measured_values, 
      mt = 1:mt
    )
    df$phase <- factor(df$phase, levels = unique(df$phase))
    
    if (design$distribution == "binomial") 
      df$trials <- rep(design$n_trials, mt)
    
    class(df) <- "data.frame"
    attr(df, "row.names") <- .set_row_names(length(df[[1]]))
  
    out[[i]] <- df
    #out[[i]] <- data.frame(
    ##  phase = rep(design$cases[[i]]$phase, length), 
    #  values = measured_values, 
    #  mt = 1:mt
    #)
  }

  if (random_names == "male") names(out) <- sample(case_names$male, n)
  if (random_names == "female") names(out) <- sample(case_names$female, n)
  if (random_names == "neutral") names(out) <- sample(case_names$neutral, n)
  if (isTRUE(random_names)) names(out) <- sample(case_names$all, n)

  attributes(out) <- .default_attributes(attributes(out))
  out
}

#' @rdname print.sc
#' @export
print.sc_desc <- function(x, digits = "auto", ...) {
  
  if (digits == "auto") digits <- 3
  
  cat("Describe Single-Case Data\n\n")
  x$descriptives[-1:-2] <- round(x$descriptives[-1:-2], digits)
  out <- as.data.frame(t(x$descriptives[-1]))
  colnames(out) <- x$descriptives$Case
  
  print(
    out[1:(2 * length(x$design) + 1), , drop = FALSE], 
    digits = digits, ...
  )
  cat("\n")
  print(
    out[-(1:(2 * length(x$design) + 1)), , drop = FALSE], 
    digits = digits, ...
  )
  .note_vars(x)
  
}

# MC Functions ------------------------------------------------------------

.opt$mc_fun <- list(
  #plm_level = function(x) {
  #  res <- plm(x)
  #  coef(res)["phaseB", "Pr(>|t|)"]
  #},
  plm_level = function(x) .plm.mt(x, type = "level p"),
  plm_slope = function(x) .plm.mt(x, type = "slope p"),
  plm_poisson_level = function(x) .plm.mt(x, count.data=TRUE, type = "level p"),
  plm_poisson_slope = function(x) .plm.mt(x, count.data=TRUE, type = "slope p"),
  hplm_level = function(x) {
    res <- summary(hplm(x, random.slopes=FALSE, ICC=FALSE)$hplm)$tTable
    res[3, 5]
  },
  hplm_slope = function(x) {
    res <- summary(hplm(x, random.slopes=FALSE, ICC=FALSE)$hplm)$tTable
    param <- (nrow(res) - 2) / 2
    res[2 + param + 1, 5]
  },
  tauU = function(x) {
    if (length(x) > 1) abort("Use 'tauU_meta' for multiple case designs.", call. = FALSE)
    res <- tau_u(x, method = "tarlow", meta_analyses = FALSE)$table[[1]]
    res[which(row.names(res) == "A vs. B - Trend A"), which(names(res) == "p")]
  },
  tauU_slope = function(x) {
    if (length(x) > 1) abort("Use 'tauU_slope_meta' for multiple case designs.", call. = FALSE)
    res <- tau_u(x, method = "tarlow", meta_analyses = FALSE)$table[[1]]
    res[which(row.names(res) == "A vs. B + Trend B - Trend A"), which(names(res) == "p")]
  },
  tauU_meta = function(x) {
    res <- tau_u(x, method = "tarlow")$Overall_tau_u
    res[which(res$Model == "A vs. B - Trend A"), which(names(res) == "p")]
  },
  tauU_slope_meta = function(x) {
    res <- tau_u(x, method = "tarlow")$Overall_tau_u
    res[which(res$Model == "A vs. B + Trend B - Trend A"), which(names(res) == "p")]
  },
  base_tau = function(x) corrected_tau(x)$p,
  rand = function(x) rand_test(x, number = 100, exclude.equal = "auto", limit = 3)$p.value,
  rand_decrease = function(x) rand_test(
    x, statistic = "Mean A-B", number = 100, exclude.equal = "auto", limit = 3)$p.value,
  rand_slope = function(x) rand_test(
    x, number = 100, statistic = "Slope B-A", exclude.equal = "auto", limit = 3)$p.value,
  rand_slope_decrease = function(x) rand_test(
    x, number = 100, statistic = "Slope B-A", exclude.equal = "auto", limit = 3)$p.value
  
)


.plm.mt <- function(data, 
                    type = "level p", 
                    model = "B&L-B", 
                    dvar = "values", pvar = "phase", mvar = "mt", 
                    count.data = FALSE) {
  n <- length(data)
  if (n > 1) {
    abort("Multiple single-cases are given. Calculations could only be applied ",
         "to a single data set.\n")
  }
  
  if (inherits(data, "list")) data <- data[[1]]
  if (ncol(data) < 3) data[, mvar] <- 1:nrow(data)
  
  y  <- data[, dvar]
  n1 <- sum(data[, pvar] == "A")
  n2 <- sum(data[, pvar] == "B")
  mt <- data[, mvar]
  D  <- c(rep(0, n1), rep(1, n2))
  
  if (model == "H-M") {
    inter <- (mt - mt[n1 + 1]) * D	
  } else if (model == "B&L-B") {
    inter <- (mt - mt[n1]) * D	
  }
  
  if (count.data) {
    full <- glm(I(round(y)) ~ 1 + mt + D + inter, family = "poisson")
  } else full <- lm(y ~ 1 + mt + D + inter)
  
  if (type == "1" || type == "level p")
    return(summary(full)$coef[3, 4])
  if (type == "2" || type == "slope p")
    return(summary(full)$coef[4, 4])
  if (type == "3" || type == "level t") 
    return(summary(full)$coef[3, 3])
  if (type == "4" || type == "slope t")
    return(summary(full)$coef[4, 3])
  if (type == "5" || type == "level B")
    return(summary(full)$coef[3, 1])
  if (type == "6" || type == "slope B")
    return(summary(full)$coef[4, 1])
  if (type == "model")
    return(full)
  abort("Wrong type:", type)
}

#' Percent exceeding the median (PEM)
#' 
#' The \code{pem} function returns the percentage of phase B data exceeding the
#' phase A median.  Additionally, a chi square test against a 50/50
#' distribution is computed.  Different measures of central tendency can be
#' addressed for alternative analyses.
#' 
#' @details The Percent Exceeding the Median is calculated as
#' the percentage of data points in phase B that exceed the median of phase A.
#' If the \code{decreasing} argument is set to \code{TRUE},
#' the percentage of data points in phase B that are below the median of
#' phase A is calculated.  The PEM is expressed as a percentage ranging
#' from 0 to 100.  Higher values indicate a greater degree of
#' improvement from phase A to phase B.
#' 
#' @inheritParams .inheritParams
#' @param binom.test Computes a binomial test for a 50/50 distribution. Default
#' is \code{binom.test = TRUE}.
#' @param chi.test Computes a Chi-square test. The default setting
#' \code{chi.test = FALSE} skips the Chi-square test.
#' @param FUN Data points are compared with the phase A median. Use this
#' argument to implement alternative measures of central tendency. Default is
#' \code{FUN = median}.
#' @param \dots Additional arguments for the \code{FUN} parameter (e.g.
#' \code{FUN = mean, trim = 0.1} will use the 10 percent trimmed arithmetic
#' mean instead of the median for comparisons). The function must take a vector
#' of numeric values and the \code{na.rm} argument and return a numeric value.
#' @author Juergen Wilbert
#' @family overlap functions
#' @examples
#' 
#' ## Calculate the PEM including the Binomial and Chi-square tests for a single-case
#' dat <- random_scdf(5, level = 0.5)
#' pem(dat, chi.test = TRUE)
#' 
#' @export
pem <- function(data, dvar, pvar, 
                decreasing = FALSE, 
                binom.test = TRUE, 
                chi.test = FALSE, 
                FUN = median, 
                phases = c(1, 2), 
                ...) {

  # set default attributes
  if (missing(dvar)) dvar <- dv(data)
  if (missing(pvar)) pvar <- phase(data)
  dv(data) <- dvar
  phase(data) <- pvar
  
  data <- .prepare_scdf(data, na.rm = TRUE)
  data <- recombine_phases(data, phases = phases)$data
  
  N <- length(data)
  
  PEM       <- rep(NA, N)
  chi       <- rep(NA, N)
  chi.df    <- rep(NA, N)
  chi.p     <- rep(NA, N)
  binom.p   <- rep(NA, N)
  positives <- rep(NA, N)
  total     <- rep(NA, N)
  
  
  for(i in 1:N) {
    A <- data[[i]][, dvar][data[[i]][, pvar] == "A"]
    B <- data[[i]][, dvar][data[[i]][, pvar] == "B"]
    if (!decreasing)
      PEM[i] <- mean(B > FUN(A,...)) * 100
    if (decreasing)
      PEM[i] <- mean(B < FUN(A,...)) * 100
    if(binom.test) {
      nB <- length(B)
      bi <- binom.test(round(PEM[i] / 100  * nB), nB, alternative = "greater")
      positives[i] <- bi$statistic
      total[i]     <- bi$parameter
      binom.p[i]   <- bi$p.value
    }
    if(chi.test) {
      nB <- length(B)
      exceeding <- PEM[i] / 100  * nB
      res <- chisq.test(c(exceeding, nB - exceeding), p = c(0.5, 0.5))
      chi[i]    <- res$statistic
      chi.df[i] <- res$parameter
      chi.p[i]  <- res$p.value
    }
  }
  stats.ma <- cbind(positives, total,binom.p)
  colnames(stats.ma) <- c("positives","total","binom.p")
  rownames(stats.ma) <- names(data)
  if(chi.test) {
    cn <- c(colnames(stats.ma),"Chi", "DF", "p")
    stats.ma <- cbind(stats.ma, chi, chi.df, chi.p)
    colnames(stats.ma) <- cn
  }
  
  PEM <- cbind(Case = revise_names(data), PEM = PEM, as.data.frame(stats.ma))
  row.names(PEM) <- NULL
  out <- list(PEM = PEM, test = stats.ma, decreasing = decreasing)
  class(out) <- c("sc_pem")
  attributes(out)[opts("phase", "dv")] <- list(pvar, dvar)
  out
}

#' @rdname transform.scdf
#' @param lag Number of values surrounding a value to calculate the average
#' @export
moving_median <- function(x, lag = 1) .moving_average(x, lag, median)

#' @rdname transform.scdf
#' @param lag Number of values surrounding a value to calculate the average
#' @export

moving_mean <- function(x, lag = 1) .moving_average(x, lag, mean)

#' @rdname transform.scdf
#' @param f the proportion of surrounding data influencing each data point.
#' @param mt A vector with measurement times.
#' @export
local_regression <- function(x, mt = 1:length(x), f = 0.2) {
  lowess(x ~ mt, f = f)$y
}

#' @export
#' @rdname transform.scdf
#' @param x A vector
#' @param first_of A logical vector
#' @param positions  A numeric vector with relative positions to the first 
#'  appearance of a TRUE value in x.
set_na_at <- function(x, first_of, positions = 0) {
  x[match(TRUE, first_of) + positions] <- NA
  x
}

#' @export
#' @rdname transform.scdf
#' @param x A vector.
#' @param at A logical vector. E.g. `phase == "A"`. The first TRUE value of that
#'   vector is the target position for centring. By default, this is the first
#'   position.
#' @param shift A value indicating a shift in measurement times for centring.
#'   E.g. `shift = 4` will centre four measurement-times after the position
#'   defined by the `at` and `part` arguments.
#' @param part A numeric value between 0 and 1. `0` refers to the first `TRUE`
#'   in the `at` vector, `1` to the last, and `0.5` to the midpoint of the
#'   sequence of `TRUE` values. E.g. if you want to centre at the middle of
#'   phase A, set `at = phase == A, part = 0.5`. Note: decimals are rounded to
#'   integers.

center_at <- function(x, at = TRUE, shift = 0, part = 0) {
  x - x[match(TRUE, at) + shift + round(sum(at, na.rm = TRUE) * part)]
}


#' @export
#' @rdname transform.scdf
#' @param x A logical vector.
#' @param positions A numeric vector with relative positions to the first 
#'  appearance of a TRUE value in x.

first_of <- function(x, positions = 0) {
  .deprecated_warning("set_na_at", "first_of")
  match(TRUE, x) + positions
}

#' @export
#' @rdname transform.scdf
across_cases <- function(...) {
  # a helper function
}

#' @export
#' @rdname transform.scdf
all_cases <- function(...) {
  # a helper function
}

#' @export
#' @rdname transform.scdf
rowwise <- function(...) {
  # a helper function
}

#' @rdname print.sc
#' @export
print.sc_overlap <- function(x, digits = "auto", ...) {
  
  if (digits == "auto") {
    digits_1 <- 0
    digits_2 <- 2
  } else {
    digits_1 <- digits
    digits_2 <- digits
  }
  
  cat("Overlap Indices\n\n")
  #cat("Design: ", x$design, "\n")
  cat(.phases_string(x$phases.A, x$phases.B),"\n\n")
  
  vars <- c("PND", "PEM", "PET", "NAP", "NAP rescaled", "PAND")
  x$overlap[, vars] <- round(x$overlap[, vars], digits_1)
  
  vars <- c(
    "IRD", "Tau_U(A)", "Tau_U(BA)", "Base_Tau", "Diff_mean", 
    "Diff_trend", "SMD", "Hedges_g"
  )
  x$overlap[, vars] <- round(x$overlap[, vars], digits_2)
  
  out <- as.data.frame(t(x$overlap[-1]))
  colnames(out) <- x$overlap$Case
  
  print(out, ...)
  .note_vars(x)
}

#' Smoothing single-case data
#'
#' *This function is superseded by the more versatile 
#' \code{\link{transform.scdf}} function.*
#' The \code{smooth_cases} function provides procedures to smooth single-case 
#' data (i.e., to eliminate noise). A moving average function (mean- or
#' median-based) replaces each data point by the average of the surrounding
#' data points step-by-step. With a local regression function, each data point
#' is regressed by its surrounding data points.
#'
#' \code{moving_median}, \code{moving_mean}, and \code{local_regression} are 
#' helper function for \code{\link{transform.scdf}} returning the smoothed 
#' values of a numeric vector.
#'
#' @inheritParams .inheritParams
#' @param method,FUN Method determining the smoothed scores. Default \code{method =
#' "median"} is a moving median function. Further possible values are:
#' \code{"mean"} and a non-parametric \code{"regression"}.
#' @param intensity For \code{method = "median"} and \code{"mean"} it
#' is the lag used for computing the average. Default is \code{intensity = 1}.
#' In case of \code{method = "regression"} it is the proportion of
#' surrounding data influencing each data point, which is \code{intensity =
#' 0.2} by default.
#' @return Returns a data frame (for each single-case) with smoothed data
#' points. See \code{\link{scdf}} to learn about the format of these data
#' frames.
#' @author Juergen Wilbert
#' @family data manipulation functions
#' @keywords internal
#' @examples
#'
#' ## Use the three different smoothing functions and compare the results
#' study <- c(
#'   "Original" = Huber2014$Berta,
#'   "Moving median" = smooth_cases(Huber2014$Berta, method = "median"),
#'   "Moving mean" = smooth_cases(Huber2014$Berta, method = "mean"),
#'   "Local regression" = smooth_cases(Huber2014$Berta, method = "regression")
#' )
#' plot(study)
#' 
#' Huber2014$Berta |>
#' transform(
#'   "compliance (moving median)" = moving_median(compliance),
#'   "compliance (moving mean)" = moving_mean(compliance),
#'   "compliance (local regression)" = local_regression(compliance, mt)
#' )
#' 
#' @export
smooth_cases <- function(data, dvar, mvar, 
                         method = "mean", 
                         intensity = NULL, 
                         FUN = NULL) {

  if (!is.null(FUN)) method <- FUN
  
  # set attributes to arguments else set to defaults of scdf
  if (missing(dvar)) dvar <- dv(data)
  if (missing(mvar)) mvar <- mt(data) 

  dv(data) <- dvar
  mt(data) <- mvar
  
  data <- .prepare_scdf(data)
  copy_attributes <- attributes(data)
  copy_names <- names(data)
  if (method %in% c("mean", "movingMean")) {
    if (is.null(intensity)) intensity <- 1
    out <- lapply(data, function(x) {
      x[, dvar] <- .moving_average(x[[dvar]], intensity, mean)
      x
    })
  }
  if (method %in% c("median", "movingMedian")) {
    if (is.null(intensity)) intensity <- 1
    out <- lapply(data, function(x) {
      x[, dvar] <- .moving_average(x[[dvar]], intensity, median)
      x
    })
  }
  if (method %in% c("regression", "localRegression")) {
    if (is.null(intensity)) intensity <- 0.2
    out <- lapply(data, function(x) {
      xval <- x[[mvar]][!is.na(x[[dvar]])]
      yval <- x[[dvar]][!is.na(x[[dvar]])]
      x[, dvar] <- lowess(yval ~ xval, f = intensity)$y
      x
    })
  }
  attributes(out) <- copy_attributes
  names(out) <- copy_names
  out
}

#' Single-case example data sets
#'
#' @docType data
#' @usage data(exampleAB, package = "scan")
#' @description The [scan] package comes with a set of fictitious and authentic
#' single-case study data. These data sets can be used to practice single-case
#' data analysis and to reproduce results from the respective publications.
#'
#' - **Beretvas2008** — Fictitious single-case intervention study from Beretvas & Chung, 2008.
#' - **Borckardt2014** — Fictitious daily pain ratings evaluating a psychological treatment from Borckardt & Nash, 2014.
#' - **byHeart2011** — Multiple-baseline (11 cases) flash card vocabulary learning (Wilbert, unpublished).
#' - **example_A24** — Number of injuries on a German autobahn before and after implementation of a speedlimit (130km/h) (Ministerium fuer Infrastruktur und Landesplanung. Land Brandenburg).
#' - **exampleA1B1A2B2** — Fictitious A1-B1-A2-B2 example dataset.
#' - **exampleA1B1A2B2_zvt** — Non-fictitious A1-B1-A2-B2 example with ZVT (intelligence measure) and D2 (concentration measure) scores.
#' - **exampleAB** — Fictitious AB example dataset with three cases.
#' - **exampleAB_50** — Fictitious AB example dataset (50 cases)s.
#' - **exampleAB_50.l2** — Level-2 data for **exampleAB_50**.
#' - **exampleAB_add** — Fictitious AB example with added covariates.
#' - **exampleAB_decreasing** — Fictitious AB example with an expected decreasing effect in the intervention phase.
#' - **exampleAB_mpd** — Fictitious example with different phase structures for each case.
#' - **exampleAB_score** — Fictitious AB example with binomial distributed score outcome.
#' - **exampleAB_simple** — Simple fictitious AB example with three cases.
#' - **exampleABAB** — Fictitious ABAB reversal design example.
#' - **exampleABC** — Fictitious ABC example dataset.
#' - **exampleABC_150** — Fictitious ABC example (150 cases).
#' - **exampleABC_50** — Fictitious ABC example (50 cases).
#' - **exampleABC_outlier** — Fictitious ABC example with outlier.
#' - **example_atd** — Fictitious AB alternating treatment design.
#' - **example_stranger** - Example for screen time of Stranger Things characters.
#' - **Grosche2011** — Multiple-baseline (three cases) from a direct-instructive reading intervention (Grosche, 2011).
#' - **Grosche2014** — Multiple-baseline (3×3 materials) reading intervention (Grosche, Lueke, & Wilbert, unpublished).
#' - **GruenkeWilbert2014** — Multiple-baseline (six cases) from a story mapping reading intervention (Gruenke, Wilbert, & Stegemann-Calder, 2013).
#' - **Huber2014** — Multiple-baseline (four cases) with DBR ratings from a behavioral compliance intervention (Huber, unpublished).
#' - **Huitema2000** — Fictitious single-case intervention study (Huitema & McKean, 2000).
#' - **Lenz2013** — Fictitious example (Lenz, 2013).
#' - **Leidig2018** — Multiple-baseline good behavior game intervention (Leidig et al., 2022).
#' - **Leidig2018_l2** — Level-2 data for **Leidig2018** (Leidig et al., 2022).
#' - **Parker2007** — Example dataset after Parker et al. (2007).
#' - **Parker2009** — Example dataset after Parker et al. (2009).
#' - **Parker2009b** — Example dataset after Parker & Vannest (2009).
#' - **Parker2011** — Example dataset after Parker et al. (2011).
#' - **Parker2011b** — Example from Parker, Vannest, & Davis (2011).
#' - **SSDforR2017** — Example from the R package SSDforR.
#' - **Tarlow2017** — Fictitious single-case intervention study (Tarlow, 2017).
#' - **Waddell2011** — Fictitious single-case intervention study (Waddell, Nassar, & Gustafson, 2011).
#'
#' @references Beretvas, S., & Chung, H. (2008). An evaluation of modified
#' R2-change effect size indices for single-subject experimental designs.
#' *Evidence-Based Communication Assessment and Intervention, 2*, 120–128.
#'
#' Borckardt, J. J., & Nash, M. R. (2014). Simulation modelling analysis for
#' small sets of single-subject data collected over time. *Neuropsychological
#' Rehabilitation, 24*, 492–506.
#'
#' Gruenke, M., Wilbert, J., & Stegemann-Calder, K. (2013). Analyzing the
#' effects of story mapping on the reading comprehension of children with low
#' intellectual abilities. *Learning Disabilities: A Contemporary Journal, 11*,
#' 51–64.
#'
#' Grosche, M. (2011). Effekte einer direkt-instruktiven Foerderung der
#' Lesegenauigkeit. *Empirische Sonderpaedagogik, 3*, 147–161.
#'
#' Huitema, B. E., & McKean, J. W. (2000). Design specification issues in
#' time-series intervention models. *Educational and Psychological Measurement,
#' 60*, 38–58.
#'
#' Lenz, A. S. (2013). Calculating Effect Size in Single-Case Research: A
#' Comparison of Nonoverlap Methods. *Measurement and Evaluation in Counseling
#' and Development, 46*(1), 64–73.
#'
#' Leidig, T., Casale, G., Wilbert, J., Hennemann, T., Volpe, R. J., Briesch,
#' A., & Grosche, M. (2022). Individual, generalized, and moderated effects of
#' the good behavior game on at-risk primary school students: A multilevel
#' multiple baseline study using behavioral progress monitoring. *Frontiers in
#' Education, 7*. https://www.frontiersin.org/articles/10.3389/feduc.2022.917138
#'
#' Parker, R. I., Hagan-Burke, S., & Vannest, K. (2007). Percentage of All
#' Non-Overlapping Data (PAND) An Alternative to PND. The Journal of Special
#' Education, 40(4), 194-204.
#'
#' Parker, R. I., Vannest, K. J., & Brown, L. (2009). The improvement rate
#' difference for single-case research. Exceptional Children, 75(2), 135-150.
#'
#' Parker, R. I., & Vannest, K. (2009). An improved effect size for single-case
#' research: Nonoverlap of all pairs. Behavior Therapy, 40(4), 357-367.
#'
#' Parker, R. I., Vannest, K. J., Davis, J. L., & Sauber, S. B. (2011).
#' Combining Nonoverlap and Trend for Single-Case Research: Tau-U. *Behavior
#' Therapy, 42*(2), 284–299. https://doi.org/10.1016/j.beth.2010.08.006
#'
#' Parker, R. I., Vannest, K. J., & Davis, J. L. (2011). Effect Size in
#' Single-Case Research: A Review of Nine Nonoverlap Techniques. Behavior
#' Modification, 35(4), 303-322. https://doi.org/10.1177/0145445511399147
#'
#' Tarlow, K. R. (2017). An Improved Rank Correlation Effect Size Statistic for
#' Single-Case Designs: Baseline Corrected Tau. *Behavior Modification, 41*(4),
#' 427–467. https://doi.org/10.1177/0145445516676750
#'
#' Waddell, D. E., Nassar, S. L., & Gustafson, S. A. (2011). Single-Case Design
#' in Psychophysiological Research: Part II: Statistical Analytic Approaches.
#' *Journal of Neurotherapy, 15*, 160–169.
#'
#' @aliases Beretvas2008 Borckardt2014 Huitema2000 Waddell2011 byHeart2011
#'   Grosche2011 Grosche2014 GruenkeWilbert2014 Huber2014 example_A24
#'   exampleAB_simple exampleAB exampleAB_add exampleAB_decreasing
#'   exampleAB_score exampleABC exampleABC_outlier exampleABC_50 exampleABAB
#'   exampleA1B1A2B2 exampleA1B1A2B2_zvt exampleAB_50 exampleAB_50.l2
#'   exampleABC_150 exampleAB_mpd example_stranger example_atd Lenz2013
#'   Leidig2018 Leidig2018_l2 SSDforR2017 Parker2011 Parker2007 Parker2009
#'   Parker2009b Parker2011b Tarlow2017
#' @author Juergen Wilbert
#' @keywords datasets
#' @name example_data_sets
NULL

#' Generate a single-case design matrix for multiple random single-cases
#'
#' Generates a parameter list used for generating multiple random single-cases.
#' This is used within the `random_scdf` function and the `power_test` function
#' and for other Monte-Carlo tasks.
#'
#' @param n Number of cases to be designed (Default is `n = 1`).
#' @param phase_design,phase.design A list defining the length and label of each
#'   phase. E.g., `phase_design = list(A1 = 10, B1 = 10, A2 = 10, B2 = 10)`. Use
#'   vectors if you want to define different values for each case `phase_design
#'   = list(A = c(10, 15), B = c(10, 15)`.
#' @param mt,MT Number of measurements (in each study). Default is `mt = 20`.
#' @param B_start,B.start Phase B starting point. The default setting `B_start =
#'   6` would assign the first five scores (of each case) to phase A, and all
#'   following scores to phase B. To assign different starting points for a set
#'   of multiple single-cases, use a vector of starting values (e.g., `B_start =
#'   c(6, 7, 8)`). If the number of cases exceeds the length of the vector,
#'   values will be recycled.
#' @param start_value,m Starting value at the first measurement. Default is
#'   `50`. When `distribution = "poission"` the start_value represents
#'   frequency. When `distribution = "binomial"` start_value must range between
#'   0 and 1 and they represent the probability of on event. To assign different
#'   start values to several single-cases, use a vector of values (e.g. `c(50,
#'   42, 56)`). If the number of cases exceeds the length of the vector, values
#'   are recycled. The `m` argument is deprecated.
#' @param s Standard deviation used to calculate absolute values from level,
#'   slope, trend effects and to calculate and error distribution from the `rtt`
#'   values. Set to `10` by default. To assign different variances to several
#'   single-cases, use a vector of values (e.g. `s = c(5, 10, 15)`). If the
#'   number of cases exceeds the length of the vector, values are recycled. if
#'   the distribution is 'poisson' or 'binomial' s is not applied.
#' @param n_trials If `distribution` (see below) is `"binomial"`, `n_trials` is
#'   the number of trials/observations/items. E.g., if you simulate accuracy
#'   data with 10 items per measurement, set `n_trials = 10`. To assign
#'   different n_trials to several single-cases, use a vector of values (e.g.
#'   `n_trials = c(10, 15, 20)`). If the number of cases exceeds the length of
#'   the vector, values are recycled.
#' @param trend Defines the effect size of a trend added incrementally to each
#'   measurement across the whole data-set. To assign different trends to
#'   several single-cases, use a vector of values (e.g. `trend = c(.1, .3,
#'   .5)`). If the number of cases exceeds the length of the vector, values are
#'   recycled. When using a 'gaussian' distribution, the `trend` parameters
#'   indicate effect size
#'   *d* changes.  When using a binomial or poisson distribution, `trend`
#'   indicates an increase in points / counts per measurement.
#' @param level A list that defines the level increase (effect size *d*) at the
#'   beginning of each phase relative to the previous phase (e.g. `list(A = 0, B
#'   = 1)`). The first element must be zero as the first phase of a single-case
#'   has no level effect (if you have one less list element than the number of
#'   phases, scan will add a leading element with 0 values). Use vectors to
#'   define variable level effects for each case (e.g. `list(A = c(0, 0), B =
#'   c(1, 2))`). When using a 'gaussian' distribution, the `level` parameters
#'   indicate effect size *d* changes. When using a binomial or poisson
#'   distribution, `level` indicates an increase in points / counts with the
#'   onset of each phase.
#' @param slope A list that defines the increase per measurement for each phase
#'   compared to the previous phase. `slope = list(A = 0, B = .1)` generates an
#'   incremental increase of 0.1 per measurement starting at the B phase. The
#'   first list element must be zero as the first phase of a single-case has no
#'   slope effect (if you have one less list element than the number of phases,
#'   scan will add a leading element with 0 values). Use vectors to define
#'   variable slope effects for each case (e.g. `list(A = c(0, 0), B = c(0.1,
#'   0.2))`). If the number of cases exceeds the length of the vector, values
#'   are recycled. When using a 'gaussian' distribution, the `slope` parameters
#'   indicate effect size *d* changes per measurement. When using a binomial or
#'   poisson distribution, `slope` indicates an increase in points / counts per
#'   measurement.
#' @param rtt Reliability of the underlying simulated measurements. Set `rtt =
#'   .8` by default. To assign different reliabilities to several single-cases,
#'   use a vector of values (e.g. `rtt = c(.6, .7, .8)`). If the number of cases
#'   exceeds the length of the vector, values are repeated. `rtt` has no effect
#'   when you're using binomial or poisson distributions.
#' @param extreme_prop,extreme.p Probability of extreme values. `extreme.p =
#'   .05` gives a five percent probability of an extreme value. A vector of
#'   values assigns different probabilities to multiple cases. If the number of
#'   cases exceeds the length of the vector, values are recycled.
#' @param extreme_range,extreme.d Range for extreme values. `extreme_range =
#'   c(-7,-6)` uses extreme values within a range of -7 and -6 . In case of a
#'   binomial or poisson distribution, `extreme_range` indicates frequencies. In
#'   case of a gaussian (or normal) distribution it indicates effect size d.
#'   Caution: the first value must be smaller than the second, otherwise the
#'   procedure will fail.
#' @param missing_prop,missing.p Portion of missing values. `missing_prop = 0.1`
#'   creates 10\% of all values as missing). A vector of values assigns
#'   different probabilities to multiple cases. If the number of cases exceeds
#'   the length of the vector, values are repeated.
#' @param distribution Distribution of the criteria varible. Default is
#'   `"normal"`. Possible values are `"normal"`, `"binomial"`, and `"poisson"`.
#' @param random_start_value If TRUE, the start_values are randomized based on
#'   the distribution.
#' @return An object of class sc_design.
#' @family mc functions
#' @author Juergen Wibert
#' @keywords datagen
#' @examples
#'  ## Create random single-case data and inspect it
#'  design <- design(
#'    n = 3, rtt = 0.75, slope = 0.1, extreme_prop = 0.1,
#'    missing_prop = 0.1
#'  )
#'  dat <- random_scdf(design, round = 1, random.names = TRUE, seed = 123)
#'  describe(dat)
#'
#'  ## And now have a look at poisson-distributed data
#'  design <- design(
#'    n = 3, B_start = c(6, 10, 14), mt = c(12, 20, 22), start_value = 10,
#'    distribution = "poisson", level = -5, missing_prop = 0.1
#'  )
#'  dat <- random_scdf(design, seed = 1234)
#'  pand(dat, decreasing = TRUE)
#' @export
design <- function(n = 1, 
                   phase_design = list(A = 5, B = 15),
                   trend = 0, 
                   level = list(0), 
                   slope = list(0),
                   start_value = 50,
                   s = 10,
                   rtt = 0.80, 
                   extreme_prop = list(0), 
                   extreme_range = c(-4, -3),
                   missing_prop = 0, 
                   distribution = c("normal", "gaussian", "poisson", 
                                    "binomial"),
                   random_start_value = FALSE,
                   n_trials = NULL,
                   mt = NULL, 
                   B_start = NULL,
                   m,
                   phase.design,
                   MT,
                   B.start,
                   extreme.p,
                   extreme.d,
                   missing.p) {
  
  ## take over deprecated variables ----
  
  if (!missing(m)) start_value <- m
  if (!missing(phase.design)) phase_design <- phase.design
  if (!missing(MT)) mt <- MT
  if (!missing(extreme.p)) extreme_prop <- extreme.p
  if (!missing(extreme.d)) extreme_range <- extreme.d
  if (!missing(missing.p)) missing_prop <- missing.p
  if (!missing(B.start)) B_start <- B.start
  
  out <- list()
  attr(out, "call") <- mget(names(formals()), sys.frame(sys.nframe()))
  
  if (is.list(start_value)) start_value <- unlist(start_value)
  if (is.list(trend)) trend <- unlist(trend)
  if (is.list(s)) s <- unlist(s)
  if (is.list(rtt)) rtt <- unlist(rtt)
  if (!is.list(phase_design)) phase_design <- as.list(phase_design)
    
  ## intial checks ----
  check_args(
    as_deprecated("m" = "start_value", "phase.design" = "phase_design", 
                  "MT" = "mt", "B.start" = "B_start", 
                  "extreme.p" = "extreme_prop", "extreme.d" = "extreme_range", 
                  "missing.p" = "missing_prop"),
    by_call(distribution),
    not(distribution == "binomial" && is.null(n_trials),
        "Binomial distributions but n_trials not defined."),
    not(distribution=="binomial" && (any(start_value>1) || any(start_value<0)), 
        "Binomial distributions but start_values outside [0,1]."),
    not(any(B_start < 1) && any(B_start >= 1), 
        "B_start with values below and above 1."),
    not(random_start_value && length(start_value) > 1,
            "Multiple start_values are given when random_start_value ",
            "is set to TRUE."),
    within(extreme_prop, 0, 1),
    within(missing_prop, 0, 1)
  )
  
  ## B_start ----
  if (!is.null(B_start)) {
    mt <- rep(mt, length.out = n)
    if (B_start[1] == "rand") {
      tmp_start <- round(as.numeric(B_start[2]) * mt)
      tmp_end <- round(as.numeric(B_start[3]) * mt)
      B_start <- round(runif(n, tmp_start, tmp_end))
    }

    if (B_start[1] < 1 && B_start[1] > 0) B_start <- round(B_start * mt) + 1
    B_start <- rep(B_start, length.out = n)

    phase_design <- rep(list(A = rep(NA, n), B = rep(NA, n)))
    for (i in 1:length(B_start)) {
      phase_design$A[i] <- B_start[i] - 1
      phase_design$B[i] <- 1 + mt[i] - B_start[i]
    }
  }

  ## start values ----
  if (random_start_value) {
    if (distribution %in% c("normal", "gaussian")) {
      start_value <- rnorm(n, start_value[1], s[1])
    }
    if (distribution == "poisson") {
      start_value <- rpois(n, start_value[1])
    }
    if (distribution == "binomial") {
      start_value <- rbinom(n, n_trials, start_value[1])
    }
  }
  
  ## recycle arguments ----
  if (length(start_value) != n) start_value <- rep(start_value, length = n)
  if (length(s) != n) s <- rep(s, length = n)
  if (length(rtt) != n) rtt <- rep(rtt, length = n)
  if (is.list(trend)) trend <- unlist(trend)

  
  trend <- .check_design(trend, n)
  level <- .check_design(level, n)
  slope <- .check_design(slope, n)
  phase_design <- .check_design(phase_design, n)

  if (length(extreme_prop) != n) {
    extreme_prop <- lapply(numeric(n), function(y) unlist(extreme_prop))
  }
  if (length(extreme_range) != n) {
    extreme_range <- lapply(numeric(n), function(y) unlist(extreme_range))
  }
  if (length(missing_prop) != n) {
    missing_prop <- lapply(numeric(n), function(y) unlist(missing_prop))
  }

  out$cases <- vector("list", n)
  out$distribution <- distribution
  out$n_trials <- n_trials
  
  ## create design case by case ----
  for (case in 1:n) {
    design <- list()
    design$phase <- names(phase_design)
    design$length <- unlist(lapply(phase_design, function(x) x[case]))
    design$rtt <- rtt[[case]]
    design$missing_prop <- missing_prop[[case]]
    design$extreme_prop <- extreme_prop[[case]]
    design$extreme_low <- extreme_range[[case]][1]
    design$extreme_high <- extreme_range[[case]][2]
    design$trend <- trend[[1]][case]
    design$level <- .design_effect(level, case, length(phase_design))
    design$slope <- .design_effect(slope, case, length(phase_design))
    design$start_value <- start_value[[case]]
    design$s <- s[[case]]

    design$start <- c(1, cumsum(design$length) + 1)[1:length(design$length)]
    design$stop <- cumsum(design$length)

    out$cases[[case]] <- design
  }
  
  class(out) <- c("sc_design")
  
  out
}

.design_effect <- function(effects, case, phase_length) {
  
  case_effects <- unlist(lapply(effects, function(x) x[case]))
  
  if (identical(case_effects, 0)) case_effects <- rep(0, phase_length)
  
  if (length(case_effects) == phase_length && case_effects[1] != 0) {
    warn("Effect for first phase is not 0. Looks like a missspecification")
  }  
    
  if (length(case_effects) == phase_length - 1) 
    case_effects <- c(0, case_effects)
  
  if (length(case_effects) != phase_length) {
    warn(
      as.character(match.call()[2]), ": the wrong number of phase ",
      "effects defined. Should be ", phase_length, " is ", 
      length(case_effects)
    )
  }  
  
  case_effects
  
}

.check_design <- function(data, n) {
  if (is.numeric(data)) data <- list(data)
  for (phase in 1:length(data)) {
    if (length(data[[phase]]) != n) {
      data[[phase]] <- rep(data[[phase]], length = n)
    }
  }
  data
}

test_that("design function returns expected output", {
  # Test case 1
  result <- design()
  expect_s3_class(result, "sc_design")
  expect_equal(length(result$cases), 1)
  expect_equal(result$distribution, "normal")
  
  # Test case 2
  result <- design(n = 3, level = list(0, 1))
  expect_s3_class(result, "sc_design")
  expect_equal(length(result$cases), 3)
  expect_equal(result$distribution, "normal")
  expect_equal(result$cases[[1]]$level, c(0, 1))
  expect_equal(result$cases[[2]]$level, c(0, 1))
  
  # Test case 3
  result <- design(n = 2, trend = c(0, 1), start_value = c(50, 60))
  expect_s3_class(result, "sc_design")
  expect_equal(length(result$cases), 2)
  expect_equal(result$distribution, "normal")
  expect_equal(result$cases[[1]]$trend, c(0))
  expect_equal(result$cases[[1]]$start_value, 50)
  expect_equal(result$cases[[2]]$trend, c(1))
  expect_equal(result$cases[[2]]$start_value, 60)
  
})

#' Load single-case data from files
#'
#' Use the `read_scdf` function to load single-case data csv, excel, or yaml
#' files.
#' 
#' The data file must be in a "long" format with one column for case
#' identifiers, one column for phase identifiers, one column for measurement
#' time, and one column for the dependent variable values. Each row represents
#' a single measurement.
#'
#' @param file Either a character string defining the file to be loaded (e.g.
#'   `"SC_Anita.csv"` (if left empty a dialog box for choosing will be opened)
#'   or a data.frame.
#' @param cvar Sets the variable name of the "case" variable. Defaults to
#'   `"case"`.
#' @param pvar Sets the variable name of the "phase" variable. Defaults to
#'   `"phase"`.
#' @param dvar Sets the variable name of the "values" variable. Defaults to
#'   `"values"`.
#' @param mvar Sets the variable name of the "mt" variable. Defaults to `"mt"`.
#' @param sort_cases,sort.labels If set TRUE, the resulting list is sorted by
#'   label names (alphabetically increasing).
#' @param phase_names,phase.names A character vector with phase names. Defaults
#'   to the phase names provided in the phase variable.
#' @param type Format of the file to be loaded. Either "csv", "xlsx", "xls",
#'   "excel", "yml" is possible. By default (NA) the type is extracted from the
#'   file extension.
#' @param na Character vector of strings to interpret as missing values.
#' @param \dots Further arguments passed to the respective read function.
#' @return Returns a single-case data frame. See \code{\link{scdf}} to learn
#'   about the format of these data frames.
#' @author Juergen Wilbert
#' @seealso [read.table()], [readRDS()]
#' @family io-functions
#' @keywords io
#' @examples
#'
#' ## Read SC-data from a file named "study1.csv" in your working directory
#' # study1 <- read_scdf("study1.csv")
#'
#' ## Read SC-data from a .csv-file with semicolon as field and comma as decimal separator
#' # study2 <- read_scdf("study2.csv", sep = ";", dec = ",")
#'
#' ## write_scdf and read_scdf
#' filename <- file.path(tempdir(), "test.csv")
#' write_scdf(exampleA1B1A2B2_zvt, filename)
#' dat <- read_scdf(filename, cvar = "case", pvar = "part", dvar = "zvt", mvar = "day")
#' res1 <- describe(exampleA1B1A2B2_zvt)$descriptives
#' res2 <- describe(dat)$descriptives
#' all.equal(res1,res2)
#'
#' @export
read_scdf <- function(file, 
                      cvar = "case", 
                      pvar = "phase", 
                      dvar = "values", 
                      mvar = "mt",
                      sort_cases = FALSE,
                      phase_names = NULL, 
                      type = NA, 
                      na = c("", "NA"),
                      sort.labels = NULL, 
                      phase.names = NULL, 
                      ...) {
  
  if (missing(file)) {
    file <- file.choose()
    cat("Load file", file, "\n\n")
  }
  
  if(!is.null(phase.names)) phase_names <- phase.names
  if(!is.null(sort.labels)) sort_cases <- sort.labels
  
  if (!inherits(file, "character")) {
    type <- "data"
  }
  
  if (is.na(type)) {
    type <- tolower(tools::file_ext(file))#substring(file, regexpr("\\.([[:alnum:]]+)$", file) + 1)
  }
  
  if (type == "csv") {
    dat <- utils::read.csv(
      file, 
      header = TRUE,
      stringsAsFactors = FALSE,
      na.strings = na,
      ...
    )
  }
  
  if (type %in% c("yml", "yaml")) return(.load_yml(file, ...))
  
  if (type %in% c("excel", "xlsx", "xls")) {
    dat <- as.data.frame(readxl::read_excel(file, na = na,...))
  }

  if (type == "data") {
    dat <- as.data.frame(file)
  }
 
  out <- as_scdf(
    object = dat, 
    sort_cases = sort_cases, 
    cvar = cvar, 
    pvar = pvar, 
    dvar = dvar, 
    mvar = mvar, 
    phase_names = phase_names   
  )
  
  message("Imported ", length(out), " cases")
  out
}

#' @rdname deprecated-functions
#' @export
readSC.excel <- function(...) {
  
  read_scdf(..., type = "excel")
  
}

#' @rdname deprecated-functions
#' @export
readSC <- function(...) {
  read_scdf(...)
}

.load_yml <- function(filename, ...) {
  
  if (!requireNamespace("yaml", quietly = TRUE)) {
    abort("You need to install the 'yaml' package first (install.packages('yaml')")
  }
  
  out <- yaml::read_yaml(filename, ...)
  
  case_names <- names(out)
  
  extract <- function(x) {
    
    if (is.null(x[["dvar"]])) x[["dvar"]] <- "values"
    if (is.null(x[["mvar"]])) x[["mvar"]] <- "mt"
    if (is.null(x[["pvar"]])) x[["pvar"]] <- "phase"
    
    dvar <- x[["dvar"]]
    pvar <- x[["pvar"]]
    mvar <- x[["mvar"]]
    
    x[["dvar"]] <- NULL
    x[["mvar"]] <- NULL
    x[["pvar"]] <- NULL
    
    phase <- numeric(0)
    values <- numeric(0)
    
    for(i in seq_along(x[[dvar]])) {
      phase <- c(
        phase, rep(names(x[[dvar]][i]), length(x[[dvar]][[i]])))
      values <- c(values, x[[dvar]][[i]])
    }
    
    x[[pvar]] <- phase
    x[[dvar]] <- values
    
    if (is.null(x[[mvar]])) x[[mvar]] <- 1:length(x[[dvar]])
    
    x <- list(as.data.frame(x))
    scdf_attr(x, "var.phase") <- pvar
    scdf_attr(x, "var.values") <- dvar
    scdf_attr(x, "var.mt") <- mvar
    class(x) <- c("scdf", "list")
    
    x
    
  }
  
  out <- lapply(out, extract)
  
  for(i in 1:length(out)) names(out[[i]]) <- case_names[i]
  
  for(i in 2:length(out)) scdf <- c(out[[i-1]], out[[i]])
  
  
  scdf
}

#' @rdname print.sc
#' @export
print.sc_smd <- function(x, digits = "auto", ...) {
  
  if (digits == "auto") digits <- 3
  
  cat("Standardized mean differences\n\n")
  x$smd[-1] <- round(x$smd[-1], digits)
  out <- as.data.frame(t(x$smd[-1]))
  colnames(out) <- x$smd$Case
  
  print(out[ , , drop = FALSE], digits = digits, ...)
  cat("\n")
  .note_vars(x)
  
}

#' (Deprecated) Plot single-case data
#'
#' This function provides a plot of a single-case or multiple single-cases.
#'
#'
#' @aliases plotSC plot.scdf
#' @inheritParams .inheritParams
#' @param ylim Lower and upper limits of the y-axis (e.g., \code{ylim = c(0,
#'   20)} sets the y-axis to a scale from 0 to 20). With multiple single-cases
#'   you can use \code{ylim = c(0, NA)} to scale the y-axis from 0 to the
#'   maximum of each case. \code{ylim} is not set by default, which makes
#'   \code{scan} set a proper scale based on the given data.
#' @param xlim Lower and upper limits of the x-axis (e.g., \code{xlim = c(0,
#'   20)} sets the x-axis to a scale from 0 to 20). With multiple single-cases
#'   you can use \code{ylim = c(0, NA)} to scale the x-axis from 0 to the
#'   maximum of each case. \code{xlim} is not set by default, which makes
#'   \code{scan} set a proper scale based on the given data.
#' @param xinc An integer. Increment of the x-axis. 1 :each mt value will be
#'   printed, 2 : every other value, 3 : every third values etc.
#' @param style Either a character with the name of a pre-implemented style or a
#'   style object. See \code{\link{style_plot}} to learn about this format.
#' @param lines A list defining one or multiple lines or curves to be plotted.
#'   The argument is passed as a list (e.g., \code{list(type = "median")}). Some
#'   of the procedures can be refined with an additional argument (e.g.,
#'   \code{lines = list(type = "mean", trim = 0.2)} adds a 20\% trimmed mean
#'   line. For multiple lines, provide a list element for each line (e.g.,
#'   \code{list( list(type = "median", col = "red"), list(type = "trend", col =
#'   "blue"))}.
#' Possible lines are: \itemize{
#' \item\code{"median"} Separate lines for phase A and B medians.
#' \item\code{"mean"} Separate lines for phase A and B means. By default it is
#' 10\%-trimmed. Other trims can be set, using a trim parameter (e.g.,
#' \code{lines = list(type = "mean", trim = 0.2)} draws a 20\%-trimmed mean line).
#' \item\code{"trend"} Separate lines for phase A and B trends.
#' \item\code{"trendA"} OLS trend line for phase A, extrapolated throughout phase B.
#' \item\code{"trendA_bisplit"} Split middle (bi-split) trend line for phase A, extrapolated throughout phase B.
#' \item\code{"trendA_trisplit"} Tukey tri-split trend line for phase A, extrapolated throughout phase B.
#' \item\code{"maxA/minA"} Line at the level of the highest or lowest phase A score.
#' \item\code{"medianA"} Line at the phase A median score.  \item\code{"meanA"}
#' Line at the phase A 10\%-trimmed mean score. Apply a different trim, by
#' using the additional argument (e.g., \code{lines = list(type = "meanA", trim = 0.2)}).
#' \item\code{"plm"} Regression lines for piecewise linear regression model.
#' \item\code{"plm.ar"} Regression lines for piecewise autoregression model.
#' The lag is specified like this: \code{lines = list(type = "plm.ar", ar = 2)}. Default lag is set to 2.
#' \item\code{"movingMean"} Draws a moving mean curve, with a specified lag:
#' \code{lines = list(type = "movingMean", lag = 2)}. Default is a lag 1 curve.
#' \item\code{"movingMedian"} Draws a moving median curve, with a specified
#' lag: \code{lines = list(type = "movingMedian", lag = 3)}. Default is a lag 1 curve.
#' \item\code{"loreg"} Draws a non-parametric local regression line. The
#' proportion of data influencing each data point can be specified using
#' \code{lines = list(type = "loreg"m f = 0.66)}. The default is 0.5.  \item\code{"lty"}
#' Use this argument to define the line type. Examples are: \code{"solid"},
#' \code{"dashed"}, \code{"dotted"}.  \item\code{"lwd"} Use this argument to
#' define the line's thickness, e.g., \code{lwd = 4}.  \item\code{"col"} Use
#' this argument to define the line's color, e.g., \code{col = "red"}.  }
#' @param marks A list of parameters defining markings of certain data points.
#' \itemize{ \item\code{"positions"} A vector or a list of vectors indicating
#' measurement-times to be highlighted. In case of a vector, the marked
#' measurement-times are the same for all plotted cases. In case of a list of
#' vectors, marks are set differently for each case. The list must have the
#' same length as there are cases in the data file.  \item\code{"col"} Color of
#' the marks.  \item\code{"cex"} Size of the marks.  } Use for example
#'   \code{marks = list(positions = c(1, 8, 15), col = "red", cex = 3)} to make
#'   the MTs one, eight and 18 appear big and red.
#' @param phase.names By default phases are labeled based on the levels of the
#'   phase variable.
#' Use this argument to specify different labels: \code{phase.names = c("Baseline",
#' "Intervention")}.
#' @param xlab The label of the x-axis. Default is \code{xlab = "Measurement
#'   time"}.
#' @param ylab The labels of the y-axis. Default is \code{ylab = "Score"}.
#' @param main Main title of the plot.
#' @param case.names Case names. If not provided, names are taken from the scdf.
#'   Set \code{case.names = ""} if you don't like to include case names.
#' @param ... Further arguments passed to the plot command.
#' @return Returns a plot of one or multiple single-cases.
#' @author Juergen Wilbert
#' @examples
#'
#' ## Request the default plot of the data from Borckhardt (2014)
#' plot(Borckardt2014)
#'
#' ## Plot the three cases from Grosche (2011) and visualize the phase A trend
#' plot(Grosche2011, style = "grid", lines = "trendA")
#'
#' ## Request the local regression line for Georg from that data set and customize the plot
#' plot(Grosche2011$Georg, style = "sienna", ylim = c(0,NA),
#'        xlab = "Training session", ylab = "Words per minute",
#'        phase.names = c("Baseline", "Intervention"), xinc = 5,
#'        lines = list(type = "loreg", f = 0.2, lty = "solid", col = "black", lwd = 3))
#'
#' ## Plot a random MBD over three cases and mark interesting MTs
#' dat <- random_scdf(design = design(3))
#' plot(dat, marks = list(positions = list(c(2,4,5),c(1,2,3),c(7,8,9)), col = "blue",
#'        cex = 1.4), style = c("grid", "annotate", "tiny"))
#'
#' @export
plot.scdf <- function(...) {
  .deprecated_warning("scplot' from the package 'scplot", "plot")
  plotSC(...)
}

#' @rdname plot.scdf
#' @export
plotSC <- function(data, dvar, pvar, mvar, 
                   ylim = NULL, xlim = NULL, xinc = 1, 
                   lines = NULL, marks = NULL, 
                   phase.names = NULL, xlab = NULL, ylab = NULL, 
                   main = "", case.names = NULL, 
                   style = getOption("scan.plot.style"), 
                   ...) {
  
  
  # prepare scdf ------------------------------------------------------------
  
  dots <- list(...)
  op <- par(no.readonly = TRUE)
  on.exit(par(op))
  
  # set attributes to arguments else set to defaults of scdf
  if (missing(dvar)) dvar <- dv(data) else dv(data) <- dvar
  if (missing(pvar)) pvar <- phase(data) else phase(data) <- pvar
  if (missing(mvar)) mvar <- mt(data) else mt(data) <- mvar
  
  data_list <- .prepare_scdf(data)
  
  N <- length(data_list)
  
  if (N > 1) par(mfrow = c(N, 1))
  
  # define style ------------------------------------------------------------
  
  if (is.list(style)) {
    ref.style <- "default"
    if ("style" %in% names(style)) ref.style <- style$style
    style <- c(style, style_plot(ref.style))
    style <- style[unique(names(style))]
  }
  
  if (is.character(style)) style <- style_plot(style)
  
  #for pre style backwards compatibility
  sty_names <- c("fill", "fill.bg", "frame", "grid", "lwd", "pch", "text.ABlag", "type")
  if (any(names(dots) %in% sty_names)) {
    abort("Using style parameters directly as arguments (like 'fill') is deprectated. ",
         "Please use the 'stlye' argument to provide these parameters. ",
         "E.g., style = list(fill = 'blue', pch = 19)")
  }
  annotations <- style$annotations
  
  if (is.na(style$frame)) style$bty <- "n"
  
  par("bg"       = style$col.bg)
  par("col"      = style$col)
  par("family"   = style$font)
  par("cex"      = style$cex)
  par("cex.lab"  = style$cex.lab)
  par("cex.axis" = style$cex.axis)
  par("las"      = style$las)
  par("bty"      = style$bty)
  par("col.lab"  = style$col.text)
  par("col.axis" = style$col.text)
  
  if (isTRUE(style$frame == ""))   style$frame <- NA
  if (isTRUE(style$grid == ""))    style$grid  <- FALSE
  if (isTRUE(style$fill.bg == "")) style$fill.bg  <- FALSE
  
  ### END: define style
  
  # Marks on the outliers from outlierandom_scdf
  if (inherits(marks, "sc_outlier")) 
    marks <- list(positions = marks$dropped.mt)
  
  # name cases
  if (!is.null(case.names)) names(data_list) <- case.names
  case.names <- names(data_list)
  
  # set x ans y axis labels
  if (is.null(xlab)) xlab <- mvar
  if (is.null(ylab)) ylab <- dvar
  if (is.null(xlab)) xlab <- "Measurement time"
  if (is.null(ylab)) ylab <- "Score"
  if (xlab == "mt") xlab <- "Measurement time"
  
  # prepare lines definitions
  lines <- .prepare_arg_lines(lines)

  # set xlim and ylim
  values.tmp <- unlist(lapply(data_list, function(x) x[, dvar]))
  mt.tmp     <- unlist(lapply(data_list, function(x) x[, mvar]))
  
  if (is.null(ylim))
    ylim <- c(min(values.tmp, na.rm = TRUE), max(values.tmp, na.rm = TRUE))
  if (is.null(xlim))
    xlim <- c(min(mt.tmp, na.rm = TRUE), max(mt.tmp, na.rm = TRUE))
  
  
  # Plotting cases ----------------------------------------------------------
  
  par(mgp = c(2, 1, 0))
  
  for(case in 1:N) {
    data <- data_list[[case]]
    
    design <- .phasestructure(data, pvar)
    
    # plot ylim
    y.lim <- ylim
    if (is.na(ylim[2])) y.lim[2] <- max(data[, dvar])
    if (is.na(ylim[1])) y.lim[1] <- min(data[, dvar])
    
    # one plot
    if (N == 1) {
      if (main != "") par(mai = c(style$mai[1:2], style$mai[3] + 0.4, style$mai[4]))
      if (main == "") par(mai = style$mai)
      plot(
        data[[mvar]], data[[dvar]], type = "n", 
        xlim = xlim, ylim = y.lim, ann = FALSE,
        #xaxp = c(xlim[1], xlim[2], xlim[2] - xlim[1]), 
        yaxt = "n", 
        xaxt = "n", ...
      )
      
      xticks_pos <- seq(xlim[1], xlim[2], 1)
      axis(side = 1, at = xticks_pos, labels = FALSE)
      text(
        x = seq(xlim[1], xlim[2], xinc),  
        y = par("usr")[3], 
        cex = style$cex.axis, 
        labels = seq(xlim[1], xlim[2], xinc), 
        srt = 0, 
        pos = 1, 
        offset = style$cex.axis, 
        xpd = TRUE
      )
      
      yticks_pos <- axTicks(2, usr = c(y.lim[1], ylim[2]))
      axis(side = 2, at = yticks_pos, labels = NA)
      text(
        x = par("usr")[1], 
        y = yticks_pos, 
        labels = yticks_pos, 
        offset = style$cex.axis, 
        srt = 0, 
        pos = 2, 
        cex = style$cex.axis, 
        xpd = TRUE
      )
      
      if (style$ylab.orientation == 0) 
        mtext(text = ylab, side = 2, line = 2, las = 0, cex = style$cex.lab)
      
      if (style$ylab.orientation == 1) {
        mtext(
          text = ylab, side = 2, line = 2, las = 1, at = max(y.lim), 
          cex = style$cex.lab
        )
      }
      
      mtext(text = xlab, side = 1, line = 2, las = 0, cex = style$cex.lab)
    } 
    
    # multple plots, first to secondlast
    if (N > 1 && case != N) {
      if (case == 1) { # first plot
        if (main != "") {
          par(mai = c(style$mai[1] * 2 / 6, style$mai[2], style$mai[3] * 3, style$mai[4]))
        }
        if (main == "") { 
          par(mai = c(style$mai[1] * 2 / 6, style$mai[2], style$mai[3] * 3, style$mai[4])) 
        }
      } else { # middle plot
        par(mai = c(style$mai[1] * 4 / 6, style$mai[2], style$mai[3] * 2, style$mai[4])) 
      }
      
      plot(
        data[[mvar]], data[[dvar]], xaxt = "n", yaxt = "n", type = "n", 
        xlim = xlim, ylim = y.lim, ann = FALSE, ...
      )
      
      yticks_pos <- axTicks(2, usr = c(y.lim[1], ylim[2]))
      axis(side = 2, at = yticks_pos, labels = NA)
      text(
        x = par("usr")[1], 
        y = yticks_pos, 
        labels = yticks_pos, 
        offset = style$cex.axis, 
        srt = 0, 
        pos = 2, 
        cex = style$cex.axis, 
        xpd = TRUE
      )
      
      if (style$ylab.orientation == 0) 
        mtext(ylab, side = 2, line = 2, las = 0, cex = style$cex.lab)
      
      if (style$ylab.orientation == 1)
        mtext(ylab, side = 2, line = 2, las = 1, at = max(y.lim), cex = style$cex.lab)
    }
    # multiple plots, last plot
    if (N > 1 && case == N) {
      par(mai = style$mai)
      plot(
        data[[mvar]], data[[dvar]], type = "n", 
        xlim = xlim, ylim = y.lim, ann = FALSE,
        xaxp = c(xlim[1], xlim[2], xlim[2] - xlim[1]),
        yaxt = "n", xaxt = "n",
        ...
      )
      
      xticks_pos <- seq(xlim[1], xlim[2], 1)
      axis(side = 1, at = xticks_pos, labels = FALSE)
      text(
        x = seq(xlim[1], xlim[2], xinc),  
        y = par("usr")[3], 
        cex = style$cex.axis, 
        labels = seq(xlim[1], xlim[2], xinc), 
        srt = 0, 
        pos = 1, 
        offset = style$cex.axis, 
        xpd = TRUE
      )
      
      yticks_pos <- axTicks(2, usr = c(y.lim[1], ylim[2]))
      axis(side = 2, at = yticks_pos, labels = NA)
      text(
        x = par("usr")[1], 
        y = yticks_pos, 
        labels = yticks_pos, 
        offset = style$cex.axis, 
        srt = 0, 
        pos = 2, 
        cex = style$cex.axis, xpd = TRUE
      )
      
      if (style$ylab.orientation == 0) 
        mtext(ylab, side = 2, line = 2, las = 0, cex = style$cex.lab)
      
      if (style$ylab.orientation == 1)
        mtext(ylab, side = 2, line = 2, las = 1, at = max(y.lim), cex = style$cex.lab)
      
      mtext(xlab, side = 1, line = 2, las = 0, cex = style$cex.lab)
      
    } 
    usr <- par("usr")
    
    # styling ------------------------------------------------------------
    
    # fill bg
    if(inherits(style$fill.bg, "character")) {
      style$col.fill.bg <- style$fill.bg
      style$fill.bg <- TRUE
    }
    
    if (isTRUE(style$fill.bg)){
      #rect(usr[1], usr[3], usr[2], usr[4], col = style$col.fill.bg, border = NA)
      
      type_phases <- unique(design$values)
      col <- rep(style$col.fill.bg, length = length(type_phases))
      
      for(i in seq_along(design$values)) {
        x <- data[design$start[i]:design$stop[i], mvar]
        y <- data[design$start[i]:design$stop[i], dvar]
        
        xmin <- min(x, na.rm = TRUE)
        xmax <- max(x, na.rm = TRUE) + 0.5
        if (i == length(design$values)) xmax <- usr[2]
        if (i > 1) xmin <- xmin - 0.5
        if (i == 1) xmin <- usr[1]
        .col <- col[which(type_phases == design$values[i])[1]]
        rect(xmin, usr[3], xmax, usr[4], col = .col, border = NA)
      }
    }
    
    # grid
    if(inherits(style$grid, "character")) {
      style$col.grid <- style$grid
      style$grid <- TRUE
    }
    
    if (isTRUE(style$grid)) 
      grid(NULL, NULL, col = style$col.grid, lty = style$lty.grid, lwd = style$lwd.grid)
    
    if (!is.na(style$frame))
      rect(usr[1],usr[3],usr[2],usr[4], col = NA, border = style$frame)
    
    if (is.na(style$frame) && isTRUE(style$fill.bg))
      rect(usr[1],usr[3],usr[2],usr[4], col = NA, border = style$col.fill.bg)
    
    if (is.na(style$frame) && !isTRUE(style$fill.bg))
      rect(usr[1],usr[3],usr[2],usr[4], col = NA, border = par("bg"))
    
    # fill array below lines
    
    if (style$fill == "" || is.na(style$fill)) style$fill <- FALSE
    
    if(inherits(style$fill, "character")) {
      style$col.fill <- style$fill
      style$fill <- TRUE
    }
    
    if (isTRUE(style$fill)) {
      for(i in 1:length(design$values)) {
        x <- data[design$start[i]:design$stop[i], mvar]
        y <- data[design$start[i]:design$stop[i], dvar]
        
        for(i in 1:length(x)) {
          x_values <- c(x[i], x[i+1], x[i+1], x[i])
          y_values <- c(y.lim[1], y.lim[1], y[i+1], y[i])
          polygon(x_values, y_values, col = style$col.fill, border = NA)      
        }
      }
    }
    
    # draw lines
    for(i in 1:length(design$values)) {
      x <- data[design$start[i]:design$stop[i], mvar]
      y <- data[design$start[i]:design$stop[i], dvar]
      if (style$col.lines != "") {
        lines(
          x, y, type = "l", pch = style$pch, lty = style$lty, lwd = style$lwd, 
          col = style$col.lines, ...
        )
      }
      if (style$col.dots != "") {
        lines(
          x, y, type = "p", pch = style$pch, lty = style$lty, lwd = style$lwd, 
          col = style$col.dots, ...
        )
      }
    }
    
    if (case == 1) title(main)
    
    # marks -------------------------------------------------------------------
    
    if (!is.null(marks)) {
      marks.cex <- 1
      marks.col <- "red"
      marks.pch <- style$pch
      
      if (any(names(marks) == "positions")) {
        marks.pos <- marks[[which(names(marks) == "positions")]]
      } else {stop("Positions of marks must be defined.")}
      
      if (any(names(marks) == "cex")) {
        marks.cex <- marks[[which(names(marks) == "cex")]]
      }
      if (any(names(marks) == "col")) {
        marks.col <- marks[[which(names(marks) == "col")]]
      }
      if (any(names(marks) == "pch")) {
        marks.pch <- marks[[which(names(marks) == "pch")]]
      }
      
      if (inherits(marks.pos, "numeric")) {
        mks <- marks.pos
      } else {
        mks <- marks.pos[[case]]
      }
      marks.x <- data[data[,mvar] %in% mks,mvar]
      marks.y <- data[data[,mvar] %in% mks,dvar]
      points(x = marks.x, y = marks.y, pch = marks.pch, cex = marks.cex, col = marks.col)
    }
    
    # annotations -------------------------------------------------------------
    
    if (!is.null(annotations)) {
      annotations.cex <- 1
      annotations.round <- 1
      annotations.col <- "black"
      annotations.pos <- 3
      annotations.offset <- 0.5
      
      if (any(names(annotations) == "cex")) {
        annotations.cex <- annotations[[which(names(annotations) == "cex")]]
      }
      if (any(names(annotations) == "col")) {
        annotations.col <- annotations[[which(names(annotations) == "col")]]
      }
      if (any(names(annotations) == "round")) {
        annotations.round <- annotations[[which(names(annotations) == "round")]]
      }
      if (any(names(annotations) == "pos")) {
        annotations.pos <- annotations[[which(names(annotations) == "pos")]]
      }
      if (any(names(annotations) == "offset")) {
        annotations.offset <- annotations[[which(names(annotations) == "offset")]]
      }
      
      annotations.label <- round(data[,dvar], annotations.round)
      ### not yet implemented
      #if (any(names(annotations) == "label")) {
      #  id <- which(names(annotations) == "label")
      #  if (annotations[[id]]=="values") {
      #  } else {
      #  }
      #}
      
      text(
        x = data[,mvar], 
        y = data[,dvar], 
        label = annotations.label, 
        col = annotations.col, 
        pos = annotations.pos, 
        offset = annotations.offset, 
        cex = annotations.cex, 
        ...
      )
    }
    
    # lines -------------------------------------------------------------------
    
    if (!is.null(lines)) {
      for(i_lines in seq_along(lines)) {
        
        line <- lines[[i_lines]]
        
        if (is.null(line[["lty"]])) line[["lty"]] <- "dashed"
        if (is.null(line[["lwd"]])) line[["lwd"]] <- 2
        if (is.null(line[["col"]])) line[["col"]] <- "black"
        
        lty.line <- line[["lty"]]
        lwd.line <- line[["lwd"]]
        col.line <- line[["col"]]
        
        if (line[["type"]] == "trend") {
          for(i in 1:length(design$values)) {
            x <- data[design$start[i]:design$stop[i], mvar]
            y <- data[design$start[i]:design$stop[i], dvar]
            reg <- lm(y~x)
            lines(
              x = c(min(x), max(x)), 
              y = c(
                reg$coefficients[1] + min(x) * reg$coefficients[2], 
                reg$coefficients[1] + max(x) * reg$coefficients[2]
              ), 
              lty = lty.line, 
              col = col.line, 
              lwd = lwd.line
            )
          }
        }
        if (line[["type"]] == "median") {
          for(i in 1:length(design$values)) {
            x <- data[design$start[i]:design$stop[i], mvar]
            y <- data[design$start[i]:design$stop[i], dvar]
            lines(
              x = c(min(x), max(x)), 
              y = c(median(y, na.rm = TRUE), median(y, na.rm = TRUE)), 
              lty = lty.line, 
              col = col.line, 
              lwd = lwd.line
            )
          }      
        }
        if (line[["type"]] == "mean") {
          if (is.null(line[["trim"]])) line[["trim"]] <- 0.1
          lines.par <- line[["trim"]]

          for(i in 1:length(design$values)) {
            x <- data[design$start[i]:design$stop[i],mvar]
            y <- data[design$start[i]:design$stop[i],dvar]
            lines(
              x = c(min(x), max(x)), 
              y = c(
                mean(y, trim = lines.par, na.rm = TRUE), 
                mean(y, trim = lines.par, na.rm = TRUE)
              ), 
              lty = lty.line, 
              col = col.line, 
              lwd = lwd.line
            )
          }
        }
        if (line[["type"]] == "trendA") {
          x <- data[design$start[1]:design$stop[1],mvar]
          y <- data[design$start[1]:design$stop[1],dvar]
          maxMT <- max(data[,mvar])
          reg <- lm(y~x)
          lines(
            x = c(min(x), maxMT), 
            y = c(
              reg$coefficients[1]  + min(x) * reg$coefficients[2], 
              reg$coefficients[1] + maxMT * reg$coefficients[2]
            ), 
            lty = lty.line, 
            col = col.line, 
            lwd = lwd.line
          )
        }
        if (line[["type"]] == "trendA_bisplit") {
          x     <- data[design$start[1]:design$stop[1],mvar]
          y     <- data[design$start[1]:design$stop[1],dvar]
          maxMT <- max(data[,mvar])
          # na.rm = FALSE for now to prevent misuse; 
          # will draw no line if NA present
          md1   <- c(median(y[1:floor(length(y)/2)], na.rm = FALSE),
                     median(x[1:floor(length(x)/2)], na.rm = FALSE))
          md2   <- c(median(y[ceiling(length(y)/2+1):length(y)], na.rm = FALSE),
                     median(x[ceiling(length(x)/2+1):length(x)], na.rm = FALSE))
          md    <- rbind(md1, md2)
          reg <- lm(md[,1]~md[,2])
          lines(
            x = c(min(x), maxMT), 
            y = c(
              reg$coefficients[1]  + min(x) * reg$coefficients[2], 
              reg$coefficients[1] + maxMT * reg$coefficients[2]
            ), 
            lty = lty.line, 
            col = col.line, 
            lwd = lwd.line
          )
        }
        if (line[["type"]] == "trendA_trisplit") {
          x     <- data[design$start[1]:design$stop[1],mvar]
          y     <- data[design$start[1]:design$stop[1],dvar]
          maxMT <- max(data[,mvar])
          # na.rm = FALSE for now to prevent misuse; 
          # will draw no line if NA present
          md1   <- c(
            median(y[1:floor(length(y) / 3)], na.rm = FALSE),
            median(x[1:floor(length(x) / 3)], na.rm = FALSE)
          )
          md2   <- c(
            median(y[ceiling(length(y) / 3 * 2 + 1):length(y)], na.rm = FALSE),
            median(x[ceiling(length(x) / 3 * 2 + 1):length(x)], na.rm = FALSE)
          )
          md    <- rbind(md1, md2)
          reg <- lm(md[,1] ~ md[,2])
          lines(
            x = c(min(x), maxMT), 
            y = c(
              reg$coefficients[1]  + min(x) * reg$coefficients[2], 
              reg$coefficients[1] + maxMT * reg$coefficients[2]
            ), 
            lty = lty.line, 
            col = col.line, 
            lwd = lwd.line
          )
        }
        if (line[["type"]] == "loreg") {
          if (is.null(line[["f"]])) line[["f"]] <- 0.5
          lines.par <- line[["f"]]
          reg <- lowess(data[,dvar] ~ data[,mvar], f = lines.par)
          lines(reg, lty = lty.line, col = col.line, lwd = lwd.line)
        }
        
        if (line[["type"]] %in% c("maxA", "pnd")) {
          x <- data[design$start[1]:design$stop[1],mvar]
          y <- data[design$start[1]:design$stop[1],dvar]
          maxMT <- max(data[,mvar])
          lines(
            x = c(min(x), maxMT), 
            y = c(max(y), max(y)), 
            lty = lty.line, 
            col = col.line, 
            lwd = lwd.line
          )		
        }
        
        if (line[["type"]] == "minA") {
          x <- data[design$start[1]:design$stop[1],mvar]
          y <- data[design$start[1]:design$stop[1],dvar]
          maxMT <- max(data[,mvar])
          lines(
            x = c(min(x), maxMT), 
            y = c(min(y), min(y)), 
            lty = lty.line, 
            col = col.line, 
            lwd = lwd.line
          )		
        }
        if (line[["type"]] == "medianA") {
          x <- data[design$start[1]:design$stop[1],mvar]
          y <- data[design$start[1]:design$stop[1],dvar]
          maxMT <- max(data[,mvar])
          
          lines(
            x = c(min(x), maxMT), 
            y = c(median(y, na.rm = TRUE), median(y, na.rm = TRUE)), 
            lty = lty.line, 
            col = col.line, 
            lwd = lwd.line
          )		
        }
        if (line[["type"]] == "meanA") {
          if (is.null(line[["trim"]])) line[["trim"]] <- 0.1
          lines.par <- line[["trim"]]
          
          x <- data[design$start[1]:design$stop[1],mvar]
          y <- data[design$start[1]:design$stop[1],dvar]
          maxMT <- max(data[,mvar])
          lines(
            x = c(min(x), maxMT), 
            y = c(
              mean(y, trim = lines.par, na.rm = TRUE), 
              mean(y, trim = lines.par, na.rm = TRUE)
            ), 
            lty = lty.line, 
            col = col.line, 
            lwd = lwd.line
          )		
        }
        
        if (line[["type"]] == "plm") {
          pr <- plm(data_list[case])
          y <- fitted(pr$full.model)
          lines(data[[mvar]], y, lty = lty.line, col = col.line, lwd = lwd.line)
        }
        if (line[["type"]] == "plm.ar") {
          if (is.null(line[["ar"]])) line[["ar"]] <-2
          lines.par <- line[["ar"]]
          pr <- plm(data_list[case], AR = lines.par)
          y <- fitted(pr$full.model)
          lines(data[[mvar]], y, lty = lty.line, col = col.line, lwd = lwd.line)
        }
        
        if (line[["type"]] == "movingMean") {
          if (is.null(line[["lag"]])) line[["lag"]] <- 1
          lines.par <- line[["lag"]]
          y <- .moving_average(data[, dvar],lines.par, mean)
          lines(data[, mvar], y, lty = lty.line, col = col.line, lwd = lwd.line)
        }
        if (line[["type"]] == "movingMedian") {
          if (is.null(line[["lag"]])) line[["lag"]] <- 1
          lines.par <- line[["lag"]]
          y <- .moving_average(data[, dvar],lines.par, median)
          lines(data[, mvar], y, lty = lty.line, col = col.line, lwd = lwd.line)
        }
      }
    }#### END: Adding help-lines
    
    
    # phase names -------------------------------------------------------------
    if (is.null(phase.names)) {
      case_phase_names <- design$values
    } else {
      case_phase_names <- phase.names
    }
    
    for(i in 1:length(design$values)) {
      mtext(
        text = case_phase_names[i], 
        side = 3, 
        at = (data[design$stop[i], mvar] - data[design$start[i], mvar]) / 2 + 
              data[design$start[i], mvar], 
        cex = style$cex.text, ...
      )
    }
    
    # line between phases -----------------------------------------------------
    if (is.null(style$text.ABlag)) {
      for(i in 1:(length(design$values) - 1)) {
        abline(
          v = data[design$stop[i] + 1, mvar] - 0.5, 
          lty = style$lty.seperators, 
          lwd = style$lwd.seperators, 
          col = style$col.seperators
        )
      }
    }
    
    if (!is.null(style$text.ABlag)) {
      for(i in 1:(length(design$values) - 1)) {
        tex <- paste(unlist(strsplit(style$text.ABlag[i], "")), collapse = "\n")
        text(
          x = data[design$stop[i] + 1, mvar] - 0.5, 
          y = (y.lim[2] - y.lim[1]) / 2 + y.lim[1], 
          labels = tex, 
          cex = 0.8, ...)
      }
      
    }
    
    # Adding case name --------------------------------------------------------
    if (length(case.names) ==  N) {
      args <- c(
        list(text = case.names[case]), 
        style$names, 
        cex = style$cex.text, 
        at = min(xlim)
      )
      args <- args[!duplicated(names(args))]
      do.call(mtext, args)
    }
    
  }
  
}

.prepare_arg_lines <- function(lines) {
  
  
  types <- list(
    "mean",
    "median", 
    "pnd", 
    "maxA", 
    "minA", 
    "medianA", 
    "trend", 
    "trendA", 
    "plm", 
    "plm.ar",
    "trendA_bisplit", 
    "trendA_trisplit",
    "loreg", 
    "movingMean", 
    "movingMedian"
  )
  
  if (!inherits(lines, "list")) {
    lines <- lapply(lines, function(x) x)
  }
  
  if(!all(sapply(lines, is.list))) {
    lines <- list(lines)
  }
  
  if (length(lines) == 1) {
    tmp_args <- unlist(lines[[1]])
    id <- which(tmp_args %in% names(types))
    if (length(id) > 1) {
      warn("More than one line type detected in one list element!")
      for (i in 2:length(id)) {
        lines <- c(lines, list(lines[[1]][id[i]]))
        lines[[1]] <- lines[[1]][-id[i]]
      }
    }
    
    tmp_args <- names(lines[[1]])
    id <- which(tmp_args %in% names(types))
    if (length(id) > 1) {
      warn("More than one line type detected in one list element!")
      for (i in 2:length(id)) {
        lines <- c(lines, list(lines[[1]][id[i]]))
        lines[[1]] <- lines[[1]][-id[i]]
      }
    }
    
  }
  
  for (i in seq_along(lines)) {
    arg_names <- names(lines[[i]])
    
    if (is.null(arg_names)) {
      names(lines[[i]]) <- rep("", length(lines[[i]]))
      arg_names <- names(lines[[i]])
    }
    
    if (any(arg_names == "mean")) {
      id <- which(arg_names == "mean")
      lines[[i]] <- c(lines[[i]], trim = as.numeric(unname(lines[[i]][id])))
      lines[[i]][id] <- names(lines[[i]])[id]
      names(lines[[i]])[id] <- "type"
    }
    
    if (any(arg_names == "loreg")) {
      id <- which(arg_names == "loreg")
      lines[[i]] <- c(lines[[i]], f = as.numeric(unname(lines[[i]][id])))
      lines[[i]][id] <- names(lines[[i]])[id]
      names(lines[[i]])[id] <- "type"
    }
    
    if (any(arg_names == "movingMedian")) {
      id <- which(arg_names == "movingMedian")
      lines[[i]] <- c(lines[[i]], lag = as.numeric(unname(lines[[i]][id])))
      lines[[i]][id] <- names(lines[[i]])[id]
      names(lines[[i]])[id] <- "type"
    }
    
    if (any(arg_names == "movingMean")) {
      id <- which(arg_names == "movingMean")
      lines[[i]] <- c(lines[[i]], lag = as.numeric(unname(lines[[i]][id])))
      lines[[i]][id] <- names(lines[[i]])[id]
      names(lines[[i]])[id] <- "type"
    }
    
    if (any(arg_names == "plm.ar")) {
      id <- which(arg_names == "plm.ar")
      lines[[i]] <- c(lines[[i]], ar = as.numeric(unname(lines[[i]][id])))
      lines[[i]][id] <- names(lines[[i]])[id]
      names(lines[[i]])[id] <- "type"
    }
    
    id <- which(arg_names == "")
    if(length(id) > 1) {
      abort("undefined lines argument(s)") # todo: specify error
    }
    if(length(id) == 1) {
      names(lines[[i]])[id] <- "type"
    }
  }
  
  lines
  
}

.default_attributes <- function(attri = NULL) {
  out <- list()
  
  if (!is.null(attri)) out <- attri
  
  out$class <- c("scdf", "list")
  
  scdf <- list()
  scdf[[opt("phase")]]   <- "phase"
  scdf[[opt("dv")]]      <- "values"
  scdf[[opt("mt")]]      <- "mt"
  out[[opt("scdf")]]     <- scdf
  
  out
}  

.deprecated_warning <- function(new, old) {
  if (isTRUE(getOption("scan.deprecated.warning"))) {
    warn(
      opt("function_deprecated_warning"), 
      "\nPlease use function '", new, "' instead of '", old, "'."
    )
  }   
} 


.moving_average <- function(x, lag = 1, FUN = mean) {
  for(i in (lag + 1):(length(x) - lag))
    x[i] <- FUN(x[(i - lag):(i + lag)], na.rm = TRUE)
  x
}

.beta_weights <- function(model) {
  b <- model$coefficients[-1]
  sx <- apply(model$model[-1], 2, sd)
  sy <- apply(model$model[ 1], 2, sd)
  c(model$coefficients, b * sx / sy)
}

.phasestructure <- function(data, pvar) {
  phases <- rle(as.character(data[, pvar]))
  phases$start <- c(1, cumsum(phases$lengths) + 1)[1:length(phases$lengths)]
  phases$stop  <- cumsum(phases$lengths)
  class(phases) <- "list"
  phases
}

.phases_string <- function(A, B) {
  nomer_s = "phase "
  nomer_p = "phases "
  a_part <- if (length(A) == 1) {
    paste0(nomer_s, A, collapse = "")
  } else {
    paste0( c(nomer_p, A[1], paste0(" + ",A[-1]) ), collapse = "")
  } 
    
  b_part <- if (length(B) == 1) {
    paste0(nomer_s, B, collapse = "")
  } else {
    paste0( c(nomer_p,B[1], paste0(" + ",B[-1])), collapse = "")
  }
    
  out <- paste0(c("Comparing ", a_part, " against ", b_part), collapse ="")
  out
}

.nice_p <- function(p, equal.sign = FALSE) {
  out <- rep(NA, length(p))
  for(i in 1:length(p)) {
    if (isTRUE(p[i] >= 0.05)) {
      out[i] <- substring(sprintf("%.2f", trunc(p[i] * 100) / 100), 2)
      if (equal.sign) out[i] <- paste0("= ", out[i])
    }
    if (isTRUE(p[i] == 1))    out[i] <- "1.00"    
    if (isTRUE(p[i] < 0.05))  out[i] <- "<.05"
    if (isTRUE(p[i] < 0.01))  out[i] <- "<.01"
    if (isTRUE(p[i] < 0.001)) out[i] <- "<.001"
  }
  out
} 

revise_names <- function(x, n) {
  
  if (missing(n)) {
    n <- length(x)
    if (!is.character(x)) x <- names(x)
  }

  if (is.null(x)) {
    x <- paste0("[case #", 1:n, "]") # paste0("Case", 1:n)
  } else {
    nonames <- which(is.na(x) | x == "")
    x[nonames] <- paste0("[case #", nonames, "]") # paste0("Case", nonames)
  }
  x
}




.std_lm <- function(model) {
  
  coef <- coef(model)
  if (isTRUE(class(coef) == "numeric")) coef <- as.matrix(coef, ncol = 1)
  intercept <- attr(attr(model$model, "terms"), "intercept")
  .sd <- function(x) sqrt(
    sum((x - mean(x, na.rm = TRUE) * intercept)^2, na.rm = TRUE)
  )
  .sd_predictors <- apply(as.matrix(model.matrix(model)), 2, .sd)
  .sd_criteria <- apply(as.matrix(model.frame(model)[, 1]), 2, .sd)
  coef_std <- coef
  for(i in 1:ncol(coef)) {
    coef_std[, i] <- coef[, i] * .sd_predictors / .sd_criteria[i]
  }
  
  coef_std
}

number_word <- function(x) {
  if (x %in% 0:9) {
    return(
      c("zero", "one", "two", "three", "four", "five", "six", "seven", 
        "eight", "nine")[x + 1]
    )
  }
  x
}

round_numeric <- function(df, digits = 0) {
  id <- which(sapply(df, is.numeric))
  if (length(id) > 0) df[, id] <- round(df[, id], digits)
  df
}

format_table <- function(df, 
                         digits = 3, 
                         min_digits = digits, 
                         na_replace = "", 
                         integer = NULL,
                         nice_p = NULL) {
  
  if (!is.null(nice_p)) df[[nice_p]] <- .nice_p(df[[nice_p]])
  
  # Apply rounding and formatting to each column
  df_rounded <- as.data.frame(lapply(names(df), function(col) {
    x <- df[[col]]
    
    if (col %in% integer && is.numeric(x)) {
      # For columns specified in the 'integer' argument, round to the nearest integer
      formatted <- format(round(x), nsmall = 0)
      formatted[is.na(x)] <- na_replace
      return(formatted)
    } else if (is.numeric(x)) {
      # For other numeric columns, round to the specified number of decimal places
      formatted <- format(round(x, digits), nsmall = min_digits)
      formatted[is.na(x)] <- na_replace
      return(formatted)
    } else {
      return(x)
    }
  }))
  
  # Convert the list back to a data frame
  names(df_rounded) <- names(df)
  row.names(df_rounded) <- row.names(df)
  df_rounded
}


str_ci <- function(ci) paste0(c("LL", "UL"), "-CI", ci * 100, "%")

str_contrasts <- function(method, contrasts) {
  c(
    paste0("Slope estimation method = ", method),
    if (identical(contrasts[[1]], contrasts[[2]])) {
      paste0("Contrasts for the level and slope effects are coded with the ", 
           contrasts[1], " phase as the reference")
    } else {
      c(paste0("Contrasts for the level-effect are coded with the ", 
               contrasts[1], " phase as the reference"),
      paste0("Contrasts for the slope-effect are coded with the ", 
             contrasts[2], " phase as the reference"))
    }
  )
}

#' @describeIn pand Print results
#' @order 2
#' @param x An object returned by [pand()]
#' @export
#' 
print.sc_pand <- function(x, ...) {
  cat("Percentage of all non-overlapping data\n\n")
  cat("Method:", x$method, "\n\n")
  cat("PAND = ", round(x$pand, 1), "%\n", sep = "")

  if (x$method == "sort") {
    cat("\u03A6 = ", round(x$phi, 3), 
        " ; \u03A6\u00b2 = ", 
        round(x$phi^2, 3), "\n\n")
  }
  
  cat(x$n, " measurements (", x$n_a, " Phase A, ", x$n_b, " Phase B) in ", x$N, " cases", sep = "")
  cat("\n")
  cat("Overlapping data: n =",x$overlaps , 
      "; percentage =", round(x$perc_overlap, 1), 
      "\n")
  
  
  if (x$method == "sort") {
    
    ma <- x$matrix_counts
    ma <- cbind(ma, total = ma[1,] + ma[2,])
    ma <- rbind(ma, total = colSums(ma))
    cat("\n")
    cat("2 x 2 Matrix of percentages\n")
    print(round(ma / x$n * 100, 1))

    cat("\n")
    cat("2 x 2 Matrix of counts\n")
    print(round(ma, 1))
    cat("\n")
    cat("\nChi-Squared test:\n")
    out <- sprintf(
      "X\u00b2 = %.3f, df = 1, p = %.3f",
      x$chi_test$statistic, 
      x$chi_test$p.value
    )
    cat(out, "\n")
    
    cat("\nFisher exact test:\n")
    out <- sprintf(
      "Odds ratio = %.3f, p = %.3f",
      x$fisher_test$estimate, 
      x$fisher_test$p.value
    )
    cat(out, "\n")
  }  
}

#' @describeIn corrected_tau Print results
#' @order 2
#' @param x An object returned by [corrected_tau()]
#' @param nice If set TRUE (default) output values are rounded and optimized for
#'  publication tables.
#' @inheritParams print.sc
#' @export
print.sc_bctau <- function(x, nice = TRUE, digits = "auto", ...) {
  
  cat("Baseline corrected tau\n\n")
  
  if (x$repeated) {
    cat("Method: Siegel repeated median regression\n")
  } else {
    cat("Method: Theil-Sen regression\n")
  }
  
  cat("Kendall's tau", x$tau_method, "applied.\n")
  if (x$continuity) {
    cat("Continuity correction applied\n")
  } else {
    cat("Continuity correction not applied.\n")
  }
  cat("\n")
  
  for(i in seq_along(x$corrected_tau)) {
    if (digits == "auto") {
      x$corrected_tau[[i]]$p <- round(x$corrected_tau[[i]]$p, 3)
      x$corrected_tau[[i]]$z <- sprintf("%.2f", x$corrected_tau[[i]]$z)
      x$corrected_tau[[i]]$tau <- sprintf("%.2f", x$corrected_tau[[i]]$tau)
    } else {
      x$corrected_tau[[i]]$p <- round(x$corrected_tau[[i]]$p, digits)
      x$corrected_tau[[i]]$z <- round(x$corrected_tau[[i]]$z, digits)
      x$corrected_tau[[i]]$tau <- round(x$corrected_tau[[i]]$tau, digits)
    }
    
    if (nice) {
      x$corrected_tau[[i]]$p <- .nice_p(x$corrected_tau[[i]]$p)
    }
    
    rownames(x$corrected_tau[[i]]) <- x$corrected_tau[[i]]$Model
    cat(names(x$corrected_tau)[i], ":\n")
    print(x$corrected_tau[[i]][,-1], ...)
    cat("\n")
 
    if (x$correction[[i]]) cat("Baseline correction should be applied.\n\n")
    if (!x$correction[[i]]) cat("Baseline correction should not be applied.\n\n")
  }
  

  
  cat("\n")

  
}

#' @describeIn corrected_tau Export results as html
#' @order 3
#' @inheritParams export
#' @export
export.sc_bctau <- function(object, 
                              caption = NA, 
                              footnote = NA, 
                              filename = NA,
                              nice = TRUE, 
                              round = 2,
                              ...) {
  
  if (is.na(caption)) {
    caption <- paste0(
      "Baseline corrected tau for variable '", 
      attr(object, opt("dv")),  "'"
    )
  }
  
  if (is.na(footnote)) {
    footnote <- if (object$repeated) {
      "Method: Siegel repeated median regression"
    } else {
      "Method: Theil-Sen regression"
    }
    
    footnote <- c(footnote, paste("Kendall's tau", object$tau_method, "applied"))
    str_cc <- if (object$continuity) {
      "Continuity correction applied"
    } else {
      "Continuity correction not applied"
    }
    footnote <- c(footnote, str_cc)
  }
  
  x <- object$corrected_tau
  
  x <- mapply(
    function(x, apply) {
      x$"Correction recommended?" <- c(ifelse(apply, "Yes", "No"), rep("", nrow(x) - 1))
      x
    },
    x = x, 
    apply = object$correction,
    SIMPLIFY = FALSE
  )
  out <- do.call(rbind, x)
 
  rows <- sapply(x, nrow)
  end <- cumsum(sapply(x, nrow)) 
  start <- end - rows + 1
  row_group <- mapply(function(start, end) start:end, start, end, SIMPLIFY = FALSE)
  names(row_group) <- names(x)
  out <- round_numeric(out, round)
  if (nice) out$p <- .nice_p(out$p)
  
  table <- .create_table(
    out,
    caption = caption,
    footnote = footnote,
    row_group = row_group,
    ...
  )
  
  if (!is.na(filename)) .save_export(table, filename)
  
  table
  
}

#' Apply a function to each element in an scdf.
#'
#' This function applies a given function to each case of a multiple case scdf,
#' returning a list of the output of each function call.
#' 
#' If `simplify` is TRUE and the function returns a vector of values, the output
#' is combined into a data frame with case names as additional columns.
#' 
#' This is particularly useful for applying statistical models or summary
#' statistics to each case in an scdf.
#'
#' @param scdf A list of inputs to apply the function to.
#' @param fn  The function to apply to each element. Use a `.` as a
#'   placeholder for the scdf (e.g. `describe(.)`).
#' @param simplify If simplify is TRUE and `fn` returns a vector of values,
#'   `batch_apply` will return a data frame case names.
#' @return A list of the output of each function call.
#' @author Juergen Wilbert
#' @family data manipulation functions
#' @keywords manip
#' @examples
#' batch_apply(exampleAB, coef(plm(.)))
#'
#' @export
batch_apply <- function(scdf, fn, simplify = FALSE) {
  fn <- substitute(fn)
  out <- vector("list", length(scdf))
  for (i in seq_along(scdf)) {
    data <- list(. = scdf[i])
    out[[i]] <- eval(fn, envir = data)
  }
  names(out) <- names(scdf)

  if (simplify) {
    out <- as.data.frame(do.call(rbind, out))
    out$case <- rep(names(scdf), each = nrow(out) / length(names(scdf)))
    out$rownames <- rownames(out)
    rownames(out) <- NULL
    out <- out[, c(ncol(out) - 1, ncol(out), 1:(ncol(out) - 2))]
  }

  out
}

test_that("batch_apply applies a function to each case and preserves names", {
  scdf <- exampleAB
  
  res <- batch_apply(scdf, mean(as.data.frame(.)$values))
  
  expect_equal(res, list(Johanna = 69.25, Karolina = 68.05, Anja = 68.95))
  expect_equal(names(res), names(scdf))
})

test_that("batch_apply simplifies vector outputs to a data frame with case and rownames", {
  
  scdf <- exampleAB
  
  res <- batch_apply(scdf,
    {x <- as.data.frame(.)
     c(
       mean = mean(x$values),
       sd   = sd(x$values)
    )},
    simplify = TRUE
  ) 
  
  expect_equal(length(unlist(res)), 12)
                                                                                                                                                   
})

#' Standardize values of an scdf file
#'
#' *This function is superseded by the much more versatile 
#' \code{\link{transform.scdf}} function (see example below).*
#' This function scales the measured values of an scdf file. It allows for mean
#' centering and standardization based on each single-case data set or a
#' scaling across all cases included in an scdf.

#'
#' @inheritParams .inheritParams
#' @param var A character string or a vector of character strings with variable
#'   names that should be scaled.
#' @param center If set TRUE, data are mean centered.
#' @param scale If set TRUE, the standard deviation is set.
#' @param m The target mean for centering.
#' @param sd The target standard deviation for scaling
#' @param grand If set TRUE, scaling is based on the mean and standard deviation
#'   of all values across all single-cases within the scdf.
#' @return An scdf with the scaled values.
#' @author Juergen Wilbert
#' @family data manipulation functions
#' @keywords internal
#' @examples
#'
#' ## Standardize a multiple case scdf and compute an hplm
#' exampleAB_50 |>
#'   standardize("values", center = TRUE, scale = TRUE) |>
#'   hplm()
#' 
#' ## The more versatile transform function supersedes standardize:
#' exampleAB_50 |>
#'   transform(values = (values - mean(all(values))) / sd(all(values))) |>
#'   hplm()
#' @export

standardize <- function(data, 
                        var, 
                        center = TRUE, 
                        scale = FALSE, 
                        m = 0, 
                        sd = 1, 
                        grand = TRUE) {
  
  data <- .prepare_scdf(data)

  N <- length(data)

  if (grand) {
    means <- c()
    sds <- c()
    for (i in 1:length(var)) {
      means <- c(
        means, 
        mean(unlist(lapply(data, function(x) x[, var[i]])), na.rm = TRUE)
      )
      sds <- c(
        sds, 
        sd(unlist(lapply(data, function(x) x[, var[i]])), na.rm = TRUE)
      )
    }

    for (case in 1:N) {
      for (i in 1:length(var)) {
        values <- data[[case]][, var[i]]
        if (center && scale) {
          data[[case]][, var[i]] <- (values - means[i]) / sds[i] * sd + m
        }
        if (center && !scale) {
          data[[case]][, var[i]] <- (values - means[i]) + m
        }
        if (!center && scale) {
          data[[case]][, var[i]] <- ((values - means[i]) / sds[i] * sd) + means[i]
        }
      }
    }
  }

  if (!grand) {
    for (case in 1:N) {
      for (i in 1:length(var)) {
        values <- data[[case]][, var[i]]
        m_case <- mean(data[[case]][, var[i]], na.rm = TRUE)
        sd_case <- sd(data[[case]][, var[i]], na.rm = TRUE)
        if (center && scale) {
          data[[case]][, var[i]] <- (values - m_case) / sd_case[i] * sd + m
        }
        if (center && !scale) {
          data[[case]][, var[i]] <- (values - m_case) + m
        }
        if (!center && scale) {
          data[[case]][, var[i]] <- ((values - m_case) / sd_case * sd) + m_case
        }
      }
    }
  }

  data
}

#' Export scan objects to html or latex
#' 
#' Export creates html files of tables or displays them directly in the viewer
#' pane of rstudio. When applied in rmarkdown/quarto, tables can also be created
#' for pdf/latex output.
#' 
#' The function uses either the kableExtra or the gt package to create the tables.
#' The default engine can be set via the option `scan.export.engine`. Additional
#' options for kable and kable_styling can be set via the options
#' `scan.export.kable` and `scan.export.kable_styling`.
#' The default options can be viewed and modified via
#' `options("scan.export.kable")` and
#' `options("scan.export.kable_styling")`.
#'
#' @param object An scdf or an object exported from a scan function.
#' @param caption Character string with table caption. If left NA (default) a
#'   caption will be created based on the exported object.
#' @param footnote Character string with table footnote. If left NA (default) a
#'   footnote will be created based on the exported object.
#' @param filename String containing the file name. If a filename is given the
#'   output will be written to that file.
#' @param cols Defines which columns are included when exporting an scdf. It is
#'   either a vector of variable names or the string "main" will select the
#'   central variables.
#' @param flip If TRUE, some objects are exported with rows and columns flipped.
#' @param round Integer passed to the digits argument used to round values.
#' @param decimals Decimal places that are reported.
#' @param select A character vector containing the names of the variables to be
#'   included. If the vector is named, the variables will be renamed
#'   accordingly.
#' @param summary If TRUE, exports the summary of an `scdf`.
#' @param ... Further Arguments passed to internal functions.
#' @return  Returns or displays a specially formatted html (or latex) file.
#' @export

export <- function (object, ...) {
  UseMethod("export")
}

.select <- function(df, select) {

  if (identical(select, "none") || 
      identical(select, "") ||
      identical(select, NA) ||
      is.null(select) ||
      identical(select, FALSE)) return(df)
  
  if (!all(select %in% names(df)) && !is.numeric(select)) {
    warn(
      "`select` arguments has variable names that are not included in " ,
      "the output table: valid names are: ", 
      paste(names(df), collapse = ", "), "."
    )
  }
  df <- df[, select]
  if (!is.null(names(select))) {
    if (is.numeric(select)) {
      select <- setNames(names(df)[select], names(select))
    }
    select <- mapply(function(x, y) ifelse(x == "",y,x), names(select), select)
    names(df) <- select
  }
  df
}

# .join_kabel <- function(kable_options) {
#   
#   default_kable <- getOption("scan.export.kable")
#   
#   tmp <- which(!(names(default_kable) %in% names(kable_options)))
#   kable_options <- c(kable_options, default_kable[tmp])
#   
#   kable_options
# } 
# 
# .join_kabel_styling <- function(kable_styling_options) {
#   
#   default_kable_styling <- getOption("scan.export.kable_styling")
#   
#   tmp <- which(
#     !(names(default_kable_styling) %in% names(kable_styling_options))
#   )
#   
#   kable_styling_options <- c(kable_styling_options, default_kable_styling[tmp])
#   
#   kable_styling_options
# } 

.save_export <- function(x, filename) {
  
  
  if (getOption("scan.export.engine") == "kable") {
    kableExtra::save_kable(x, filename, zoom = 2)
  }
  
  if (getOption("scan.export.engine") == "gt") {
    gt::gtsave(x, filename)
  }
  
}

.add_footnote <- function(x, footnote) {
  if (length(footnote) > 1) {
    footnote <- paste0(
      footnote, 
      collapse = getOption("scan.export.footnote.collapse")
    ) |> paste0(".")
  }
  if (inherits(x, "kableExtra")) {
    footnote <- gsub("<br>", "\n", footnote)
    out <- kableExtra::footnote(
      x, general = footnote, threeparttable = TRUE,
      footnote_as_chunk = TRUE
    )
  }
}

.create_table <- function(x, 
                          options = getOption("scan.export.kable"), 
                          kable_styling_args = getOption("scan.export.kable_styling"), 
                          caption = NULL,
                          footnote = NULL,
                          align = NULL,
                          ...) {
  
  if (getOption("scan.export.engine") == "gt") {
    table <- export_table_gt(
      x, 
      title = caption, 
      footnote = footnote,
      ...
    )
    return(table)
  }
  
  rownames(x) <- NULL
  
  if (!is.null(align)) options$align <- align
  
  if (is.null(options$align))  
    options$align <- c("l", rep("c",  ncol(x) - 1))
  
  if (is.null(options$caption))  
    options$caption <- caption
  
  options$x <- x
  table <- do.call(kable, options)
  kable_styling_args$kable_input <- table
  table <- do.call(kable_styling, kable_styling_args)
  
  if (!is.null(footnote)) {
    if (!identical(footnote, NA) && !identical(footnote, ""))
      table <- .add_footnote(table, footnote)
  }
  
  table
}

###### gt ####

export_table_gt <- function(x, 
                            title = NULL, 
                            footnote = NULL, 
                            spanner = NULL,
                            row_group = NULL,
                            rownames = FALSE,
                            cols_label = NULL,
                            decimals = NULL,
                            prefix_title = getOption("scan.export.title.prefix"),
                            fmt_markdown = FALSE,
                            ...) {
          
  
  while(TRUE) {
    id <- which(duplicated(names(x)))
    if (length(id) == 0) break 
    names(x)[id] <- paste0(" ", names(x)[id], " ")
    
  }
  
  if (!is.null(title) && title != "") title <- paste0("*", title, "*")
  if (!is.null(footnote) && 
      !identical(footnote, "") && 
      !identical(footnote, NA)) {
    footnote <- paste0(
      "*Note.* ", 
      paste0(footnote, collapse = getOption("scan.export.footnote.collapse")), 
      "."
    )
  }
  
  if (!inherits(x, "data.frame")) {
    x <- as.data.frame(x)
    rownames(x) <- NULL
  }
  if (rownames && !is.null(rownames(x))) x <- cbind(" " = rownames(x), x)
  
  out <- do.call(gt::gt, list(data = x))|> gt_apa_style()
  
  if (!is.null(title)) {
    if (!is.null(prefix_title)) {
      out <- gt::tab_header(out, title = prefix_title, subtitle = gt::md(title))
    } else {
      out <- gt::tab_header(out, title = gt::md(title))
    }
      
  }
  if (!is.null(row_group)) {
    for(i in length(row_group):1)
      out <- gt::tab_row_group(
        out, label = names(row_group)[i], rows = row_group[[i]]
      )
    for(i in length(row_group):1)  
      out <- gt::tab_style(
        out, style = gt::cell_text(align = "center"),
        locations = gt::cells_row_groups(groups = names(row_group)[i])
      )
  }
  if (!is.null(spanner)) {
    for(i in seq_along(spanner)) {
      out <- gt::tab_spanner(
        out, 
        label = names(spanner)[i], 
        columns = spanner[[i]]
      )  
    }
  }
  
  if (!is.null(cols_label)) out <- gt::cols_label(out, .list = cols_label)
  if (!is.null(footnote) && !identical(footnote, "") && !identical(footnote, NA)) 
    out <- gt::tab_footnote(out, gt::md(footnote))
  if (!is.null(decimals)) out <- gt::fmt_number(out, decimals = decimals)
  if (fmt_markdown) out <- gt::fmt_markdown(out, columns = everything())
  out
}

gt_apa_style <- function(gt_tbl) {
  gt_tbl  |> 
    gt::tab_options(
      table.border.bottom.color = "white",
      #table.border.bottom.width = 3,
      
      table.border.top.color = "white",
      #table.border.top.width = 3,
      
      table_body.border.bottom.color = "black",
      table_body.border.bottom.width = 3,
      
      table_body.border.top.color = "black",
      table_body.border.top.width = 3,
      
      table_body.hlines.width = 0,
      
      heading.align = "left",
      heading.border.bottom.width = 3,
      heading.border.bottom.color = "black",
      heading.title.font.size = "100%",
      heading.subtitle.font.size = "100%",
      column_labels.border.bottom.width = 2,
      column_labels.border.bottom.color = "black",
      column_labels.border.top.width = 3,
      column_labels.border.top.color = "black",
      
      row_group.border.bottom.color = "white",
      row_group.border.bottom.style = NULL,
      row_group.border.bottom.width = NULL
      
    )  |> 
    gt::opt_table_font(font = "times") |> 
    gt::cols_align(align = "center") |> 
    gt::cols_align(align = "left", columns = 1)
}

rename_phase_duplicates <- function(phase) {
  ph_str <- rle(as.character(phase))
  ts <- table(ph_str$values)
  new_phase <- ph_str$values
  for(i in 1:length(ts)) {
    if (ts[i] > 1) {
      new_phase[which(ph_str$values == names(ts[i]))] <- 
        paste0(names(ts[i]), "(", 1:ts[i], ")")
    }
  }
  new_phase <- mapply(
    function(x,y) rep(x, y), 
    new_phase, ph_str$lengths, 
    SIMPLIFY = FALSE, USE.NAMES = FALSE
  ) |> unlist()
  as.factor(new_phase)
}

rownames_to_first_column <- function(x, name = "Rownames") {
  x <- cbind(.rownames = rownames(x), x)
  rownames(x) <- NULL
  names(x)[1] <- name
  x
}

#' Print methods for scan objects
#'
#' @param x Object 
#' @param digits The minimum number of significant digits to be use. 
#' If set to "auto" (default), values are predefined.
#' @param ... Further parameters passed to the print function
#' @name print.sc
#' @keywords internal
NULL


.note_vars <- function(x) {
  v <- any(attr(x, opt("dv")) != "values")
  p <- attr(x, opt("phase")) != "phase"
  m <- attr(x, opt("mt")) != "mt"
  if (v || p || m) { 
    cat(
      "\nThe following variables were used in this analysis:\n'", 
      paste0(attr(x, opt("dv")), collapse = "/ "), 
      "' as dependent variable, '", 
      paste0(attr(x, opt("phase")), collapse = "/ "), 
      "' as phase variable, and '", 
      paste0(attr(x, opt("mt")), collapse = "/ "), 
      "' as measurement-time variable.\n", 
      sep = ""
    )
  }
}

#' Replacing missing measurement points in single-case data
#'
#' The `fillmissing()` function replaces missing measurements in single-case
#' data. It linearly interpolates missing data points between two existing
#' measurements for all variables except the measurement time and phase. The
#' measurement time variable is filled with the missing time points. The phase
#' variable is copied from the previous measurement time point. If mt values are
#' missing (`NA`), they are also interpolated if `interpolate_na = TRUE`.
#'
#' @details The `fill_missing()` function is designed to handle single-case data
#' with missing measurement points. It performs linear interpolation to estimate
#' the missing values based on the existing data points. The function iterates
#' through each single-case in the provided single-case data frame (scdf) and
#' identifies gaps in the measurement time variable. For each gap, it calculates
#' the step size for linear interpolation and fills in the missing values for
#' all target variables (i.e., all variables except the measurement time and
#' phase). The interpolated data points are then added to the single-case data
#' frame, and the final result is sorted by measurement time. This function is
#' particularly useful for preparing single-case data for further analysis, such
#' as calculating overlap indices or conducting randomization tests, where
#' continuous measurement times are required. It ensures that the data is
#' complete by filling in the missing measurement points in a systematic manner.

#' 
#' @inheritParams .inheritParams
#' @param interpolate_na If set `TRUE`, `NA` values in the measurement time variable are
#'   also interpolated. Default is `TRUE`.
#' @return A single-case data frame with interpolated missing data points.
#' @author Juergen Wilbert
#' @family data manipulation functions
#' @keywords manip
#' @examples
#'
#' ## In his study, Grosche (2011) could not realize measurements each
#' ## single week for all participants. During the course of 100 weeks,
#' ## about 20 measurements per person at different times were administered.
#'
#' ## Fill missing values in a single-case dataset with discontinuous
#' ## measurement times
#' Grosche2011filled <- fill_missing(Grosche2011)
#' study <- c(Grosche2011[2], Grosche2011filled[2])
#' names(study) <- c("Original", "Filled")
#' study
#' 
#' ## An example with multiple interpolated variables
#' 
#' rolf_n <- exampleAB_add
#' rolf_n[[1]] <- rolf_n[[1]][-c(3,7,8),]
#' rolf_f <- fill_missing(rolf_n)
#' study1 <- c("original" = exampleAB_add, "interpolated" = rolf_f)
#' study1
#' 
#' ## Example with missing NAs in measurement time
#' Maggie2 <- random_scdf(design(level = list(0,1)), seed = 123)
#' Maggie2_n <- Maggie2
#' Maggie2_n[[1]][c(5,12:14,20), "mt"] <- NA
#' Maggie2_f <- fill_missing(Maggie2_n)
#' study2 <- c("original" = Maggie2, "interpolated" = Maggie2_f)
#' study2
#'
#' @export
fill_missing <- function(data, 
                         dvar, 
                         mvar, 
                         pvar,
                         interpolate_na = TRUE) {

  # set attributes to arguments else set to defaults of scdf
  if (missing(dvar)) dvar <- dv(data) else dv(data) <- dvar
  if (missing(mvar)) mvar <- mt(data) else mt(data) <- mvar
  if (missing(pvar)) pvar <- phase(data) else phase(data) <- pvar
  
  #  preserve attributes
  source_attributes <- attributes(data)
  
  data <- .prepare_scdf(data)
  N <- length(data)
  
  for(i_case in 1:N) {
    dat <- data[[i_case]]
    
    # get variable names of all variables that are not mvar or pvar
    target_vars <- names(dat)[!(names(dat) %in% c(mvar, pvar))]
    
    # interpolate missing measurement times
    if (interpolate_na) {
      dat[[mvar]] <- .interpolate(dat[[mvar]]) |> round()
    }
    
    new_dat <- dat[0, , drop = FALSE]  # empty data frame to store new data

    for(i_row in 1:(nrow(dat) - 1)) {
      
      # skip if current or next measurement time is NA
      if (is.na(dat[[mvar]][i_row]) || is.na(dat[[mvar]][i_row + 1])) next
      
      # check for missing measurement times
      if (dat[i_row + 1, mvar] - dat[i_row, mvar] == 1) next
      
      # interpolate missing values for all target variables
      new_rows <- (dat[i_row, mvar] + 1) : (dat[i_row + 1, mvar] - 1)
      
      tmp <- dat[rep(NA_integer_, length(new_rows)), , drop = FALSE]
      tmp[pvar] <- dat[i_row, pvar]
      new_mt <- .interpolate(
        c(dat[[mvar]][i_row], tmp[[mvar]], dat[[mvar]][i_row + 1] )
      ) |> round()
      tmp[[mvar]] <- new_mt[2:(length(new_mt) -1)]
      
      for(i_col in target_vars) {
        new_values <- .interpolate(
          c(dat[[i_col]][i_row], tmp[[i_col]], dat[[i_col]][i_row + 1] )
        )
        tmp[[i_col]] <- new_values[2:(length(new_values) -1)]
      }
      new_dat <- rbind(new_dat, tmp)
    }
    # combine original and new data
    new_dat <- rbind(dat, new_dat)
    
    # sort by measurement time and store
    data[[i_case]] <- new_dat[sort.list(new_dat[[mvar]]), ]
  }
  
  # restore attributes
  attributes(data) <- source_attributes

  return(data)
}

.interpolate <- function(mt) {
  
  if (!any(is.na(mt))) return(mt)

  x <- mt
  n <- length(x)
  na <- is.na(x)
  idx <- seq_len(n)
  
  # Only interpolate NAs that have a following non-NA value
  ok <- na & (idx < max(which(!na)))
  
  x[ok] <- approx(x = idx[!na], y = x[!na], xout = idx[ok], method = "linear")$y
  
  return(x)
}

#' Trend analysis for single-cases data
#'
#' The `trend()` function provides an overview of linear trends in single case
#' data. By default, it provides the intercept and slope of a linear and
#' quadratic regression of measurement time on scores. Models are calculated
#' separately for each phase and across all phases. For more advanced use, you
#' can add regression models using the R-specific formula class.
#'
#' @details The function computes separate regression models for each phase and
#'   for the whole data. By default two models are computed: a linear model and
#'   a quadratic model. Additionally, custom models can be specified using the
#'   `model` argument. The measurement time variable is adjusted such that the
#'   first measurement time point of each phase is set to the value specified in
#'   the `first_mt` argument (default = 0). This means that if `first_mt = 0`,
#'   the first measurement time point of each phase is set to 0, if `first_mt =
#'   1`, the first measurement time point of each phase is set to 1, and so on.
#'   This adjustment allows for a more intuitive interpretation of the
#'   regression coefficients, especially the intercept, which then represents
#'   the estimated value at the beginning of each phase.
#'
#' @inheritParams .inheritParams
#' @param first_mt A numeric setting the value for the first measurement-time.
#'   Default = 0.
#' @param offset (Deprecated. Please use first_mt). An offset for the first
#'   measurement-time of each phase. If `offset = 0`, the phase measurement is
#'   handled as MT 1. Default is `offset = -1`, setting the first value of MT to
#'   0.
#' @param model A string or a list of (named) strings each depicting one
#'   regression model. This is a formula expression of the standard R class. The
#'   parameters of the model are `values`, `mt` and `phase`.
#' @return A list of class `sc_trend` containing:
#' \item{trend}{A matrix containing the results (Intercept, B and beta)
#'   of separate regression models for phase A, phase B, and the whole data.}
#' \item{first_mt}{Numeric argument from function call (see arguments
#' section).}
#' @author Juergen Wilbert
#' @seealso [describe()]
#' @family regression functions
#' @examples
#'
#' ## Compute the linear and squared regression for a random single-case
#' design <- design(slope = 0.5)
#' matthea <- random_scdf(design)
#' trend(matthea)
#'
#' ## Besides the linear and squared regression models compute two custom models:
#' ## a) a cubic model, and
#' ## b) the values predicted by the natural logarithm of the
#' ## measurement time.
#' design <- design(slope = 0.3)
#' ben <- random_scdf(design)
#' trend(
#'   ben,
#'   model = list("Cubic" = values ~ mt^3, "Log Time" = values ~ log(mt)),
#'   first_mt = 1 # must be set to 1 because log(0) would be -Inf
#' )
#'
#' @export
trend <- function(data, dvar, pvar, mvar, 
                  offset = "deprecated",
                  first_mt = 0,
                  model = NULL) {

  check_args(
    has_length(data, 1, "trend can not be applied to more than one case.")
  )
  
  if (is.numeric(offset)) first_mt <- offset + 1
  
  # set attributes to arguments else set to defaults of scdf
  if (missing(dvar)) dvar <- dv(data) else dv(data) <- dvar
  if (missing(pvar)) pvar <- phase(data) else phase(data) <- pvar
  if (missing(mvar)) mvar <- mt(data) else  mt(data) <- mvar
  
  data <- .prepare_scdf(data)
  data <- data[[1]]
  
  design <- rle(as.character(data[, pvar]))$values
  while(any(duplicated(design))) {
    design[anyDuplicated(design)] <- paste0(
      design[anyDuplicated(design)], 
      ".phase", 
      anyDuplicated(design)
    )
  }
  
  phases <- .phasestructure(data, pvar = pvar)
  
  formulas <- c(
    "Linear" = formula(paste0(dvar, " ~ ", mvar)) , 
    "Quadratic" = formula(paste0(dvar, " ~ I(", mvar, "^2)"))
  )
  formulas_names <- c("Linear", "Quadratic")
  if(!is.null(model)) {
    formulas <- c(formulas, model)
    formulas_names <- names(formulas)#c(formulas_names, names(model))
  }
  tmp <- length(design) + 1
  rows <- paste0(paste0(rep(formulas_names, each = tmp), "."), c("ALL", design))
  
  ma <- matrix(NA, nrow = length(rows), ncol = 3)
  row.names(ma) <- rows
  colnames(ma) <- c("Intercept", "B", "Beta")
  ma <- as.data.frame(ma)
  
  for(i_formula in 1:length(formulas)) {
    data_phase <- data
    mvar_correction <- min(data_phase[[mvar]], na.rm = TRUE) - first_mt
    data_phase[[mvar]] <- data_phase[[mvar]] - mvar_correction
    
    .row <- which(rows == paste0(formulas_names[i_formula], ".ALL"))
    ma[.row, 1:3] <- .beta_weights(lm(formulas[[i_formula]], data = data_phase))
    for(p in 1:length(design)) {
      data_phase <- data[phases$start[p]:phases$stop[p], ]
      mvar_correction <- min(data_phase[[mvar]], na.rm = TRUE) - first_mt
      data_phase[[mvar]] <- data_phase[[mvar]] - mvar_correction 
      .row <- which(rows == paste0(formulas_names[i_formula], ".", design[p]))
      ma[.row, 1:3] <- .beta_weights(lm(formulas[[i_formula]], data=data_phase))
    }
  }
  
  out <- list(
    trend = ma, 
    first_mt = first_mt,
    formulas = formulas, 
    offset = offset, 
    design = design
  )
  class(out) <- c("sc_trend")
  attributes(out)[opts("phase", "mt", "dv")] <- list(pvar, mvar, dvar)
  out
}

test_that("trend", {
  
  scdf <- exampleA1B1A2B2$Moritz
  result <- trend(
    scdf,
    model = list("Cubic" = values ~ I(mt^3), "Log Time" = values ~ log(mt)),
    first_mt = 1
  )
  values <- result$trend |> sum() |> round(3)
  expect_equal(values, 386.582)

})

#' Select a subset of cases from an scdf
#' 
#' This function allows to select a subset of cases from an scdf by
#' specifying either the case names or their numeric indices. Negative
#' selection is also supported.
#'
#' @inheritParams .inheritParams
#' @param ... Selection criteria. Either numeric, objectnames, or as characters.
#' @return An scdf with a subset of cases.
#' @author Juergen Wilbert
#' @export
#' @family data manipulation functions
#' @examples
#' select_cases(exampleAB, Johanna, Karolina)
#' select_cases(exampleAB, c(Johanna, Karolina))
#' select_cases(exampleAB, 1,2)
#' select_cases(exampleAB, 1:2)
#' select_cases(exampleAB, -Johanna)
#' select_cases(exampleAB, -c(Johanna, Karolina))
#' v <- c("Moritz", "Jannis")
#' select_cases(exampleA1B1A2B2, v)
#' @export
select_cases <- function(scdf, ...) {
  cases <- as.list(substitute(list(...)))[-1]
  
  nl <- as.list(seq_along(scdf))
  names(nl) <- names(scdf)
  
  selection <- lapply(cases, function(x) scdf[eval(x, envir = nl, enclos = parent.frame())])
  
  out <- selection[[1]]
  if (length(selection) > 1) 
    for(i in 2:length(selection)) out <- c(out, selection[[i]])
  out
}

#' Reliable change index
#'
#' The `rci()` function computes indices of reliable change (Wise, 2004)
#' and corresponding descriptive statistics.
#' 
#' The reliable change index (RCI) indicates whether the change in scores from
#' phase A to phase B is statistically reliable, that is, whether the change is
#' larger than would be expected due to measurement error alone. The RCI can be
#' calculated using different methods (Jacobsen & Truax, 1991; Christensen &
#' Mendoza, 1986). The most common method is that of Jacobsen and Truax (1991) 
#' which uses the standard error of measurement in phase A to
#' calculate the RCI. Christensen and Mendoza (1986) proposed an alteration of
#' the RCI which uses the standard error of the difference between phase A and
#' phase B means. Both methods are implemented in the `rci()` function. 
#' 
#' @inheritParams .inheritParams
#' @param rel Reliability of the measure, used to compute the standard error.
#' @param ci Width of confidence interval as a decimal. Default is `ci = 0.95`
#'   applying a 95 percent confidence interval.
#' @param graph If set `TRUE`, a box plot of phase A and B scores is displayed.
#'   `graph = FALSE` by default.
#' @author Juergen Wilbert
#' @references Christensen, L., & Mendoza, J. L. (1986). A method of assessing
#'   change in a single subject: An alteration of the RC index. *Behavior
#'   Therapy, 17*, 305-308.
#'
# #'   Hageman, W. J. J., & Arrindell, W. A. (1993). A further refinement of the
# #'   reliable change (RC) index by improving the pre-post difference score:
# #'   Introducing RCID. *Behaviour Research and Therapy, 31*, 693-700.
#'
#'   Jacobson, N. S., & Truax, P. (1991). Clinical Significance: A statistical
#'   approach to defining meaningful change in psychotherapy research.
#' *Journal of Consulting and Clinical Psychology, 59*, 12-19.
#'
#'   Wise, E. A. (2004). Methods for analyzing psychotherapy outcomes: A review
#'   of clinical significance, reliable change, and recommendations for future
#'   directions. *Journal of Personality Assessment, 82*, 50 - 59.
#'
#' @examples
#'
#' ## Report the RCIs of the first case from the byHeart data and include a graph
#' rci(byHeart2011[1], graph = TRUE, rel = 0.8)
#'
#' @export
rci <- function(data, dvar, pvar, 
                rel, 
                ci = 0.95, 
                graph = FALSE, 
                phases = c(1, 2)) {
  
  check_args(
    at_most(length(data), 1, 
            "RCI can not be applied to more than one case."),
    within(rel, 0, 1),
    within(ci, 0, 1),
    is_logical(graph)
  )
  
  # set attributes to arguments else set to defaults of scdf
  if (missing(dvar)) dvar <- dv(data) else dv(data) <- dvar
  if (missing(pvar)) pvar <- phase(data) else phase(data) <- pvar
  
  data <- .prepare_scdf(data, na.rm = TRUE)
  data <- recombine_phases(data, phases = phases)$data
  
  N <- length(data)
  
  #if(N > 1) {
  #  abort("Multiple single-cases are given. Calculations can only be applied to one single-case data set.\n")
  #}
  
  A <- lapply(data, function(x) x[, dvar][x[, pvar] == "A"])
  B <- lapply(data, function(x) x[, dvar][x[, pvar] == "B"])
  A <- unlist(A)
  B <- unlist(B)
  sA <- sd(A, na.rm = TRUE)
  sB <- sd(B, na.rm = TRUE)
  mA <- mean(A, na.rm = TRUE)
  mB <- mean(B, na.rm = TRUE)
  nA <- sum(!is.na(A))
  nB <- sum(!is.na(A))
  n <- nA + nB
  seA <- sA * sqrt(1 - rel)
  seB <- sB * sqrt(1 - rel)
  stand_dif <- (mB-mA)/sd(c(A,B))
  se_dif <- sqrt(2*seA^2)
  
  cor_a_b <- rel
  xA <- mA
  xB <- mB
  rel_A <- rel
  rel_B <- rel
  
  rdd <- (sA^2*rel_A + sB^2*rel_B - 2*sA*sB*cor_a_b) / 
         (sA^2 + sB^2 - 2*sA*sB*cor_a_b)
  
  rci_jacobsen <- (mB - mA) / seA
  rci_christensen <- (mB - mA) / se_dif
  rci_hageman <- (xB - xA) * rdd + (mB - mA) * (1 - rdd) / 
    sqrt(seA^2 + seB^2)
  
  descriptives_ma <- matrix(
    c(nA, nB, mA, mB, sA, sB, seA, seB), 2, 4, 
    dimnames = list(c("A-Phase", "B-Phase"), c("n", "mean", "SD", "SE"))
  ) 
  z <- qnorm(ci + 0.5 * (1 - ci))
  ci_ma <- matrix(
    NA, 2, 2, byrow = TRUE, 
    dimnames = list(c("A-Phase", "B-Phase"), c("Lower", "Upper"))
  )
  ci_ma[1,1] <- mA - z * seA
  ci_ma[1,2] <- mA + z * seA
  ci_ma[2,1] <- mB - z * seB
  ci_ma[2,2] <- mB + z * seB
  
  if(graph) {
    dat <- cbind(ci_ma[1, ], ci_ma[2, ])
    colnames(dat) <- c("A-Phase", "B-Phase")
    main <- sprintf("%d%% confidence interval (rtt = %.2f)", ci * 100, rel)
    boxplot(dat, ylab = "Mean", main = main)
  }
  
  rci_ma <- matrix(
    c(rci_jacobsen, rci_christensen), 2, 1, 
    dimnames = list(
      c(
        "Jacobson et al.", 
        "Christensen and Mendoza"), 
        #"Hageman and Arrindell"), 
      "RCI"
    )
  )
  out <- list(
    rci = rci_ma, 
    stand_dif = stand_dif, 
    se_dif = se_dif,
    conf = ci_ma, 
    conf_percent = ci, 
    reliability = rel, 
    descriptives = descriptives_ma,
    rdd = rdd
  ) 
  class(out) <- c("sc_rci")
  attributes(out)[opts("phase", "dv")] <- list(pvar, dvar)
  out
}

#' Multivariate Piecewise linear model / piecewise regression
#'
#' The [mplm()] function computes a multivariate piecewise regression model. The
#' function automatically creates the regression formula based on the provided
#' data and the selected options. The default model includes trend, level, and
#' slope effects for each dependent variable. The regression formula can be
#' changed by providing a custom formula to the `update` argument.
#' 
#' The function currently only supports single-case data (i.e., one case per
#' dataset). For multilevel piecewise regression models, please use the [hplm] 
#' function.
#'
#' @inheritParams .inheritParams
#' @order 1
#' @param formula Defaults to the standard piecewise regression model. The
#'   parameter phase followed by the phase name (e.g., `phaseB`) indicates the
#'   level effect of the corresponding phase. The parameter 'inter' followed by
#'   the phase name (e.g., `interB`) adresses the slope effect based on the
#'   method provide in the model argument (e.g., `"B&L-B"`). The formula can be
#'   changed for example to include further variables into the regression model.
#' @param update An easier way to change the regression formula (e.g., `. ~ . +
#'   newvariable`).
#' @param na.action Defines how to deal with missing values.
#' @param ... Further arguments passed to the [lm()] function.
#' @return 
#'  |  |  |
#'  | --- | --- |
#'  | `model` | Character string from function call (see arguments above). |
#'  | `contrast` | List with contrast definitions. |
#'  | `full.model` | Full regression model list. |
#'  | `formula` | 	Formula of the mplm model. | 
#' @author Juergen Wilbert
#' @family regression functions
#' @examples
#' res <- mplm(Leidig2018$`1a1`,
#'   dvar = c("academic_engagement", "disruptive_behavior")
#' )
#' print(res)
#' ## also report standardized coefficients:
#' print(res, std = TRUE)
#' @export

mplm <- function(data, dvar, mvar, pvar, 
                 model = c("W", "H-M", "B&L-B", "JW"),
                 contrast = c("first", "preceding"),
                 contrast_level = c(NA, "first", "preceding"),
                 contrast_slope = c(NA, "first", "preceding"),
                 trend = TRUE, 
                 level = TRUE, 
                 slope = TRUE, 
                 formula = NULL, 
                 update = NULL, 
                 na.action = na.omit, 
                 ...) {
 
  # set attributes to arguments else set to defaults of scdf
  if (missing(dvar)) dvar <- dv(data) else dv(data) <- dvar
  if (missing(pvar)) pvar <- phase(data) else phase(data) <- pvar
  if (missing(mvar)) mvar <- mt(data) else mt(data) <- mvar
  
  check_args(
    by_call(model),
    by_call(contrast),
    by_call(contrast_level),
    by_call(contrast_slope),
    is_true(length(dvar) > 1, "At least two dependent variables must be defined in the argument 'dvar'.")
  )
  model <- model[1]
  contrast <- contrast[1]
  contrast_level <- contrast_level[1]
  contrast_slope <- contrast_slope[1]
  
  if (model == "JW") {
    contrast_level <- "preceding"
    contrast_slope <- "preceding"
    model <- "B&L-B"
  }
  
  if (is.na(contrast_level)) contrast_level <- contrast
  if (is.na(contrast_slope)) contrast_slope <- contrast
  
  data <- .prepare_scdf(data)

  N <- length(data)
  if (N > 1) {
    abort("Procedure could not be applied to more than one case.")
  }

  ### model definition
  tmp_model <- .add_dummy_variables(
    data = data, 
    model = model, 
    contrast_level = contrast_level, 
    contrast_slope = contrast_slope
  )
  
  data <- tmp_model$data[[1]]

  if (is.null(formula)) {
    formula <- .create_fixed_formula(
      dvar = "y", 
      mvar = mvar, 
      slope = slope, 
      level = level, 
      trend = trend, 
      var_phase = tmp_model$var_phase, 
      var_inter = tmp_model$var_inter
    )
  }

  if (!is.null(update)) formula <- update(formula, update)

  y <- as.matrix(data[, dvar])
 
  full <- lm(formula, data = data, na.action = na.action, ...)
  full$coef_std <- .std_lm(full)
  
  null <- lm(y ~ 1, na.action = na.action, ...)
  
  out <- structure(
    list(
      model = model, 
      contrast = list(level = contrast_level, slope = contrast_slope), 
      full.model = full,
      null_model = null,
      formula = formula
    ),
    class = c("sc_mplm")
  )

  attributes(out)[opts("phase", "mt", "dv")] <- list(pvar, mvar, dvar)
  out
}

test_that("anova wrappers delegate to underlying models and match numeric results", {
  # sc_plm
  mod0 <- plm(exampleAB$Johanna, level = FALSE, slope = FALSE)
  mod1 <- plm(exampleAB$Johanna)
  
  res_wrap_plm <- anova(mod0, mod1)
  res_base_plm <- anova(mod0$full.model, mod1$full.model)
  
  expect_equal(res_wrap_plm, res_base_plm)

  # sc_mplm
  data_mplm <- Leidig2018$`1a1`
  m0 <- mplm(
    data_mplm,
    update = . ~ 1,
    dvar = c("academic_engagement", "disruptive_behavior")
  )
  m1 <- mplm(
    data_mplm,
    trend = FALSE,
    dvar = c("academic_engagement", "disruptive_behavior")
  )
  m2 <- mplm(
    data_mplm,
    dvar = c("academic_engagement", "disruptive_behavior")
  )
  
  res_wrap_mplm <- anova(m0, m1, m2)
  res_base_mplm <- anova(m0$full.model, m1$full.model, m2$full.model)
  
  expect_equal(res_wrap_mplm, res_base_mplm)
  
  # sc_hplm
  h0 <- hplm(Leidig2018, trend = FALSE, slope = FALSE, level = FALSE)
  h1 <- hplm(Leidig2018, trend = FALSE)
  h2 <- hplm(Leidig2018)
  
  res_wrap_hplm <- anova(h0, h1, h2)
  res_base_hplm <- anova(h0$hplm, h1$hplm, h2$hplm)
  row.names(res_base_hplm) <- NULL  # match wrapper behavior
  
  expect_equal(res_wrap_hplm, res_base_hplm)
  
})

test_that("mplm", {
  data <- Leidig2018$`1a1`
  result <- mplm(
    data,
    dvar = c("academic_engagement", "disruptive_behavior")
  )

  expect_s3_class(result, "sc_mplm")
 
  expect_equal(
    round(sum(coef(result$full.model)), 3),
    4.575
  )
  
  expect_equal(
    round(sum(result$full.model$coef_std), 4),
    -0.0102
  )
  
})

#' Import scdf – RStudio Addin
#' @export
import_scdf <- function() {
  #suppressPackageStartupMessages({
  #  library(shiny)
  #  library(bslib)
  #})

  # ui ------
  ui <- miniUI::miniPage(
    shiny::tags$head(
      shiny::tags$style(shiny::HTML("
      body { zoom: 0.8; }
    "))
    ),
    miniUI::gadgetTitleBar("scdf import"),
    miniUI::miniContentPanel(
      shiny::verbatimTextOutput("status"),
      shiny::fillRow(
        flex = c(1,1),
        shiny::wellPanel(
          shiny::fileInput("file", "Choose file",
                           accept = c(".csv", ".xlsx", ".xls", ".rds", ".txt")),
          shiny::textInput("na", "NA strings (comma-sep.)", '"", NA'),
          shiny::selectInput("sep", "CSV separator", 
                      c("comma" = ",", "semicolon" = ";", "tab" =  "\t"), 
                      selected = 1),
          shiny::textInput("obj_name", "Save as (object name)", "dat_scdf")
        ),
        shiny::wellPanel(
          shiny::uiOutput("map_ui")
        )
      ),
      shiny::tableOutput("preview")
    )
  )
  
  # server ------
  server <- function(input, output, session) {
    raw_df  <- shiny::reactiveVal(NULL)
    scdfobj <- shiny::reactiveVal(NULL)
    fname   <- shiny::reactiveVal(NULL)
    ftype   <- shiny::reactiveVal(NULL)
    filepath <- shiny::reactiveVal()
    
    output$status <- shiny::renderText("First choose a file")
    
    
    
    shiny::observeEvent(input$file, load_file())
    shiny::observeEvent(input$sep, load_file())
    
    load_file <- function() {
      
      path <- input$file$datapath
      if (length(path) == 0) return()
      
      fname(input$file$name)
      ext  <- tolower(tools::file_ext(input$file$name))
      typ  <- "auto"
      ftype(typ)
      filepath(path)
      nas <- eval(str2lang(paste0("c(",input$na, ")")))
      
      # rds
      
      if (ext == "rds") {
        tb <- try(
          readRDS(path), silent = TRUE
        )
        if (inherits(tb, "try-error")) { 
          output$status <- shiny::renderText(as.character(tb))
          return() 
        }
        if (inherits(tb, "scdf")) {
          if (rstudioapi::isAvailable()) {
            tmp <- tempfile(pattern = "scdf_", fileext = ".rds")
            saveRDS(tb, tmp)
            rstudioapi::sendToConsole(
              sprintf('%s <- readRDS("%s")', input$obj_name, tmp),
              execute = TRUE
            )
          }
          shiny::stopApp(tb)
        } else {
          output$status <- shiny::renderText("Not a valid scdf file")
          return()
        }
        
      }
      
      # Excel
      if (ext %in% c("xlsx","xls")) {
        tb <- try(
          readxl::read_excel(path = path, na = nas), silent = TRUE
        )
        if (inherits(tb, "try-error")) { 
          output$status <- shiny::renderText(as.character(tb))
          return() 
        }
        raw_df(as.data.frame(tb))
      }
      
      # CSV
      if (ext %in% c("csv","txt")) {
        sep <- input$sep
        dec <- if (sep == ";") "," else "."
        args <- list(file = path, header = TRUE, 
                     stringsAsFactors = FALSE, 
                     na.strings = nas,
                     sep = sep, dec = dec)
        
        tb <- try(do.call(utils::read.csv, args), silent = TRUE)
        if (inherits(tb, "try-error")) { 
          output$status <- shiny::renderText(as.character(tb))
          return() 
        }
        raw_df(tb)
      }
      scdfobj(NULL)
      msg <- "Loaded table (see below). Map columns and click Done."
      if (ncol(raw_df()) == 1 && ext %in% c("csv","txt")) {
        msg <- "Check CSV separator. File not loaded correctly (see beow)."
      }
      
      output$status <- shiny::renderText(msg)
    }
    
    # Mapping UI
    output$map_ui <- shiny::renderUI({
      df <- raw_df()
      if (is.null(df)) return()#shiny::helpText("Not applicable for rds or file not loaded."))
      cols <- names(df)
      sel <- function(hints) { 
        hit <- which(tolower(cols) %in% hints)
        if (length(hit)) cols[hit[1]] else cols[1] 
      }
      shiny::tagList(
        shiny::selectInput("cvar", "Case",  cols, selected = sel(c("case", "id", "subject"))),
        shiny::selectInput("pvar", "Phase", cols, selected = sel(c("phase","condition","segment"))),
        shiny::selectInput("dvar", "Dependent variable",cols, selected = sel(c("values","score","y", "outcome"))),
        shiny::selectInput("mvar", "Mesaurement-time",    cols, selected = sel(c("mt","time","t","day", "session")))
      )
    })
    
    # Preview
    output$preview <- shiny::renderTable({
      if (!is.null(scdfobj())) utils::head(scdfobj()[[1]])
      else if (!is.null(raw_df())) utils::head(raw_df())
    }, striped = TRUE, spacing = "s")
    
    # Handle Done/Cancel
    shiny::observeEvent(input$done, {
      nm <- input$obj_name
      # Build from mapping if needed
      df <- raw_df()
      if (!is.null(df) && all(nzchar(c(input$cvar, input$pvar, input$dvar, input$mvar)))) {
        obj <- try(as_scdf(
          object = df,
          cvar = input$cvar, pvar = input$pvar, dvar = input$dvar, mvar = input$mvar
        ), silent = TRUE)
        if (!inherits(obj, "try-error")) {
          
          # code <- paste0(
          #   nm, " <- read_scdf(", deparse(input$file$datapath), ", ",
          #   "cvar = ",  deparse(input$cvar), ", ",
          #   "pvar = ",  deparse(input$pvar), ", ",
          #   "dvar = ",  deparse(input$dvar), ", ",
          #   "mvar = ",  deparse(input$mvar), ", ",
          #   "na = c(", input$na,
          # ")")
          
          if (rstudioapi::isAvailable()) {
            tmp <- tempfile(pattern = "scdf_", fileext = ".rds")
            saveRDS(obj, tmp)
            rstudioapi::sendToConsole(
              sprintf('%s <- readRDS("%s")', nm, tmp),
              execute = TRUE, echo = FALSE 
            )
            
            rstudioapi::sendToConsole(nm, execute = FALSE)
          }
            
        }
      }
      shiny::stopApp(obj)
    })
    shiny::observeEvent(input$cancel, shiny::stopApp(invisible(NULL)))
  }
  
  viewer <- shiny::dialogViewer("scdf import", width = 600, height = 400)
  shiny::runGadget(ui, server, viewer = viewer)
}

#' @rdname export
#' @export
export.sc_overlap <- function(object, 
                              caption = NA, 
                              footnote = NULL, 
                              filename = NA,
                              round = 2,
                              decimals = 2,
                              flip = FALSE,
                              ...) {
  
  if (is.na(caption)) caption <- paste0(
    "Overlap indices. ",
    .phases_string(object$phases.A, object$phases.B),
    collapse = ""
  )
  
  if (is.null(footnote)) footnote <- c(
    "PND = Percentage Non-Overlapping Data",
    "PEM = Percentage Exceeding the Median",
    "PET = Percentage Exceeding the Trend",
    "NAP = Nonoverlap of all pairs",
    "NAP-R = NAP rescaled",
    "PAND = Percentage all nonoverlapping data",
    "IRD = Improvement rate difference",
    "Tau U (A + B - trend A) = Parker's Tau-U",
    "Tau U (A + B - trend A + trend B) = Parker's Tau-U",
    "Base Tau = Baseline corrected Tau",
    "Delta M = Mean difference between phases",
    "Delta Trend = Trend difference between phases",
    "SMD = Standardized Mean Difference",
    "Hedges g = Corrected SMD"
  )
  
  out <- object$overlap
  
  cn <- colnames(out)
  colnames(out)[which(cn == "NAP rescaled")] <- "NAP-R"
  colnames(out)[which(cn == "Tau_U(A)")] <- "Tau-U (A + B - trend A)"
  colnames(out)[which(cn == "Tau_U(BA)")] <- "Tau-U (A + B - trend A + trend B)"
  colnames(out)[which(cn == "Base_Tau")] <- "Base Tau"
  colnames(out)[which(cn == "Diff_mean")] <- "Delta M"
  colnames(out)[which(cn == "Diff_trend")] <- "Delta Trend"
  colnames(out)[which(cn == "Hedges_g")] <- "Hedges g"
  
  if (isTRUE(flip)) {
    cases <- out$Case
    out[-2:-1] <- round(out[-2:-1], round)
    names_par <- colnames(out)[-1]
    out <- t(out[-2:-1]) |> as.data.frame()
    out <- cbind(Statistic = rownames(out), out)
    colnames(out) <- c("Statistic", cases)
  }
  
  table <- .create_table(
    out, 
    caption = caption,
    footnote = footnote,
    decimals = decimals,
    ...
  )
 
  # finish ------------------------------------------------------------------
  
  if (!is.na(filename)) .save_export(table, filename)
  
  table
}

check_scdf <- function(object, message = FALSE) {
  results <- .check_scdf(object)
  if (isTRUE(results)) {
    if (message) message("No errors or warnings.")
    return(invisible(TRUE))
  }
  if(length(results$warnings) > 0) {
    warn(results$warnings)
  }
  
  if(length(results$errors) > 0) {
    abort(results$errors)
  }
}

.check_scdf <- function(object) {
  
  errors <- character()
  warnings <- character()
  
  # check class
  if (!inherits(object, "scdf")) {
    msg <- paste0("Class is not 'scdf'.")
    errors <- c(errors, msg)
  }
  
  scdf_attributes <- attr(object, opt("scdf"))
  
  if (is.null(scdf_attributes)) {
    msg <- paste0("Attribute '", opt("scdf"), "' missing.")
    errors <- c(errors, msg)
    return(errors)
  } 
  
  var_phase <- scdf_attributes[[opt("phase")]]
  var_mt <- scdf_attributes[[opt("mt")]]
  var_dv <- scdf_attributes[[opt("dv")]]
  
  # check all elements are data frames
  if (!all(unlist(lapply(object, function(x) is.data.frame(x))))) {
    msg <- "Not all list-elements are data frames.\n"
    errors <- c(errors, msg)
  }
  
  # check all cases have phase var
  if (!all(unlist(lapply(object, function(x) {var_phase %in% names(x)})))) {
    msg <-paste0(
      "scdf has phase variable ",
      "'", var_phase, "' ",
      "but not all cases have a variable named ",
      "'", var_phase, "'."
    )
    errors <- c(errors, msg)
  }
  
  # check all cases have mt var
  if (!all(unlist(lapply(object, function(x) {var_mt %in% names(x)})))) {
    msg <-paste0(
      "scdf has measurement-time variable ",
      "'", var_mt, "' ",
      "but not all cases have a variable named ",
      "'", var_mt, "'."
    )
    errors <- c(errors, msg)
  }
  
  # check all cases have dv var
  if (!all(unlist(lapply(object, function(x) {var_dv %in% names(x)})))) {
    msg <-paste0(
      "scdf has dependent variable ",
      "'", var_dv, "' ",
      "but not all cases have a variable named ",
      "'", var_dv, "'."
    )
    errors <- c(errors, msg)
  }
  
  # check phase design is identical
  phases_1 <- rle(as.character(object[[1]][[var_phase]]))$values
  .ident_phases <- lapply(object, 
    function(x) identical(rle(as.character(x[[var_phase]]))$values, phases_1)
  )
  .ident_phases <- unlist(.ident_phases)
  if (!all(.ident_phases)) {
    msg <- "Phase design is not identical for all cases."
    warnings <- c(warnings, msg)
  }
  
  
  if (length(errors) == 0 && length(warnings) == 0) return(TRUE)
  
  if (length(errors) > 0) errors <- paste(errors, collapse = "\n  ")
  if (length(warnings) > 0) warnings <- paste(warnings, collapse = "\n  ")
  
  return(list(errors = errors, warnings = warnings))

}

test_that("convert", {
  
  fx <- tempfile()
  fy <- tempfile()
  convert(exampleABC, file = fx, silent = TRUE)
  convert(exampleABC, inline = TRUE, file = fy, silent = TRUE)
  
  source(fx)
  x <- study
  
  source(fy)
  y <- study
  
  expect_equal(x,y)
  
  convert(Leidig2018, file = fx, silent = TRUE)
  source(fx)
  
  expect_equal(scan:::check_scdf(study), TRUE)
  
  ne <- new.env()

  eval(parse(text = convert(exampleAB_score, silent = TRUE)), ne)
  expect_equal(exampleAB_score, ne$study)
  
})

#' Samples random names
#' 
#' Generates random names for cases. Names are sampled from a predefined list
#' of neutral, male, female, or mixed names.
#' 
#' This function is useful for anonymizing case names in datasets or for generating
#' random identifiers.
#'
#' @param n Number of names to sample.
#' @param type "neutral", "male", "female", or "mixed" type of names to sample.
#' @param seed A seed for the random number generator. If provided, the sampling will be reproducible.
#' @return A character vector with random names.
#' @examples 
#' sample_names(3)
#' @export
sample_names <- function(n = 1, type = "neutral", seed = NULL) {
  
  if (!is.null(seed)) set.seed(seed)
  out <- character(0)
  if (type == "neutral") out <- sample(case_names$neutral, n)
  if (type == "female") out <- sample(case_names$female, n)
  if (type == "male") out <- sample(case_names$male, n)
  if (type == "mixed") out <- sample(case_names$all, n)
  
  out
}

#' @describeIn between_smd Print results
#' @inheritParams print.sc
#' @order 2
#' @param x An object returned by `baseline_smd()`.
#' @export
print.sc_bcsmd <- function(x, digits = 2, ...) {
  cat("Between-Case Standardized Mean Difference\n")

  cat("\nMethod:", x$method)
  
  for (i in 1:length(x$models)) {
    x$models[[i]] <- round_numeric(x$models[[i]], digits)
    names(x$models[[i]])[4:5] <- str_ci(x$ci)
    cat("\n")
    cat(names(x$models)[i])
    cat("\n\n")
    print(x$models[[i]], row.names = FALSE)
  }
 
}

#' @describeIn between_smd export results
#' @inheritParams export
#' @order 3
#' @param x An object returned by `baseline_smd()`.
#' @export
export.sc_bcsmd <- function(object, 
                            caption = NA, 
                            footnote = NA, 
                            filename = NA,
                            digits = 2, 
                            round = 2,
                            ...) {
  
  if (is.na(caption)) caption <- c("Between-Case Standardized Mean Difference")
  
  if (is.na(footnote)) {
    footnote <- c(
      if (object$method == "MCMCglmm") "CI = credible interval" else "CI = confidence interval",
      "LL = lower limit",
                  "UL = upper limit",
                  paste0("Method: ", object$method))
  }

  for (i in 1:length(object$models)) {
    object$models[[i]] <- round_numeric(object$models[[i]], digits)
    names(object$models[[i]])[4:5] <- c("LL", "UL")
  }
  
  out <- do.call(rbind, object$models)
  
  if (getOption("scan.export.engine") == "gt") {
    spanner <- list("CI" = 4:5)
    names(spanner) <- gsub(
      "CI", paste0("CI(", object$ci * 100, "%)"), x = names(spanner)
    )
  }
  
  if (length(object$models) == 2) {
    rows <- nrow(object$models[[1]])
    row_group <- list(
      "Base model" = 1:rows,
      "Full model" = (rows + 1) : (rows * 2)
    )
    names(row_group) <- names(object$models)
  } else {
    row_group <- NULL
  }
  
  table <- .create_table(
    out, 
    row_group = row_group,
    spanner = spanner,
    caption = caption,
    footnote = footnote,
    digits = digits,
    ...
  )
  
  # finish ------------------------------------------------------------------
  
  if (!is.na(filename)) .save_export(table, filename)
  table
}

#' Pipe
#'
#' Several functions in scan are designed to work with pipes `\%>\%`. This
#' pipe function is directly imported from magrittr.
#'
#' @importFrom magrittr %>%
#' @details Since R 4.1 a pipe operator |> is included in base R. This pipe
#'   operator can be used along with scan perfectly fine.
#' @name %>%
#' @rdname pipe
#' @export
#' @keywords internal
NULL

#' Transform every single case of a single case data frame
#'
#' Takes an scdf and applies transformations to each individual case. This is
#' useful to calculate or modify new variables.
#'
#' This function is a method of the generic `transform()` function.  
#' Unlike the generic version, expressions are evaluated **serially**:  
#' the result of one expression is used as the basis for subsequent computations.  
#'
#' Several helper functions can be used inside expressions:
#'
#' - **`n()`**: returns the number of measurements in a case.  
#' - **`all_cases()`**: extracts the values of a variable across all cases.  
#'   Takes an expression as argument.  
#'   For example:  
#'   - `mean(all_cases(values))` calculates the mean of `values` across all cases.  
#'   - `mean(all_cases(values[phase == "A"]))` calculates the mean of all values where `phase == "A"`.  
#' - **`rowwise()`**: applies a calculation separately to each row.  
#'   Example: `rowwise(sum(values, mt, na.rm = TRUE))`.  
#' - **`across_cases()`**: creates new variables or replaces existing ones across all cases.  
#'   Example: `across_cases(values_ranked = rank(values, na.last = "keep"))` 
#'   ranks the `values` variable across all cases and creates a new variable `values_ranked`.
#'   
#' @param _data An scdf.
#' @param ... Expressions.
#' @return An scdf.
#' @author Juergen Wilbert
#' @family data manipulation functions
#' @export
#' @examples
#' ## Creates a single-case with frequency distributions. The proportion and
#' ## percentage of the frequencies are calculated with transform:
#' design <- design(
#'  n = 3,
#'  level = 5,
#'  distribution = "binomial",
#'  n_trials = 20,
#'  start_value = 0.5
#' )
#' study <- random_scdf(design)
#' transform(study, proportion = values/trials, percentage = proportion * 100)
#'
#' ## Z standardize the dependent variable and add two new variables:
#' exampleAB |>
#'   transform(
#'     values = scale(values),
#'     mean_values = mean(values),
#'     sd_values = sd(values)
#'   )
#'
#' ## Use `all` to calculate global variables.
#' exampleAB |>
#'   transform(
#'     values_center_case = values - mean(values[phase == "A"]),
#'     values_center_global = values - mean(all(values[phase == "A"])),
#'     value_dif = values_center_case - values_center_global
#'   )
#'
#' ## Use `across_cases` to calculate or replace a variable with values from
#' ## all cases. E.g., standardize the dependent variable:
#' exampleABC |>
#'   transform(
#'     across_cases(values = scale(values))
#'   )
#'
#' ## Rank transform the values based on all cases vs. within each case:
#' exampleABC |>
#'   transform(
#'     across_cases(values_across = rank(values, na.last="keep")),
#'     value_within = rank(values, na.last="keep")
#'   )
#'
#' ## Three helper functions to smooth the data
#' Huber2014$Berta |>
#' transform(
#'   "compliance (moving median)" = moving_median(compliance),
#'   "compliance (moving mean)" = moving_mean(compliance),
#'   "compliance (local regression)" = local_regression(compliance, mt)
#' )
#'
#' ## Function first_of() helps to set NAs for specific phases.
#' ## E.g., you want to replace the first two values of phase A and the first
#' ## value of phase B and its preceding value.
#'
#' byHeart2011 |>
#'   transform(
#'     values = set_na_at(values, phase == "A", 0:1),
#'     values = set_na_at(values, phase == "B", -1:0)
#'   )
#' @export
transform.scdf <- function(`_data`, ...) {
  expressions <- substitute(list(...))
  original_attr <- attributes(`_data`)
  .df <- as.data.frame(`_data`)
  
  for(i_expression in 2:length(expressions)) {
    
    if(startsWith(deparse(expressions[[i_expression]]), "across_cases(")) {
      .list_env <- as.list(.df)
     
      # across cases
      .list_env$across_cases <- function(...) {
        exp_across <- substitute(list(...))
        for (i in 2:length(exp_across)) {
          new <- eval(exp_across[c(1,i)], .list_env)
          .df[[names(new)]] <- new[[1]]
          .list_env[[names(new)]] <- new[[1]]
        }
        .df
      }
      
      new <- eval(expressions[c(1,i_expression)], .list_env, parent.frame())
      .df <- new[[1]]
     
      `_data` <- suppressMessages(as_scdf(.df))
     
      #attributes(`_data`) <- original_attr
    } else {
      # For across case calculations
      .df <- as.data.frame(`_data`)
      
      # case by case loop
      for(i_case in seq_along(`_data`)) {
        .list_env <- as.list(`_data`[[i_case]])
        
        ## additional functions all(), all_cases() and n()
        .list_env$all_cases <- .list_env$all <- function(x) {
          x <- substitute(x)
          eval(x, .df)
        }
        .list_env$n <- function() {nrow(`_data`[[i_case]])}
        .list_env$rowwise <- function(expr) {
          expr <- substitute(expr)
          out <- c()
          for(i_row in 1:nrow(`_data`[[i_case]])) {
            out <- c(out, eval(expr, `_data`[[i_case]][i_row, ]))
          }
          out
        }
        
        # compute new variable
        new <- eval(expressions[c(1,i_expression)], .list_env, parent.frame())
        
        # Add new variable
        `_data`[[i_case]][[names(new)]] <- new[[1]]
        
        
      }
      .df <- as.data.frame(`_data`)
    }
  }
  `_data`
}

#' (Deprecated) Create styles for single-case data plots
#'
#' The `style_plot` function is used to create graphical styles for a
#' single-case plot
#'
#' @param style A character string or a vector of character strings with
#'   predefined styles.
#' @param ... Further arguments passed to the plot command.
#'
#' @return Returns a list to be provided for the style argument of the
#'   [plot.scdf()] function.
#' @return \itemize{ \item\code{fill} If set, the area under the line is filled
#'   with the given color (e.g., \code{fill = "tomato"}). Use the standard R
#'   command colors() to get a list of all possible colours. \code{fill} is
#'   empty by default. \item\code{annotations} A list of parameters defining
#'   annotations to each data point. This adds the score of each MT to your
#'   plot.
#' \itemize{
#' \item\code{"pos"} Position of the annotations: 1 = below, 2 =
#' left, 3 = above, 4 = right.
#' \item\code{"col"} Color of the annotations.
#' \item\code{"cex"} Size of the annotations.
#' \item\code{"round"} Rounds the values to the specified decimal.}
#' \item\code{annotations = list(pos = 3, col =
#' "brown", round = 1)} adds scores rounded to one decimal above the data point
#'   in brown color to the plot. \item\code{"names"} A list of parameters
#'   defining the depiction of phase names (e.g. \code{names = list(cex = 0.8,
#'   col = "red", side = 1)}: cex for size, col for color, and side for
#'   position). See \code{\link{mtext}} for more details. \item\code{"lwd"}
#'   Width of the plot line. Default is \code{lwd = 2}. \item\code{"pch"} Point
#'   type. Default is \code{pch = 17} (triangles). Other options are for
#'   example: 16 (filled circles) or "A" (uses the letter A). \item\code{"main"}
#'   Main title of the plot. \item\code{"mai"} Sets the margins of the plot.
#'   \item\code{"bty"} Shape of the frame surrounding the inner plot
#'   \item\code{"fill.bg"} Background color of the plot. If a vector is
#'   provided, these colors will be assigned to phases (each phase name becomes
#'   a color). \item\code{"grid"} Color of a grid. \item\code{"text.ABlag"} Text
#'   displayed between phases. \item\code{"cex.axis"} Size of the axis
#'   annotations \item\code{"las"} Orientation of the axis annotations
#'   \item\code{"col.lines"} Color of the lines \item\code{"col.dots"} Color of
#'   the dots \item\code{"col.seperator"} Color of the phase seperating lines
#'   \item\code{"col.bg"} Color of the outer plot \item\code{"col"} General
#'   color setting for the plot \item\code{"col.text"} Color of all labels of
#'   the plot. }
#' @details \code{style_plot("")} will return a list of predefined styles.
#'   Predefined styles can be combined \code{style_plot(style = c("grid2",
#'   "tiny"))} where settings of a latter style overwrite settings of the
#'   former. Additional style paramters are set following the style argument and
#'   can be combined with those: \code{style_plot(style = "grid2", fill =
#'   "grey50", pch = 18)}.
#' @author Juergen Wilbert
#' @seealso [plot.scdf()]
#' @examples
#' newstyle <- style_plot(style = "default")
#' newstyle$text.ABlag <- c("START", "END")
#' newstyle$col.dots <- ""
#' newstyle$annotations <- list(cex = 0.6, col = "grey10", offset = 0.4)
#' newstyle$names <- list(cex = 0.8, col = "blue", side = 1, adj = 1, line = -1, at = 31)
#' newstyle$fill.bg <- c("grey99", "grey95", "grey90")
#' plot(exampleABC, style = newstyle, main = "Example Plot")
#'
#' @export
style_plot <- function(style = "default", ...) {
  
  .deprecated_warning("scplot' from the package 'scplot", "style_plot")
  
  new <- list(...)
  if (identical(style, "")) {
    cat("Available styles: \n")
    cat(paste(names(opt("style")), collapse = ", "), "\n")
    return(invisible(NULL))
  }
  if (!(all(style %in% names(opt("style"))))) {
    cat("Style '", paste(style), "' is unknown.\n", sep = "")
    cat("Available styles: \n")
    cat(paste(names(opt("style")), collapse = ", "), "\n")
    return(invisible(NULL))
  }

  styles <- list()
  for (i in rev(style)) styles <- c(styles, opt("style")[[i]])

  out <- c(styles, opt("style")$default)
  out <- out[unique(names(out))]
  if (is.list(new)) {
    out <- c(new, out)
    out <- out[unique(names(out))]
  }
  invisible(out)
}

#' Hierarchical piecewise linear model / piecewise regression for multiple cases
#'
#' The [hplm()] function computes a hierarchical piecewise regression model. It
#' extends the standard piecewise regression model to multiple cases by
#' estimating fixed and random effects. The function uses the `lme` function of
#' the `nlme` package to fit linear mixed-effects models. The model can include
#' random intercepts and random slopes for level, trend, and treatment effects.
#' Additionally, it allows for the inclusion of autoregressive structures and
#' unequal variances across phases. The function also provides options for
#' likelihood ratio tests to compare models with and without random slope
#' effects, as well as the calculation of intraclass correlations (ICC) to
#' assess the proportion of variance attributable to between-case differences.
#' This function is particularly useful for analyzing data from multiple
#' single-case experimental designs (SCEDs) where observations are nested within
#' cases.
#'
#' @section Model specification: The fixed effects part of the model can be
#'   specified using the `fixed` argument, while the random effects part can be
#'   specified using the `random` argument. If not provided, default formulas
#'   based on the specified model type (e.g., "B&L-B") are created. The function
#'   also allows for the inclusion of autoregressive structures through the `ar`
#'   argument and unequal variances across phases through the
#'   `unequal_variances` argument.
#' @section Random slopes: By setting the `random.slopes` argument to TRUE, the
#'   model will include random slope effects for level, trend, and treatment
#'   effects. This allows for individual differences in how cases respond to
#'   these effects.
#' @section Likelihood ratio tests: If the `lr.test` argument is set to TRUE,
#'   the function will perform likelihood ratio tests to compare models with and
#'   without random slope effects. This helps to determine whether including
#'   random slopes significantly improves model fit.
#' @section Intraclass correlation: If the `ICC` argument is set to TRUE, the
#'   function will calculate the intraclass correlation coefficient (ICC) to
#'   assess the proportion of variance attributable to between-case differences.
#'   This provides insight into the degree of similarity among observations
#'   within the same case.
#'
#' @inheritParams .inheritParams
#' @order 1
#' @param method Method used to fit your model. Pass `"REML"` to maximize the
#'   restricted log-likelihood or `"ML"` for maximized log-likelihood. Default
#'   is `"ML"`.
#' @param control A list of settings for the estimation algorithm, replacing the
#'   default values passed to the function `lmeControl` of the `nlme` package.
#' @param random.slopes If `random.slopes = TRUE` random slope effects of the
#'   level, trend, and treatment parameter are estimated.
#' @param lr.test If set TRUE likelihood ratio tests are calculated comparing
#'   model with vs. without random slope parameters.
#' @param ICC If `ICC = TRUE` an intraclass-correlation is estimated.
#' @param random_trend If TRUE, includes a random trend trend effect.
#' @param random_level If TRUE, includes a random level trend effect.
#' @param random_slope If TRUE, includes a random slope trend effect.
#' @param fixed Defaults to the fixed part of the standard piecewise regression
#'   model. The parameter phase followed by the phase name (e.g., phaseB)
#'   indicates the level effect of the corresponding phase. The parameter
#'   'inter' followed by the phase name (e.g., interB) adresses the slope effect
#'   based on the method provide in the model argument (e.g., "B&L-B"). The
#'   formula can be changed for example to include further L1 or L2 variables
#'   into the regression model.
#' @param random The random part of the model. Defaults to a random intercept
#'   model. The formula can be changed to include random slope effects for
#'   level, trend, and treatment effects.
#' @param ar Maximal lag of autoregression. Modelled based on the
#'   Autoregressive-Moving Average (ARMA) function.
#' @param unequal_variances Logical. If set TRUE, estimations are weighted by
#'   phase variances.
#' @param update.fixed An easier way to change the fixed model part (e.g., `. ~
#'   . + newvariable`).
#' @param data.l2 A data frame providing additional variables at Level 2. The
#'   scdf File has to have names for all cases and the Level 2 data frame has to
#'   have a column named 'cases' with the names of the cases the Level 2
#'   variables belong to.
#' @param ... Further arguments passed to the lme function.
#' @return An object of class `sc_hplm`.
#' - `model` | List containing infromation about
#' - `N` | Number of single-cases. 
#' - `formula` | A list containing the fixed and the random formulas of the hplm model.
#' - `hplm` | Object of class lme contaning the multilevel model. 
#' - `model.0` | Object of class lme containing the zero model. 
#' - `ICC` | List containing intraclass correlation and test parameters. 
#' - `model.without` | Object of class gls containing the fixed effect model. 
#' - `contrast` | List with contrast definitions.
#' @author Juergen Wilbert
#' @family regression functions
#' @examples
#'
#' ## Compute hplm model on a MBD over fifty cases (restricted log-likelihood)
#' hplm(exampleAB_50, method = "REML", random.slopes = FALSE)
#'
#' ## Analyzing with additional L2 variables
#' Leidig2018 |>
#'   add_l2(Leidig2018_l2) |>
#'   hplm(update.fixed = .~. + gender + migration + ITRF_TOTAL*phaseB,
#'        slope = FALSE, random.slopes = FALSE, lr.test = FALSE
#'   )
#'
#' @export

hplm <- function(data, dvar, pvar, mvar, 
                 model = c("W", "H-M", "B&L-B", "JW"),
                 contrast = c("first", "preceding"),
                 contrast_level = NA,
                 contrast_slope = NA,
                 method = c("ML", "REML"), 
                 control = list(opt = "optim"), 
                 random.slopes = FALSE, 
                 lr.test = FALSE, 
                 ICC = TRUE, 
                 trend = TRUE, 
                 level = TRUE, 
                 slope = TRUE, 
                 random_trend = FALSE, 
                 random_level = FALSE, 
                 random_slope = FALSE, 
                 fixed = NULL, 
                 random = NULL, 
                 ar = 0,
                 unequal_variances = FALSE,
                 update.fixed = NULL, 
                 data.l2 = NULL, 
                 ...) {

  args_add <- list(...)
  
  check_args(
    by_call(model),
    by_call(method),
    by_call(contrast),
    not(ar > 0 && "correlation" %in% names(args_add), 
        "ar and correlation arguments defined at the same time."
    ),
    not(unequal_variances && "weights" %in% names(args_add), 
        "unequal_variances and weights arguments defined at the same time."
    )
  )
  
  if (model == "JW") {
    contrast_level <- "preceding"
    contrast_slope <- "preceding"
    model <- "B&L-B"
  }
  
  if (is.na(contrast_level)) contrast_level <- contrast
  if (is.na(contrast_slope)) contrast_slope <- contrast

  if (random.slopes) {
    random_trend <- trend
    random_level <- level
    random_slope <- slope
  }
  
  # set attributes to arguments else set to defaults of scdf
  if (missing(dvar)) dvar <- dv(data) else dv(data) <- dvar
  if (missing(pvar)) pvar <- phase(data) else phase(data) <- pvar
  if (missing(mvar)) mvar <- mt(data) else mt(data) <- mvar
  
  dat <- .prepare_scdf(data)
  
  N <- length(dat)
  out <- list()
  out$model$interaction.method  <- model
  out$model$contrast.method     <- contrast
  out$model$estimation.method   <- method
  out$model$lr.test             <- lr.test
  out$model$random.slopes       <- random.slopes
  out$model$ICC                 <- ICC
  out$N                         <- N

# interaction and dummy coding and L2 --------------------------------------

  tmp_model <- .add_dummy_variables(
    data = dat, model = model, 
    contrast_level = contrast_level, contrast_slope = contrast_slope
  )
  
  dat <- tmp_model$data

  dat <- as.data.frame(dat, l2 = data.l2)

# create formulas ---------------------------------------------------------

  if (is.null(fixed)) {
    fixed <- .create_fixed_formula(
      dvar, mvar, slope, level, trend, tmp_model$var_phase, tmp_model$var_inter
    )
  }
  if (!is.null(update.fixed)) fixed <- update(fixed, update.fixed)
  
  
  if (any(random_trend, random_level, random_slope)) {
    random.slopes <- TRUE
  }
  
  if (!random.slopes && is.null(random)) 
    random <- as.formula("~1|case")
  
  if (is.null(random)) {
    random <- .create_random_formula(
      mvar, 
      random_slope, random_level, random_trend, 
      tmp_model$var_phase, 
      tmp_model$var_inter
    )
  }
  out$formula <- list(fixed = fixed, random = random)

  if (ar > 0) {
    args_add$correlation <- corARMA(
      form = as.formula(paste0("~", mvar, " | case")), 
      p = ar, q = 0
    )
  }
  
  if (unequal_variances) {
    args_add$weights <- varIdent(form = as.formula(paste0("~ 1 | ", pvar)))
  }
  
  args <- list(
    fixed = fixed,
    random = random,
    data = dat,
    na.action = na.omit,
    method = method,
    control = control,
    keep.data = FALSE
  )
  args <- c(args, args_add)
 
# lme hplm model ----------------------------------------------------------
  out$hplm <- do.call(lme, args) 
  
  # check:
  #out$hplm <- lme(
  #  fixed = fixed,
  #  random = random,
  #  data = dat,
  #  na.action = na.omit,
  #  method = method,
  #  control = control,
  #  keep.data = FALSE,
  #  ...
  #)

  out$hplm$call$fixed <- fixed
  out$hplm$call$random <- random
  out$hplm$call$data <- str2lang("dat")
  
# LR tests ----------------------------------------------------------------

  if (lr.test) {
    pred_rand    <- unlist(strsplit(as.character(random[2]), "\\|"))[1]
    pred_rand_id <- unlist(strsplit(as.character(random[2]), "\\|"))[2]
    pred_rand    <- unlist(strsplit(pred_rand, "\\+"))
    pred_rand    <- trimws(pred_rand)
    pred_rand_id <- trimws(pred_rand_id)

    if (length(pred_rand) == 1) {
      abort("LR Test not applicable with only one random effect.")
    }
    
    random_ir <- list(formula(gsub("1", "-1", paste(random, collapse = " "))))
    for(i in 2:length(pred_rand)) {
      random_ir[[i]] <- formula(
        paste0("~", paste0(pred_rand[!pred_rand %in% pred_rand[i]], 
        collapse = " + "), " | ", pred_rand_id)
      )
    }
    out$random_ir$restricted <- list()
    
    # lme
    for(i in 1:length(random_ir)) {
      
      args_restricted <- args
      args_restricted$random <- random_ir[i]
      
      out$random_ir$restricted[[i]] <- do.call(lme, args_restricted)
      out$random_ir$restricted[[i]]$call$data <- str2lang("dat")
      # check:
      # out$random_ir$restricted[[i]] <- lme(
      #   fixed = fixed, 
      #   random = random_ir[i], 
      #   data = dat,
      #   na.action = na.omit, 
      #   method = method, 
      #   control=control,
      #   keep.data = FALSE, 
      #   ...
      # )
      out$random_ir$restricted[[i]]$call$data <- str2lang("dat")
      out$random_ir$restricted[[i]]$call$fixed <- fixed
    }
    out$LR.test <- list()
    
    # LR test
    for(i in 1:length(random_ir)) {
      out$LR.test[[i]] <- anova(out$random_ir$restricted[[i]], out$hplm)
    }
    attr(out$random_ir, "parameters") <- c("Intercept", pred_rand)
  }
  
# ICC ---------
  
  if (ICC) {
    .formula.null <- as.formula(paste0(dvar, " ~ 1"))
    out$model.0 <- lme(
      .formula.null, random =~1|case, data = dat, 
      method = method, na.action = na.omit, control = control
    )
    out$model.0$call$fixed <- .formula.null
    
    vc <- as.numeric(VarCorr(out$model.0))
    out$ICC$value <- vc[1] / (vc[1] + vc[2])	
    out$model.without <- gls(
      .formula.null, data = dat, method = method, 
      na.action = na.omit, control = control
    )
    out$model.without$call$model <- .formula.null
    dif <- anova(out$model.0, out$model.without)
    out$ICC$L <- dif$L.Ratio[2]
    out$ICC$p <- dif$"p-value"[2]
  } 
  
# out ---
  
  out$model$fixed  <- fixed
  out$model$random <- random
  out$model$ar <- ar
  out$model$unequal_variances <- unequal_variances
  out$contrast <- list(level = contrast_level, slope = contrast_slope)
  
  class(out) <- c("sc_hplm")
  attributes(out)[opts("phase", "mt", "dv")] <- list(pvar, mvar, dvar)
  out
}

test_that("hplm", {
  
  scdf <- exampleA1B1A2B2
  result <- hplm(scdf)
  values <- round(sum(residuals(result$hplm)^2), 3)
  expect_equal(values, 3009.384)

  result <- hplm(scdf, contrast = "preceding")
  values <- round(sum(residuals(result$hplm)^2), 3)
  expect_equal(values, 3005.251)
  
  result <- hplm(
    exampleAB_50, 
    model = "W",
    method = "REML",
    unequal_variances = TRUE,
    random_trend = TRUE,
    control = lmeControl(msMaxIter = 50, apVar = FALSE, returnObject = TRUE),
    ar = 1
  )
  values <- round(sum(residuals(result$hplm)^2))
  expect_equal(values, 34923)
  
})

#' @rdname print.sc
#' @param nice If set TRUE (default) output values are rounded and optimized for
#' publication tables.
#' @export
print.sc_design <- function(x, ...) {
  
  cat("A scdf design matrix\n\n")
  cat("Number of cases:", length(x$cases), "\n")
  cat("Distribution: ", x$distribution, "\n")
  cat("Start values: ", unique(
    sapply(x$cases, function(x) {x$start_value[1]})), "\n")
  if (x$distribution %in% c("normal", "gaussian")) {
    cat("SD = ", unique(sapply(x$cases, function(x) {x$s[1]})), "\n")
    cat("rtt = ", unique(sapply(x$cases, function(x) {x$rtt[1]})), "\n")
  }
  cat("Phase design: ", unique(sapply(x$cases, function(x) {
    paste0(x$phase, "=", x$length, collapse = " ")
  })), "\n")
  cat("Trend effect: ", unique(sapply(x$cases, function(x) {x$trend[1]})), "\n")
  cat("Level effect: ", unique(sapply(x$cases, function(x) {
    paste0(x$phase[-1], "=", x$level[-1], collapse = " ")
  })), "\n")
  cat("Slope effect: ", unique(sapply(x$cases, function(x) {
    paste0(x$phase[-1], "=", x$slope[-1], collapse = " ")
  })), "\n")
  cat("Missing proportion: ", unique(sapply(x$cases, function(x) {x$missing_prop})), "\n")
  
  ext_p <- unique(sapply(x$cases, function(x) {x$extreme_prop}))
  cat("Extreme proportion: ", ext_p, "\n")
  if (ext_p != 0) {
    cat("Extreme range: ", unique(sapply(x$cases, function(x) {
      paste0(x$extreme_low, "/", x$extreme_high, collapse = " ")
    })), "\n")
  }
  
}

#' @rdname export
#' @export
export.sc_pem <- function(object, 
                          caption = NA, 
                          footnote = NA, 
                          filename = NA,
                          round = 2,
                          ...) {
  
  if (is.na(caption)) caption <- c("Percent Exceeding the Median")
  
  object$PEM <- round_numeric(object$PEM, round)
  
  table <- .create_table(
    object$PEM, 
    caption = caption,
    footnote = footnote,
    ...
  )
  
  # finish ------------------------------------------------------------------
  
  if (!is.na(filename)) .save_export(table, filename)
  table
}

#' Overlap indices for single-case data
#'
#' The `overlap` function provides the most common overlap indices for
#' single-case data and some additional statistics.
#' 
#' It computes PND, PEM, PET, NAP, PAND, IRD, Tau-U, mean difference,
#' difference in trend, SMD, and Hedges-g for each single-case included in an
#' scdf.
#'
#' @inheritParams .inheritParams
#' @details See corresponding functions of PND, PEM, PET, NAP, PAND for
#'   calculation. Tau_U(A) reports "A vs. B - Trend A" whereas Tau_U(BA) reports
#'   "A vs. B + Trend B - Trend A". Base_Tau is baseline corrected tau
#'   (correction applied when autocorrelation in phase A is significant).
#'   Diff_mean is the mean difference. Diff_trend is the difference in the
#'   regression estimation of the dependent variable on measurement-time (`x ~
#'   mt`) for each phase. SMD is the mean difference divided by the standard
#'   deviation of phase A. Hedges_g is the mean difference divided by the pooled
#'   standard deviation: \eqn{\sqrt{ (n_A - 1)sd_A^2 + (n_B - 1)sd_B^2 \over n_A
#'   + n_B - 2 }} with a hedges correction applied: \eqn{Hedges_g * (1 -
#'   \frac{3}{4n - 9} ) )}.
#' @return
#'  |  |  |
#'  | --- | --- |
#'  | `overlap` | A data frame consisting of the following indices for each single-case for all cases: PND, PEM, PET, NAP, PAND, IRD, Tau-U (A vs. B - Trend A), Diff_mean, Diff_trend, SMD, Hedges-g. |
#'  | `phases.A` | Selection for A phase. |
#'  | `phases.B` | Selection for B phase. |
#'  | `design` | Phase design. |
#' @family overlap functions
#' @author Juergen Wilbert
#' @seealso [pnd()], [pem()], [pet()], [nap()], [pand()], [ird()], [tau_u()], [corrected_tau()]
#' @keywords overlap
#' @examples
#'
#' ## Display overlap indices for one single-case
#' overlap(Huitema2000, decreasing = TRUE)
#'
#' ## Display overlap indices for six single-cases
#' overlap(GruenkeWilbert2014)
#'
#' ## Combining phases for analyszing designs with more than two phases
#' overlap(exampleA1B1A2B2, phases = list(c("A1","A2"), c("B1","B2")))
#'
#' @export
overlap <- function(data, dvar, pvar, mvar, 
                    decreasing = FALSE, 
                    phases = c(1, 2)){

  # set attributes to arguments else set to defaults of scdf
  if (missing(dvar)) dvar <- dv(data) else dv(data) <- dvar
  if (missing(pvar)) pvar <- phase(data) else phase(data) <- pvar
  if (missing(mvar)) mvar <- mt(data) else mt(data) <- mvar
  
  data_list <- .prepare_scdf(data)
  
  keep <- recombine_phases(data_list, phases = phases)
  data_list <- keep$data
  
  designs <- lapply(keep$designs, function(x) x$values)
  designs <- sapply(designs, function(x) paste0(x, collapse = "-"))
  
  N <- length(data_list)

  case_names <- revise_names(data_list)

  vars <- c(
    "PND", "PEM", "PET", "NAP", "NAP rescaled", "PAND", "IRD", "Tau_U(A)", 
    "Tau_U(BA)", "Base_Tau",  "Diff_mean", "Diff_trend", "SMD", "Hedges_g"
  )
  df <- as.data.frame(matrix(nrow = N, ncol = length(vars)))
  colnames(df) <- vars
  df <- data.frame(Case = case_names, Design = designs, df, check.names = FALSE)
  
  for(i in 1:N) {
    data <- data_list[i]
    df$PND[i] <- pnd(data, decreasing = decreasing)$PND
    df$PEM[i] <- pem(data, 
      decreasing = decreasing, binom.test = FALSE, chi.test = FALSE)$PEM$PEM
    df$PET[i] <- pet(data, decreasing = decreasing)$PET$PET
    df$NAP[i] <- nap(data, decreasing = decreasing)$nap[[1, "NAP"]]
    df$"NAP rescaled"[i] <- nap(
      data, decreasing = decreasing)$nap[[1, "NAP Rescaled"]]
    df$PAND[i] <- pand(data, decreasing = decreasing)$pand
    df$IRD[i] <- ird(data, decreasing = decreasing)$ird
    df$`Tau_U(A)`[i] <- tau_u(data)$table[[1]]["A vs. B - Trend A", "Tau"]
    df$`Tau_U(BA)`[i] <- tau_u(data)$table[[1]]["A vs. B + Trend B - Trend A", "Tau"]
    df$Base_Tau[i] <- corrected_tau(data)$tau
    
    data <- data[[1]]
    A <- data[data[, pvar] == "A", dvar]
    B <- data[data[, pvar] == "B", dvar]
    mtA <- data[data[, pvar] == "A", mvar]
    mtB <- data[data[, pvar] == "B", mvar]
    nA <- sum(!is.na(A))
    nB <- sum(!is.na(B))    
    n <- nA + nB
    mA <- mean(A, na.rm = TRUE)
    mB <- mean(B, na.rm = TRUE)    
    sdA <- sd(A, na.rm = TRUE)
    sdB <- sd(B, na.rm = TRUE)    
    
    
    df$Diff_mean[i] <- mB - mA
    df$SMD[i] <- (mB - mA) / sdA
    
    sd_hg <- sqrt(
      ( (nA - 1) * sdA^2 + (nB - 1) * sdB^2) 
      / 
      (nA + nB - 2) 
    )  
    
    df$Hedges_g[i] <- (mB - mA) / sd_hg
    df$Hedges_g[i] <- df$Hedges_g[i] * (1 - (3 / (4 * n - 9)))
    
    df$Diff_trend[i] <- coef(lm(B ~ I(mtB - mtB[1] + 1)))[2] - 
                        coef(lm(A ~ I(mtA - mtA[1] + 1)))[2]
    
  }
  
  out <- list(
    overlap = df, 
    phases.A = keep$phases_A, 
    phases.B = keep$phases_B 
    #design = keep$design[[1]]$values
  )
  
  atr <- scdf_attr(data_list)
  for(i in seq_along(atr)) attr(out, names(atr)[i]) <- atr[[i]]
  class(out) <- c("sc_overlap")
  
  out
}

test_that("overlap", {
  data <- exampleAB
  
  result <- overlap(data = data)
  expect_s3_class(result, "sc_overlap")
  expect_s3_class(result$overlap, "data.frame")
  expect_identical(nrow(result$overlap), 3L)
  expect_identical(ncol(result$overlap), 16L)
  expect_identical(names(result$overlap), 
                   c("Case", "Design", "PND", "PEM", "PET", "NAP", "NAP rescaled", 
                     "PAND", "IRD",  "Tau_U(A)", "Tau_U(BA)", "Base_Tau", "Diff_mean", 
                     "Diff_trend", "SMD", "Hedges_g"))
  expect_equal(round(result$overlap[1, "Diff_mean"], 4), 19.5333)
  expect_equal(round(result$overlap[1, "SMD"], 4), 8.1108)
  expect_equal(round(result$overlap[1, "Hedges_g"], 4), 2.3478)
  expect_equal(round(result$overlap[1, "Diff_trend"], 3), 1.525)
  expect_equal(
    round(sum(apply(result$overlap[,3:15], 2, function(x) sum(x), simplify = TRUE)), 3),
    1829.948
  )
})

#' A Shiny app for scan
#'
#' Run a Shiny app with most of the scan functions.
#' 
#' @return This function launches a Shiny application.
#' @author Juergen Wilbert
#' @param scdf If you provide an *scdf* here, it will be loaded at startup.
#' @param quiet If TRUE (default) does not report shiny messages in the console.
#' @param browser c("external","viewer") 
#' @param theme Bootstrap 5 theme. Default is "cerulean".
#' @param \dots Further arguments passed to the `shiny::runApp()` function.
#' @details This function launches a shiny application.
#'   You need to have `scplot` and `shiny` installed.
#'   These packages are suggested but not necessarily installed along with scan.
#'   `shinyscan()` will ask to install missing packages.
#'
#' @export
shinyscan <- function(scdf = NULL,
                      quiet = TRUE, 
                      browser = c("external", "viewer"),
                      theme = "cerulean",
                      ...) {
  
  browser <- browser[1]
  
  
  
  miss <- c()
  if (!requireNamespace("scplot", quietly = TRUE)) miss <- c(miss, "scplot")
  if (!requireNamespace("shiny",  quietly = TRUE)) miss <- c(miss,  "shiny")
  
  if (length(miss) > 0) {
    notify("shinyscan needs: ", paste(miss, collapse = ", "))
    ans <- utils::askYesNo("Install missing packages now?")
    if (isTRUE(ans)) {
      install.packages(miss) 
    } else {
      abort("Packages missing.")
    }
  }
  
  browser <- match.arg(browser)
  
  options(scan.shiny.theme = theme)
  
  old_opt <- if (identical(browser, "external")) {
    options(shiny.launch.browser = TRUE)
  } else {
    options(shiny.launch.browser = FALSE)
  }

  if (inherits(scdf, "scdf")) {
    old_opt <- options(scan.shinyscan.initial = scdf)
  }
  
  on.exit(options(old_opt), add = TRUE)
  
  app_dir <- system.file("shiny_scan", package = "scan")
  if (app_dir == "") 
    abort("Cannot find inst/shiny_scan in installed package.")
  
  shiny::runApp(app_dir, quiet = quiet, ...)
  
}

#' (Deprecated) Rank-transformation of single-case data files
#'
#' *This function is superseded by the more versatile 
#' \code{\link{transform.scdf}} function.*
#'
#' @inheritParams .inheritParams
#' @param var A string or string vector with the names of the variables to be
#'   ranked.
#' @param grand If TRUE, ranks will be calculated across all cases. If FALSE
#'   ranks are calculated within each case.
#' @param ... Additional parameters passed to the \code{\link{rank}} function.
#' @return An \code{scdf} object where the values of the variable(s) are
#'   replaced with ranks.
#' @author Juergen Wilbert
#' @family data manipulation functions
#' @examples
#' # The ranks function is deprecated. Please use transform:
#' res1 <- ranks(Huber2014, var = "compliance")
#' res2 <- transform(Huber2014, across_cases(compliance = rank(compliance, na.last="keep")))
#' identical(res1, res2)
#' 
#' res1 <- ranks(Huber2014, var = "compliance", grand = FALSE)
#' res2 <- transform(Huber2014, compliance = rank(compliance, na.last="keep"))
#' identical(res1, res2)
#' @export
#' @keywords internal

ranks <- function(data, var, grand = TRUE, ...) {
  
  if(identical(grand, FALSE)) {
    for(i in 1:length(data)) {
      for(v in var)
        data[[i]][, v] <- rank(data[[i]][, v], na.last = "keep", ...)
    }
  }

  if(isTRUE(grand)) {
    ATTRIBUTES <- attributes(data)
    data       <- as.data.frame(data)
    data$case  <- factor(data$case, levels = unique(data$case))
    for (v in var) data[, v] <- rank(data[, v], na.last = "keep", ...)
    data <- split(data, data$case)
    data <- lapply(data, function(x) x[, -1])
    attributes(data) <- ATTRIBUTES
  }
    
  data
}

#' @rdname print.sc
#' @export
#' 
print.sc_pet <- function(x, digits = 3, ...) {
  cat("Percent Exceeding the Trend\n\n")
  cat("\n")
  
  print(x$PET, row.names = FALSE, digits = digits, ...)
  cat("\n")
  
  if (x$decreasing) {
    cat("Assumed decreasing values in the B-phase.\n\n")
    cat("Binom.test: alternative hypothesis: true probability < 50%\n")
    cat(
      sprintf(
        "PET CI: Percent of values less than lower %d%% confidence threshold (single sided)\n", 
        x$ci * 100))
  } else {
    cat("Binom.test: alternative hypothesis: true probability > 50%\n")
    cat(
      sprintf(
        "PET CI: Percent of values greater than upper %d%% confidence threshold (single sided)\n", 
        x$ci * 100))
  }
  
}	

#' Shift values in a single-case data file
#'
#' *This function has been superseded by the much more versatile 
#' \code{\link{transform.scdf}} function.*
#' Shifting the values might be helpful in cases where the measurement time
#' is given as a time variable (see example below).
#'
#' @inheritParams .inheritParams
#' @param value Number by which to shift the values
#' @param var Character string with the name of the target variable.
#' Defaults to the measurement time variable.
#'
#' @return A scdf with shifted data
#' @family data manipulation functions
#' @keywords internal
#' @export
#'
#' @examples
#' ### Shift the measurement time for a better estimation of the intercept
#' ex <- shift(example_A24, value = -1996)
#' plm(ex)
#' 
#' # Please use transform instead:
#' example_A24 |>
#'   transform(year = year - 1996) |>
#'   plm()
shift <- function(data, value, var) {
  .deprecated_warning("transform", "shift")
  if (missing(var)) var <- mt(data)
  for (i in 1:length(data)) {
    data[[i]][[var]] <- data[[i]][[var]] + value
  }
  data
}

#' Add Dummy Variables for Piecewise Linear Models
#' 
#' Adds dummy variables to an scdf for calculating piecewise linear models.
#' 
#' This function creates dummy variables for phase levels and phase slopes
#' according to the specified piecewise regression model. It supports different
#' contrast coding schemes for both level and slope contrasts.
#' 
#' @inheritParams .inheritParams
#' @examples
#' add_dummy_variables(
#'  scdf = exampleABC, 
#'  model = "W", 
#'  contrast_level = "first", 
#'  contrast_slope = "first"
#' )
#' @export
add_dummy_variables <- function(scdf, 
                                model = c("W", "H-M", "B&L-B"), 
                                contrast_level = c("first", "preceding"),
                                contrast_slope = c("first", "preceding")) {
  
  check_args(
    by_call(model),
    by_call(contrast_level),
    by_call(contrast_slope)
  )
  
  out <- .add_dummy_variables(
    data = scdf, 
    model = model, 
    contrast_level = contrast_level, 
    contrast_slope = contrast_slope
  )
  out$data
}


.add_dummy_variables <- function(data, model, 
                                dvar = dv(data), 
                                pvar = phase(data), 
                                mvar = mt(data),
                                contrast_level,
                                contrast_slope) {
      

  for(case in 1:length(data)) {
    dat_inter <- .create_dummies(
      data[[case]], 
      model = model, 
      dvar = dvar, 
      pvar = pvar, 
      mvar = mvar, 
      contrast_level = contrast_level, 
      contrast_slope = contrast_slope
    )
    
    data[[case]][, mvar] <- dat_inter$mt
    data[[case]] <- cbind(data[[case]], dat_inter[, -1])
    n_Var <- (ncol(dat_inter) - 1) / 2
    var_inter <- names(dat_inter)[(ncol(dat_inter) - n_Var + 1):ncol(dat_inter)]
    var_phase <- names(dat_inter)[2:(n_Var + 1)]
  }
  out <- list(
    data = data, 
    var_inter = var_inter, 
    var_phase = var_phase
  )
  out
}

.create_dummies <- function(data, 
                            dvar, 
                            pvar, 
                            mvar,
                            model,
                            contrast_level,
                            contrast_slope) {

  mt <- data[[mvar]]
  n  <- nrow(data)
  
  if (model == "W") mt <- mt - mt[1]
  
  out <- data.frame(mt = mt)
  design <- rle(as.character(data[[pvar]]))
  
  # dummy phases
  
  if (identical(contrast_level, "first")) {
    contr <- contr.treatment(nlevels(data[[pvar]]))
  } else if (identical(contrast_level, "preceding")) {
    contr <- contr.cum(nlevels(data[[pvar]]))
  } else {
    abort("Wrong declaration of level contrast: ", contrast_level)
  }
  rownames(contr) <- levels(data[[pvar]])
  colnames(contr) <- levels(data[[pvar]])[-1]
  contrasts(data[[pvar]])<- contr 
  
  add <- .create_phase_dummies(data[[pvar]])
  out <- cbind(out, add)

  # dummy slopes
  
  if (identical(contrast_slope, "first")){
    contr <- contr.treatment(nlevels(data[[pvar]]))
  } else if (identical(contrast_slope, "preceding")) {
    contr <- contr.cum(nlevels(data[[pvar]]))
  } else {
    abort("Wrong declaration of slope contrast: ", contrast_slope)
  }
  
  rownames(contr) <- levels(data[[pvar]])
  colnames(contr) <- levels(data[[pvar]])[-1]
  
  add <- .create_slope_dummies(data[[pvar]], mt, contr, model)
  out <- cbind(out, add)
  
  out
}

contr.cum <- function(n) {
  out <- c() 
  for(i in 1:(n - 1)) out <- c(out, c(rep(0, i), rep(1, n - i)))
  matrix(out, ncol = n - 1)
}

.create_phase_dummies <- function(fac, 
                          contrast, 
                          var_name = getOption("scan.string.dummy.phase")) {
  
  if (missing(contrast)) contrast <- contrasts(fac)
  if(is.null(colnames(contrast))) {
    dummy_names <- paste0(var_name, 1:ncol(contrast))
  } else {
    dummy_names <- paste0(var_name, colnames(contrast))
  }
  
  vec_len <- length(fac)
  df <- list()
  
  for(i in 1:ncol(contrast)) {
    df[[dummy_names[i]]] <- numeric(vec_len)
    for(j in 1:nrow(contrast)) {
      id <- which(fac == levels(fac)[j])
      df[[dummy_names[i]]][id] <- contrast[j,i]
    }
  }
  
  df <- as.data.frame(df)
  attr(df, "formula") <- formula(paste(
      c(".", "~", paste(c(".", dummy_names), collapse = "+")), 
      collapse = " "
  ))
  df
}

.create_slope_dummies <- function(phase, 
                                  mt, 
                                  contrast, 
                                  model, 
                                  var_name = getOption("scan.string.dummy.slope")) {
  
  if(is.null(colnames(contrast))) {
    dummy_names <- paste0(var_name, 1:ncol(contrast))
  } else {
    dummy_names <- paste0(var_name, colnames(contrast))
  }
  
  df <- list()
  
  for(i in 1:ncol(contrast)) {
    df[[dummy_names[i]]] <- 0
    phase_str <- rle(contrast[,i])
    class(phase_str) <- "list"
    phase_str$start <- c(1, cumsum(phase_str$lengths) + 1)[1:length(phase_str$lengths)]
    phase_str$stop  <- cumsum(phase_str$lengths)

    for(j in 1:length(phase_str$lengths)) {
      selection_phases <- levels(phase)[phase_str$start[j]:phase_str$stop[j]]
      id <- which(phase %in% selection_phases)
      
      mt_dummy <- (mt[id] - mt[id[1]])
      if (model %in% c("B&L-B")) 
        df[[dummy_names[i]]][id] <- (mt_dummy + 1) * phase_str$values[j] #check: (mt[1:length(id)] - mt[1] + 1) * phase_str$values[j]
      
     
      if (model %in% c("H-M", "W"))
        df[[dummy_names[i]]][id] <- mt_dummy * phase_str$values[j] #check: (mt[1:length(id)] - mt[1]) * phase_str$values[j]
    }
  }
  
  df <- as.data.frame(df)
  attr(df, "formula") <- formula(paste(
    c(".", "~", paste(c(".",dummy_names), collapse = "+")), 
    collapse = " "
  ))
  df
}

#' Tau-U for single-case data
#'
#' This function calculates indices of the Tau-U family as proposed by Parker et
#' al. (2011a). It allows to calculate Tau-U values for single cases as well
#' as overall Tau-U values across several single cases by applying a meta
#' analysis.
#'
#' @order 1
#' @inheritParams .inheritParams
#' @param method `"complete"` (default), `"parker"` or `"tarlow"`. The
#'   `"parker"` calculates the number of possible pairs as described in Parker
#'   et al. (2011) which might lead to tau-U values greater than 1. `"tarlow"`
#'   follows an online calculator and R code developed by Tarlow (2017).
#' @param meta_analyses If TRUE, a meta analysis is conducted.
#' @param meta_weight_method String to specify the method for calculating the
#'   weights of the studies. Either "tau" or "z".
#' @param ci Confidence intervals level. If NULL or NA, no confidence intervals
#'   are calculated.
#' @param ci_method String to specify the method for calculating the standard
#'   error of tau. Either "tau", "z", or "s" (not recommended).
#' @param tau_method Character with values "a" or "b" (default) indicating
#'   whether Kendall Tau A or Kendall Tau B is applied. Ignored for methods
#'   'tarlow' and 'parker'.
#' @param continuity_correction If TRUE, a continuity correction is applied for
#'   calculating p-values of correlations (here: S will be reduced by one before
#'   calculating Z). Ignored for methods 'tarlow' and 'parker'.
#' @return
#' \item{table}{A data frame containing statistics from the Tau-U
#' family, including: Pairs, positive and negative comparisons, S, and Tau}
#' \item{matrix}{The matrix of comparisons used for calculating the
#' statistics.} \item{tau_u}{Tau-U value.}
#' @details Tau-U is an inconsistently operationalized construct. Parker et al.
#'   (2011b) describe a method which may result in Tau-U outside the \[-1;1\]
#'   interval. A different implementation of the method (provided at
#'   http://www.singlecaseresearch.org/calculators/tau-u) uses tau-b (instead of
#'   tau-a as in the original formulation by Parker). Bossart et. al (2018)
#'   describe inconsistencies in the results from this implementation as well.
#'   Another problems lies in the calculation in overall Tau-U values from
#'   several single cases. The function presented here applies a meta-analysis
#'   to gain the overall values. Each tau value is weighted by the inverse of
#'   the variance (ie. the tau standard error). The confidence intervals for
#'   single cases are calculated by Fisher-Z transforming tau, calculating the
#'   confidence intervals, and inverse transform them back to tau (see Long &
#'   Cliff, 1997).
#' @author Juergen Wilbert
#' @family overlap functions
#' @references Brossart, D. F., Laird, V. C., & Armstrong, T. W. (2018).
#'   Interpreting Kendall’s Tau and Tau-U for single-case experimental designs.
#'   \emph{Cogent Psychology, 5(1)}, 1–26.
#'   https://doi.org/10.1080/23311908.2018.1518687.
#'
#'   Long, J. D., & Cliff, N. (1997). Confidence intervals for Kendall’s tau.
#'   \emph{British Journal of Mathematical and Statistical Psychology}, 50(1),
#'   31–41. https://doi.org/10.1111/j.2044-8317.1997.tb01100.x
#'
#'   Parker, R. I., Vannest, K. J., & Davis, J. L. (2011a). Effect Size in
#'   Single-Case Research: A Review of Nine Nonoverlap Techniques.
#'   \emph{Behavior Modification}, 35(4), 303–322. https://doi.org/10/dsdfs4
#'
#'   Parker, R. I., Vannest, K. J., Davis, J. L., & Sauber, S. B. (2011b).
#'   Combining Nonoverlap and Trend for Single-Case Research: Tau-U.
#'   \emph{Behavior Therapy, 42}(2), 284–299.
#'   https://doi.org/10.1016/j.beth.2010.08.006
#'
#'   Tarlow, K. R. (2017, March). Tau-U for single-case research (R code).
#'   Retrieved from http://ktarlow.com/stats/
#'
#' @examples
#'
#' tau_u(Grosche2011$Eva)
#'
#' ## Replicate  tau-U calculation from Parker et al. (2011)
#' bob <- scdf(c(A = 2, 3, 5, 3, B = 4, 5, 5, 7, 6), name = "Bob")
#' res <- tau_u(bob, method = "parker")
#' print(res, complete = TRUE)
#'
#' ## Request tau-U for all single-cases from the Grosche2011 data set
#' tau_u(Grosche2011)
#' @export

tau_u <- function(data, dvar, pvar, 
                  method = c("complete", "parker", "tarlow"), 
                  phases = c(1, 2), 
                  meta_analyses = TRUE,
                  ci = 0.95,
                  ci_method = c("z", "tau", "s"),
                  meta_weight_method = c("z", "tau"),
                  tau_method = c("b", "a"), 
                  continuity_correction = FALSE) {

  # validity check ----
  check_args(
    by_call(tau_method),
    by_call(ci_method),
    by_call(method),
    by_call(meta_weight_method),
    within(ci, 0, 1)
  )
  
  if (method == "parker") {
    #message("method = 'parker' ignores the tau_method argument.")
    tau_method <- "a"
    continuity_correction <- FALSE
  }
  
  if (method == "tarlow") {
    #message("method = 'tarlow' ignores the tau_method argument.")
    tau_method <- "a"
    continuity_correction <- TRUE
  }

  # prepare scdf ----
  if (missing(dvar)) dvar <- dv(data)
  if (missing(pvar)) pvar <- phase(data)
  phase(data) <- pvar
  dv(data) <- dvar
  
  data <- .prepare_scdf(data)
  data <- recombine_phases(data, phases = phases)$data
  
  # define "out" data structure ----
  N <- length(data)
  out <- list(
    table = list(),
    tau_u = list(),
    method = method,
    tau_method = tau_method,
    phases = phases,
    n_cases = N,
    continuity_correction = continuity_correction,
    Overall_tau_u = NA,
    meta_analyses = meta_analyses,
    meta_weight_method = meta_weight_method,
    ci = ci,
    ci_method = ci_method
  )
  
  # define tau table data structure -----
  row_names <- c(
    "A vs. B", 
    "Trend A", 
    "Trend B", 
    "A vs. B - Trend A",
    "A vs. B + Trend B", 
    "A vs. B + Trend B - Trend A"
  )
  col_names <- c(
    "pairs", "pos", "neg", "ties", "S", "D", "Tau", "CI lower", "CI upper",
    "SD_S", "VAR_S", "SE_Tau", "Z", "p"
  )
  
  template_table_tau <- as.data.frame(matrix(
    NA, length(row_names), length(col_names), 
    dimnames = list(row_names, col_names)
  ))

  # tau-U for each case -----
  for (case in 1:N) {
    
    table_tau <- template_table_tau
    
    # Extract A and B phase values
    .isA <- data[[case]][[pvar]] == "A"
    .isB <- data[[case]][[pvar]] == "B"
    A <- data[[case]][.isA, dvar]
    B <- data[[case]][.isB, dvar]
    
    # drop NA
    A <- A[!is.na(A)]
    B <- B[!is.na(B)]
    
    #
    AB <- c(A, B)
    nA <- length(A)
    nB <- length(B)
    nAB <- nA + nB
    
    # create tau matrix -----------------------------------------------------
    AvApos <- 0
    AvAneg <- 0
    AvAtie <- 0
    BvBpos <- 0
    BvBneg <- 0
    BvBtie <- 0
    AvBpos <- 0
    AvBneg <- 0
    AvBtie <- 0
    
    for(i in 1:(nA-1)) {
      AvApos <- AvApos + sum(A[i] < A[(i+1):nA])
      AvAneg <- AvAneg + sum(A[i] > A[(i+1):nA])
      AvAtie <- AvAtie + sum(A[i] == A[(i+1):nA])
    }
    
    for(i in 1:(nB-1)) {
      BvBpos <- BvBpos + sum(B[i] < B[(i+1):nB])
      BvBneg <- BvBneg + sum(B[i] > B[(i+1):nB])
      BvBtie <- BvBtie + sum(B[i] == B[(i+1):nB])
    }
    
    AvBpos <- sum(vapply(A, function(x) x < B, FUN.VALUE = logical(nB)))
    AvBneg <- sum(vapply(A, function(x) x > B, FUN.VALUE = logical(nB)))
    AvBtie <- sum(vapply(A, function(x) x == B, FUN.VALUE = logical(nB)))
    
    # Kendall tau analyses ----------------------------------------------------

    if (method == "complete" && tau_method == "a") {
      tau_s <- list(
        AvB = kendall_tau(AB, c(rep(0, nA), rep(1, nB)), tau_method = "a", continuity_correction = continuity_correction), 
        AvA = kendall_tau(A, 1:nA, tau_method = "a", continuity_correction = continuity_correction), 
        BvB = kendall_tau(B, 1:nB, tau_method = "a", continuity_correction = continuity_correction), 
        AvB_A = kendall_tau(AB, c(nA:1, rep(nA + 1, nB)), tau_method = "a", continuity_correction = continuity_correction), 
        AvB_B = kendall_tau(AB, c(rep(0, nA), (nA + 1):nAB), tau_method = "a", continuity_correction = continuity_correction), 
        AvB_B_A = kendall_tau(AB, c(nA:1, (nA + 1):nAB), tau_method = "a", continuity_correction = continuity_correction)
      )
    } else {
      tau_s <- list(
        AvB = kendall_tau(AB, c(rep(0, nA), rep(1, nB)), tau_method = "b", continuity_correction = continuity_correction), 
        AvA = kendall_tau(A, 1:nA, tau_method = "b", continuity_correction = continuity_correction), 
        BvB = kendall_tau(B, 1:nB, tau_method = "b", continuity_correction = continuity_correction), 
        AvB_A = kendall_tau(AB, c(nA:1, rep(nA + 1, nB)), tau_method = "b", continuity_correction = continuity_correction), 
        AvB_B = kendall_tau(AB, c(rep(0, nA), (nA + 1):nAB), tau_method = "b", continuity_correction = continuity_correction), 
        AvB_B_A = kendall_tau(AB, c(nA:1, (nA + 1):nAB), tau_method = "b", continuity_correction = continuity_correction)
      )
    }

    # n ----------------------------
    
    table_tau$n <- lapply(tau_s, \(.) .$N)|>unlist()
    
    # pairs -------------------------------------------------------------------
    
    AvB_pair <- nA * nB
    AvA_pair <- nA * (nA - 1) / 2
    BvB_pair <- nB * (nB - 1) / 2
    ABvAB_pair <- nAB * (nAB - 1) / 2
    
    table_tau$pairs <- c(
      AvB_pair, # A vs. B
      AvA_pair, # A vs. A
      BvB_pair, # B vs. B
      AvB_pair + AvA_pair, # A vs. B - A vs. A
      AvB_pair + BvB_pair, # A vs. B + B vs. B
      AvB_pair + AvA_pair + BvB_pair # A vs. B + B vs. B - A vs. A
    )
    
    if (method == "parker") {
      table_tau$pairs[4] <- AvB_pair # A vs. B - A vs. A
      table_tau$pairs[6] <- ABvAB_pair # A vs. B + B vs. B - A vs. A
    }
    
    # pos/neg/ties ------------------------------------------------------------
    
    table_tau$pos <- c(
      AvBpos,
      AvApos,
      BvBpos,
      AvBpos + AvAneg,
      AvBpos + BvBpos,
      AvBpos + BvBpos + AvAneg
    )
    
    table_tau$neg <- c(
      AvBneg,
      AvAneg,
      BvBneg,
      AvBneg + AvApos,
      AvBneg + BvBneg,
      AvBneg + BvBneg + AvApos
    )
    
    table_tau$ties <- c(
      AvBtie,
      AvAtie,
      BvBtie,
      AvBtie + AvAtie,
      AvBtie + BvBtie,
      AvBtie + BvBtie + AvAtie
    )
    
    # S -----------------------------------------------------------------
    
    table_tau$S <- lapply(tau_s, \(.) .$S)|>unlist()
   
    # D ----------------------------------------------------------------------
    
    if (method == "complete" && tau_method == "b") {
      table_tau$D <- lapply(tau_s, \(.) .$D)|>unlist()
      table_tau$D[1] <- table_tau$pairs[1] - table_tau$ties[1] / 2
    } else {
      table_tau$D <- table_tau$pairs
    }

    # tau -----------------------------------------------------------
    
    table_tau$Tau <- table_tau$S / table_tau$D

    # SD and VAR --------------------------------------------------------------
    
    if (method == "tarlow") {
      table_tau$SD_S <- lapply(tau_s, \(.) .$sdS)|>unlist()
    } else {
      table_tau$SD_S <- c(
        sqrt((nA * nB) * (nA + nB + 1) / 12) * 2,
        kendall_tau(1:nA, 1:nA, tau_method = tau_method)$sdS,
        kendall_tau(1:nB, 1:nB, tau_method = tau_method)$sdS,
        tau_s$AvB_A$sdS,
        tau_s$AvB_B$sdS,
        tau_s$AvB_B_A$sdS
      )
    }
    
    table_tau$VAR_S <- table_tau$SD_S^2

    # Z, p, se ----------------------------------------
    
    table_tau$Z <- lapply(tau_s, \(.) .$z)|>unlist()
    table_tau$p <- lapply(tau_s, \(.) .$p)|>unlist()
    
    table_tau$SE_Tau <- table_tau$Tau / table_tau$Z
    
    # confidence intervals --------------------
    if (!is.na(ci) && !is.null(ci)) {
      if (ci_method == "s") {
        see <- qnorm((1 - ci) / 2, lower.tail = FALSE)
        S <- table_tau$S
        if (continuity_correction) S <- S - 1
        cis <- list(
          tau_ci_lower = (S - table_tau$SD_S * see) / table_tau$D, 
          tau_ci_upper = (S + table_tau$SD_S * see) / table_tau$D
        )
      } else {
        cis <- .tau_ci(
          table_tau$Tau, table_tau$n, ci = ci, se_method = ci_method
        )
      }
    } else {
      cis <- list(tau_ci_lower = NA, tau_ci_upper = NA)
    }
    
    table_tau$`CI lower` <-  cis$tau_ci_lower
    table_tau$`CI upper` <-  cis$tau_ci_upper

    out$table[[case]] <- table_tau
    out$tau_u[[case]] <- c(
      "A vs. B - Trend A" = 
        table_tau["A vs. B - Trend A", "Tau"]
    )
  }
  
  # Meta analysis ----------------------------------------------------
  
  if (meta_analyses) {
    out$Overall_tau_u <- .meta_tau_u(
      out$table, ci = ci, se_method = meta_weight_method
    )
  } else {
    out$Overall_tau_u <- NA
  }
  
  # return ------------------------------------------------------------------
  
  names(out$table) <- names(data)
  names(out$tau_u) <- names(data)
  
  class(out) <- c("sc_tauu")
  attributes(out)[opts("phase", "dv")] <- list(pvar, dvar)
  out
}


.meta_tau_u <- function(tau_matrix, ci, se_method) {
  
  ci_z <- qnorm((1 - ci) / 2, lower.tail = FALSE)
  
  .meta <- function(model) {
    tau <- sapply(tau_matrix, function(x) x[model, "Tau"])
    n <- sapply(tau_matrix, function(x) x[model, "n"])
    out <- data.frame(Model = model)
  
    res <- .meta_tau(tau, n, ci = ci, se_method = se_method)      

    out$Tau_U <- res$tau
    out$se <- res$se
    out$'CI lower' <- res$lower
    out$'CI upper' <- res$upper
    out$z <- res$z
    out$p <- res$p
    
    out
  }
  
  out <- data.frame(
    Model = character(4), 
    Tau_U = numeric(4),
    se = numeric(4),
    'CI lower' = numeric(4),
    'CI upper' = numeric(4),
    z = numeric(4),
    p = numeric(4),
    check.names = FALSE
  )
  
  models <- c("A vs. B", 
              "A vs. B - Trend A", 
              "A vs. B + Trend B", 
              "A vs. B + Trend B - Trend A"
            )
  
  for(i in 1:length(models)) out[i,] <- .meta(models[i])
  
  out
}

test_that("tau-u", {
  
  data <- Grosche2011
  result <- tau_u(data = data)
  expect_s3_class(result, "sc_tauu")
  expect_identical(sum(round(result$Overall_tau_u$p,5)), 1.44159)
  expect_identical(sum(round(unlist(result$Overall_tau_u[, -1]),5)), 7.57242)
  expect_identical(round(result$table[[1]]|>sum(), 3), 5295.933)

  result <- tau_u(
    Huber2014, 
    method = "parker", ci = 0.9, ci_method = "tau"
  )
  
  expect_identical(unlist(result$table)|>sum(na.rm = TRUE)|>round(0), 439300)

  result <- tau_u(
    Huber2014, 
    meta_weight_method = "z"
  )
  expect_identical(sum(round(unlist(result$Overall_tau_u[, -1]),5)), 22.51747)
  
  
})

#' @rdname print.sc
#' @export
#' 
print.sc_pnd <- function(x, ...) {
  cat("Percent Non-Overlapping Data\n\n")
  out <- data.frame(
    Case = x$case.names, 
    PND = paste0(round(x$PND, 2),"%"), 
    "Total" = x$n.B, 
    "Exceeds" = round(x$PND / 100 * x$n.B)
  )
  print(out, row.names = FALSE)
  cat("\nMean  :", round(mean(x$PND, na.rm = TRUE), 2),"%\n")
}	

#' @rdname export
#' @export
export.sc_pnd <- function(object, 
                          caption = NA, 
                          footnote = NA, 
                          filename = NA,
                          select = c("Case", "PND", "Total", "Exceeds"),
                          round = 2,
                          ...) {
  
  if (is.na(caption)) caption <- c("Percentage Non-Overlapping Data")
  
  out <- data.frame(
    Case = object$case.names, 
    PND = object$PND, 
    Total = object$n.B, 
    Exceeds = round(object$n.B * object$PND / 100)
  )
  out <- .select(out, select)
  
  out <- round_numeric(out, round)
  
  table <- .create_table(
    out, 
    caption = caption,
    footnote = footnote,
    ...
  )
  
  # finish ------------------------------------------------------------------
  
  if (!is.na(filename)) .save_export(table, filename)
  
  table
}

#' Check function arguments
#' 
#' This function checks function arguments based on a set of predefined tests.
#' It is intended for internal use within other functions to validate their
#' arguments.
#' #' @param ... A set of expressions defining the tests to be performed on the
#' function arguments. Each expression should use one of the predefined test
#' functions available within the scope of `check_args()`.
#' @return This function does not return a value. If any of the tests fail,
#' it raises an error with a descriptive message.
#' @keywords internal
check_args <- function(...) {
  
  if (!opt("check_arguments")) {
    return()
  }
 
  expressions <- substitute(list(...))

  env <- new.env(parent = parent.frame()) 
 
  env$call <- as.call(
    as.list(parse(text = gsub("^[^:]+::", "", deparse(sys.call(-1)))))
  )[[1]]
  
  env$is_true <- function(condition, ..., .warning = FALSE) {
    
    if (isFALSE(condition)) {
      message <- paste0(...)
      if (length(message) == 0) {
        message <- paste0(
          "Argument ", as.character(match.call()[2]), " is ill defined."
        )
      }
      return(list(pass = FALSE, msg = message, warning = .warning) )
    } else {
      return(list(pass = TRUE))
    }
  }
  
  env$has_length <- function(arg, l, msg, .warning = FALSE) {
    if (missing(msg)) 
      msg <- paste0(
        "Argument ", as.character(match.call()[2]), " not of length ", l, "."
      )
    env$is_true(length(arg) == l, msg)
  }
  
  env$not <- function(condition, ..., .warning = FALSE) env$is_true(!condition, ...)
  
  env$one_of <- function(arg, ..., .warning = FALSE) {
    match <- c(...)
    msg <- paste0("'", match, "'")
    if (length(match) == 2) msg <- paste0(msg, collapse = " or ")
    if (length(match) > 2) {
      msg <- paste0("one of ", paste0(msg, collapse = ", "))
    }
    env$is_true(
      arg %in% match, 
      "Argument ", as.character(match.call()[2]), " is not ", msg, "."
    )
  }
  
  env$by_call <- function(arg, .warning = FALSE) {
    
    args <- formals(as.character(env$call[[1]]))
   
    id <- which(names(args) == as.character(match.call()[2]))
    if (length(id) == 0) stop("by_call has no matching arg.")
    match <- eval(args[[id]])
    msg <- paste0("'", match, "'")
    if (length(match) == 2) msg <- paste0(msg, collapse = " or ")
    if (length(match) > 2) msg <- paste0("one of ", paste0(msg, collapse = ", "))
    arg_name <- as.character(match.call()[2])
    if (identical(arg, match)) {
      caller_env <- parent.frame(n = 4)
      caller_env[[arg_name]] <- arg[1]
      return(list(pass = TRUE))
    }
  
    env$is_true(
      arg %in% match, 
      "Argument ", arg_name, " is not ", msg, "."
    )
  }
  
  env$within <- function(arg, lower, upper, .warning = FALSE) {
    env$is_true(
      arg >= lower && arg <= upper, 
      "Argument ",
      as.character(match.call()[2]), 
      " is not within ", lower, " and ", upper, " (is ", arg, ")"
    ) 
  }
  
  env$at_least <- function(arg, lower, msg, .warning = FALSE) {
    
    if (missing(msg)) {
      msg <- paste0(
        "Argument ",
        as.character(match.call()[2]), 
        " is not greater or equal to ", lower, " (is ", arg, ")"
      )
    }
    
    env$is_true(arg >= lower, msg) 
  }
  
  env$at_most <- function(arg, upper, msg, .warning = FALSE) {
    if(missing(msg)) {
      msg <- paste0(
        "Argument ",
        as.character(match.call()[2]), 
        " is not less or equal to ", upper, " (is ", arg, ")"
      )
    }
    env$is_true(arg <= upper, msg) 
  }
  
  env$by_class <- function(param, class, ..., .warning = FALSE) {
    env$is_true(
      inherits(param, class), 
      "Argument ", 
      as.character(match.call()[2]), " is not of class ", class, "."
    )
  }
  
  env$is_logical <- function(param, .warning = FALSE) {
    env$is_true(
      is.logical(param), 
      "Argument ", as.character(match.call()[2]), " is not logical."
    )
  }
  
  env$is_deprecated <- function() {
    .call <- env$call
    defaults <- formals(as.character(.call[[1]]))
    .call <- as.list(.call)
    id_deprecated <- names(defaults)[
      sapply(defaults, function(x) identical(x, "deprecated")) |> which()
    ]
    id <- which(names(.call) %in% id_deprecated)
  
    if (length(id) > 0) {
      message <- paste0(
        if (length(id) > 1) "Arguments " else "Argument ", 
        paste0("'", names(.call)[id], "'", collapse = ", "),
        if (length(id) > 1) " are " else " is ", 
        "deprecated"
      )
      return(list(pass = FALSE, msg = message, warning = TRUE) )
    }
    return(list(pass = TRUE))
  }
  
  env$as_deprecated <- function(...) {
    
    test_args <- list(...)
    .call <- env$call
    defaults <- formals(as.character(.call[[1]]))
    .call <- as.list(.call)
    deprecated_args <- test_args[which(names(test_args) %in% names(.call))]
    
    if (length(deprecated_args) > 0) {
      message <- paste0(
        "Deprecated ",
        if (length(deprecated_args) > 1) "arguments " else "argument ", 
        paste0("'", names(deprecated_args), "'", collapse = ", "),
        ". Use ",
        paste0("'", unlist(deprecated_args), "'", collapse = ", "),
        " instead"
      )
      return(list(pass = FALSE, msg = message, warning = TRUE) )
    }
    return(list(pass = TRUE))
  }
  
  
  out <- vector("list", length(expressions))
  
  if (length(out) > 0) {
    for(i in 2:length(expressions)) {
      out[i - 1] <- eval(expressions[c(1, i)], envir = env)
    }
  }  
  
  out[length(out)] <- list(eval(str2lang("is_deprecated()"), envir = env))

  out <- out[sapply(out, function(x) if (!isTRUE(x$pass)) TRUE else FALSE)]

  error_msg <- lapply(out, function(x) {
    if (!x$warning) x$msg else NULL
  }) |> unlist()
  
  warning_msg <- lapply(out, function(x) {
    if (x$warning) x$msg else NULL
  }) |> unlist()
  
  if (length(warning_msg) > 0) {
    #warning_msg <- paste0(1:length(warning_msg), ": ", warning_msg)
    warn(warning_msg)
  }
  
  if (length(error_msg) > 0) {
    #error_msg <- paste0(1:length(error_msg), ": ", error_msg)
    abort(error_msg)
  }
}

# 

utils::globalVariables(
  c(
    "by_class", "by_call", "not","within", "one_of", "has_length", "is_true",
    "at_least", "at_most", "is_logical", "as_deprecated"
  )
)

#' List of old deprecated function names
#' 
#' This is a list with functions names that have been replaced by new names
#' which are in line with R syntax guidelines. The old function names still 
#' work. They are wrappers that call the new function.
#' 
#' @param ... Arguments passed through to the new function.
#' @name deprecated-functions
#' @keywords internal
NULL

#' @rdname deprecated-functions
#' @export
tauUSC <- function(...) {
  .deprecated_warning("tau_u", "tauUSC")
  tau_u(...)
}

#' @rdname deprecated-functions
#' @export
power_testSC <- function(...) {
  .deprecated_warning("power_test", "power_testSC")
  power_test(...)
}

#' @rdname deprecated-functions
#' @export
fillmissingSC <- function(...) {
  .deprecated_warning("fill_missing", "fillmissingSC")
  fill_missing(...)
}

#' @rdname deprecated-functions
#' @export
overlapSC <- function(...) {
  .deprecated_warning("overlap", "overlapSC")
  overlap(...)
}

#' @rdname deprecated-functions
#' @export
randSC <- function(...) {
  .deprecated_warning("rand_test", "randSC")
  rand_test(...)
}

#' @rdname deprecated-functions
#' @export
rand.test <- function(...) {
  .deprecated_warning("rand_test", "rand.test")
  rand_test(...)
}

#' @rdname deprecated-functions
#' @export
rciSC <- function(...) {
  rci(...)
}


#' @rdname deprecated-functions
#' @export
rSC <- function(...) {
  .deprecated_warning("random_scdf", "rSC")
  random_scdf(...)
}

#' IRD - Improvement rate difference
#'
#' `ird()` calculates the robust improvement rate difference as proposed by
#' Parker and colleagues (2011).
#'
#' The adaptation of the improvement rate difference for single-case phase
#' comparisons was developed by Parker and colleagues (2009). A variation called
#' robust improvement rate difference was proposed by Parker and colleagues in
#' 2011. This function calculates the robust improvement rate difference. It
#' follows the formula suggested by Pustejovsky (2019). For a multiple case
#' design, ird is based on the overall improvement rate of all cases which is
#' the average of the irds for each case.
#' 
#' @inheritParams .inheritParams
#' @family overlap functions
#' @references Parker, R. I., Vannest, K. J., & Brown, L. (2009). The
#'   improvement rate difference for single-case research. Exceptional Children,
#'   75(2), 135-150.
#'
#'   Parker, R. I., Vannest, K. J., & Davis, J. L. (2011). Effect Size in
#'   Single-Case Research: A Review of Nine Nonoverlap Techniques. Behavior
#'   Modification, 35(4), 303-322. https://doi.org/10.1177/0145445511399147
#'
#'   Pustejovsky, J. E. (2019). Procedural sensitivities of effect sizes for
#'   single-case designs with directly observed behavioral outcome measures.
#'  *Psychological Methods*, *24(2)*, 217-235.
#'   https://doi.org/10.1037/met0000179
#'
#' @order 1
#' @export
ird <- function(data, dvar, pvar,
                decreasing = FALSE,
                phases = c(1, 2)) {
  check_args(
    by_class(decreasing, "logical")
  )
  
  # set attributes to arguments else set to defaults of scdf
  if (missing(dvar)) dvar <- dv(data) else dv(data) <- dvar
  if (missing(pvar)) pvar <- phase(data) else phase(data) <- pvar
  
  data <- .prepare_scdf(data, na.rm = TRUE)
  recombined_data <- recombine_phases(data, phases = phases)
  
  data <- recombined_data$data
  
  casenames <- revise_names(data)

  pa <- pand(data, method = "minimum", decreasing = decreasing)
  
  ird <- 1 - ( ( (pa$n^2) / (2 * pa$n_a * pa$n_b) ) * (1 - (pa$pand / 100)) )
  
  out <- list(
    ird = ird,
    decreasing = decreasing,
    phases = recombined_data$phases,
    n_cases = length(data)
  )
  
  class(out) <- "sc_ird"
  attributes(out)[opts("phase", "dv")] <- list(pvar, dvar)
  out
  
}


#' @describeIn ird Print results
#' @order 2
#' @param x An object returned by [ird()]
#' @param digits The minimum number of significant digits to be use. 
#' @export
#' 
print.sc_ird <- function(x, digits = 3, ...) {
  cat("Improvement rate difference =",  round(x$ird, digits))
  if (x$decreasing) {
    cat("\nAssumed decreasing values in Phase B.\n\n")
  }
}

#' @describeIn ird Export results to html
#' @order 3
#' @inheritParams export
#' @export
export.sc_ird <- function(object, 
                           caption = NA, 
                           footnote = NA, 
                           filename = NA,
                           round = 3,
                           ...) {
  
  if (is.na(caption)) {
    caption <- paste0(
      "Improvement rate difference for variable '", 
      attr(object, opt("dv")),  "'"
    )
  }
  
  if (is.na(footnote)) {
    if (object$decreasing) {
      footnote <- "Assumed decreasing values in Phase B"
    }
  }
  
  
  out <- data.frame("IRD" = round(object$ird, round))
  table <- .create_table(
    out,
    caption = caption,
    footnote = footnote,
    ...
  )
  
  if (!is.na(filename)) .save_export(table, filename)
  
  table

}  

#' Handling outliers in single-case data
#'
#' Identifies and drops outliers within a single-case data frame (scdf).
#' Outliers can be identified based on mean average deviation (MAD),
#' standard deviation (SD), confidence intervals (CI), or Cook's Distance
#' from a Piecewise Linear Regression Model.
#'
#' @inheritParams .inheritParams
#' @param method Specifies the method for outlier identification. Set `method =
#'   "MAD"` for mean average deiviation, `method = "SD"` for standard
#'   deviations, `method = "CI"` for confidence intervals, `method = "Cook"` for
#'   Cook's Distance based on the Piecewise Linear Regression Model.
#' @param criteria Specifies the criteria for outlier identification. Based on
#'   the `method` setting.
#' @return 
#'  |  |  |
#'  | --- | --- |
#'  | `data` | A single-case data frame with substituted outliers. |
#'  | `dropped.n` | A list with the number of dropped data points for each single-case. |
#'  | `dropped.mt` | A list with the measurement-times of dropped data points for each single-case (values are based on the mt variable of each single-case data frame). |
#'  | `sd.matrix` | A list with a matrix for each case with values for the upper and lower boundaries based on the standard deviation. |
#'  | `ci.matrix` | A list with a matrix for each single-case with values for the upper and lower boundaries based on the confidence interval. |
#'  | `cook` | A list of Cook's Distances for each measurement of each single-case. |
#'  | `criteria` | Criteria used for outlier analysis. |
#'  | `N` | Number of single-cases. |
#'  | `case.names` | Case identifier. |
#' @details For `method = "SD"`, `criteria = 2` would refer t0 two standard
#'   deviations. For `method = "MAD"`, `criteria = 3.5` would refer to 3.5 times
#'   the mean average deviation. For `method = "CI"`, `criteria = 0.99` would
#'   refer to a 99 percent confidence interval. For `method = "cook"`, `criteria
#'   = "4/n"` would refer to a Cook's Distance greater than 4/n.
#' @author Juergen Wilbert
#' @family data manipulation functions
#' @keywords manip
#' @examples
#'
#' ## Identify outliers using 1.5 standard deviations as criterion
#' susanne <- random_scdf(level = 1.0)
#' res_outlier <- outlier(susanne, method = "SD", criteria = 1.5)
#' res_outlier
#'
#' ## Identify outliers in the original data from Grosche (2011)
#' ## using Cook's Distance greater than 4/n as criterion
#' res_outlier <- outlier(Grosche2011, method = "Cook", criteria = "4/n")
#' res_outlier
#'
#' @export
outlier <- function(data, dvar, pvar, mvar, 
                    method = c("MAD", "Cook", "SD", "CI"),
                    criteria = 3.5) {
  
  if (length(criteria) == 2) {
    method <- criteria[1]
    criteria <- criteria[2]
  }
  
  check_args(
    one_of(method, c("MAD", "Cook", "SD", "CI"))
  )

  method <- method[1]
  
  # set defaults attributes
  if (missing(dvar)) dvar <- dv(data) 
  if (missing(pvar)) pvar <- phase(data) 
  if (missing(mvar)) mvar <- mt(data) 
  
  dv(data) <- dvar
  phase(data) <- pvar
  mt(data) <- mvar
  
  data_list <- .prepare_scdf(data)
 
  out <- list()
  
  N <- length(data_list)
  case.names <- names(data_list)
  dropped.mts <- list()
  dropped.n <- list()
  ci.matrix <- list()
  sd.matrix <- list()
  mad.matrix <- list()
  cook <- list()
  
  for(i in 1:N) {
    data <- data_list[[i]]
    
    phases <- rle(as.character(data[, pvar]))$value
    values <- lapply(phases, function(x) data[data[, pvar] == x, dvar])

# CI ----------------------------------------------------------------------
    
    if (identical(method, "CI")) {
      cut_off <- as.numeric(criteria)
      mat <- matrix(NA, length(values), ncol = 5)
      colnames(mat) <- c("phase","m","se","lower", "upper")
      rownames(mat) <- names(values)
      filter <- c()
      fac <- qnorm((1 - cut_off) / 2, lower.tail = FALSE)
      
      for(p in 1:length(values)) {
        x <- values[[p]]
        mat[p,"m"] <- mean(x)
        mat[p,"se"] <- sd(x)/sqrt(length(x))
        mat[p,"lower"] <- mean(x) - fac * (sd(x)/sqrt(length(x)))
        mat[p,"upper"] <- mean(x) + fac * (sd(x)/sqrt(length(x)))
        filter <- c(filter, (x < mat[p, "lower"]) | (x > mat[p, "upper"]))
      }
      mat <- as.data.frame(mat)
      mat$phase <- phases
      ci.matrix[[i]] <- mat
    }

# MAD ---------------------------------------------------------------------

    if (identical(method, "MAD")) {
      fac <- as.numeric(criteria)
      mat <- matrix(NA, length(values), ncol = 5)
      colnames(mat) <- c("phase", "md", "mad", "lower", "upper")
      filter <- c()
      for(p in 1:length(values)) {
        x <- values[[p]]
        mat[p,"md"] <- median(x)
        mat[p,"mad"] <- mad(x,constant = 1)
        mat[p,"lower"] <- median(x) - fac * mad(x)
        mat[p,"upper"] <- median(x) + fac * mad(x)
        filter <- c(filter, (x < mat[p,"lower"]) | (x > mat[p,"upper"]))
      }
      mat <- as.data.frame(mat)
      mat$phase <- phases
      mad.matrix[[i]] <- mat
    }	

# SD ----------------------------------------------------------------------

    if (identical(method, "SD")) {
      SD <- as.numeric(criteria)
      mat <- matrix(NA, length(values), ncol = 5)
      colnames(mat) <- c("phase", "m", "sd", "lower", "upper")
      filter <- c()
      for(p in 1:length(values)) {
        x <- values[[p]]
        mat[p,"m"] <- mean(x)
        mat[p,"sd"] <- sd(x)
        mat[p,"lower"] <- mean(x) - SD * sd(x)
        mat[p,"upper"] <- mean(x) + SD * sd(x)
        filter <- c(filter, (x < mat[p,"lower"]) | (x > mat[p, "upper"]))
      }
      mat <- as.data.frame(mat)
      mat$phase <- phases
      sd.matrix[[i]] <- mat
    }		

# Cook --------------------------------------------------------------------

    if (identical(method, "Cook")) {
      
      if (criteria == "4/n") 
        cut_off <- 4/nrow(data)
      else 
        cut_off <- as.numeric(criteria)

      reg <- plm(data_list[i], dvar = dvar, pvar = pvar, mvar = mvar)$full.model
      
      cd <- cooks.distance(reg)
      filter <- cd >= cut_off
      cook[[i]] <- data.frame(Cook = round(cd, 2), MT = data_list[[i]][, mvar])
    }		
    
    dropped.mts[[i]] <- data_list[[i]][filter, mvar]
    dropped.n[[i]]   <- sum(filter)
    
    data_list[[i]] <- data_list[[i]][!filter, ]
  }
  
  out$data <- data_list
  out$dropped.mt <- dropped.mts
  out$dropped.n <- dropped.n
  out$ci.matrix <- ci.matrix
  out$sd.matrix <- sd.matrix
  out$mad.matrix <- mad.matrix
  out$cook <- cook
  out$criteria <- criteria
  out$N <- N
  out$case.names <- case.names
  
  class(out) <- c("sc_outlier")
  attributes(out)[opts("phase", "mt", "dv")] <- list(pvar, mvar, dvar)
  out
}

#' Subset cases, rows, and variables of an scdf
#'
#' This function is mainly used to filter rows by a logical expression. It has
#' also arguments to filter variables and cases.
#' 
#' For subsetting rows, a logical expression can be provided in the `subset`
#' argument. Missing values are treated as FALSE. The expression is evaluated
#' within each single-case data frame of the scdf.
#'
#' @param x An scdf object.
#' @param ... not implemented
#' @param subset Logical expression indicating rows to keep: missing values are
#'   taken as false.
#' @param select Expression, indicating columns to select from an scdf.
#' @param cases Expression, indicating cases to keep from an scdf.
#'
#' @return An scdf.
#' @export
#'
#' @examples
#' exampleAB |>
#'   subset((values < 60 & phase == "A") | (values >= 60 & phase == "B"))
#' subset(exampleAB_add, select = c(-cigarrets, -depression))
#' subset(exampleAB, cases = c(Karolina, Johanna))
#' subset(exampleA1B1A2B2, phase %in% c("A1", "B2"), cases = Pawel:Moritz)
subset.scdf <- function(x, subset, select, cases, ...) {
  scdf <- x
  scdf_attributes <- attributes(scdf)
  if (missing(subset)) subset <- TRUE else subset <- substitute(subset)
  if (missing(select)) select <- TRUE
  if (missing(cases)) cases <- TRUE
  
  #if (is.numeric(eval(subset))) subset <- eval(subset)
   
  nl <- as.list(seq_along(scdf))
  names(nl) <- names(scdf)
  scdf <- scdf[eval(substitute(cases), envir = nl, enclos = parent.frame())]
  
   
  for(i in 1:length(scdf)) {
    x <- scdf[[i]]
    # select vars
    nl <- as.list(seq_along(x))
    names(nl) <- names(x)
    columns <- eval(substitute(select), nl, parent.frame())
    # subset rows
    if (isTRUE(subset)) rows <- TRUE
    if (is.call(subset)) {
      rows <- eval(subset, x, parent.frame())
      if (is.logical(rows))
        rows <- rows & !is.na(rows)
    }
    #print(rows)
    scdf[[i]] <- x[rows, columns, drop = FALSE]
  }
  
  scdf
}

#' @rdname print.sc
#' @export
#' 
print.sc_rci <- function(x, digits = 3, ...) {
  
  cat("Reliable Change Index\n\n")
  cat("Mean Difference = ", round(x$descriptives[2, 2] - x$descriptives[1, 2], digits), "\n")
  cat("Standardized Difference = ", round(x$stand_dif, digits), "\n")
  cat("Standard error of differences = ", round(x$se_dif, digits), "\n")
  cat("Reliability of measurements = ", round(x$reliability, digits), "\n")
  #cat("Reliability of difference scores = ", round(x$rdd, digits), "\n")
  cat("\n")
  cat("Descriptives:\n")
  print(round(x$descriptives, digits))
  cat("\n")
  cat(round(x$conf_percent * 100, digits), "% Confidence Intervals:\n")
  print(round(x$conf, digits))
  cat("\n")
  cat("Reliable Change Indices:\n")
  print(round(x$rci, digits))
  cat("\n")
}

#' Export Descriptive Statistics Table
#' 
#' @rdname export
#' @export
export.sc_desc <- function(object, 
                           caption = NA, 
                           footnote = NA, 
                           filename = NA,
                           flip = FALSE, 
                           round = 2,
                           ...) {
  # prepare table -----------------------------------------------------------
  if (is.na(caption)) caption <- "Descriptive statistics"

  if (is.na(footnote)) {
    footnote <- c(
      "n = Number of measurements",
      "Missing = Number of missing values",
      "M = Mean",
      "Median = Median",
      "SD = Standard deviation",
      "MAD = Median average deviation",
      "Min = Minimum",
      "Max = Maximum",
      "Trend = Slope of dependent variable regressed on measurement-time"
    )
  }
  
  object$descriptives[-1:-2] <- round(
    object$descriptives[-1:-2], round
  )
  
  if (flip) {
    out <- as.data.frame(t(object$descriptives[-1]))
    colnames(out) <- object$descriptives$Case
    rownames(out) <- gsub("mis", "Missing", rownames(out))
    rownames(out) <- gsub("med", "Median", rownames(out))
    rownames(out) <- gsub("min", "Min", rownames(out))
    rownames(out) <- gsub("max", "Max", rownames(out))
    rownames(out) <- gsub("trend", "Trend", rownames(out))
    rownames(out) <- gsub("\\.", " ", rownames(out))
    out <- cbind(Parameter = rownames(out), out)

    table <- .create_table(
      out, 
      caption = caption,
      footnote = footnote,
      ...
    )
  }
  
  if (!flip) {
    n_phases <- length(object$design)
    out <- object$descriptives
    colnames(out) <- c("Case", "Design", rep(object$design, 9))
    spannerpos <- 3 + (0:9 * n_phases)
    
    table <- .create_table(
      out, 
      caption = caption,
      footnote = footnote,
      spanner = list(
        "n" = spannerpos[1]:(spannerpos[1] + n_phases - 1),
        "Missing" = spannerpos[2]:(spannerpos[2] + n_phases - 1),
        "M" = spannerpos[3]:(spannerpos[3] + n_phases - 1),
        "Median" = spannerpos[4]:(spannerpos[4] + n_phases - 1),
        "SD" = spannerpos[5]:(spannerpos[5] + n_phases - 1),
        "MAD" = spannerpos[6]:(spannerpos[6] + n_phases - 1),
        "Min" = spannerpos[7]:(spannerpos[7] + n_phases - 1),
        "Max" = spannerpos[8]:(spannerpos[8] + n_phases - 1),
        "Trend" = spannerpos[9]:(spannerpos[9] + n_phases - 1)
      ),
      ...
    )
    
    if (getOption("scan.export.engine") == "kable") {
      table <- add_header_above(
        table,
        c(
          " " = 2, "n" = n_phases,
          "Missing" = n_phases,
          "M" = n_phases,
          "Median" = n_phases,
          "SD" = n_phases,
          "MAD" = n_phases,
          "Min" = n_phases,
          "Max" = n_phases,
          "Trend" = n_phases
        )
      )
    }
  }
  
  # finish ------------------------------------------------------------------
  
  if (!is.na(filename)) .save_export(table, filename)
  
  table
}

#' Summary function for an scdf object
#' 
#' Provides a summary of an `scdf` object, including the number of cases,
#' measurements per case, and design information.
#' 
#' @details The summary includes:
#' - Total number of cases in the `scdf`.
#' - A table listing each case with the number of measurements and design.
#' - Variable names with annotations for phase, measurement-time, and dependent variable.
#' - Additional information and author details if available.
#' @author Juergen Wilbert
#' @param object scdf
#' @param all_cases IF TRUE, more that 10 cases are summarized
#' @param ... not in use
#' @export
summary.scdf <- function(object, all_cases = FALSE, ...) {

  N <- length(object)
  
  if (!all_cases) max_cases <- 10 else max_cases <- N
  if (max_cases > N) max_cases <- N
  
  if(N > 1) {
    cat("#A single-case data frame with", number_word(N), "cases\n\n")
  } else {
    cat("#A single-case data frame with one case\n\n")
  }
  
  designs <- lapply(
    object, function(x) 
    paste0(
      rle(as.character(x[[phase(object)]]))$values, 
      collapse = "-"
    )
  )
  
  rows <- lapply(object, nrow)

  out <- data.frame(
    " " = format(revise_names(object), justify = "left")[1:max_cases],
    Measurements = unname(unlist(rows))[1:max_cases], 
    Design = unname(unlist(designs))[1:max_cases],
    check.names = FALSE
  )

  print(out, row.names = FALSE)
  
  if (N > max_cases) cat("... [skipped", N - max_cases, "cases]\n")
  
  cat("\n", sep = "")
  
  cat("Variable names:\n")
  names <- names(object[[1]])
  id_dv <- which(names == scdf_attr(object, opt("dv")))
  id_phase <- which(names == phase(object))
  id_mt <- which(names == scdf_attr(object, opt("mt")))
  names[id_phase] <- paste(names[id_phase], "<phase variable>")
  names[id_mt] <- paste(names[id_mt], "<measurement-time variable>")
  names[id_dv] <- paste(names[id_dv], "<dependent variable>")
  cat(names[c(
    id_dv, id_phase, id_mt, (1:length(names))[-c(id_dv, id_phase, id_mt)]
  )], sep = "\n")
  cat("\n")
  
  
  if(!is.null(scdf_attr(object, "info"))) {
    cat("Note:", scdf_attr(object, "info"), "\n")
  }
  
  if(!is.null(scdf_attr(object,"author"))) {
    cat("\nAuthor of data:", scdf_attr(object, "author"), "\n")
  }
  
  class(object) <- "scdf_summary"
  invisible(object)
  
}

#' @rdname export
#' @export
export.scdf <- function(object, 
                        summary = FALSE,
                        caption = NA, 
                        footnote = NA, 
                        filename = NA,
                        cols, 
                        round = 2, 
                        ...) {
  
  if (summary) {
    sink(nullfile())
    out <- summary(object)
    sink()
    return(export(out))
  }
  
  if (is.na(footnote)) {
    if (!is.null(scdf_attr(object, "info"))) 
      footnote <- scdf_attr(object, "info")
    if (!is.null(scdf_attr(object, "author"))) {
      footnote <- paste(footnote, "\nAuthor:", scdf_attr(object, "author"))
    }
  }
  
  N <- cases <- length(object)
  
  if (is.na(caption)) {
    if (N > 1) {
      caption <- paste("Single case data frame with", number_word(N), "cases")
    } else {
      caption <- paste("Single case data frame for case", names(object))
    }
    
  }
      
  if (missing(cols)) {
    cols <- names(object[[1]])
  }
  if (identical(cols, "main")) {
    cols <- c(phase(object), dv(object), mt(object))
  }
  
  names(object) <- revise_names(object)
  
  max_row <- max(unlist(lapply(object, nrow)))
  for (i in 1:cases) {
    n_row <- nrow(object[[i]])
    object[[i]][, phase(object)] <- as.character(object[[i]][, phase(object)])
    if (n_row < max_row) {
      object[[i]][(n_row + 1):max_row, names(object[[i]])] <- ""
    }
  }
  rows <- max_row
  out <- lapply(object[1:cases], function(x) x[1:rows, cols])
  names <- lapply(out, names)
  out <- as.data.frame(out)
  names(out) <- unlist(names[1:cases])
  
  n_vars <- ncol(out) / N
  
  if (N > 1) {
    spanner <- setNames(vector("list", N), names(object))
    for(i in 1:N) {
      spanner[[i]] <- ((i - 1) * n_vars + 1) : ((i - 1) * n_vars + n_vars)
    }
  } else {
    spanner <- NULL
  } 
  
  opts <- options()
  options(scan.export.kable = c(
    list(align = rep("c", ncol(out))), 
    getOption("scan.export.kable")
  ))
  
  
  table <- .create_table(
    out, 
    caption = caption,
    footnote = footnote,
    spanner = spanner,
    ...
  )
  
  if (getOption("scan.export.engine") == "kable") {
    if (N > 1) { 
      case_names <- rep(n_vars, N)
      names(case_names) <- names(object)
      table <- add_header_above(table, case_names)
    }
  }
  
  # finish ------------------------------------------------------------------
  
  if (!is.na(filename)) .save_export(table, filename)
  
  options(opts)
  
  table
}

#' @rdname print.sc
#' @export
#' 
print.sc_pem <- function(x, ...) {
  cat("Percent Exceeding the Median\n\n")
  print(x$PEM, digits = 3, row.names = FALSE)
  cat("\n")
  if (x$decreasing) {
    cat("Assumed decreasing values in the B-phase.\n\n")
    cat("Alternative hypothesis: true probability < 50%\n")
  } else {
    cat("Alternative hypothesis: true probability > 50%\n")
  }
}

mc_function <- function(name) {
  
  if(missing(name)) name <- ""
  id <- which(names(opt("mc_fun")) == name)
  
  if (length(id) == 1) {
    return(opt("mc_fun")[id])
  }
  
  abort("Unknown function. Valid values are: ",
       paste(names(opt("mc_fun")), collapse = ", "))
  
}

#' ANOVA Table for Piecewise Linear Models
#' 
#' Model comparison for piecewise regression models fitted with `plm()`,
#' `hplm()`, or `mplm()` using likelihood ratio tests.
#' 
#' @aliases anova.sc_plm anova.sc_hplm anova.sc_mplm
#' @param object An object containing the results returned by `plm()`,
#' `hplm()`, or `mplm()`.
#' @param ... additional objects for model comparison.
#' @return An object of class `anova` containing the results of the model
#'  comparison.
#' @details
#' The function performs likelihood ratio tests to compare nested piecewise
#' regression models. It extracts the underlying model from the sc_plm, sc_hplm,
#' or sc_mplm object and passes them to the generic anova() function
#' for model comparison.
#'  
#' @examples
#' ## For glm models with family = "gaussian"
#' mod1 <- plm(exampleAB$Johanna, level = FALSE, slope = FALSE)
#' mod2 <- plm(exampleAB$Johanna)
#' anova(mod1, mod2)
#' ## For glm models with family = "poisson"
#' mod0 <- plm(example_A24, formula = injuries ~ 1, family = "poisson")
#' mod1 <- plm(example_A24, trend = FALSE, family = "poisson")
#' anova(mod0, mod1, mod2)
#' ## For glm with family = "binomial"
#' mod0 <- plm(
#'   exampleAB_score$Christiano, 
#'   formula = values ~ 1, 
#'   family = "binomial", 
#'   var_trials = "trials"
#' )
#' mod1 <- plm(
#'   exampleAB_score$Christiano, 
#'   trend = FALSE, 
#'   family = "binomial", 
#'   var_trials = "trials"
#' )
#' anova(mod0, mod1)
#' @export
anova.sc_plm <- function(object, ...) {
  models <- list(...)
  models <- lapply(models, function(x) {
    if (inherits(x, "sc_plm")) x$full.model else x
  })
  models <- c(list(object$full.model), models)
  
  do.call(anova, models)
}

#' @rdname anova.sc_plm
#' @examples
#' ## For multilevel models:
#' mod0 <- hplm(Leidig2018, trend = FALSE, slope = FALSE, level = FALSE)
#' mod1 <- hplm(Leidig2018, trend = FALSE)
#' mod2 <- hplm(Leidig2018)
#' anova(mod0, mod1, mod2)
#' @export
anova.sc_hplm <- function(object, ...) {
  models <- list(...)
  models <- lapply(models, function(x) {
    if (inherits(x, "sc_hplm")) x$hplm else x
  })

  models <- c(list(object$hplm), models)

  id <- lapply(models, function(x) inherits(x, "lme")) |> unlist()
  
  str_models <- paste0("models[[", which(id), "]]", collapse = ", ")
  
  id <- which(!id)
  
  str_args <- if (length(id > 0)) {
    paste0(names(id), " = ", models[[id]], collapse = ", ")
  } else {
    NULL
  }
  
  str <- paste0(c(str_models, str_args), collapse = ", ")
  out <- paste0("anova(", str, ")")
  out <- eval(str2lang(out))
  row.names(out) <- NULL
  out
}

#' @rdname anova.sc_plm
#' @examples
#' ## For mplm
#' mod0 <- mplm(
#'   Leidig2018$`1a1`, 
#'   update = . ~  1, dvar = c("academic_engagement", "disruptive_behavior")
#' )
#' mod1 <- mplm(
#'   Leidig2018$`1a1`, 
#'   trend = FALSE, 
#'   dvar = c("academic_engagement", "disruptive_behavior")
#' )
#' mod2 <- mplm(
#'   Leidig2018$`1a1`, 
#'   dvar = c("academic_engagement", "disruptive_behavior")
#' )
#' 
#' anova(mod0, mod1, mod2)
#' @export
anova.sc_mplm <- function(object, ...) {
  models <- list(...)
  models <- lapply(models, function(x) {
    if (inherits(x, "sc_mplm")) x$full.model else x
  })
  models <- c(list(object$full.model), models)
  
  do.call(anova, models)
  
}

## version 0.1.0-2

#' Throws a message
#'
#' Messages can be shown with different levels
#' of detail and as warnings. This is a convenience wrapper for internal use.
#'
#' @param ... The message. This can be a character string or a
#'   combination of character strings and variables.
#' @param detail The level of detail for the message. Messages with a detail
#'   level higher than the current detail level set in the options will not be
#'   shown. The default is 1.
#' @param warning If TRUE, the message will be shown as a warning. If FALSE
#'   (default), the message will be shown as an informational message.
#' @keywords internal

notify <- function(... , type = "!", detail = 1, warning = FALSE) {
  
  
  if (detail > getOption("wmisc.msg.detail", default = 1))
    return(invisible(NULL))
  
  msg <- paste0(...)
  msg <- gsub("\\n", "\n  ", msg)
  if (type == "h1") {
    msg <- paste0("\u256D\u2500 ", msg)
    msg <- cli::col_blue(msg)
    type <- ""
  }
  
  max_char <- getOption("wmisc.msg.max", default = 100)
  if (nchar(msg) > max_char) {
    msg <- paste0(substring(msg, 1, max_char), "... [truncated]")
  }
  msg <- setNames(msg, type)
  if (!warning) {
    rlang::inform(msg)
  } else {
    rlang::warn(msg)
  }
  invisible(NULL)
}

#' Throws a warning
#' This is a convenience wrapper for internal use.
#' @param ... The message. This can be a character string or a 
#'  combination of character strings and variables.
#' @keywords internal
warn <- function(...) {
  notify(..., warning = TRUE)
}

#' Throws an error
#'
#' This is a convenience wrapper for internal use.
#'
#' @param ... The message. This can be a character string or a
#'   combination of character strings and variables.
#' @param class An optional class or vector of classes to add to the error.
#' @param call If TRUE, the call will be included in the error message. Default
#'   is FALSE.
#' @keywords internal
abort <- function(..., class = NULL, call = FALSE) {
  
  call <- if (isTRUE(call)) {
    rlang::caller_env()
  } else if (identical(call, FALSE) || is.null(call)) {
    NULL
  } else {
    call
  }
  
  msg <- paste0(...)
  rlang::abort(message = msg, class = class, call = call)
  
  invisible(NULL)
}

#' Empirical power analysis for single-case data
#'
#' Conducts a Monte-Carlo study on the test-power and alpha-error probability of
#' a statistical function.
#'
#' Based on a [design()] object, a large number of single-cases are generated
#' and re-analysed with a provided statistical function. The proportion of
#' significant analyses is the test power. In a second step, a specified effect
#' of the design object is set to 0 and again single-cases are generated and
#' re-analysed. The proportion of significant analyses is the alpha error
#' probability.
#'
#' @param design An object returned from the `design` function.
#' @param method A (named) list that defines the methods the power analysis is
#'   based on. Each element can contain a function (that takes an scdf file and
#'   returns a p value) or a character string (the name of predefined
#'   functions). default `method = list("plm_level", "rand", "tauU")` computes a
#'   power analysis based on [tau_u()], [rand_test()] and [plm()] analyses.
#'   (Further predefined functions are: "plm_slope", "plm_poisson_level",
#'   "plm_poisson_slope", "hplm_level", "hplm_slope", "base_tau".
#' @param effect Either "level" or "slope". The respective effect of the
#'   provided design is set to 0 when computing the alpha-error proportion.
#' @param n_sim Number of sample studies created for the the Monte-Carlo study.
#'   Default is `n = 100`. Ignored if design_is_one_study = FALSE.
#' @param design_is_one_study If TRUE, the design is assumed to define all cases
#'   of one study that is repeatedly randomly created `n_sim` times. If FALSE,
#'   the design is assumed to contain all cases from which a random sample is
#'   generated. This is useful for very specific complex simulation studies.
#'   In this case, n_sim indicates how many random samples are drawn from the
#'   provided design. Default is TRUE.
#' @param alpha_test Logical. If TRUE, alpha error is calculated.
#' @param power_test Logical. If TRUE, power is calculated.
#' @param binom_test Shortcut. When set TRUE, binom_test_power is set to 0.80,
#'   binom_test_alpha is set to 0.05, and binom_test_correct is set to 0.875.
#' @param binom_test_power Either FALSE or a value. If a value is provided, a
#'   binomial test is calculated testing if the power is greater than the
#'   provided value.
#' @param binom_test_alpha Either FALSE or a value. If a value is provided, a
#'   binomial test is calculated testing if the alpha error proportion is less
#'   than the provided value.
#' @param binom_test_correct Either FALSE or a value. If a value is provided, a
#'   binomial test is calculated testing if the correct proportion is greater
#'   than the provided value.
#' @param ci Either FALSE or a value. If a value is provided, confidence
#'   intervals at the provided level are calculated for power, alpha error, and
#'   correct proportions.
#' @param alpha_level Alpha level used to calculate the proportion of
#'   significant tests. Default is `alpha_level = 0.05`.
#' @author Juergen Wilbert
#' @return A data frame with the power, alpha error, and correct proportions for
#'  each provided method. If binomial tests are requested, p values for these
#'  tests are also provided. If confidence intervals are requested, these
#'  are also provided.
#' @seealso [random_scdf()], [design()]
#' @examples
#'
#' ## Assume you want to conduct a single-case study with 15 measurements
#' ## (phases: A = 6 and B = 9) using a highly reliable test and
#' ## an expected level effect of d = 1.4.
#' ## A (strong) trend effect is trend = 0.05. What is the power?
#' ## (Note: n_sims is set to 10. Set n_sims to 1000 for a serious calculation.)
#' design <- design(
#'   n = 1, phase_design = list(A = 6, B = 9),
#'   rtt = 0.8, level = 1.4, trend = 0.05
#' )
#' power_test(design, n_sim = 10)
#'
#' ## Would you achieve higher power by setting up a MBD with three cases?
#' design <- design(
#'   n = 3, phase_design = list(A = 6, B = 9),
#'   rtt = 0.8, level = 1.4, trend = 0.05
#' )
#' power_test(design, n_sim=10, method=list("hplm_level", "rand", "tauU_meta"))
#' @export

power_test <- function(design,
                       method = c("plm_level", "rand", "tauU"), 
                       effect = "level",
                       n_sim = 100,
                       design_is_one_study = TRUE,
                       alpha_test = TRUE,
                       power_test = TRUE,
                       binom_test = FALSE,
                       binom_test_alpha = FALSE,
                       binom_test_power = FALSE,
                       binom_test_correct = FALSE,
                       ci = FALSE,
                       alpha_level = 0.05) {
  
  
  starttime <- proc.time()
  
  mc_fun <- unlist(
    lapply(
      method, 
      function(x) if (inherits(x, "character")) mc_function(x) else x),
      recursive = FALSE
  )
  
  # return object
  out <- data.frame(Method = names(mc_fun))
  
  # power calculation ----------
  
  if (power_test) {
    out$Power <- .mc_scdf(
      design = design,
      n_sim = n_sim,
      alpha_level = alpha_level,
      mc_fun = mc_fun,
      design_is_one_study = design_is_one_study
    )
  } else
    out$Power <- NA
  
  # alpha error calculation ----------
  
  if (alpha_test) {
    
    #level <- any(sapply(design$cases, function(x) x$level) != 0)
    #slope <- any(sapply(design$cases, function(x) x$slope) != 0)
    
    design_no_effect <- design
    if (effect == "level") {
      design_no_effect$cases <- lapply(design_no_effect$cases,  function(x) {
        x$level <- rep(0, length = length(x$length))
        x
      })
    }
    
    if (effect == "slope") {
      design_no_effect$cases <- lapply(design_no_effect$cases, function(x) {
        x$slope <- rep(0, length = length(x$length))
        x
      })
    }
    
    out$"Alpha Error" <- .mc_scdf(
      design = design_no_effect,
      n_sim = n_sim,
      alpha_level = alpha_level,
      mc_fun = mc_fun,
      design_is_one_study = design_is_one_study
    )
  } else {
    out$"Alpha Error" <- NA
  }
  
  out$"Alpha:Beta" <- sprintf("1:%.1f", (100 - out$Power) / out$"Alpha Error")
  out$Correct <- (out$Power + (100 - out$"Alpha Error")) / 2
  
  if (binom_test) {
    binom_test_alpha <- 0.05
    binom_test_power <- 0.80
    binom_test_correct <- 0.875
  }
  
  if (is.numeric(binom_test_power)) {
    b_test <- function(x) {
      x <- binom.test(round(x / 100 * n_sim), 
        n_sim, 
        p = binom_test_power, 
        alternative = "greater"
      )
      round(x$p.value)
    }
    out[["p_power"]] <- sapply(out$"Power", b_test)
  }
  
  if (is.numeric(binom_test_alpha)) {
    b_test <- function(x) {
      x <- binom.test(round(x / 100 * n_sim), 
        n_sim, 
        p = binom_test_alpha, 
        alternative = "less"
      )
      round(x$p.value)
    }
    out[["p_alpha"]] <- sapply(out$"Alpha Error", b_test)
  }
  
  if (is.numeric(binom_test_correct)) {
    b_test <- function(x) {
      x <- binom.test(round(x / 100 * n_sim * 2), 
        n_sim * 2, 
        p = binom_test_correct, 
        alternative = "greater"
      )
      round(x$p.value, 3)
    }
    out[["p_correct"]] <- sapply(out$Correct, b_test)
  }

  if (is.numeric(ci)) {
    b_test <- function(x) {
      x <- binom.test(round(x / 100 * n_sim), 
        n_sim, 
        p = binom_test, 
        conf.level = ci, 
        alternative = "two.sided"
      )
      x$conf.int * 100
    }
    
    out[["Power lower"]] <- sapply(out$Power, function(x) b_test(x)[1])
    out[["Power upper"]] <- sapply(out$Power, function(x) b_test(x)[2])
    out[["Alpha Error lower"]] <- sapply(out$"Alpha Error", 
                                         function(x) b_test(x)[1])
    out[["Alpha Error upper"]] <- sapply(out$"Alpha Error", 
                                         function(x) b_test(x)[2])
      
    b_test <- function(x) {
      x <- binom.test(round(x / 100 * n_sim * 2), 
        n_sim * 2, 
        p = binom_test, 
        alternative = "two.sided"
      )
      x$conf.int * 100
    }
    out[["Correct lower"]] <- sapply(out$Correct, function(x) b_test(x)[1])
    out[["Correct upper"]] <- sapply(out$Correct, function(x) b_test(x)[2])
  }
  
  attr(out, "computation_duration") <- proc.time() - starttime
  attr(out, "binom_test_alpha") <- binom_test_alpha
  attr(out, "binom_test_power") <- binom_test_power
  attr(out, "binom_test_correct") <- binom_test_correct
  attr(out, "ci") <- ci
  class(out) <- c("sc_power")
  out
}

.mc_scdf <- function(design, 
                     alpha_level = NA, 
                     n_sim, 
                     mc_fun, 
                     design_is_one_study) {

  # Genrate random sample ----------------------------------------------------
  rand_sample <- list()
  
  if (design_is_one_study) {
    for(i in 1:n_sim) rand_sample[[i]] <- random_scdf(design = design)
  }
  
  if (!design_is_one_study) {
    tmp <- random_scdf(design = design)
    for (i in seq_along(tmp)) rand_sample[[i]] <- tmp[i]
  }
  
  # analyse random sample ---------------------------------------------------
  
  test_function <- function(func) {
    p <- sapply(rand_sample, func)
    mean(p <= alpha_level, na.rm = TRUE) * 100
  }

  out <-  sapply(mc_fun, test_function) 

  # return
  out
}

#' Baseline corrected tau
#'
#' Kendall's tau correlation for the dependent variable and the phase variable
#' is calculated after correcting for a baseline trend.
#'
#' @inheritParams .inheritParams
#' @param alpha Sets the p-value at and below which a baseline correction is
#'   applied.
#' @param continuity If TRUE applies a continuity correction for calculating p
#' @param tau_method Character with values "a" or "b" (default) indicating
#'   whether Kendall Tau A or Kendall Tau B is applied.
#' @details This method has been proposed by Tarlow (2016). The baseline data
#'   are checked for a significant autocorrelation (based on Kendall's Tau). If
#'   so, a non-parametric Theil-Sen regression is applied for the baseline data
#'   where the dependent values are regressed on the measurement time. The
#'   resulting slope information is then used to predict data of the B-phase.
#'   The dependent variable is now corrected for this baseline trend and the
#'   residuals of the Theil-Sen regression are taken for further calculations.
#'   Finally, Kendall's tau is calculated for the dependent variable and the
#'   dichotomous phase variable. The function here provides two extensions to
#'   this procedure: The more accurate continuity correction is applied when
#'   `continuity = TRUE`.
#'
#' @family regression functions
#' @references Tarlow, K. R. (2016). An Improved Rank Correlation Effect Size
#'   Statistic for Single-Case Designs: Baseline Corrected Tau. *Behavior
#'   Modification, 41(4)*, 427–467. https://doi.org/10.1177/0145445516676750
#' @export
#'
#' @examples
#' dat <- scdf(c(A = 33,25,17,25,14,13,15, B = 15,16,16,5,7,9,6,5,3,3,8,11,7))
#' corrected_tau(dat)
corrected_tau <- function(data, dvar, pvar, mvar, 
                          phases = c(1, 2), 
                          alpha = 0.05, 
                          continuity = FALSE, 
                          tau_method = c("b", "a")) {
  
  # validity check ----
  check_args(
    by_call(tau_method),
    within(alpha, 0, 1),
    is_logical(continuity)
    #is_logical(repeated)
  )
  
  # prepare scdf ----
  if (missing(dvar)) dvar <- dv(data) else dv(data) <- dvar
  if (missing(pvar)) pvar <- phase(data) else phase(data) <- pvar
  if (missing(mvar)) mvar <- mt(data) else mt(data) <- mvar

  data <- .prepare_scdf(data, na.rm = TRUE)
  data <- recombine_phases(data, phases = phases)$data

  corr_tau <- function(data) {
    
    rowsA <- which(data[[pvar]] == "A")
    rowsB <- which(data[[pvar]] == "B")
    A_data <- data[rowsA, ]
    B_data <- data[rowsB, ]
    
    if (length(unique(A_data[[dvar]])) == 1) {
      warn(
        "All phase A values are identical. ",
        "Autocorrelation can not be calculated and is set to NA."
      )
      auto_tau <- list(tau = NA, z = NA, p = NA)
    } else if (length(A_data[[dvar]]) < 3 ) {
      warn(
        "Need at least three data points in phase A. ",
        "Autocorrelation can not be calculated and is set to NA."
      )
      auto_tau <- list(tau = NA, z = NA, p = NA)
    } else {
      auto_tau <- kendall_tau(
        A_data[[dvar]], 
        A_data[[mvar]], 
        continuity_correction = continuity,
        tau_method = tau_method
      )
    }
    
    formula  <- as.formula(paste0(dvar, "~", mvar))
    
    fit_ts <- theil_sen(formula, data = A_data)
    data$fit <- fit_ts$intercept + data[[mvar]] * fit_ts$slope
    
    y <- as.numeric(factor(data[[pvar]]))
    
    base_corr_tau <- kendall_tau(
      x = data[[dvar]] - data$fit, 
      y = y, 
      continuity_correction = continuity, 
      tau_method = tau_method
    )
    
    x <- data[[dvar]]
    
    uncorrected_tau <- kendall_tau(
      x = data[[dvar]], 
      y = y, 
      continuity_correction = continuity, 
      tau_method = tau_method
    )
    
    if (is.na(auto_tau$p)) {
      corr_applied <- FALSE
    } else {
      if (auto_tau$p <= alpha) {
        corr_applied <- TRUE
      } else {
        corr_applied <- FALSE
      }
    }
    
    if (corr_applied) tau <- base_corr_tau else tau <- uncorrected_tau
    
    df <- data.frame(
      Model = c("Baseline autocorrelation", 
                "Uncorrected tau", 
                "Baseline corrected tau"),
      tau = c(auto_tau$tau, uncorrected_tau$tau, base_corr_tau$tau),
      z = c(auto_tau$z, uncorrected_tau$z, base_corr_tau$z),
      p = c(auto_tau$p, uncorrected_tau$p, base_corr_tau$p),
      check.names = FALSE
    )
    
    df
    
  }
  
  x <- lapply(data, corr_tau)

  out <- list(
    tau = sapply(x, function(x) if(is.na(x$p[1]) || x$p[1] > alpha) x$tau[2] else x$tau[3]), 
    p = sapply(x, function(x) if(is.na(x$p[1]) || x$p[1] > alpha) x$p[2] else x$p[3]), 
    corrected_tau = x,
    auto_tau = sapply(x, function(x) x$tau[1]),
    tau_corrected = sapply(x, function(x) x$tau[3]),
    tau_uncorrected = sapply(x, function(x) x$tau[2]),    
    correction = sapply(x, function(x) if(is.na(x$p[1]) || x$p[1] > alpha) FALSE else TRUE),
    alpha = alpha,
    continuity = continuity,
    repeated   = FALSE,
    tau_method = tau_method,
    data = data
  )
  
  class(out) <- c("sc_bctau")
  attributes(out)[opts("phase", "mt", "dv")] <- list(pvar, mvar, dvar)
  out
}

theil_sen <- function(formula, data) {
  
  y <- model.frame(formula, data)[[1]]
  x <- model.frame(formula, data)[[2]]
  
  # Pairwise slopes (exclude ties in x)
  idx <- combn(length(y), 2)
  dx <- x[idx[2, ]] - x[idx[1, ]]
  dy <- y[idx[2, ]] - y[idx[1, ]]
  keep <- dx != 0
  
  #if (!any(keep)) abort("All x values are identical; slope is undefined.")

  #fitted <- intercept + slope * x
  #resid <- y - fitted
  
  slope <- median(dy[keep] / dx[keep])
  intercept <- median(y - slope * x)
  
  out <- list(
    intercept = intercept,
    slope = slope
  )
  
  out
}

.prepare_scdf <- function(data, na.rm = FALSE) {
  
  # check class scdf validity
  if (opt("rigorous_class_check")) {
    check_scdf(data)
  }

  pvar <- phase(data)
  mvar <- mt(data)
  dvar <- dv(data)
  
  # revise names -----
  names(data) <- revise_names(names(data), length(data))
  
  
  for(case in 1:length(data)) {
    
    if (inherits(data[[case]], "tbl_df")) {
      class(data[[case]]) <- "data.frame"
      message("Found tibble within scdf and changed it to data.frame.")
    }
    
    vars <- names(data[[case]])
    if (na.rm) data[[case]] <- data[[case]][!is.na(data[[case]][[dvar]]), ]
    if (!is.factor(data[[case]][[pvar]])) {
      data[[case]][[pvar]] <- factor(
        data[[case]][[pvar]], levels = unique(data[[case]][[pvar]])
      )
    }
    data[[case]][[pvar]] <- droplevels(data[[case]][[pvar]])
  }
  data
}

#' Convert scdf to R code
#'
#' Converts an scdf object into R code that can be used to recreate the object.
#' 
#' This function is useful for exporting scdf objects for sharing or
#' documentation purposes. The generated R code can be sourced to recreate the
#' original scdf object.
#'
#' @inheritParams .inheritParams
#' @param file A filename for exporting the syntax. If an empty string (default),
#'  the syntax is not written to a file.
#' @param study_name Character string. Name of the study object.
#' @param case_name Character string. Name of the scdf objects.
#' @param inline If TRUE, phase definition is in an online version.
#' @param indent Integer. Indentation.
#' @param silent If TRUE, syntax is not printed to the console.
#' @return Returns a string (invisible).
#' @author Juergen Wilbert
#' @keywords io
#' @family io-functions
#' @examples
#' filename <- tempfile()
#' convert(exampleABC, file = filename)
#' source(filename)
#' all.equal(study, exampleABC)
#' unlink(filename)
#' @export
convert <- function(scdf,
                    file = "",
                    study_name = "study",
                    case_name = "case",
                    inline = FALSE,
                    indent = 2,
                    silent = FALSE) {
  check_args(
    by_class(scdf, "scdf")
  )

  if (length(scdf) == 1) case_name <- study_name

  scdf_string <- c()
  
  var_phase <- phase(scdf)
  var_dv <- dv(scdf)
  var_mt <- mt(scdf)

  sindent <- strrep(" ", indent)

  for (i_case in seq_along(scdf)) {

    body_string <- format_body(
      case = scdf[i_case],
      inline,
      indent
    )

    def_string <- format_definition(
      var_dv,
      var_phase,
      var_mt,
      sindent,
      casename = names(scdf)[i_case],
      inline
    )

    phase_design <- format_phase_design(
      inline,
      design = rle(as.character(scdf[[i_case]][[var_phase]])),
      def_string,
      sindent
    )

    scdf_string[i_case] <- paste0(
      case_name, i_case, " <- scdf(",
      body_string,
      phase_design,
      def_string, "\n)"
    )
  }

  scdf_string <- paste0(scdf_string, collapse = "\n\n")

  study_string <- format_study(
    n_cases = length(scdf),
    scdf_attr(scdf),
    case_name,
    indent,
    sindent,
    study_name
  )

  complete_string <- paste(
    scdf_string,
    if (!is.null(study_string)) study_string,
    sep = "\n\n"
  )

  if (!silent) cat(complete_string, "\n")
  if (file != "") cat(complete_string, "\n", file = file)
  
  invisible(complete_string)
}

format_phase_design <- function(inline, design, def_string, sindent) {
  if (!inline) {
    phase_design <- paste0(
      design$values, " = ", design$lengths,
      collapse = ", "
    )
    phase_design <- paste0("phase_design = c(", phase_design, ")")
    phase_design <- paste0(",\n", sindent, phase_design)
  } else {
    phase_design <- paste0(",\n")
  }

  phase_design
}

format_body <- function(case,
                        inline,
                        indent) {

  attr_case <- scdf_attr(case)
  case <- case[[1]]
  var_names <- names(case)[!names(case) %in% attr_case[[opt("phase")]]]
  
  
  vars <- c()

  sindent <- strrep(" ", indent)
  
  for (i in seq_along(var_names)) {
    values <- case[, var_names[i]]

    if (var_names[i] == attr_case[[opt("mt")]]) {
      if (all(seq_along(values) == values)) next
    }

    if (is.numeric(values)) {
      values_string <- paste0(values, collapse = ", ")
    }
    if (!is.numeric(values)) {
      values_string <- paste0('\"', values, '\"', collapse = ", ")
    }

    if (var_names[i] == attr_case[[opt("dv")]] && inline) {
      x <- split(
        case[[attr_case[[opt("dv")]]]], 
        case[[attr_case[[opt("phase")]]]]
      )
      x <- mapply(
        function(x, n) {
          paste0(n, " = ", paste0(x, collapse = ", "), collapse = "")
        },
        x, names(x)
      )

      values_string <- paste0(
        x,
        collapse = paste0(",\n", strrep(" ", indent * 2))
      )
      vars[i] <- paste0(
        var_names[i], " = c(\n", strrep(" ", 2 * indent), values_string, "\n",
        strrep(" ", indent * 1), ")"
      )
    } else {
      values_string <- paste0(
        strwrap(values_string, exdent = 2 * indent),
        collapse = "\n"
      )
      vars[i] <- paste0(
        var_names[i], " = c(\n", sindent, sindent, values_string, "\n",
        sindent, ")"
      )
    }
  }
  vars <- vars[!is.na(vars)]
  paste0("\n", sindent, vars, collapse = ",")
}

format_definition <- function(var_dv,
                              var_phase,
                              var_mt,
                              sindent,
                              casename,
                              inline) {
  def_string <- character(0)
  if (var_dv != "values") {
    def_string <- c(
      def_string,
      paste0(sindent, 'dvar = \"', var_dv, '\"')
    )
  }
  if (var_phase != "phase") {
    def_string <- c(
      def_string,
      paste0(sindent, 'pvar = \"', var_phase, '\"')
    )
  }
  if (var_mt != "mt") {
    def_string <- c(
      def_string,
      paste0(sindent, 'mvar = \"', var_mt, '\"')
    )
  }
  if (!is.null(casename) && !is.na(casename)) {
    def_string <- c(
      def_string,
      paste0(sindent, 'name = \"', casename, '\"')
    )
  }
  def_string <- paste0(def_string, collapse = ", \n")

  if (!inline && def_string != "")
    def_string <- paste0(",\n", def_string)

  def_string
}

format_study <- function(n_cases,
                         attr_scan,
                         case_name,
                         indent,
                         sindent,
                         study_name) {
  con_string <- NULL

  if (n_cases > 1 ||
      !is.null(attr_scan$author) ||
      !is.null(attr_scan$info)) {
    con_string <- paste0(case_name, 1:n_cases, collapse = ", ")
    con_string <- paste0(
      strwrap(con_string, exdent = indent, width = 80, simplify = TRUE),
      collapse = "\n"
    )
    if (!is.null(attr_scan$info)) {
      con_string <- paste0(
        con_string, ", \n", sindent, "info = \"", attr_scan$info, "\""
      )
    }
    if (!is.null(attr_scan$author)) {
      con_string <- paste0(
        con_string, ", \n", sindent, "author = \"", attr_scan$author, "\""
      )
    }
    con_string <- paste0(study_name, " <- c(\n", sindent, con_string, "\n)")
  }
  con_string
}

#' Truncate single-case data
#'
#' #' *This function is superseded by the more versatile 
#' \code{\link{transform.scdf}} function.*
#' This function truncates data points at the beginning and / or end of each
#' phase in each case.
#'
#' @inheritParams .inheritParams
#' @param truncate A list with a vector of two (beginning and end) values for
#'   each phase defining the number of data points to be deleted. For lists of
#'   single-case data frames, the truncation is adapted to the length of each
#'   phase for each single case.
#' @param na If FALSE, the truncated measurement times are deleted. If TRUE, NAs
#'   are set for the dependent variable.
#' @return A truncated data frame (for each single-case).
#' @author Juergen Wilbert
#' @family data manipulation functions
#' @keywords internal
#' @examples
#'
#' ## Truncate the first two data points of both phases and compare the two 
#' ## data sets
#' study <- c(
#'   "Original" = byHeart2011[1],
#'   "Selected" = truncate_phase(
#'     byHeart2011[1], truncate = list(A = c(2, 0), B = c(2, 0))
#'   )
#' )
#' plot(study)
#' @export
truncate_phase <- function(data, dvar, pvar, 
                           truncate = list(A = c(0, 0), B = c(0, 0)), 
                           na = TRUE) {

  # set attributes to arguments else set to defaults of scdf
  if (missing(pvar)) 
    pvar <- phase(data) else phase(data) <- pvar
  if (missing(dvar)) 
    dvar <- dv(data) else dv(data) <- dvar

  data <- .prepare_scdf(data)

  N <- length(data)

  for (i in 1:N) {
    phases <- rle(as.character(data[[i]][[pvar]]))
    phases$start <- c(1, cumsum(phases$lengths) + 1)[1:length(phases$lengths)]
    phases$stop <- cumsum(phases$lengths)
    class(phases) <- "list"
    deselect <- c()
    if (length(phases$values) != length(truncate)) {
      abort("Please provide truncation values for each phase.")
    }
    for (ph in 1:length(phases$values)) {
      if (truncate[[ph]][1] > 0) {
        deselect <- c(
          deselect, 
          phases$start[ph]:(phases$start[ph] + truncate[[ph]][1] - 1)
        )
      }
      if (truncate[[ph]][2] > 0) {
        deselect <- c(
          deselect, 
          (phases$stop[ph] - truncate[[ph]][2] + 1):phases$stop[ph]
        )
      }
    }

    cat("Deletet measurements per case:\n\n")
    cat(paste0(names(data)[i], ": "))
    cat(deselect)
    cat("\n")
    if (!na) data[[i]] <- data[[i]][-deselect, ]
    if (na) data[[i]][deselect, dvar] <- NA
  }
  
  data
}

#' @describeIn outlier Print results
#' @order 2
#' @param x An object returned by [outlier()]
#' @export
#' @inheritParams print.sc
print.sc_outlier <- function(x, digits = "auto", ...) {
  
  cat("Outlier Analysis for Single-Case Data\n\n")
  
  if (x$criteria[1] == "CI") {
    names(x$ci.matrix) <- x$case.names
    cat("Criteria: Exceeds", as.numeric(x$criteria[2]) * 100, 
        "% Confidence Interval\n\n")
    print(x$ci.matrix)
  }
  
  if (x$criteria[1] == "SD") {
    names(x$sd.matrix) <- x$case.names
    cat("Criteria: Exceeds", x$criteria[2], "Standard Deviations\n\n")
    print(x$sd.matrix)
  }
  
  if (x$criteria[1] == "MAD") {
    names(x$mad.matrix) <- x$case.names
    cat("Criteria: Exceeds", x$criteria[2], "Mean Average Deviations\n\n")
    print(x$mad.matrix)
  }
  
  if (x$criteria[1] == "Cook") {
    cat("Criteria: Cook's Distance based on piecewise-regression exceeds", 
        x$criteria[2],"\n\n")
  }
  
  for(i in 1:length(x$dropped.n)) {
    cat("Case",x $case.names[i],": Dropped", x$dropped.n[[i]], "\n")
  }
  cat("\n")
}

#' @describeIn outlier Export html results
#' @order 3
#' @inheritParams export
#' @export
export.sc_outlier <- function(object, 
                              caption = NA, 
                              footnote = NA, 
                              filename = NA,
                              ...) {
  
  if (is.na(caption)) {
    caption <- paste0(
      "Outlier Analysis for variable '", 
      attr(object, opt("dv")),  "'"
    )
  }
  
  if (is.na(footnote)) {
  }
  
  out <- capture.output(
    print(object)
  ) [-1:-2]
  
  out <- data.frame(
    "Outlier Analysis" = gt::html(paste(out, collapse = "  <br>")),
    check.names = FALSE
  )
  table <- .create_table(
    out,
    caption = caption,
    footnote = footnote,
    fmt_markdown = TRUE,
    ...
  )
  
  if (!is.na(filename)) .save_export(table, filename)
  
  table
  
}

#' Bayesian Piecewise Linear Model (bplm)
#'
#' Computes a bayesian (hierarchical) piecewise linear model based on a Markov
#' chain Monte Carlo sampler. The function automatically creates the fixed and
#' random part of the regression model.
#'
#' The function uses the [MCMCglmm::MCMCglmm()] function to fit the model. The
#' default model includes fixed trend, level, and slope effects as well as a
#' random intercept for each single-case. The fixed part of the model can be
#' changed by providing a custom formula to the `fixed` argument or by using the
#' `update_fixed` argument. The random part of the model can be changed by
#' providing a custom formula to the `random` argument or by setting the
#' `random_trend`, `random_level`, or `random_slope` arguments to TRUE.
#'
#' @inheritParams .inheritParams
#' @order 1
#' @param random_trend If TRUE, includes a random trend effect.
#' @param random_level If TRUE, includes a random level effect.
#' @param random_slope If TRUE, includes a random slope effect.
#' @param fixed A formula that overwrites the automatically created fixed part
#'   of the regression model that defaults to the standard piecewise regression
#'   model. The parameter phase followed by the phase name (e.g., phaseB)
#'   indicates the level effect of the corresponding phase. The parameter
#'   'inter' followed by the phase name (e.g., interB) adresses the slope effect
#'   based on the method provide in the model argument (e.g., "B&L-B"). The
#'   formula can be changed for example to include further L1 or L2 variables
#'   into the regression model.
#' @param random A formula that overwrites the automatically created random part
#'   of the regression model.
#' @param update_fixed An easier way to change the fixed model part (e.g., `. ~
#'   . + newvariable`).
#' @param ... Further arguments passed to the [MCMCglmm::MCMCglmm()] function.
#' @return An object of class `sc_bplm` with element:
#' - `model` - List containing information about the applied model. 
#' - `N` - Number of single-cases.
#' - `formula` - A list containing the fixed and the random formulas of the hplm model. 
#' - `mcmglmm` -  Object of class MCMglmm.
#' - `contrast` -  List with contrast definitions.
#' @author Juergen Wilbert
#' @family regression functions
#' @examples
#' # plm regression
#' bplm(example_A24)
#'
#' # Multilevel plm regression with random intercept
#' bplm(exampleAB_50, nitt = 5000)
#'
#' # Adding a random slope
#' bplm(exampleAB_50, random_level = TRUE, nitt = 5000)
#'
#' # Custom fixed formula
#' bplm(exampleAB_50, update_fixed = values ~ -1 + mt + phaseB +
#'   interB, nitt = 5000)
#' @export

bplm <- function(data, dvar, pvar, mvar, 
                 model = c("W", "H-M", "B&L-B"),
                 contrast_level = c("first", "preceding"),
                 contrast_slope = c("first", "preceding"),
                 trend = TRUE, 
                 level = TRUE, 
                 slope = TRUE, 
                 random_trend = FALSE, 
                 random_level = FALSE, 
                 random_slope = FALSE, 
                 fixed = NULL, 
                 random = NULL, 
                 update_fixed = NULL, 
                 ...) {

  check_args(
    by_call(model),
    by_call(contrast_level),
    by_call(contrast_slope)
  )
  
  # set attributes to arguments else set to defaults of scdf
  if (missing(dvar)) dvar <- dv(data) else dv(data) <- dvar
  if (missing(pvar)) pvar <- phase(data) else phase(data) <- pvar
  if (missing(mvar)) mvar <- mt(data) else mt(data) <- mvar
  
  dat <- .prepare_scdf(data)
  
  N <- length(dat)
  out <- list()
  out$model$interaction.method  <- model
  out$N                         <- N

# interaction and dummy coding and L2 --------------------------------------

  dat_dummies <- .add_dummy_variables(
    data = dat, 
    model = model, 
    contrast_level = contrast_level, 
    contrast_slope = contrast_slope
  )
  dat <- as.data.frame(dat_dummies$data)

# create formulas ---------------------------------------------------------

  if (is.null(fixed)) {
    fixed <- .create_fixed_formula(
      dvar, 
      mvar, 
      slope, 
      level, 
      trend, 
      dat_dummies$var_phase, 
      dat_dummies$var_inter
    )
  }
  if (!is.null(update_fixed)) fixed <- update(fixed, update_fixed)
  
  if (is.null(random) && N > 1) random <- ~case
  
  if (any(random_trend, random_level, random_slope)) {
    random <- .create_random_formula(
      mvar, 
      random_slope, 
      random_level, 
      random_trend, 
      dat_dummies$var_phase, 
      dat_dummies$var_inter,
      syntax = "mcmc"
    )
  }
  
  out$formula <- list(fixed = fixed, random = random)
  
# mcmcglmm model ----------------------------------------------------------
 
  out$mcmcglmm <- MCMCglmm(
    fixed = fixed,
    random = random,
    data = dat,
    verbose = FALSE,
    ...
  )
  
# out ----
  
  out$model$fixed  <- fixed
  out$model$random <- random
  
  
  out$contrast <- list(level = contrast_level, slope = contrast_slope)
  
  class(out) <- c("sc_bplm")
  
  attributes(out)[opts("phase", "mt", "dv")] <- list(pvar, mvar, dvar)

  out
}

#' @describeIn hplm Print results
#' @inheritParams print.sc
#' @param casewise Returns the effect estimations for each case
#' @param bcsmd If TRUE, reports between-case standardized mean differences.
#' @order 2
#' @param x An object returned by [hplm()]
#' @export
print.sc_hplm <- function(x, 
                          digits = 3, 
                          bcsmd = FALSE, 
                          casewise = FALSE,
                          ...) {
  
  out <- .output_hplm(x, casewise = casewise, bcsmd = bcsmd)
  
  cat("Hierarchical Piecewise Linear Regression\n\n")
  cat("Estimation method", x$model$estimation.method,"\n")
  cat("Contrast model: ", out$model, "\n", sep = "")
  cat(x$N, "Cases\n\n")

  cat("AIC = ", out$AIC, ", BIC = ", out$BIC, "\n", sep = "")
  if (!is.null(out$icc)) cat(out$icc, "\n")
  
  cat("\nFixed effects (", out$formula$fixed, ")\n\n", sep = "")
  print(round_numeric(out$fixed, digits))

  cat("\nRandom effects (", out$formula$random ,")\n\n", sep = "")
  print(format_table(out$random, digits = digits, integer = "df"), ...)
  
  if (!is.null(out$correlation)) {
    cat("\nCorrelation:\n")
    print(out$correlation)
  }
  
  if (!is.null(out$bc_smd)) {
    cat("\nBetween-Case Standardized Mean Difference\n\n")
    print(out$bc_smd, digits = digits, row.names = FALSE)
  }
  
  if (!is.null(out$casewise)) {
    cat("\nCasewise estimation of effects\n\n")
    print(out$casewise, row.names = FALSE)
  }
}

#' @describeIn hplm Export results as html table (see [export()])
#' @order 3
#' @inheritParams export
#' @export
export.sc_hplm <- function(object, 
                           caption = NA, 
                           footnote = NA, 
                           filename = NA,
                           round = 2,
                           nice = TRUE,
                           casewise = FALSE,
                           ...) {
  
  if (is.na(caption)) {
    caption <- paste0(
      "Hierarchical Piecewise Linear Regression predicting variable '", 
      attr(object, opt("dv")),  "'"
    )
  }
  
  footnote <- c(
    paste0("Estimation method ", object$model$estimation.method),
    str_contrasts(object$model$interaction.method, object$contrast),
    paste0("N = ", object$N, " cases")
  )
  
  if (casewise) {
    out <- .export_casewise(object, caption, footnote, filename, round)
    return(out)
  }
  
  results <- .output_hplm(object)
  
  out <- results$fixed
  dat_random <- results$random
  
  if (nice) {
    out$p <- .nice_p(out$p)
    if (!is.null(dat_random$p)) dat_random$p <- .nice_p(dat_random$p)
  }
  
  out[, ] <- lapply(out[, ], function(x)
    if (inherits(x, "numeric")) as.character(round(x, round)) else x
  )
  out <- cbind(Predictors = rownames(out), out, stringsAsFactors = FALSE)
  rownames(out) <- NULL
  dat_random[, ] <- lapply(dat_random, function(x)
    if (inherits(x, "numeric")) as.character(round(x, round)) else x
  )
  dat_random <- cbind(
    " " = rownames(dat_random), 
    dat_random, 
    stringsAsFactors = FALSE
  )
  rownames(dat_random) <- NULL
  
  nrow_out <- nrow(out)
  nrow_random <- nrow(dat_random)
  tmp_row <- (nrow_out + 1):(nrow_out + nrow_random + 1 + 3)
  out[tmp_row, ] <- ""
  
  tmp_row <- (nrow_out + 1):(nrow_out + nrow_random + 1)
  out[tmp_row, 1:ncol(dat_random)] <- rbind(
    colnames(dat_random), 
    dat_random, 
    stringsAsFactors = FALSE
  )
  
  out[nrow_out + nrow_random + 2, 1:2] <- c(
    "AIC", as.character(round(results$AIC, 1))
  )
  out[nrow_out + nrow_random + 3, 1:2] <- c(
    "BIC", as.character(round(results$BIC, 1))
  )
  if (!is.null(object$ICC)) {
    out[nrow_out + nrow_random + 4, 1:4] <-
      c(
        "ICC", 
        as.character(round(object$ICC$value, 2)),
        paste0("L = ", round(object$ICC$L, 1)),
        paste0("p ", .nice_p(object$ICC$p))
      )
  }
  
  table <- .create_table(
    out,
    caption = caption,
    footnote = footnote,
    row_group = list(
      "Fixed effects" = 1: nrow_out,
      "Random effects" = (nrow_out + 1) : (nrow(out) - 3),
      "Model" = (nrow(out) - 2) : nrow(out)
    )
  )
  
  if (getOption("scan.export.engine") == "kable") {
    table <- table |>
      #pack_rows("Fixed effects", 1, nrow_out, indent = FALSE) |>
      pack_rows("\nRandom effects", nrow_out + 1, nrow(out), indent = FALSE) |>
      pack_rows("\nModel", nrow(out) - 2, nrow(out), indent = FALSE) |>
      #row_spec(nrow_out + nrow(dat_random) + 1, hline_after = TRUE) |>
      row_spec(nrow_out, hline_after = TRUE)
  }
  
  if (!is.na(filename)) .save_export(table, filename)
  
  table
  
}

.export_casewise <- function(object, 
                             caption = NA, 
                             footnote = NA, 
                             filename = NA,
                             round = 2) {
  
  out <- coef(object, casewise = TRUE)
  
  if (getOption("scan.export.engine") == "kable") {
    table <- .create_table(
      out,
      caption = caption,
      footnote = footnote
    )
  }
  
  if (getOption("scan.export.engine") == "gt") {
    table <- export_table_gt(
      out, title = caption, footnote = footnote, 
      decimals = round
    )
  }
  
  if (!is.na(filename)) .save_export(table, filename)
  
  table
}


.output_hplm <- function(x, casewise = FALSE, bcsmd = FALSE) {
  
  out <- list()
  
  summary_model <- summary(x$hplm)
  
  out$AIC <- summary_model$AIC
  out$BIC <- summary_model$BIC
  
  out$formula$fixed <- deparse(x$model$fixed)
  out$formula$random <- deparse(x$model$random)
  
  out$model <- paste0(
    x$model$interaction.method, " / ", 
    paste0(names(x$contrast), ": ",x$contrast, collapse = ", ")
  )
  
  if (x$model$ICC) {
    out$icc <- sprintf("ICC = %.3f; L = %.1f; p = %.3f", 
                       x$ICC$value, x$ICC$L, x$ICC$p)
  }
  
  # fixed ----
  
  md <- as.data.frame(summary_model$tTable)
  colnames(md) <- c("B", "SE", "df", "t", "p")
  row.names(md) <- rename_predictors(row.names(md), x)
  out$fixed <- md
 
  
  # random -----
  
  sd <- round(as.numeric(VarCorr(x$hplm)[,"StdDev"]), 3)
  md <- data.frame("SD" = sd)
  row.names(md) <- rename_predictors(names(VarCorr(x$hplm)[, 2]), x)
  
  if (x$model$lr.test) {
    if (is.null(x$LR.test[[1]]$L.Ratio)) {
      x$LR.test[[1]]$L.Ratio <- NA
      x$LR.test[[1]]$"p-value" <- NA
      x$LR.test[[1]]$df <- NA
    }
    
    md$L  <- c(unlist(lapply(x$LR.test, function(x) x$L.Ratio[2])), NA)
    md$df <- c(unlist(lapply(x$LR.test, function(x) x$df[2] - x$df[1])), NA)
    md$p  <- c(unlist(lapply(x$LR.test, function(x) x$"p-value"[2])), NA)
  }
  
  out$random <- md
  
  # correlation ----
  
  cov_matrix <- getVarCov(x$hplm)
  var_sd <- sqrt(diag(cov_matrix))
  cor_matrix <- cov_matrix / (var_sd %o% var_sd)
  cor_matrix <- as.data.frame(round(cor_matrix[,], 2))
  row.names(cor_matrix) <- rename_predictors(rownames(cor_matrix), x)
  colnames(cor_matrix) <- rename_predictors(colnames(cor_matrix), x)
  cor_matrix[upper.tri(cor_matrix, diag = TRUE)] <- ""
  cor_matrix <- cor_matrix[-1,-ncol(cor_matrix), drop = FALSE]
 
  if (nrow(cor_matrix) > 0) out$correlation <- cor_matrix
  
  # casewise ------
  
  if (casewise) out$casewise <- coef(x, casewise = TRUE)
  
  # bcsmd ----
  
  if (bcsmd) {
    out$bc_smd <- between_smd(x)$models[[1]]
  }
  
  out
  
}

#' Descriptive statistics for single-case data
#'
#' The `describe()` function provides common descriptive statistics for
#' single-case data.
#' 
#' It computes the number of measurements, number of missing values, mean,
#' median, standard deviation, median average deviation, minimum, maximum, and
#' trend (slope of dependent variable regressed on measurement-time) for each
#' phase of each single-case included in an scdf.
#'
#' @inheritParams .inheritParams
#' @details n = number of measurements; mis = number of missing vaues; m = mean;
#'   md = median; sd = standard deviation; mad = median average deviation; min =
#'   minimum; max = maximum; trend = weight of depended variable regressed on
#'   time (values ~ mt).
#' @return A list containing a data frame of descriptive statistics
#'   (descriptives); the cse design (design); the number of cases (N).
#' @author Juergen Wilbert
#' @seealso [overlap()], [plot.scdf()]
#' @examples
#'
#' ## Descriptive statistics for a study of three single-cases
#' describe(Grosche2011)
#'
#' ## Descriptives of a three phase design
#' describe(exampleABC)
#'
#' ## Write descriptive statistics to .csv-file
#' study <- describe(Waddell2011)
#' write.csv(study$descriptives, file = tempfile())
#' @export
describe <- function(data, dvar, pvar, mvar) {

  # set defaults attributes
  if (missing(dvar)) dvar <- dv(data) 
  if (missing(pvar)) pvar <- phase(data) 
  if (missing(mvar)) mvar <- mt(data) 
  
  dv(data) <- dvar
  phase(data) <- pvar
  mt(data) <- mvar
  
  data_list <- .prepare_scdf(data)
  N <- length(data_list)

  designs <- lapply(
    data_list, 
    function(x) rle(as.character(x[[pvar]]))$values
  )
  phase_designs <- sapply(designs, function(x) paste0(x, collapse = "-"))
  
  for(i in 1:N) {
    data_list[[i]][[pvar]] <- rename_phase_duplicates(data_list[[i]][[pvar]])
  }
  
  designs <- lapply(
    data_list, 
    function(x) rle(as.character(x[[pvar]]))$values
  )
  design <- unique(unlist(designs))
  
  vars <- c("n", "mis", "m", "md", "sd", "mad", "min", "max", "trend")
  vars <- paste0(rep(vars, each = length(design)), ".", design)
  
  desc <- as.data.frame(matrix(nrow = N, ncol = length(vars)))
  colnames(desc) <- vars
  desc <- data.frame(
    Case = names(data_list), 
    Design = phase_designs, 
    desc,
    check.names = FALSE
  )

  for (case in 1:N) {
    data <- data_list[[case]]
    phases <- .phasestructure(data, pvar = pvar)
    for (i in 1:length(phases$values)) {

      x <- data[[mvar]][phases$start[i]:phases$stop[i]]
      y <- data[[dvar]][phases$start[i]:phases$stop[i]]

      phase <- phases$values[i]
      desc[case, paste0("n.", phase)] <- length(y)
      desc[case, paste0("mis.", phase)] <- sum(is.na(y), na.rm = TRUE)
      desc[case, paste0("m.", phase)] <- mean(y, na.rm = TRUE)
      desc[case, paste0("md.", phase)] <- median(y, na.rm = TRUE)
      desc[case, paste0("sd.", phase)] <- sd(y, na.rm = TRUE)
      desc[case, paste0("mad.", phase)] <- mad(y, na.rm = TRUE)
      desc[case, paste0("min.", phase)] <- min(y, na.rm = TRUE)
      desc[case, paste0("max.", phase)] <- max(y, na.rm = TRUE)
      desc[case, paste0("trend.", phase)] <- coef(lm(y ~ I(x - x[1] + 1)))[2]
    }
  }

  out <- list(
    descriptives = desc,
    design = design,
    N = N
  )
  class(out) <- c("sc_desc")
  attributes(out)[opts("phase", "mt", "dv")] <- list(pvar, mvar, dvar)

  out
}

#' @rdname export
#' @export
export.sc_power <- function(object, caption = NA, footnote = NA, filename = NA,
                            round = 3,
                            ...) {
  
  if (is.na(caption)) {
    caption <- c("Test power in percent")
  }
  
  if (is.na(footnote)) {
  }

  out <- object
  class(out) <- "data.frame"
 
  out <- round_numeric(out, round)
  
  table <- .create_table(
    out, 
    caption = caption,
    footnote = footnote,
    ...
  )
  
  # finish ------------------------------------------------------------------
  
  if (!is.na(filename)) .save_export(table, filename)
  table
}

.tau_z <- function(tau) {
  0.5 * log((1 + tau)/(1 - tau))
}

.inv_tau_z <- function(tau) {
  if (identical(tau, Inf)) {
    return(1)
  }
  if (identical(tau, -Inf)) {
    return(-1)
  }
  (exp(2 * tau) - 1) / (exp(2 * tau) + 1)
}

.tau_ci <- function(tau, n, ci, se_method) {
  z <- qnorm((1 - ci) /2, lower.tail = FALSE)
  
  if (se_method == "tau") se_z <- sqrt(0.437 / (n - 4))
  if (se_method == "z") se_z <- 1 / sqrt(n - 3)
  
  tau_z <- .tau_z(tau)
  tau_z_ci_lower <- tau_z - z * se_z
  tau_z_ci_upper <- tau_z + z * se_z
  
  list(
    tau = .inv_tau_z(tau_z),
    n = n,
    tau_ci_lower = .inv_tau_z(tau_z_ci_lower),
    tau_ci_upper = .inv_tau_z(tau_z_ci_upper),
    tau_z = tau_z,
    tau_z_ci_lower = tau_z_ci_lower,
    tau_z_ci_upper = tau_z_ci_upper,
    se_z = se_z,
    se_method = se_method
  )
  
}

.meta_tau <- function(tau, n, ci, se_method) {
  ci_z <- qnorm((1 - ci) / 2, lower.tail = FALSE)
  tau_z <- .tau_z(tau)
  
  if (se_method == "tau") se_z <- sqrt(0.437 / (n - 4))
  if (se_method == "z") se_z <- 1 / sqrt(n - 3)
  
  weight <- 1 / se_z^2
  TE <- sum(tau_z * weight) / sum(weight)
  
  # Common effect estimate (Cooper & Hedges, 1994, p. 265-6)
  seTE <- sqrt(1 / sum(weight))
  
  # return
  list(
    tau = .inv_tau_z(TE),
    lower = .inv_tau_z(TE - ci_z * seTE),
    upper = .inv_tau_z(TE + ci_z * seTE),
    z = TE / seTE,
    p = 2 * pnorm(abs(TE/ seTE), lower.tail = FALSE),
    se = seTE,
    ci = ci,
    se_method = se_method
  )
}

#' @rdname print.sc
#' @param duration If TRUE the duration for computation is printed.
#' @export
print.sc_power <- function(x, duration = FALSE, digits = 1, ...) {
  
  cat("Test-Power in percent:\n\n")
  
  class(x) <- "data.frame"
  
  out <- x
  
  ci <- attr(x, "ci")
  binom_test <- attr(x, "binom_test")
  binom_test_power <- attr(x, "binom_test_power")
  binom_test_alpha <- attr(x, "binom_test_alpha")
  binom_test_correct <- attr(x, "binom_test_correct")

  if (is.numeric(ci)) {
    ci_p <- attr(x, "ci") * 100
    ci_str <- c(paste0((100 - ci_p) / 2, "%"), paste0(ci_p + (100 - ci_p) / 2, "%"))
    
    out <- x[c("Method", "Power", "Power lower",  "Power upper", 
               "Alpha Error", "Alpha Error lower", "Alpha Error upper", 
               "Alpha:Beta", 
               "Correct", "Correct lower",  "Correct upper")]
    names(out)[c(3, 4, 6, 7, 10, 11)] <- rep(ci_str, 3)
  } else {
    out <- out[c("Method", "Power", "Alpha Error", "Alpha:Beta", "Correct")]
  }
  
  if (is.numeric(binom_test_power)) {
      out$p_power <- x$p_power
      names(out)[which(names(out) == "p_power")] <- paste0("p Power>=", binom_test_power*100)
  }

  if (is.numeric(binom_test_alpha)) {
    out$p_alpha <- x$p_alpha
    names(out)[which(names(out) == "p_alpha")] <- paste0("p Alpha Error<=", binom_test_alpha*100)
  }
    
  if (is.numeric(binom_test_power)) {
    out$p_correct <- x$p_correct
    names(out)[which(names(out) == "p_correct")] <- paste0("p Correct>=", binom_test_correct*100)
  }

  print(out, row.names = FALSE, digits = digits)
  
  if (duration) 
    cat(
      "\nComputation duration is", 
      round(attr(x, "computation_duration")[3], 1), 
      "seconds.\n"
    )
  
}

#'Percentage of all non-overlapping data
#'
#'The `pand()` function calculates the percentage of all non-overlapping data
#'(PAND; Parker, Hagan-Burke, & Vannest, 2007), an index to quantify a level
#'increase (or decrease) in performance after the onset of an intervention.
#'
#'PAND was proposed by Parker, Hagan-Burke, and Vannest in 2007. The authors
#'emphasize that PAND is designed for application in a multiple case design with
#'a substantial number of measurements, technically at least 20 to 25, but
#'preferably 60 or more. PAND is defined as 100% minus the percentage of data
#'points that need to be removed from either phase in order to ensure nonoverlap
#'between the phases. Several approaches have been suggested to calculate PAND,
#'leading to potentially different outcomes. In their 2007 paper, Parker and
#'colleagues present an algorithm for computing PAND. The algorithm involves
#'sorting the scores of a time series, including the associated phases, and
#'comparing the resulting phase order with the original phase order using a
#'contingency table. To account for ties, the algorithm includes a randomization
#'process where ties are randomly assigned to one of the two phases.
#'Consequently, executing the algorithm multiple times could yield different
#'results. It is important to note that this algorithm does not produce the same
#'results as the PAND definition provided earlier in the same paper. However, it
#'offers the advantage of allowing the calculation of an effect size measure
#'`phi`, and the application of statistical tests for frequency distributions.
#'Pustejovsky (2019) presented a mathematical formulation of Parker's original
#'definition for comparing two phases of a single case: \deqn{PAND =
#'\frac{1}{m+n}max\{(i+j)I(y^A_{i}<y^B_{n+1-j}\}} This formulation provides
#'accurate results for PAND, but the original definition has the drawback of an
#'unknown distribution under the null hypothesis, making a statistical test
#'difficult. The `pand()` function enables the calculation of PAND using both
#'methods. The first approach (`method = "sort"`) follows the algorithm
#'described above, with the exclusion of randomization before sorting to avoid
#'ambiguity. It calculates a phi measure and provides the results of a
#'chi-squared test and a Fisher exact test. The second approach (`method =
#'"minimum"`) applies the aforementioned formula. The code of this function is
#'based on the code of the `SingleCaseES` package (function `calc_PAND`). For a
#'multiple case design, overlaps are calculated for each case, summed, and then
#'divided by the total number of measurements. No statistical test is conducted
#'for this method.
#'
#'@inheritParams .inheritParams
#'@param method Either `"sort"`" or `"minimum"`. See details.
#'@order 1
#'@return 
#'  |  |  |
#'  | --- | --- |
#'  | `pand` | Percentage of all non-overlapping data. |
#'  | `method` | Calculation method. |
#'  | `phi` | Effect size Phi based on expected and observed values. | 
#'  | `perc_overlap` | Percentage of overlapping data points. | 
#'  | `overlaps` | Number of overlapping data points. |
#'  | `n` | Number of data points. |
#'  | `N` | Number of cases. | 
#'  | `n_a` | Number of data points in phase A. | 
#'  | `n_b` | Number of data points in phase B. | 
#'  | `matrix` | 2x2 frequency matrix of phase A and B comparisons. |
#'  | `matrix_counts` | 2x2 counts matrix of phase A and B comparisons. |
#'  | `chi_test` | A Chi-squared analysis of expected and observed data (chisq.test()). | 
#'  | `fisher_test` | A Fisher exact test analysis of expected and observed data (fisher.test()). | 
#'@author Juergen Wilbert
#'@family overlap functions
#'@references Parker, R. I., Hagan-Burke, S., & Vannest, K. (2007). Percentage
#'  of All Non-Overlapping Data (PAND): An Alternative to PND. *The Journal of
#'  Special Education, 40*, 194-204.
#'
#'  Parker, R. I., & Vannest, K. (2009). An Improved Effect Size for Single-Case
#'  Research: Nonoverlap of All Pairs. *Behavior Therapy, 40*, 357-367.
#'
#'  Pustejovsky, J. E. (2019). Procedural sensitivities of effect sizes for
#'  single-case designs with directly observed behavioral outcome measures.
#'  *Psychological Methods*, *24(2)*, 217-235.
#'  https://doi.org/10.1037/met0000179
#'
#'  Pustejovsky JE, Chen M, Swan DM (2023). SingleCaseES: A Calculator for
#'  Single-Case Effect Sizes. R package version 0.7.1.9999,
#'  https://jepusto.github.io/SingleCaseES/.
#' @examples
#' ## REplication of the Parker et al. 2007 example
#' pand(Parker2007)
#'
#' ## Calculate the PAND with an expected decrease of phase B scores
#' cubs <- scdf(c(20,22,24,17,21,13,10,9,20,9,18), B_start = 5)
#' pand(cubs, decreasing = TRUE)
#'
#'@export
pand <- function(data, dvar, pvar, 
                 decreasing = FALSE, 
                 phases = c(1, 2),
                 method = c("sort", "minimum")) {
  
  check_args(
    by_call(method)
  )
  
  # set default attributes
  if (missing(dvar)) dvar <- dv(data)
  if (missing(pvar)) pvar <- phase(data)
  
  dv(data) <- dvar
  phase(data) <- pvar
  
  data <- .prepare_scdf(data, na.rm = TRUE)
  data <- recombine_phases(data, phases = phases)$data
  
  N <- length(data)
  values_a <- lapply(data, function(x) x[x[[pvar]] == "A", dvar])
  values_b <- lapply(data, function(x) x[x[[pvar]] == "B", dvar])
  n_all_a <- length(unlist(values_a))
  n_all_b <- length(unlist(values_b))
  n <- n_all_a + n_all_b
  
  if (method == "sort") {
    
    # phase order per case as found in data -----
    
    phases_data <- lapply(data, function(x) x[[pvar]]) |> unlist()
    
    # phase order when sorted by values within case ----
    phases_sorted <- lapply(data, function(x) {
      x <- x[sample(1:nrow(x)),]
      x[[pvar]][order(x[[dvar]], x[[pvar]], decreasing = decreasing)]
    }) |> unlist() 
    
    #phases_sorted <- lapply(data, function(x) {
    #  x <- x[sample(1:nrow(x)),]
    #  x[[pvar]][sort.list(x[[dvar]],decreasing = decreasing)]
    #}) |> unlist()
    
    mat_counts <- table(phases_data, phases_sorted)
    mat_propotions <- prop.table(mat_counts)
    pand <- (mat_propotions[1,1] + mat_propotions[2,2]) * 100
    overlaps <- mat_counts[1,2] + mat_counts[2,1]
    perc_overlap <- overlaps / n * 100
    
    chi_test <- suppressWarnings(chisq.test(mat_counts, correct = FALSE))
    
    phi <- sqrt(chi_test$statistic / n)
    
    out <- list(
      pand = pand, 
      method = method,
      phi = unname(phi), 
      perc_overlap = perc_overlap, 
      overlaps = overlaps, 
      n = n, 
      N = N, 
      n_a = n_all_a, 
      n_b = n_all_b, 
      matrix = mat_propotions, 
      matrix_counts = mat_counts, 
      chi_test = chi_test,
      fisher_test = suppressWarnings(fisher.test(mat_counts))
    )
    
  }
  
  if (method == "minimum") {
    
    .pand_pustejowski <- function(values_a, values_b) {
      
      if (decreasing) {
        values_a <- -1 * values_a
        values_b <- -1 * values_b
      }
      n_a <- length(values_a)
      n_b <- length(values_b)
      x <- c(-Inf, sort(values_a))
      y <- c(sort(values_b), Inf)
      grid <- expand.grid(a = 1:(n_a + 1), b = 1:(n_b + 1))
      grid$no_overlap <- mapply(
        function(a, b) x[a] < y[b], 
        a = grid$a, 
        b = grid$b
      )
      grid$overlap <- grid$a + n_b - grid$b
      
      nonoverlaps <- max(grid$overlap * grid$no_overlap)
  
      list(
        pand = nonoverlaps/(n_a + n_b), 
        nonoverlaps = nonoverlaps, 
        length_a = n_a, 
        length_b = n_b
      )
    }
    
    casewise <- mapply(
      .pand_pustejowski, 
      values_a = values_a, 
      values_b = values_b,
      SIMPLIFY = FALSE, 
      USE.NAMES = TRUE
    )
    nonoverlaps <- lapply(casewise, function(x) x$nonoverlaps) |> 
      unlist() |> 
      sum()
    
    out <- list(
      pand = nonoverlaps / n * 100,
      overlaps = n - nonoverlaps,
      perc_overlaps = 100 - (nonoverlaps / n * 100),
      n = n, 
      N = N, 
      n_a = n_all_a, 
      n_b = n_all_b, 
      casewise = casewise,
      method = method
    )
  }

  class(out) <- c("sc_pand")
  attributes(out)[opts("phase", "dv")] <- list(pvar, dvar)
  out
}


#' @describeIn pand Export results as html table (see [export()])
#' @inheritParams export
#' @order 3
#' @export
export.sc_pand <- function(object, 
                           caption = NA, 
                           footnote = NA, 
                           filename = NA,
                           round = 1,
                           ...) {
  
  if (is.na(caption)) {
    caption <- c("Percentage of all non-overlapping data (PAND)")
  }
  
  if (is.na(footnote)) {
    footnote <- c(
      paste0("PAND = ", round(object$pand, 1), "%"),
      paste0("\u03A6 = ", round(object$phi, 3)), 
      paste0("\u03A6\u00b2 = ", round(object$phi^2, 3)), 
      paste0("Number of cases: ", object$N), 
      sprintf("\u03C7\u00B2 = %.2f, df = 1, p = %.3f; ",
              object$chi_test$statistic, 
              object$chi_test$p.value
      ),
      sprintf(
        "Fisher exact test: Odds ratio = %.2f, p = %.3f",
        object$fisher_test$estimate, 
        object$fisher_test$p.value
      )
    )
  }
  
  object$matrix <- rbind(object$matrix, object$matrix[1,] + object$matrix[2,])
  object$matrix_counts <- rbind(
    object$matrix_counts, object$matrix_counts[1,] + object$matrix_counts[2,]
  )
  object$matrix <- cbind(object$matrix, object$matrix[,1] + object$matrix[,2])
  object$matrix_counts <- cbind(
    object$matrix_counts, object$matrix_counts[,1] + object$matrix_counts[,2]
  )  
  out <- as.data.frame(
    round(rbind(object$matrix * 100, object$matrix_counts), round)
  )
  out <- cbind(
    data.frame(
      " " = rep(c("Real", " ", " "), 2), Phase = rep(c("A", "B", "Total"), 2)
    ), 
    out
  )
  names(out) <- c(" ", "  ", "A", "B", "Total")

  opts <- options(knitr.kable.NA = "")
  options(scan.export.kable = c(
    list(align = c("l", "r", "c", "c", "c")), 
    getOption("scan.export.kable")
  ))
  
  
  table <- .create_table(
    out, 
    caption = caption,
    footnote = footnote,
    spanner = list("Expected" = 3:5),
    row_group = list("Percentage" = 1:3, "Counts" = 4:6)
  )
  
  if (getOption("scan.export.engine") == "kable") {
    style <- "border-bottom: 1px solid; text-align: center"
    table <- table  |> 
      add_header_above(c(" " = 2, "Expected" = 3))  |> 
      pack_rows(index = c("Percentage" = 3, "Counts" = 3), label_row_css = style)  |> 
      column_spec(1:2, bold = TRUE)
  }
  
  # finish ------------------------------------------------------------------
  
  if (!is.na(filename)) .save_export(table, filename)
  
  options(opts)
  
  table
}

test_that("pand", {
  
  data <- scdf(c(A = 20, 20, 26, 25, 22, 23, B = 28, 25, 24, 27, 30, 30, 29))
  result <- pand(data = data)
  expect_s3_class(result, "sc_pand")
  expect_identical(round(result$pand,5) ,84.61538)
  expect_equal(round(result$phi, 4) , 0.6905)
  
  result <- pand(Parker2009, method = "minimum")
  expect_equal(round(result$pand, 5) , 86.95652)
  result <- pand(Parker2009, method = "sort")
  expect_equal(result$pand , 100)
  
})

#' Rescales values of an scdf
#'
#' This function scales the measured values of an scdf file. It allows for mean
#' centering and standardization across all cases included in an scdf.
#'
#' @inheritParams .inheritParams
#' @param ... Names of variables to be standardized. If none are given, all numeric
#'  variables are standardized.
#' @param m The target mean. If set NULL, it is not changed.
#' @param sd The target standard deviation. If set NULL, it is not changed.
#' @return An scdf with the scaled values.
#' @author Juergen Wilbert
#' @family data manipulation functions
#' @keywords manip
#' @examples
#'
#' ## Standardize a multiple case scdf and compute an hplm
#' exampleAB_50 |>
#'   rescale(values, mt) |>
#'   hplm()
#' @export
rescale <- function(data, 
                    ...,
                    m = 0, 
                    sd = 1) {

  data <- .prepare_scdf(data)

  N <- length(data)
  
  vars <-  list(...) |> substitute() |> sapply(deparse)
  vars <- vars[-1]
  
  if (length(vars) == 0) {
    vars <- sapply(data[[1]], function(x) is.numeric(x)) |> which()
    vars <- names(data[[1]])[vars]
    message(paste0("Rescaled ", paste0(vars, collapse = ", ")))
  }
  
  m_sd <- vapply(vars, function(x) {
      y <- unlist(lapply(data, function(case) case[[x]]))
      c(mean(y, na.rm = TRUE), sd(y, na.rm = TRUE))
    },
    FUN.VALUE = double(2)
  ) 

  for (case in 1:N) {
    for (i in 1:length(vars)) {
      values <- data[[case]][, vars[i]]
      if (!is.null(m) && !is.null(sd)) {
        data[[case]][, vars[i]] <- (values - m_sd[1, i]) / m_sd[2, i] * sd + m
      }
      if (!is.null(m) && is.null(sd)) {
        data[[case]][, vars[i]] <- (values - m_sd[1, i]) + m
      }
      if (is.null(m) && !is.null(sd)) {
        data[[case]][, vars[i]] <- ((values - m_sd[1, i]) / m_sd[2, i] * sd) + m_sd[1, i]
      }
    }
  }

  data
}

#' Nonoverlap of all Pairs (NAP)
#'
#' The [nap()] function calculates the nonoverlap of all pairs (NAP; Parker &
#' Vannest, 2009).  NAP summarizes the overlap between all pairs of phase A and
#' phase B data points.  If an increase of phase B scores is expected, a
#' non-overlapping pair has a higher phase B data point. The NAP equals
#' *number of pairs showing no overlap / number of pairs* where ties are
#' counted as half non-overlaps.  Because NAP can take values between 0 and 100
#' percent where values below 50 percent indicate an inverse effect, an nap
#' rescaled from -100 to 100 percent where negative
#' values indicate an inverse effect is also displayed (\eqn{nap_{rescaled} = 2
#' * nap - 100}).
#' 
#' If a decrease of phase B scores is expected, set the argument
#' `decreasing = TRUE`.
#'
#' @inheritParams .inheritParams
#' @return 
#'  |  |  |
#'  | --- | --- |
#'  | `nap` | A data frame with NAP and additional values for each case. |
#'  | `N` | Number of cases. |
#'  
#' @author Juergen Wilbert
#' @family overlap functions
#' @references Parker, R. I., & Vannest, K. (2009). An improved effect size for
#'   single-case research: Nonoverlap of all pairs. *Behavior Therapy*, *40*,
#'   357-367.
#' @examples
#'
#' ## Calculate NAP for a study with  lower expected phase B scores
#' ## (e.g. aggressive behavior)
#' gretchen <- scdf(c(A = 12, 14, 9, 10, B = 10, 6, 4, 5, 3, 4))
#' nap(gretchen, decreasing = TRUE)
#'
#' ## Request NAP for all cases from the Grosche2011 scdf
#' nap(Grosche2011)
#' 
#' ## Calculate NAP for phase 1 and phase 3 of an ABAB design
#' nap(exampleABAB, phases = c(1, 3))
#' 
#'
#' @export

nap <- function(data, dvar, pvar,
                decreasing = FALSE,
                phases = c(1, 2)) {
  # set attributes to arguments else set to defaults of scdf
  if (missing(dvar)) dvar <- dv(data) else dv(data) <- dvar
  if (missing(pvar)) pvar <- phase(data) else phase(data) <- pvar
  
  data <- .prepare_scdf(data, na.rm = TRUE)
  data <- recombine_phases(data, phases = phases)$data
  
  casenames <- revise_names(data)
  
  .nap <- function(data) {
    
    values <- split(data[[dvar]], data[[pvar]])
    pairs <- length(values$A) * length(values$B)
    
    if (!decreasing) {
      pos <- sum(unlist(lapply(values$A, function(x) values$B > x)))
    }
    
    if (decreasing) {
      pos <- sum(unlist(lapply(values$A, function(x) values$B < x)))
    }
    
    ties <- sum(unlist(lapply(values$A, function(x) x == values$B)))
    
    non_overlaps <- pos + (0.5 * ties)
    
    nap <- non_overlaps / pairs
    
    test <- wilcox.test(
      values$A, values$B,
      alternative = if (decreasing) "greater" else "less",
      exact = FALSE
    )
    #nap <- (pairs - test$statistic) / pairs
    d <- 3.464 * (1 - sqrt((1 - nap) / 0.5))
    r <- d / sqrt(d^2 + 4)
    data.frame(
      NAP = nap * 100,
      "NAP Rescaled" = 2 * (nap * 100) - 100,
      Pairs = pairs,
      "Non-overlaps" = non_overlaps,
      Positives = pos,
      Ties = ties,
      w = test$statistic,
      p = test$p.value,
      d = d,
      "R\u00B2" = r^2,
      check.names = FALSE
    )
 
  }  
  
  x <- lapply(data, .nap)
  nap <- do.call(rbind, x)
  nap <- cbind(Case = casenames, nap)
  rownames(nap) <- NULL
  
  out <- list(nap = nap)
  class(out) <- c("sc_nap")
  attributes(out)[opts("phase", "dv")] <- list(pvar, dvar)
  out
}

#' Creating a long format data frame from several single-case data frames
#' (scdf).
#'
#' The as.data.frame function transposes an scdf into one long data frame.
#' Additionally, a data frame can be merged that includes level 2 data of the
#' subjects. This might be helpful to prepare data to be used with other
#' packages than scan.
#'
#' @param x An scdf object.
#' @param l2 A data frame providing additional variables at Level 2. The scdf
#'   has to have names for all cases and the Level 2 data frame has to have a
#'   column with corresponding case names.
#' @param id Variable name of the Level 2 data frame that contains the case
#'   names. Defaults to "case".
#' @param ... Not implemented.
#' @return Returns one data frame with data of all single-cases structured by
#'   the case variable.
#' @author Juergen Wilbert
#' @family data manipulation functions
#' @keywords manip
#' @examples
#'
#' ## Convert the list of three single-case data frames from Grosche (2011)
#' ### into one long data frame
#' Grosche2011
#' Grosche2011_long <- as.data.frame(Grosche2011)
#' Grosche2011_long
#'
#' ## Combine an scdf with data for l2
#' Leidig2018_long <- as.data.frame(Leidig2018, l2 = Leidig2018_l2)
#' names(Leidig2018_long)
#' summary(Leidig2018_long)
#'
#' @export
as.data.frame.scdf <- function(x, ..., l2 = NULL, id = "case") {
  if (!is.null(l2)) x <- add_l2(x, l2)

  label <- revise_names(x)
  outdat <- vector()

  for (i_case in seq_along(x)) {
    x[[i_case]]$case <- label[i_case]
    outdat <- rbind(outdat, x[[i_case]])
  }

  outdat <- cbind(outdat[, ncol(outdat)], outdat[, -ncol(outdat)])
  colnames(outdat)[1] <- id
  outdat[[1]] <- factor(outdat[[1]], levels = label, labels = label)
  attr(outdat, "scdf") <- attr(x, "scdf")
  outdat
}

#' @describeIn plm Print
#' @order 2
#' @inheritParams print.sc
#' @param lag_max Maximum lag to be reported for autocorrelation of residuals.
#'   Default is `3`. Set `FALSE` for no report of autocorrelations.
#' @param ci Print confidence intervals. Either FALSE, TRUE or a number 
#'   between 0 and 1 (0.90 for a 90% intervals).
#' @param q Logical. If set `TRUE`, Yule's Q is reported.
#' @param r_squared Either "delta", "partial", or "none".
#' @examples
#' ## Print
#' plm(exampleAB$Johanna) |> 
#'   print(ci = 0.9, r_squared = c("delta", "partial"))
#' @export
#' 
print.sc_plm <- function(x, 
                         lag_max = 3, 
                         ci = 0.95, 
                         q = FALSE, 
                         r_squared = getOption("scan.rsquared"), 
                         ...) {
  cat("Piecewise Regression Analysis\n\n")
  cat(
    "Contrast model: ", 
    x$model, " / ", 
    paste0("level = ", x$contrast$level, ", slope = " ,x$contrast$slope),
    "\n\n", sep = ""
  )
  
  cat("Fitted a", x$family, "distribution.\n")		
  
  results <- .output_plm(
    x, ci = ci, q = q, lag_max = lag_max, format = "print", 
    r_squared = r_squared
  )
  
  if (x$ar > 0) {
    cat("Autocorrelated residuals up to lag", x$ar, "were modeled\n\n")
  }
  
  cat(results$fit, "; AIC = ", results$aic, "\n\n", sep = "")
  
  print(results$table)
  cat("\n")
  
  if (!is.null(results$autocorrelation)) {
    cat("Autocorrelations of the residuals\n")
    print(results$autocorrelation, row.names = FALSE)
    cat(results$ljung, "\n\n")
  }
  
  cat("Formula:", results$formula, "\n")
}

#' @describeIn plm Export results as html table (see [export()])
#' @order 3
#' @inheritParams export
#' @param ci Print confidence intervals. Either FALSE, TRUE or a number 
#'   between 0 and 1 (0.90 for a 90% intervals).
#' @param q Logical. If set `TRUE`, Yule's Q is reported.
#' @param r_squared Either "delta", "partial", or "none".
#' @examples
#' ## Export
#' plm(exampleAB$Johanna) |> export()
#' 
#' @export
export.sc_plm <- function(object, 
                          caption = NA, 
                          footnote = NA, 
                          filename = NA,
                          nice = TRUE,
                          ci = 0.95,
                          q = FALSE,
                          round = 2,
                          r_squared = getOption("scan.rsquared"), 
                          ...) {
  
  if (is.na(caption)) {
    caption <- paste0(
      "Piecewise-regression model predicting '", attr(object, opt("dv")), "'"
    )
  }
  
  results <- .output_plm(
    object, ci = ci, q = q, format = "export", 
    round = round,
    r_squared = r_squared
  )
 
  out <- results$table
  out <- rownames_to_first_column(out, "Parameter")
  
  if (nice) out$p <- .nice_p(out$p)
  
  if (is.na(footnote)) footnote <- c(
    paste0(results$fit), 
    paste0("AIC = ", round(results$aic)),
    "LL = lower limit",
    "UL = upper limit",
    str_contrasts(object$model, object$contrast)
  )
  
  if (getOption("scan.export.engine") == "gt") {
    spanner <- list("CI" = 3:4)
    if (object$family %in% c("poisson", "nbinomial")) {
      spanner[[" CI "]]  <- 9:10
      if (q) spanner[["  CI  "]]  <- 12:13
    }
    names(spanner) <- gsub(
      "CI", paste0("CI(", ci * 100, "%)"), x = names(spanner)
    )
  }

  table <- .create_table(
    out, 
    caption = caption,
    footnote = footnote,
    spanner = spanner
  )
  
  if (getOption("scan.export.engine") == "kable") {
    spanner <- c(" " = 2, "CI" = 2, " " = 4)
  
    if (object$family %in% c("poisson", "nbinomial")) {
      spanner <- c(spanner, "CI" = 2)
      if (q) spanner <- c(spanner, " " = 1, "CI" = 2)
    }
    
    names(spanner) <- gsub(
      "CI", paste0("CI(", ci * 100, "%)"), x = names(spanner)
    )
    
    table <- add_header_above(table, spanner)
    
  }

  
  # finish ------------------------------------------------------------------
  
  if (!is.na(filename)) .save_export(table, filename)
  
  table
}

.output_plm <- function(x, 
                        ci, q, 
                        lag_max = 0, 
                        format, 
                        round = 3,
                        r_squared) {
  
  out <- list()
  
  report_r_squared <- if (is.null(x$r.squares) || identical(x$r.squares, NA)) 
    FALSE else TRUE
  
  out$aic <- x$full.model$aic
  if (!is.numeric(out$aic)) out$aic <- NA
  if (x$family == "poisson" || x$family == "binomial") {
    chi <- x$full$null.deviance - x$full$deviance
    df <- x$full$df.null - x$full$df.residual
    out$fit <- sprintf(
      "X\u00b2(%d) = %.2f; p = %0.3f", 
      df, chi, 1 - pchisq(chi, df = df)
    )
  } else {
    out$fit <- if (x$F.test["df1"] == 0) {
      "Null model" 
    } else {
      sprintf(
       "F(%d, %d) = %.2f; p = %0.3f; R\u00b2 = %0.3f; Adjusted R\u00b2 = %0.3f",
       x$F.test["df1"], x$F.test["df2"], x$F.test["F"],
       x$F.test["p"],   x$F.test["R2"],  x$F.test["R2.adj"]
      )
    }
  }
  
  ## coef table ----
  
  model_summary <- summary(x$full.model)
  
  if (x$ar == 0) out$table <- model_summary$coefficients
  if (x$ar  > 0) out$table <- model_summary$tTable
  
  inter_incl <- if (attr(x$full.model$terms, "intercept")) TRUE else FALSE
 
  out$table <- round(out$table, round)
  out$table <- as.data.frame(out$table)
  names(out$table)[1:4] <- c("B", "SE", "t", "p")
  
  
  if (identical(ci, FALSE)) {
    if (x$family == "poisson" || x$family == "binomial") {
      OR <- exp(out$table[, "B"])
      out$table <- cbind(out$table[, c("B", "SE", "t", "p")],  "OR" = round(OR$table, 3))
      if (q) {
        Q <- (OR - 1) / (OR + 1)
        out$table <- cbind(out$table, "Q" = round(Q, 2))
      }
    }
  }
  
  ## ci ----
  if (!identical(ci, FALSE)) {
    if (isTRUE(ci)) ci <- 0.95
    
    #str_ci <- paste0(round(c((1 - ci) / 2, ci + ((1 - ci) / 2)) * 100, 2), "%")
    
    if (format == "print") str_ci <- str_ci(ci)
    if (format == "export") str_ci <- c("LL", "UL")
    
    ci <- suppressMessages(confint(x$full, level = ci))
    
    if (length(ci) == 2) ci <- matrix(ci, ncol = 2)
    param_filter <- apply(
      ci, 1, function(x) if(!all(is.na(x))) TRUE else FALSE
    )
    ci <- round(ci[param_filter, ], round)
    if (nrow(out$table) == 1) ci <- t(ci)
    out$table <- cbind(
      out$table[, 1, drop = FALSE], 
      ci, 
      out$table[, 2:4, drop = FALSE]
    )
    
    names(out$table) <- c("B", str_ci, "SE", "t", "p")
    
    if (x$family == "poisson" || x$family == "binomial") {
      OR <- exp(out$table[, 1:3])
      out$table <- cbind(out$table[, 1:6], round(OR, round))
      names(out$table)[7:9] <- c("OR", paste0(" ", str_ci)) 
      
      if (q) {
        Q <- (OR - 1) / (OR + 1)
        out$table <- cbind(out$table, round(Q, round))
        names(out$table)[10:12] <- c("Q", paste0("  ", str_ci))  
      }
      
    }
  } 
  
  ## r squared ----
  
  if ("delta" %in% r_squared && report_r_squared) {
    x$r.squares <- x$r.squares[c(param_filter[-1])]
    out$table[["delta R\u00b2"]] <- c(
      rep("", inter_incl), format(round(x$r.squares, round))
    )
  }

  if ("partial" %in% r_squared && report_r_squared) {
    r_squared_partial <- out$table$t^2 /(out$table$t^2 + model_summary$df.residual)
    if (inter_incl) r_squared_partial <- r_squared_partial[-1]
    out$table[["partial R\u00b2"]] <- c(
      rep("", inter_incl), format(round(r_squared_partial, round))
    )
  }

  rownames(out$table) <- rename_predictors(rownames(out$table), x)
  
  ## autocorrelation ----
  
  if (x$family == "gaussian" && lag_max > 0) {
    
    cr <- acf(
      residuals(x$full.model), 
      lag.max = lag_max,plot = FALSE
    )$acf[2:(1 + lag_max)]
    
    cr <- round(cr, 2)
    out$autocorrelation <- data.frame(lag = 1:lag_max, cr = cr)
    
    bj <- Box.test(residuals(x$full.model), lag_max, type = "Ljung-Box")
    out$ljung <- sprintf(
      "Ljung-Box test: X\u00b2(%d) = %.2f; p = %0.3f", 
      bj$parameter, bj$statistic, bj$p.value
    )
  }
  
  ## formula ----
  
  if (x$family == "binomial" && !x$dvar_percentage) {
    x$formula[2] <- str2expression(paste0(x$formula[2], "/", x$var_trials))
  }
  out$formula <- deparse(x$formula)
  if (x$family == "binomial") {
    out$formula <- paste0(out$formula, "\n", "weights = ", x$var_trials)
  }
  out
}

recombine_phases <- function(data, 
                             phases = c(1, 2), 
                             set_phases = TRUE, 
                             phase_names = c("A", "B"),
                             pvar) {
      
  if (missing(pvar)) pvar <- phase(data)
  
  source_attributes <- attributes(data)
  
  warning <- character(0)
  
  if (inherits(phases, c("character", "numeric", "integer"))) {
    if (!length(phases) == 2) {
      abort("phases argument not set correctly.")
    }    
    phases_A <- phases[1]
    phases_B <- phases[2]
  }
  
  if (inherits(phases, "list")) {
    phases_A <- phases[[1]]
    phases_B <- phases[[2]]
  }
  
  phases_total <- c(phases_A, phases_B)
  
  N <- length(data)
  design_list <- list()
  dropped_cases <- numeric(0)
  
  for(case in 1:N) {
    
    design <- rle(as.character(data[[case]][[pvar]]))
    
    if (inherits(phases_total, "character")) {
      select_A <- which(design$values %in% phases_A)
      select_B <- which(design$values %in% phases_B)
    } else {
      select_A <- phases_A
      select_B <- phases_B
    }
    
    
    if (!inherits(phases_total, "character")) {
      if (any(phases_total > length(design$values))) {
        warning <- c(
          warning, 
          paste0("Phase(s) not found. Case ", case, " dropped.\n")
        )
        dropped_cases <- c(dropped_cases, case)
        next
      }
    }
    
    if (inherits(phases_total, "character")) {
      
      tmp <- sapply(phases_total, function(x) sum(x == design$values) > 1)
      if (any(tmp)) {
        abort(paste0(
          "Selected phase ", paste(names(tmp[tmp])), " occure several times. ",
          "Please give number of phases instead of characters."
        ))
      }
      
      tmp <- sapply(phases_total, function(x) any(x == design$values))
      if (!all(tmp)) {  
        warning <- c(
          warning, 
          paste0("Phase(s) ",  names(tmp[!tmp]), " not found. Case ", case, 
            " dropped.\n")
        )
      
        dropped_cases <- c(dropped_cases, case)
        next
      }
      
    }
    
    design$start <- c(1, cumsum(design$lengths) + 1)[1:length(design$lengths)]
    design$stop <- cumsum(design$lengths)
    class(design) <- "list"
    
    A <- unlist(lapply(select_A, function(x) design$start[x]:design$stop[x]))
    B <- unlist(lapply(select_B, function(x) design$start[x]:design$stop[x]))
    
    data[[case]][[pvar]] <- as.character(data[[case]][[pvar]])
    
    if (set_phases) {
      if (identical(phase_names, "auto")) {
        phase_names <- sapply(phases, function(x) {
          if (is.numeric(x)) {
            paste0(design$values[x], collapse = "")
          } else {
            paste0(x, collapse = "")
          }
        })          
      }
      data[[case]][A ,pvar] <- phase_names[1]
      data[[case]][B ,pvar] <- phase_names[2]
    }
    data[[case]] <- data[[case]][c(A, B),]
    design_list[[case]] <- design
  }
  
  if (length(warning) > 0) warn(paste0(warning, collapse = "  "))
  if (length(dropped_cases > 0)) {
    data <- data[-dropped_cases]
    design_list <- design_list[-dropped_cases]
    source_attributes$names <- source_attributes$names[-dropped_cases]
    if (length(data) == 0) abort("No case remained.")
  }
  
  attributes(data) <- source_attributes
  
  original_phases <- lapply(design_list, function(x) x$values)
  
  new_phases <- lapply(original_phases, function(x) {
    out <- if (is.numeric(phases_A)) {
      paste0(x[phases_A], collapse = "")
    } else {
      paste0(x[which(phases_A %in% x)], collapse = "")
    }
    out <- if (is.numeric(phases_B)) {
      c(out, paste0(x[phases_B], collapse = ""))
    } else {
      c(out, paste0(x[which(phases_B %in% x)], collapse = ""))
    }
    out
  })
  
  #browser()
  
  out <- list(
    data = data, 
    designs = design_list, 
    N = N, 
    phases_A = phases_A, 
    phases_B = phases_B,
    phases = list(original = original_phases, new = new_phases)
  )
  class(out) <- c("sc_keepphases")
  
  out
}

#' @describeIn cdc Print results
#' @order 2
#' @inheritParams print.sc
#' @export
print.sc_cdc <- function(x, nice = TRUE, ...) {
  
  cat("Conservative Dual Criterion\n\n")
  cat("N cases = ", x$N, "\n\n")
  
  if (nice) x$cdc_p <- .nice_p(x$cdc_p)
  out <- data.frame(
    Case = x$case_names,
    "nB improve" = x$cdc_be,
    "nB" = x$cdc_b,
    "binom p" = x$cdc_p,
    "CDC Evaluation" = x$cdc,
    check.names = FALSE
  )
  print(out, row.names = FALSE)
  cat("\n")
  if (x$decreasing) {
    cat("Assuming an expected decrease in phase B.\n")
    cat("Alternative hypothesis (Binomial test): true probability < 50%\n")
  } else {
    cat("Assuming an expected increase in phase B.\n")
    cat("Alternative hypothesis (Binomial test): true probability > 50%\n")
  }
  if (x$N > 1) {
    cat("Overall evaluation of all MBD instances:  ",x$cdc_all,"\n")
  }
  
}

#' @describeIn cdc Export html results
#' @order 3
#' @inheritParams export
#' @export
export.sc_cdc <- function(object, 
                          caption = NA, 
                          footnote = NA, 
                          filename = NA,
                          nice = TRUE,
                          ...) {
  
  
  if (is.na(caption)) {
    caption <- paste0(
      "Conservative Dual Criterion for variable '", 
      attr(object, opt("dv")),  "'"
    )
  }
  
  if (is.na(footnote)) {
    if (object$decreasing) {
      footnote <- c(
        "Assuming an expected decrease in phase B.",
        "Alternative hypothesis (Binomial test): true probability < 50%"
      )
    } else {
      footnote <- c(
        "Assuming an expected increase in phase B",
        "Alternative hypothesis (Binomial test): true probability > 50%"
      )
    }
    if (object$N > 1) {
      footnote <- c(
        footnote, 
        paste0("Overall evaluation of all MBD instances:  ",object$cdc_all)
      )
    }
  }
  
  if (nice) object$cdc_p <- .nice_p(object$cdc_p)
  out <- data.frame(
    Case = object$case_names,
    "nB improve" = object$cdc_be,
    "nB" = object$cdc_b,
    "binom p" = object$cdc_p,
    "CDC Evaluation" = object$cdc,
    check.names = FALSE
  )
  
  table <- .create_table(
    out,
    caption = caption,
    footnote = footnote,
    ...
  )
  
  if (!is.na(filename)) .save_export(table, filename)
  
  table
  
}

#' Fetches elements from scan objects
#'
#' @description The fetch function is a getter function for scan objects
#' returned from regression functions such as `plm()`, `hplm()`, `bplm()`, and
#' `mplm()`. It allows users to extract specific elements from these objects,
#' such as the fitted model.
#'
#' @aliases fetch.sc_plm fetch.sc_hplm fetch.sc_bplm fetch.sc_mplm
#' @param object Object returned from a scan function.
#' @param what Element/part to be extracted. Currently, only "model" is
#'   supported to extract the fitted regression model.
#' @param ... Further parameters passed to the function.
#' @return An object of the respective regression model class.
#' @author Juergen Wilbert
#' @family regression functions
#' @keywords regression
#' @examples
#' # plm regression
#' model1 <- plm(example_A24)
#' fetch(model1, what = "model") |> summary()
#' # Multilevel plm regression
#' model2 <- hplm(exampleAB_50)
#' fetch(model2, what = "model") |> summary()
#' # Bayesian plm regression
#' model3 <- bplm(exampleAB_50, nitt = 5000)
#' fetch(model3, what = "model") |> summary()
#' @export
fetch <- function (object, what, ...) {
  
  UseMethod("fetch")
  
}

#' @exportS3Method fetch sc_plm
fetch.sc_plm <- function(object, what = "model", ...) {
  if (what == "model") {
    return(object$full.model)
  }
}


#' @exportS3Method fetch sc_hplm
fetch.sc_hplm <- function(object, what = "model", ...) {
  if (what == "model") {
    return(object$hplm)
  }
  
}


#' @exportS3Method fetch sc_bplm
fetch.sc_bplm <- function(object, what = "model", ...) {
  if (what == "model") {
    return(object$mcmcglmm)
  }
}

#' @exportS3Method fetch sc_mplm
fetch.sc_mplm <- function(object, what = "model", ...) {
  if (what == "model") {
    return(object$full.model)
  }
}

#' @describeIn autocorr Print method
#' @order 2
#' @param x An object returned by [autocorr()]
#' @export
#' @inheritParams print.sc
print.sc_ac <- function(x, digits = "auto", ...) {
  
  if (digits == "auto") digits <- 2
  
  cat("Autocorrelations\n\n")
  
  x <- x$autocorr
  for (i in 1:length(x)) {
    x[[i]][, -1] <- round(x[[i]][, -1], digits)
    cat(names(x)[i], "\n")
    print(x[[i]], row.names = FALSE)
    cat("\n")
  }
}

#' @describeIn autocorr Export results to html
#' @order 3
#' @inheritParams export
#' @export
export.sc_ac <- function(object, 
                         caption = NA, 
                         footnote = NA, 
                         filename = NA,
                         round = 3,
                         ...) {
  
  if (is.na(caption)) {
    caption <- paste0(
      "Autocorrelations for variable '", 
      attr(object, opt("dv")),  "'"
    )
  }
 
  x <- object$autocorr
  out <- do.call(rbind, object$autocorr)
  rows <- sapply(object$autocorr, nrow)
  end <- cumsum(sapply(object$autocorr, nrow)) 
  start <- end - rows + 1
  row_group <- mapply(function(start, end) start:end, start, end, SIMPLIFY = FALSE)
  names(row_group) <- names(object$autocorr)
  out <- round_numeric(out, round)
  table <- .create_table(
    out,
    caption = caption,
    footnote = footnote,
    row_group = row_group
  )
  
  if (!is.na(filename)) .save_export(table, filename)
  
  table
}

#' Percentage of non-overlapping data (PND)
#' 
#' This function returns the percentage of non-overlapping data.  Due to its
#' error-proneness the PND should not be used, but \code{\link{nap}} or
#' \code{\link{pand}} instead (see Parker & Vannest, 2009).
#' 
#' PND is calculated by determining the number of data points in phase B that
#' exceed the highest data point in phase A (or are lower than the lowest data
#' point in phase A for decreasing data) divided by the total number of data
#' points in phase B. This value is then multiplied by 100 to get a percentage
#' value.
#' 
#' @inheritParams .inheritParams
#' @param decreasing If you expect data to be lower in the B phase, set
#' \code{decreasing = TRUE}. Default is \code{decreasing = FALSE}.
#' @return \item{PND}{Percentage of non-overlapping data.}
#' @author Juergen Wilbert
#' @family overlap functions
#' @examples
#' 
#' ## Calculate the PND for multiple single-case data
#' pnd(GruenkeWilbert2014)
#' 
#' @export
pnd <- function(data, dvar, pvar, decreasing = FALSE, phases = c(1, 2)) {

  # set attributes to arguments else set to defaults of scdf
  if (missing(dvar)) dvar <- dv(data) else dv(data) <- dvar
  if (missing(pvar)) pvar <- phase(data) else phase(data) <- pvar
  
  data <- .prepare_scdf(data, na.rm = TRUE)
  data <- recombine_phases(data, phases = phases)$data
  
  pnd <- c()
  n.B <- c()
  
  for(i in 1:length(data)) {
    A <- data[[i]][, dvar][data[[i]][, pvar] == "A"]
    B <- data[[i]][, dvar][data[[i]][, pvar] == "B"]
    n.B[i] <- length(B)
    if (!decreasing) pnd[i] <- sum(B > max(A)) /  n.B[i] * 100
    if (decreasing) pnd[i] <- sum(B < min(A)) /  n.B[i] * 100
  }
  
  out <- list(PND = pnd, case.names = names(data), n.B = n.B)
  class(out) <- c("sc_pnd")
  attributes(out)[opts("phase", "dv")] <- list(pvar, dvar)
  out
}

#' Add level-2 data to an scdf
#'
#' Merges variables with corresponding case names from a data.frame with an scdf.
#'
#' This function is mostly used in combination with the [hplm()] function.
#' It adds level-2 variables to each single-case data frame in an scdf based on
#' matching case names.
#'
#' @inheritParams .inheritParams
#' @param data_l2 A level 2 dataset.
#' @param cvar Character string with the name of the "case" variable in the L2
#'   dataset (default is 'case').
#' @return An scdf with added level-2 variables.
#' @seealso [hplm()]
#' @family data manipulation functions
#' @keywords transform
#' @examples
#' ## Example with the default case variable name 'case'
#' Leidig2018 |> add_l2(Leidig2018_l2)
#' ## Example with a different case variable name in the L2 data
#' Leidig2018_l2_renamed <- Leidig2018_l2
#' names(Leidig2018_l2_renamed)[2] <- "subject"
#' Leidig2018 |> add_l2(Leidig2018_l2_renamed, cvar = "subject")
#' @author Juergen Wilbert
#' @export
add_l2 <- function(scdf,
                   data_l2,
                   cvar = "case") {
  
  check_args(
    by_class(scdf, "scdf"),
    by_class(data_l2, "data.frame"),
    by_class(cvar, "character"),
    is_true(!is.null(names(scdf)), "scdf must have casenames."),
    is_true(any(names(data_l2) == cvar), paste0("cvar '", cvar, "' not found in data_l2.")),
    is_true(!any(is.na(names(scdf))), "scdf has missing casenames.")
  )
  
  for (i in seq_along(scdf)) {

    casename <- names(scdf)[i]
    id <- which(data_l2[[cvar]] == casename)

    if (length(id) == 0) warn("No matching case in L2 data found for ", casename)
    if (length(id) > 1) abort("Multiple matches for a casename in the L1 dataset for ", casename)

    if (length(id) == 1) {
      scdf[[i]] <- cbind(
        scdf[[i]],
        data_l2[id, -which(names(data_l2) == cvar), drop = FALSE],
        row.names = NULL
      )
    }
  }

  scdf
}

test_that("combination", {
  
  scdf <- exampleABC
  
  l2 <- data.frame(
    case = c("Marie", "Rosalind", "Lise"),
    gender = c(0,0,2),
    age = c(23,35,12)
  )
  combined <- scdf |> add_l2(l2)
  
  expect_equal(length(unlist(combined)), 450)
  
})

test_that("error", {
  
  scdf <- exampleABC
  l2 <- data.frame(
    case = c("Marie", "Heinz", "Lise"),
    gender = c(0,0,2),
    age = c(23,35,12)
  )
  
  expect_warning(scdf |> add_l2(l2))
  
  l2 <- data.frame(
    name = c("Marie", "Rosalind", "Lise"),
    gender = c(0,0,2),
    age = c(23,35,12)
  )
  
  expect_error(scdf$Marie |> add_l2(l2[1,]))
  
})

kendall_tau <- function(x, y, 
                        tau_method = "b", 
                        continuity_correction = TRUE) {
  
  N <- length(x)
  
  if (N < 3) {
    warn("could not calculate p-values for tau. Less than three data points.")
  }
  
  if (all(x == x[1]) || all(x == x[2])) {
    warn("could not calculate tau. Variance is zero.")
  }
  
  .sort <- sort.list(x)
  x <- x[.sort]
  y <- y[.sort]
  C <- 0
  D <- 0
  
  for(i in 1:(N - 1)) {
    C <- C + sum(y[(i + 1):N] > y[i] & x[(i + 1):N] > x[i])
    D <- D + sum(y[(i + 1):N] < y[i] & x[(i + 1):N] > x[i])
  }
  
  tie_x <- rle(x)$lengths
  tie_y <- rle(sort(y))$lengths
  
  ti <- sum(vapply(tie_x, FUN = function(x) x * (x - 1) / 2, FUN.VALUE = numeric(1)))
  ui <- sum(vapply(tie_y, FUN = function(x) x * (x - 1) / 2, FUN.VALUE = numeric(1)))
  
  S  <- C - D
  n0 <- N * (N - 1) / 2
  
  if (tau_method == "a") {
    Den <- n0
    tau   <- S / Den
    se <- sqrt((2 * N + 5) / Den) / 3
    
    #out$varS <- (2 * (2 * N + 5)) / (9 * N * (N - 1))
    #out$sdS <- sqrt(out$varS)
    
    sdS <- S / (3 * S / sqrt( N * (N - 1) * (2* N + 5) / 2 ))
    varS <- sdS^2
    if (!continuity_correction) {
      z <- 3 * S / sqrt( N * (N - 1) * (2* N + 5) / 2 )
    }
    if (continuity_correction)  {
      z <- 3 * (sign(S) * (abs(S) - 1)) / sqrt( N * (N - 1) * (2* N + 5) / 2 ) 
    }
    
  }
  
  if (tau_method == "b") {
    
    Den <- sqrt( (n0 - ti) * (n0 - ui) )
    tau <- S / Den
    
    v0 <- N * (N - 1) * (2 * N + 5)
    vt <- sum(vapply(tie_x, function(x) (x * (x - 1)) * (2 * x + 5), FUN.VALUE = numeric(1)))
    vu <- sum(vapply(tie_y, function(x) (x * (x - 1)) * (2 * x + 5), FUN.VALUE = numeric(1)))
    v1 <- sum(vapply(tie_x, function(x) (x * (x - 1)), FUN.VALUE = numeric(1))) * 
      sum(vapply(tie_y, function(x) (x * (x - 1)), FUN.VALUE = numeric(1)))
    v2 <- sum(vapply(tie_x, function(x) (x * (x - 1)) * (x - 2), FUN.VALUE = numeric(1))) * 
      sum(vapply(tie_y, function(x) (x * (x - 1)) * (x - 2), FUN.VALUE = numeric(1)))
    
    varS <- (v0 - vt - vu) / 18 + 
      (v1 / (2 * N * (N - 1))) +  
      (v2 / (9 * N * (N - 1) * (N - 2)))
    
    sdS <- sqrt(varS)
    se  <- sdS / Den
    
    if (!continuity_correction) z <- S / sdS #out$tau.b / out$se
    if (continuity_correction)  z <- (sign(S) * (abs(S) - 1)) / sdS 
  }
  
  #out$tau   <- S / n0
  #out$tau.b <- S / sqrt( (n0 - ti) * (n0 - ui) )
  #out$D <- out$S / out$tau.b
  #out$varS_tau_a <- (2 * (2 * N + 5)) / (9 * N * (N - 1))
  #out$sdS_tau_a <- sqrt((2 * (2 * N + 5)) / (9 * N * (N - 1)))

  p <- pnorm(abs(z), lower.tail = FALSE) * 2
  
  if (is.infinite(z)) {
    p <- NA
    tau <- NA
  }
  
  list(
    N  = N,
    n0 = n0,
    ti = ti,
    ui = ui,
    nC = C,
    nD = D,
    S  = S,
    D = Den,
    tau = tau,
    varS = varS,
    sdS = sdS,
    se = se,
    z = z,
    p = p
  )
  
}

#' @rdname export
#' @export
export.scdf_summary <- function(object, 
                                caption = NA, 
                                footnote = NA, 
                                filename = NA,
                                round = 2, 
                                ...) {
  
  if (is.na(footnote)) {
    footnote <- c(
      if (!is.null(scdf_attr(object, "info"))) scdf_attr(object, "info"),
      if (!is.null(scdf_attr(object, "author"))) paste("Author:", scdf_attr(object, "author"))
    )
    
    names <- names(object[[1]])
    id_dv <- which(names == dv(object))
    id_phase <- which(names == phase(object))
    id_mt <- which(names == mt(object))
    names[id_phase] <- paste(names[id_phase], "(phase variable)")
    names[id_mt] <- paste(names[id_mt], "(measurement-time variable)")
    names[id_dv] <- paste(names[id_dv], "(dependent variable)")
    str_vars <- paste(names[c(
      id_dv, id_phase, id_mt, (1:length(names))[-c(id_dv, id_phase, id_mt)]
    )], collapse = "<br>")
    footnote <- c(footnote, paste0("<br><br><b>Variable names:</b><br>", str_vars, collapse = "<br>")
    )
  }
  
  N <- cases <- length(object)
  
  if (is.na(caption)) {
    caption <- if(N > 1) {
      paste("A single-case data frame with", number_word(N), "cases")
    } else {
      paste("A single-case data frame with one case")
    }
  }
  
  designs <- lapply(object, function(x) {
    paste0(
      rle(as.character(x[[phase(object)]]))$values, "(",
      rle(as.character(x[[phase(object)]]))$lengths, ")",
      collapse = "-")
  })
  
  rows <- lapply(object, nrow)
  
  out <- data.frame(
    Case = format(revise_names(object), justify = "left")[1:min(N, 10)],
    Measurements = unname(unlist(rows))[1:min(N, 10)], 
    Design = unname(unlist(designs))[1:min(N, 10)],
    check.names = FALSE
  )
  
  opts <- options()
  options(scan.export.kable = c(
    list(align = rep("c", ncol(out))), 
    getOption("scan.export.kable")
  ))
  
  table <- .create_table(
    out, 
    caption = caption,
    footnote = footnote,
    ...
  )
  
  # finish ------------------------------------------------------------------
  
  if (!is.na(filename)) .save_export(table, filename)
  
  options(opts)
  
  table
}

#' Extract coefficients from plm/hplm objects
#' 
#' @param object plm or hplm object
#' @param ... not implemented
#'
#' @return data frame with coefficient table
#' @export
#'
#' @examples
#' coefficients(plm(exampleAB$Johanna))
#' 
coef.sc_plm <- function(object, ...) {
  summary(object$full.model)$coef
}

#' @describeIn hplm Extract model coefficients
#' @order 4
#' @param casewise Returns the estimations for each case separately
#' @inheritParams coef.sc_plm
#' @export
coef.sc_hplm <- function(object, casewise = FALSE, ...) {
  if (casewise) {
    out <- coef(object$hplm)
    names(out) <- rename_predictors(names(out), object)
    out <- rownames_to_first_column(out, "Case")
    return(out)
  } else {
    summary(object$hplm)$tTable
  }
}

.onAttach <- function(lib, pkg, ...) {
  packageStartupMessage(
    "\033[34m", 
    opt("startup_message"),
    "\033[31m",
    opt("tip")[[sample(1:length(opt("tip")), 1)]]
  )
  
}	

.onLoad <- function(lib, pkg, ...) {
  # global options ----------------------------------------------------------
  op <- options()
  op_scan <- list(
    scan.print.cases = "fit",
    scan.print.rows   = 15,
    scan.print.cols   = "all",
    scan.print.digits = 2,
    scan.print.long   = FALSE,
    scan.print.bar   = "\u2502", #"|",
    scan.print.scdf.name = TRUE,
    scan.plot.style = "grid",
    scan.rename.predictors = "full",
    scan.rsquared = "delta",
    scan.string.dummy.slope = "inter",
    scan.string.dummy.phase = "phase",
    scan.deprecated.warning = TRUE,
    scan.export.engine = "kable",
    scan.export.footnote.collapse = "; ",
    scan.export.title.prefix = md("**Table**"),
    scan.export.kable = list(digits = 2, linesep ="", booktab = TRUE),
    scan.export.kable_styling = list(
      bootstrap_options = c("bordered", "condensed"), 
      full_width = FALSE, position = "left",
      latex_options = c("hold_position"),
      htmltable_class = "lightable-classic"
    ),
    scan.shiny.theme = "cerulean"
  )
  
  toset <- !(names(op_scan) %in% names(op))
  if (any(toset)) options(op_scan[toset])
  
  invisible()
  
}



.onAttach()

#' @rdname print.sc
#' @export
#' 
print.sc_trend <- function(x, digits = 3, ...) {
  x$trend <- round(x$trend, digits)
  cat("Trend for each phase\n\n")
  print(x$trend)
  cat("\n")
  cat("Note. Measurement-times start at", x$first_mt, "for each phase\n")
  .note_vars(x)
}

#' @describeIn rand_test Print results
#' @order 2
#' @param x An object returned by [rand_test()]
#' @export
#' @inheritParams print.sc
print.sc_rand <- function(x, ...) {
  
  cat("Randomization Test\n\n")
  if (x$N > 1) cat("Combined test for", number_word(x$N), "cases.\n\n")
  
  cat(.phases_string(x$phases.A, x$phases.B), "\n")
  
  cat("Statistic: ", x$statistic, "\n\n")
  
  if (is.na(x$startpoints[1])) {
    cat("Minimal length of each phase:", "A =", x$limit[1], ", B =", x$limit[2], "\n")
  } else {
    cat("Possible starting points of phase B: ", x$startpoints, "\n")
  }
  cat("Observed statistic = ", x$observed.statistic, "\n")
  
  if (x$auto.corrected.number) {
    cat("\nWarning! The assigned number of random permutations exceeds the",
        "number of possible permutations.", 
        "\nAnalysis is restricted to all possible permutations.\n")
  }
  if (x$complete) {
    cat("\nDistribution based on all", x$possible.combinations, 
        "possible combinations.\n")
  } else 
    cat("\nDistribution based on a random sample of all", 
        x$possible.combinations, 
        "possible combinations.\n")
  
  cat("n   = ", x$number,"\n")
  cat("M   = ", mean(x$distribution),"\n")
  cat("SD  = ", sd(x$distribution),"\n")
  cat("Min = ", min(x$distribution),"\n")
  cat("Max = ", max(x$distribution),"\n")
  cat("\n")
  #cat("Probability of observed statistic based on distribution:\n")
  
  
  if (x$testdirection == "greater") {
    if (x$exclude.equal) cat("Probability of a higher value than the observed statistic:\n")
    if (!x$exclude.equal) cat("Probability of an equal or higher value than the observed statistic:\n")  
  } else {
    if (x$exclude.equal) cat("Probability of a lower value than the observed statistic:\n")
    if (!x$exclude.equal) cat("Probability of an equal or lower value than the observed statistic:\n")  
  }
  
  if (x$p.value == 0) {
    cat("p   < ", format(1/x$number, scientific = FALSE), "\n")
  } else {
    cat("p   = ", x$p.value, "\n")
  }
  
  if (x$number > 3 && x$number < 5001) {
    sh <- shapiro.test(x$distribution)
    cat(sprintf("\nShapiro-Wilk Normality Test: W = %0.3f; p = %0.3f",sh[[1]], sh$p.value))
    if (sh$p.value > .05) {
      cat("  (Hypothesis of normality maintained)\n")
    } else {
      cat("  (Hypothesis of normality rejected)\n")
    }
  } else {
    cat("\nSample size must be between 3 and 5000 to perform a Shapiro-Wilk Test.\n")
  }
  
  cat("\nProbabilty of observed statistic based on the assumption of normality:\n")
  cat(sprintf("z = %0.4f, p = %0.4f (single sided)\n", x$Z, x$p.Z.single))
  
}

#' @describeIn rand_test Export html results
#' @order 3
#' @inheritParams export
#' @export
export.sc_rand <- function(object, 
                          caption = NA, 
                          footnote = NA, 
                          filename = NA,
                          ...) {
  
  if (is.na(caption)) {
    caption <- paste0(
      "Randomization Test for variable '", 
      attr(object, opt("dv")),  "'"
    )
  }
  
  if (is.na(footnote)) {
  }
  
  out <- capture.output(
    print(object)
  ) [-1:-2]
  
  out <- data.frame(
    "Randomization test" = gt::html(paste(out, collapse = "  <br>")),
    check.names = FALSE
  )
  table <- .create_table(
    out,
    caption = caption,
    footnote = footnote,
    fmt_markdown = TRUE,
    ...
  )
  
  if (!is.na(filename)) .save_export(table, filename)
  
  table
  
}

#' Set analysis variables in an scdf object
#' 
#' This function allows to set or change the dependent variable,
#' measurement-time variable, and phase variable in an scdf object.
#'
#' @inheritParams .inheritParams
#' @param dvar Character string. Name of the dependent variable.
#' @param mvar Character string. Name of the measurement-time variable.
#' @param pvar Character string. Name of the phase variable.
#' @family data manipulation functions
#' @return An \code{scdf} object with updated variable settings.
#' @author Juergen Wilbert
#' @examples 
#' exampleAB_add |>
#'   set_dvar("depression") |>
#'   describe()
#' @export
set_vars <- function(data, dvar, mvar, pvar) {
  if (!missing(dvar)) dv(data) <- dvar
  if (!missing(mvar)) mt(data) <- mvar
  if (!missing(pvar)) phase(data) <- pvar
  
  data
}

#'@rdname set_vars
#'@export
set_dvar <- function(data, dvar) {
  dv(data) <- dvar
  data
}

#'@rdname set_vars
#'@export
set_mvar <- function(data, mvar) {
  mt(data) <- mvar
  data
}

#'@rdname set_vars
#'@export
set_pvar <- function(data, pvar) {
  phase(data) <- pvar
  data
}

#' Autocorrelation within and across phases
#'
#' The autocorr function calculates autocorrelations within each phase and
#' across all phases.
#' 
#' Autocorrelations are computed using the [acf()] function from the stats
#' package. For each single-case in the scdf object, a data frame is returned
#' containing the autocorrelations for each phase and for all phases up to the
#' specified lag.
#'
#' @inheritParams .inheritParams
#' @param lag_max,lag.max The lag up to which autocorrelations will be computed.
#' @param ... Further arguments passed to the [acf()] function.
#' @return A data frame containing separate autocorrelations for each phase and
#'   for all phases (for each single-case). If `lag_max` exceeds the length
#'   of a phase minus one, NA is returned for this cell.
#' @author Juergen Wilbert
#' @seealso [acf()]
#' @family regression functions
#' @keywords regression
#' @examples
#' ## Compute autocorrelations for a list of four single-cases up to lag 2.
#' autocorr(Huber2014, lag_max = 2)
#' @concept Autocorrelation
#' @concept Serial correlation
#' @export
autocorr <- function(data, dvar, pvar, mvar,
                     lag_max = 3,
                     lag.max,
                     ...) {
  if (!missing(lag.max)) lag_max <- lag.max

  # set defaults attributes
  if (missing(dvar)) dvar <- dv(data) else dv(data) <- dvar
  if (missing(pvar)) pvar <- phase(data) else phase(data) <- pvar
  if (missing(mvar)) mvar <- mt(data) else mt(data) <- mvar

  data <- .prepare_scdf(data)

  case_names <- revise_names(data)
  var_lag <- paste0("Lag ", 1:lag_max)

  ac <- list()
  for (case in seq_along(data)) {
    phases <- .phasestructure(data[[case]], pvar = pvar)

    while (any(duplicated(phases$values))) {
      phases$values[anyDuplicated(phases$values)] <- paste0(
        phases$values[anyDuplicated(phases$values)],
        "_phase",
        anyDuplicated(phases$values)
      )
    }

    df <- data.frame(Phase = c(phases$values, "all"))
    for (phase in seq_along(phases$values)) {
      y <- data[[case]][phases$start[phase]:phases$stop[phase], dvar]
      if (length(y) - 1 < lag_max) lag <- length(y) - 1 else lag <- lag_max

      .tmp <- acf(y, lag.max = lag, plot = FALSE, ...)$acf[-1]
      df[phase, var_lag[1:lag]] <- .tmp
    }

    y <- data[[case]][[dvar]]
    if (length(y) - 1 < lag_max) lag <- length(y) - 1 else lag <- lag_max

    .tmp <- acf(y, lag.max = lag, plot = FALSE, ...)$acf[-1]
    df[length(phases$values) + 1, var_lag[1:lag]] <- .tmp

    ac[[case]] <- df
  }

  names(ac) <- case_names

  out <- list(
    autocorr = ac,
    dvar = dvar
  )
  class(out) <- c("sc_ac")
  attributes(out)[opts("phase", "mt", "dv")] <- list(pvar, mvar, dvar)
  out
}

test_that("main", {
  
  res <- autocorr(exampleABC)
  testsum <- sum(
    unlist(lapply(res$autocorr, function(x) unlist(x[,-1]))), na.rm = TRUE
  )
  
  expect_equal(round(testsum,4), -0.3649)
  
})

#' Conservative Dual-Criterion Method
#'
#' The `cdc()` function applies the Conservative Dual-Criterion Method (Fisher,
#' Kelley, & Lomas, 2003) to scdf objects. It compares phase B data points to
#' both phase A mean and trend (OLS, bi-split, tri-split) with an additional
#' increase/decrease of .25 SD. A binomial test against a 50/50 distribution is
#' computed and p-values below .05 are labelled "systematic change".
#'
#'
#' @inheritParams .inheritParams
#' @param trend_method Method used to calculate the trend line. Default is
#'   `trend_method = "OLS"`. Possible values are: `"OLS"`, `"bisplit"`, and
#'   `"trisplit"`. `"bisplit"`, and `"trisplit"` should only be used for cases
#'   with at least five data-points in both relevant phases.
#' @param conservative The CDC method adjusts the original mean and trend lines
#'   by adding (expected increase) or subtracting (expected decrease) an
#'   additional .25 SD before evaluating phase B data. Default is the CDC method
#'   with `conservative = .25`. To apply the Dual-Criterion (DC) method, set
#'   `conservative = 0`.
#' @return 
#'  |  |  |
#'  | --- | --- |
#'  | `cdc` | CDC Evaluation based on a p-value below .05. |
#'  | `cdc_exc` | Number of phase B datapoints indicating expected change. |
#'  | `cdc_nb` | Number of phase B datapoints. |
#'  | `cdc_p` | P value of Binomial Test. |
#'  | `cdc_all` | Overall CDC Evaluation based on all instances/cases of a Multiple Baseline Design. |
#'  | `N` | Number of cases. |
#'  | `decreasing` | Logical argument from function call (see Arguments above). |
#'  | `conservative` | Numeric argument from function call (see Arguments above). |
#'  | `case_names` | Assigned name of single-case. |
#'  | `phases` | - |
#' @author Timo Lueke
#' @references Fisher, W. W., Kelley, M. E., & Lomas, J. E. (2003). Visual Aids
#'   and Structured Criteria for Improving Visual Inspection and Interpretation
#'   of Single-Case Designs. *Journal of Applied Behavior Analysis, 36*,
#'   387-406. https://doi.org/10.1901/jaba.2003.36-387
#' @family overlap functions
#' @keywords overlap
#' @examples
#'
#' ## Apply the CDC method (standard OLS line)
#' design <- design(n = 1, slope = 0.2)
#' dat <- random_scdf(design, seed = 42)
#' cdc(dat)
#'
#' ## Apply the CDC with Koenig's bi-split and an expected decrease in phase B.
#' cdc(exampleAB_decreasing, decreasing = TRUE, trend_method = "bisplit")
#'
#' ## Apply the CDC with Tukey's tri-split, comparing the first and fourth phase
#' cdc(exampleABAB, trend_method = "trisplit", phases = c(1,4))
#'
#' ## Apply the Dual-Criterion (DC) method (i.e., mean and trend without
#' ##shifting).
#' cdc(
#'  exampleAB_decreasing,
#'  decreasing = TRUE,
#'  trend_method = "bisplit",
#'  conservative = 0
#' )
#'
#'
#' @export
cdc <- function(data,
                dvar,
                pvar,
                mvar,
                decreasing = FALSE,
                trend_method = c("OLS", "bisplit", "trisplit"),
                conservative = .25,
                phases = c(1, 2)) {

  check_args(
    by_class(decreasing, "logical"),
    by_call(trend_method),
    within(conservative, 0, 1)
  )

  # set attributes to arguments else set to defaults of scdf
  if (missing(dvar)) dvar <- dv(data) else dv(data) <- dvar
  if (missing(pvar)) pvar <- phase(data) else phase(data) <- pvar
  if (missing(mvar)) mvar <- mt(data) else mt(data) <- mvar

  data  <- .prepare_scdf(data, na.rm = TRUE)
  data  <- recombine_phases(data, phases = phases)$data

  n_cases <- length(data)
  cdc_na  <- rep(NA, n_cases)  # total data points in phase A
  cdc_nb  <- rep(NA, n_cases)  # total data points in phase B
  cdc_exc <- rep(NA, n_cases)  # exceeding data points in phase B
  cdc     <- rep(NA, n_cases)  # CDC rule evaluation of change
  cdc_p   <- rep(NA, n_cases)  # binomial p (50/50)
  cdc_all <- NA          # CDC rule evaluation of all "cases"

  for (i in 1:n_cases) {
    A <- data[[i]][data[[i]][, pvar] == "A", ]
    B <- data[[i]][data[[i]][, pvar] == "B", ]
    cdc_na[i] <- nrow(A)
    cdc_nb[i] <- nrow(B)

    if ((cdc_na[i] < 5 || cdc_nb[i] < 5) && trend_method != "OLS") {
      abort(
        "The selected method for trend estimation should not be applied ",
        "with less than five data points per phase."
      )
    }

    if (trend_method == "bisplit") {
      x <- A[[mvar]]
      y <- A[[dvar]]
      # na.rm = FALSE for now to prevent misuse; will draw no line if NA present
      md1 <- c(
        (median(y[1:floor(length(y) / 2)], na.rm = FALSE)),
        median(x[1:floor(length(x) / 2)], na.rm = FALSE)
      )
      md2 <- c(
        (median(y[ceiling(length(y) / 2 + 1):length(y)], na.rm = FALSE)),
        median(x[ceiling(length(x) / 2 + 1):length(x)], na.rm = FALSE)
      )
      md <- as.data.frame(rbind(md1, md2))
      names(md) <- c(dvar, mvar)
      formula <- as.formula(paste0(dvar, "~", mvar))
      model <- lm(formula, data = md, na.action = na.omit)
    }

    if (trend_method == "trisplit") {
      x <- A[[mvar]]
      y <- A[[dvar]]
      # na.rm = FALSE for now to prevent misuse; will draw no line if NA present
      md1 <- c(
        (median(y[1:floor(length(y) / 3)], na.rm = FALSE)),
        median(x[1:floor(length(x) / 3)], na.rm = FALSE)
      )
      md2 <- c(
        (median(y[ceiling(length(y) / 3 * 2 + 1):length(y)], na.rm = FALSE)),
        median(x[ceiling(length(x) / 3 * 2 + 1):length(x)], na.rm = FALSE)
      )
      md <- as.data.frame(rbind(md1, md2))
      names(md) <- c(dvar, mvar)
      formula <- as.formula(paste0(dvar, "~", mvar))
      model <- lm(formula, data = md, na.action = na.omit)
    }

    if (trend_method == "OLS") {
      formula <- as.formula(paste0(dvar, "~", mvar))
      model <- lm(formula, data = A, na.action = na.omit)
    }

    trnd <- predict(model, B, se.fit = TRUE)

    if (!decreasing) {
      cdc_exc[i] <- sum(
        B[[dvar]] > trnd$fit + (conservative * sd(A[[dvar]])) &
        B[[dvar]] > (mean(A[[dvar]]) + (conservative * sd(A[[dvar]])))
      )
      cdc_p[i] <- binom.test(
        cdc_exc[i], cdc_nb[i], alternative = "greater"
      )$p.value
      cdc[i] <- if (cdc_p[i] < .05) "systematic change" else "no change"
    } else {
      cdc_exc[i] <- sum(
        B[[dvar]] < trnd$fit - (conservative * sd(A[[dvar]])) &
        B[[dvar]] < (mean(A[[dvar]]) - (conservative * sd(A[[dvar]])))
      )
      cdc_p[i] <- binom.test(
        cdc_exc[i], cdc_nb[i], alternative = "greater"
      )$p.value
      cdc[i] <- if (cdc_p[i] < .05) "systematic change" else "no change"
    }
    
    cdc_all <- if (length(cdc_p[cdc_p > .05]) / length(cdc_p) <= .25) {
      "systematic change"
    } else {
      "no change"
    }
  }

  out <- list(
    cdc = cdc,
    cdc_be = cdc_exc,
    cdc_b = cdc_nb,
    cdc_p = cdc_p,
    cdc_all = cdc_all,
    N = n_cases,
    decreasing = decreasing,
    trend_method = trend_method,
    conservative = conservative,
    case_names = revise_names(data)
  )
  class(out) <- c("sc_cdc")
  attributes(out)[opts("phase", "mt", "dv")] <- list(pvar, mvar, dvar)
  out
}

#' Randomization Tests for single-case data
#'
#' The `rand_test` function computes a randomization test for single or multiple
#' baseline single-case data.  The function is based on an algorithm from the
#' `SCRT` package (Bulte & Onghena, 2009, 2012), but rewritten and extended for
#' the use in AB designs.
#'
#' # Details
#'
#' ## Predefinded statisic
#'
#' Use the `statistic` argument to choose a predefnied statistic. The
#' following comparisons are possible: \itemize{ \item`Mean A-B`: Uses
#' the difference between the mean of phase A and the mean of phase B. This is
#' appropriate if a decrease of scores was expected for phase B.
#' \item`Mean B-A`: Uses the difference between the mean of phase B and
#' the mean of phase A. This is appropriate if an increase of scores was
#' expected for phase B.  \item`Mean |A-B|`: Uses the absolute value of
#' the difference between the means of phases A and B.  \item`Median
#' A-B`: The same as `Mean A-B`, but based on the median.
#' \item`Median B-A`: The same as `Mean B-A`, but based on the
#' median.  \item`SMD hedges / SMD glass`: Standardizes mean difference of B-A
#' as Hedges's g or Glass' delta. \item`NAP`: Non-overlap of all pairs.
#' \item`W-test`: Wilcoxon-test statistic W.\item`T-test`: T-test statistic t.}
#'
#' ## Create own statistic function
#'
#' Use the `statistic_function` argument to proved your own function in a list.
#' This list must have an element named `statistic` with a function that takes
#' two arguments `a` and `b` and returns a single numeric value.  E.g.
#' `list(statistic = function(a, b) mean(a) - mean(b)`. A second element of the
#' list is named `aggregate` which takes a function with one numeric argument
#' that returns a numeric argument. This function is used to aggregate the
#' values of a multiple case design. If you do not provide this element, it uses
#' the default `function(x) sum(x)/length(x)`. The third optional argument is
#' `name` which provides a name for your user function.
#'
#' @inheritParams .inheritParams
#' @param statistic Defines the statistic on which the comparison of phases A
#'   and B is based on. Default setting is `statistic = "Mean B-A"`. See
#'   details.
#' @param statistic_function A list with a user defined function to calculate
#'   the statistic. When set, overwrites the `statistic` argument. See details.
#' @param number Sample size of the randomization distribution. The exactness of
#'   the p-value can not exceed \eqn{1/number} (i.e., `number = 100` results in
#'   p-values with an exactness of one percent). Default is `number = 500`. For
#'   faster processing use `number = 100`. For more precise p-values set `number
#'   = 1000`).
#' @param complete If TRUE, the distribution is based on a complete permutation
#'   of all possible starting combinations. This setting overwrites the number
#'   Argument. The default setting is FALSE.
#' @param limit Minimal number of data points per phase in the sample. The first
#'   number refers to the A-phase and the second to the B-phase (e.g., `limit =
#'   c(5,3)`). If only one number is given, this number is applied to both
#'   phases. Default is `limit = 5`.
#' @param startpoints Alternative to the `limit`-parameter `startpoints` exactly
#'   defines the possible start points of phase B (e.g., `startpoints = 4:9`
#'   restricts the phase B start points to measurements 4 to 9. `startpoints`
#'   overruns the `limit`-parameter. If a list of start points is provided, each
#'   single-case can have different start points (e.g., `startpoints = list(4:9,
#'   5:10)` for a two-case design). Default is `startpoints = NA`, which uses
#'   the `limit`-parameter to define possible start points.
#' @param exclude.equal If set to \code{exclude.equal = FALSE}, which is the
#'   default, random distribution values equal to the observed distribution are
#'   counted as null-hypothesis conform. That is, they decrease the probability
#'   of rejecting the null-hypothesis (increase the p-value).
#'   \code{exclude.equal} should be set to `TRUE` if you analyse one single-case
#'   design (not a multiple baseline data set) to reach a sufficient power. But
#'   be aware, that it increases the chance of an alpha-error. If
#'   \code{exclude.equal = "auto" } is used, the function automatically sets
#'   \code{exclude.equal = TRUE} for single-case designs and to \code{FALSE} for
#'   multiple-case designs.
#' @param graph If `graph = TRUE`, a histogram of the resulting distribution is
#'   plotted. It is `FALSE` by default. *Note: use the more versatile
#'   [plot_rand()] function instead.*
#' @param output (deprecated and not implemented)
#' @param seed A seed number for the random generator. Default is `NULL`.
#' @return An object of class `sc_rand`. It is a list containing the following
#' elements:
#' \item{statistic}{Character string from function call (see
#'   \code{Arguments} above).} \item{N}{Number of single-cases.}
#'   \item{n1}{Number of data points in phase A.} \item{n2}{Number of data
#'   points in phase B.} \item{limit}{Numeric from function call (see
#'   \code{Arguments} above).}
#' \item{startpoints}{A vector defining the start points passed from the
#' function call (see `Arguments` above).} \item{p.value}{P-value of the
#' randomization test for the given data.} \item{number}{Sample size of
#' randomization distribution from function call (see \code{Arguments} above).}
#' \item{complete}{Logical argument from function call (see \code{Arguments}
#' above).} \item{observed.statistic}{Test statistic observed for the given
#' single-case data. (see \code{statistic} in the \code{Arguments} above.)}
#' \item{Z}{Z-value of observed test statistic.} \item{p.z.single}{Probability
#' of z-value.} \item{distribution}{Test statistic distribution from randomized
#' data sets.} \item{possible.combinations}{Number of possible combinations
#' under the given restrictions.} \item{auto.corrected.number}{`TRUE`
#' indicates that a corrected number of combinations was used. This happens, if
#' the number of possible combinations (under the given restrictions) undercuts
#' the requested `number` of combinations.} \item{ecxlude.equal}{see argument
#' above}
#' \item{testdirection}{Indicates the direction of the test ("greater" or
#' "less").}
#' @author Juergen Wilbert
#' @references Bulte, I., & Onghena, P. (2009). Randomization tests for
#'   multiple-baseline designs: An extension of the SCRT-R package.
#' *Behavior Research Methods, 41*, 477-485.
#'
#'   Bulte, I., & Onghena, P. (2012). *SCRT: Single-Case Randomization Tests*.
#'   Available from: \url{https://CRAN.R-project.org/package=SCRT}
#' @examples
#'
#' ## Compute a randomization test on the first case of the byHeart2011 data and include a graph
#' rand_test(byHeart2011[1], statistic = "Median B-A", graph = TRUE, seed = 123)
#'
#' ## Compute a randomization test on the Grosche2011 data using complete permutation
#' rand_test(Grosche2011, statistic = "Median B-A", complete = TRUE, limit = 4, seed = 123)
#'
#' @export
rand_test <- function (data, dvar, pvar, 
                       statistic = c("Mean B-A", "Mean A-B", "Median B-A", 
                                     "Median A-B", "Mean |A-B|", "Median |A-B|",
                                     "SMD hedges", "SMD glass", 
                                     "W-test", "T-test", 
                                     "NAP", "NAP decreasing",
                                     "Slope B-A","Slope A-B"), 
                       statistic_function = NULL,
                       number = 500, 
                       complete = FALSE, 
                       limit = 5, 
                       startpoints = NA, 
                       exclude.equal = FALSE, 
                       phases = c(1, 2), 
                       graph = FALSE, 
                       output = NULL, 
                       seed = NULL) {

  
  check_args(
    
    by_call(statistic)
  )
  
  if(!is.null(seed)) set.seed(seed)
  
  # set attributes to arguments else set to defaults of scdf
  if (missing(dvar)) dvar <- dv(data)
  if (missing(pvar)) pvar <- phase(data) 
  dv(data) <- dvar
  phase(data) <- pvar
  
  data <- .prepare_scdf(data, na.rm = TRUE)
  
  keep <- recombine_phases(data, phases = phases)
  data <- keep$data
  
  a   <- lapply(data, function(x) x[x[, pvar] == "A", dvar])
  b   <- lapply(data, function(x) x[x[, pvar] == "B", dvar])
  obs <- lapply(data, function(x) x[, dvar])
  mts  <- lapply(data, nrow)
  n_cases   <- length(data)
  
  testdirection <- "greater"
  
  if (identical(exclude.equal, "auto")) exclude.equal <- n_cases == 1
  
# starting points ---------------------------------------------------------

  if (length(limit) == 1) limit[2] <- limit[1]
  #obs.B.start <- unlist(lapply(a, function(x) length(x) + 1))
  
  if (is.na(startpoints[1])) {
    pos_startpts <- lapply(mts, function(x) (limit[1] + 1):(x - limit[2] + 1))
  } else {
    pos_startpts <- lapply(mts, function(x) startpoints)
  }
  
  ### posible combinations
  
  possible_combinations <- lapply(pos_startpts, length)
  possible_combinations <- cumprod(unlist(possible_combinations))[n_cases]	
  
  auto_corrected_number <- FALSE
  if (!complete && possible_combinations <= number) {
    auto_corrected_number <- TRUE
    complete <- TRUE
  }
  
  if (!complete) {
    startpts <- lapply(pos_startpts, function(x) sample(x, number, replace = TRUE))
    startpts <- matrix(unlist(startpts), nrow = number, ncol = n_cases)
  }
  if (complete) {
    startpts <- expand.grid(pos_startpts)
    number   <- nrow(startpts)
  }
  
# Sample Random A and B phases ---------------------------------------------
  
  rnd_a <- vector("list", number)
  for (i in 1:number) {
    ascores <- vector("list", n_cases)
    for (case in 1:n_cases)
      ascores[[case]] <- data[[case]][1:(startpts[i, case] - 1), dvar]
    rnd_a[[i]] <- ascores
  }

  rnd_b <- vector("list", number)
  for (i in 1:number) {
    ascores <- vector("list", n_cases)
    for (case in 1:n_cases)
      ascores[[case]] <- data[[case]][startpts[i, case]:mts[[case]], dvar]
    rnd_b[[i]] <- ascores
  }
  
# Functions for phase differences -----------------------------------------

  if (!is.null(statistic_function)) {
    
    if (is.null(statistic_function$aggregate)) 
      statistic_function$aggregate <- function(x) sum(x) / length(x)
    
    res <- rand_test_statistic(rnd_a, rnd_b, a, b,
      statistic = statistic_function$statistic,
      args_statistic = NULL,
      aggregate = statistic_function$aggregate
    )
    
    if (is.null(statistic_function$name)) statistic_function$name <- ""
    statistic <- paste0("user defined function ", statistic_function$name)
  }
  
  if (statistic == "SMD hedges") {
    res <- rand_test_statistic(rnd_a, rnd_b, a, b,
      statistic = .opt$rand_test$smd,
      args_statistic = list(method = "hedges"),
      aggregate = function(x) sqrt(sum(x^2) / length(x))
    )
  }  
  
  if (statistic == "SMD glass") {
    res <- rand_test_statistic(rnd_a, rnd_b, a, b,
      statistic = .opt$rand_test$smd,
      args_statistic = list(method = "glass"),
      aggregate = function(x) sqrt(sum(x^2) / length(x))
    )
  }  
  
  if (statistic == "W-test") {
    res <- rand_test_statistic(rnd_a, rnd_b, a, b, 
      statistic = .opt$rand_test$u_test,
      args_statistic = NULL,
      aggregate = function(x) sum(x) / length(x)
    )
    testdirection <- "less"
  }   

  if (statistic == "T-test") {
    res <- rand_test_statistic(rnd_a, rnd_b, a, b, 
      statistic = .opt$rand_test$t_test,
      args_statistic = NULL,
      aggregate = function(x) sum(x) / length(x)
    )
  } 
  
  if (statistic == "NAP") {
    res <- rand_test_statistic(rnd_a, rnd_b, a, b, 
      statistic = .opt$rand_test$nap,
      args_statistic = list(decreasing = FALSE),
      aggregate = function(x) sum(x) / length(x)
    )
  } 
  
  if (statistic == "NAP decreasing") {
    res <- rand_test_statistic(rnd_a, rnd_b, a, b, 
      statistic = .opt$rand_test$nap,
      args_statistic = list(decreasing = TRUE),
      aggregate = function(x) sum(x) / length(x)
    )
  } 
  
  if (statistic == "Mean B-A") {
    res <- rand_test_statistic(rnd_a, rnd_b, a, b, 
      statistic = .opt$rand_test$average,
      args_statistic = list(
        fn = function(x) {sum(x) / length(x)}, method = "B-A"
      ),
      aggregate = function(x) sum(x) / length(x)
    )
  } 
  
  if (statistic == "Mean A-B") {
    res <- rand_test_statistic(rnd_a, rnd_b, a, b, 
      statistic = .opt$rand_test$average,
      args_statistic = list(
        fn = function(x) {sum(x) / length(x)}, method = "A-B"
      ),
      aggregate = function(x) sum(x) / length(x)
    )
  } 
  
  if (statistic == "Mean |A-B|") {
    res <- rand_test_statistic(rnd_a, rnd_b, a, b, 
      statistic = .opt$rand_test$average,
      args_statistic = list(
        fn = function(x) {sum(x) / length(x)}, method = "abs"
      ),
      aggregate = function(x) sum(x) / length(x)
    )
  } 

  if (statistic == "Median B-A") {
    res <- rand_test_statistic(rnd_a, rnd_b, a, b, 
      statistic = .opt$rand_test$average,
      args_statistic = list(
        fn = function(x) {median(x)}, method = "B-A"
      ),
      aggregate = function(x) median(x)
    )
  } 
  
  if (statistic == "Median A-B") {
    res <- rand_test_statistic(rnd_a, rnd_b, a, b, 
      statistic = .opt$rand_test$average,
      args_statistic = list(
        fn = function(x) {median(x)}, method = "A-B"
      ),
      aggregate = function(x) median(x)
    )
  } 
  
  if (statistic == "Median |A-B|") {
    res <- rand_test_statistic(rnd_a, rnd_b, a, b, 
      statistic = .opt$rand_test$average,
      args_statistic = list(
        fn = function(x) {median(x)}, method = "abs"
      ),
      aggregate = function(x) median(x)
    )
  }   

  if (statistic == "Slope B-A") {
    res <- rand_test_statistic(rnd_a, rnd_b, a, b,
      statistic = .opt$rand_test$slope,
      args_statistic = list(method = "B-A"),
      aggregate = function(x) sum(x) / length(x)
    )
  }   
  
  if (statistic == "Slope A-B") {
    res <- rand_test_statistic(rnd_a, rnd_b, a, b,
      statistic = .opt$rand_test$slope,
      args_statistic = list(method = "B-A"),
      aggregate = function(x) sum(x) / length(x)
    )
  }   
  
# p value -----------------------------------------------------------------
  
  if (testdirection == "greater") {
    test <- if (!exclude.equal) res$dist >= res$obs_stat else res$dist > res$obs_stat
  } else {
    test <- if (!exclude.equal) res$dist <= res$obs_stat else res$dist < res$obs_stat
  }
  
  p_value <- sum(test) / number
  
# return ------------------------------------------------------------------

  if (graph){
    h <- hist(res$dist, plot = FALSE)
    lab <- paste0(round(h$counts / length(res$dist) * 100, 0), "%")
    xlim <- c(min(h$breaks,na.rm = TRUE), max(h$breaks, na.rm = TRUE))
    ylim <- round(max(h$counts * 1.2))
    if (res$obs_stat < xlim[1]) xlim[1] <- res$obs_stat
    if (res$obs_stat > xlim[2]) xlim[2] <- res$obs_stat
    hist(
      res$dist, xlab = statistic, labels = lab, xlim = xlim, ylim = c(0, ylim), 
      ylab = "Frequency", main = "Random distribution", col = "lightgrey"
    )
    abline(v = res$obs_stat, lty = 2, lwd = 2, col = "grey") 
    if (p_value < 0.5) pos <- 2 else pos <- 4
    text(res$obs_stat, ylim, "observed", pos = pos)
  }
  
  z <- (res$obs_stat - mean(res$dist, na.rm = TRUE)) / sd(res$dist, na.rm = TRUE)
  p_z_single <- if (testdirection == "greater") 1 - pnorm(z) else pnorm(z)
    
  possible_combinations <- cumprod(unlist(lapply(pos_startpts, length)))[n_cases]

  out <- list(
    statistic = statistic, 
    phases.A = keep$phases_A, 
    phases.B = keep$phases_B, 
    N = n_cases, 
    n1 = length(unlist(a)), 
    n2 = length(unlist(b)), 
    limit = limit, 
    startpoints = startpoints, 
    p.value = p_value, 
    number = number, 
    complete = complete, 
    observed.statistic = res$obs_stat, 
    Z = z, 
    p.Z.single = p_z_single, 
    distribution = res$dist, 
    distribution_startpoints = startpts,
    possible.combinations = possible_combinations, 
    auto.corrected.number = auto_corrected_number,
    exclude.equal = exclude.equal,
    testdirection = testdirection
  )
  
  class(out) <- c("sc_rand")
  attributes(out)[opts("phase", "dv")] <- list(pvar, dvar)
  attr(out, "casenames") <- names(data)
  out

}

rand_test_statistic <- function(rnd_a, rnd_b, a, b, 
                                statistic,
                                args_statistic,
                                aggregate) {
  
  n_cases <- length(rnd_a[[1]])
  number <- length(rnd_a)
  rnd_a <- unlist(rnd_a,recursive = FALSE)
  rnd_b <- unlist(rnd_b,recursive = FALSE)

  dat <- mapply(statistic, a = rnd_a, b = rnd_b, MoreArgs = args_statistic)
  ma <- matrix(dat, ncol = n_cases, nrow = number, byrow = TRUE)   
  dist <- apply(ma, 1, aggregate)
  
  dat <- mapply(statistic, a = a, b = b, MoreArgs = args_statistic)
  list(
    obs_stat = aggregate(dat), 
    dist = dist
  )
}  


# function derived from: https://stackoverflow.com/questions/40141738/is-there-a-fast-estimation-of-simple-regression-a-regression-line-with-only-int
.estimate_slope <- function (x, y) {
  
  ## centring
  yc <- y - sum(y) / length(y)
  xc <- x - sum(x) / length(x)
  
  ## fitting an intercept-free model: yc ~ xc + 0
  xty <- c(crossprod(xc, yc))
  xtx <- c(crossprod(xc))
  
  # slope
  xty / xtx
  
}

.opt$rand_test$statistic_slope <- statistic <- function(a, b, method) {
  slope_b <- .estimate_slope(1:length(b), b)
  slope_a <- .estimate_slope(1:length(a), a)
  if (method == "B-A") slope_b - slope_a else slope_a - slope_b
}

.opt$rand_test$nap <- function(a, b, decreasing) {
  
  if (!decreasing) pos <- sum(unlist(lapply(a, function(x) b > x)))
  if (decreasing) pos <- sum(unlist(lapply(a, function(x) b < x)))
  
  ties <- sum(unlist(lapply(a, function(x) x == b)))
  non_overlaps <- pos + (0.5 * ties)
  pairs <- length(a) * length(b)
  nap <- non_overlaps / pairs * 100
}

.opt$rand_test$u_test <- function(a, b) {
  suppressWarnings(wilcox.test(a,b)$statistic)
}  

.opt$rand_test$t_test <- function(a, b) {
  suppressWarnings(t.test(b, a)$statistic)
}  

.opt$rand_test$smd <- function(a, b, method) {
  # Hedges'g + Durlak correction or Glass' Delta
  
  mean_a <- sum(a) / length(a)
  mean_b <- sum(b) / length(b)
  sd_a <- sd(a)
  sd_b <- sd(b)
  n_a <- length(a)
  n_b <- length(b)
  
  if (method == "hedges") {
    dat <- (mean_b - mean_a) / 
      sqrt(((n_a - 1) * sd_a^2 + (n_b - 1) * sd_b^2) / (n_a + n_b - 2)) *
      (((n_a+n_b) - 3) / ((n_a+n_b) - 2.25) * sqrt(((n_a+n_b) - 2) / (n_a+n_b)))
  } else if (method == "glass") {
    dat <- (mean_b - mean_a) / sd_a
  }
}

.opt$rand_test$average <- function(a, b, fn, method) {
  if (method == "B-A") {
    fn(b) - fn(a)
  } else if (method == "A-B") {
    fn(a) - fn(b)
  } else if (method == "abs") {
    abs(fn(b) - fn(a))
  }
}

test_that("rand_test", {
  
  scdf <- exampleAB
  
  results <- rand_test(scdf, complete = TRUE)

  expect_identical(
    round(
      with(
        results, 
        p.value + observed.statistic + Z + p.Z.single + mean(distribution)
      ), 
    8),
    40.97299008
  )
  
  userstat <- list(
    statistic = function(a, b) sum(b)/length(b) - sum(a)/length(a), 
    aggregate = function(x) sum(x)/length(x),
    name = "mean B - A"
  )
  
  results_user <- rand_test(
    exampleAB, statistic_function = userstat , complete = TRUE
  )
  
  results_user$statistic <- results$statistic
  expect_identical(results_user, results)
  
})

#' @describeIn mplm Print results
#' @order 2
#' @inheritParams print.sc
#' @param x Object returned from [mplm()]. 
#' @param digits The minimum number of significant digits to be use. 
#' If set to "auto" (default), values are predefined.
#' @param std If TRUE, a table with standardized estimates is included.
#' @export
print.sc_mplm <- function(x, digits = "auto", std = FALSE, ...) {
  
  if (digits == "auto") digits <- 3
  
  out <- .output_mplm(x, std = std)
  
  cat("Multivariate piecewise linear model\n\n")
  cat(
    "Dummy model: ", x$model, " ", 
    paste0("level = ", x$contrast$level, ", slope = " ,x$contrast$slope), "\n", 
    sep = ""
  )
  cat("Type", out$anova$type, "MANOVA", "\n")
  cat(out$fit_string, "\n")
  #" MANOVA Tests: ", out$anova$test, " test statistic", 
  cat("\n")
  out$coef <- round_numeric(out$coef, digits)
  print(out$coef, ...)
  cat("\n")
  if (out$std) cat("Coefficients are standardized\n")
  #cat(out$note, "\n", sep = "")
  cat("Formula: ")
  print(x$formula, showEnv = FALSE)
  
  #.note_vars(x)
}

.output_mplm <- function(x, std) {
  
  out <- list()
  
  out$coef <- if (isTRUE(std)) { 
    x$full.model$coef_std
  } else {
    x$full.model$coefficients
  }
  
  if (inherits(out$coef, "numeric"))  {
    out$coef <- as.data.frame(out$coef)
    names(out$coef) <- attr(x, opt("dv"))
  }
  
  row.names(out$coef) <- rename_predictors(row.names(out$coef), x)
 
  global_test <- anova(x$full.model, x$null_model)
  out$f_test <- c(
    pillai = global_test$Pillai[2],
    df1 = global_test$`num Df`[2],
    df2 = global_test$`den Df`[2],
    f = global_test$`approx F`[2],
    p = global_test$`Pr(>F)`[2]

  )
  out$fit_string <- sprintf(
    "Pillai = %.2f; F(%d, %d) = %.2f; p = %0.3f", 
    out$f_test["pillai"], out$f_test["df1"], out$f_test["df2"], 
    out$f_test["f"], out$f_test["p"]
  )
  ### anova
  
  out$anova <- car::Anova(x$full.model, type = 3)
  
  if (!is.null(out$anova$terms)) {
    out$anova$terms <- rename_predictors(out$anova$terms, x)
  } else {
    row.names(out$anova) <- rename_predictors(row.names(out$anova), x)
  }

  test <- out$anova$test
  repeated <- out$anova$repeated
  ntests <- length(out$anova$terms)
  coef <- matrix(NA, ntests, 4)
  if (!repeated) 
    SSPE.qr <- qr(out$anova$SSPE)
  for (term in 1:ntests) {
    eigs <- Re(
      eigen(
        qr.coef(
          if (repeated) qr(out$anova$SSPE[[term]]) else SSPE.qr, 
          out$anova$SSP[[term]]
        ), 
        symmetric = FALSE
      )$values
    )
    coef[term, 1:4] <- switch(
      test, 
      Pillai = Pillai(eigs, out$anova$df[term], out$anova$error.df), 
      Wilks = Wilks(eigs, out$anova$df[term], out$anova$error.df), 
      `Hotelling-Lawley` = HL(eigs, out$anova$df[term], out$anova$error.df), 
      Roy = Roy(eigs, out$anova$df[term], out$anova$error.df)
    )
  }
  ok <- coef[, 2] >= 0 & coef[, 3] > 0 & coef[, 4] > 0
  ok <- !is.na(ok) & ok
  coef <- cbind(
    out$anova$df, 
    coef, 
    pf(coef[ok, 2], coef[ok, 3], coef[ok, 4], lower.tail = FALSE)
  )
  rownames(coef) <- out$anova$terms
  colnames(coef) <- c(
    "df", "stat", "F", "num Df", "den Df", "p"
  )
  
  coef <- as.data.frame(coef) 
  names(coef)[2] <- test
  
  out$std <- std
  out$coef <- cbind(out$coef, coef[, c(2, 3, 6)]) 
  out$note = paste0(
    "df = ", round(coef[1, 1], 1), 
    "; num Df = ", coef[1, 4], 
    "; den Df = ", coef[1, 5]
  )

  out
  
}

Roy <- function (eig, q, df.res) {
  p <- length(eig)
  test <- max(eig)
  tmp1 <- max(p, q)
  tmp2 <- df.res - tmp1 + q
  c(test, (tmp2 * test)/tmp1, tmp1, tmp2)
}

HL <- function (eig, q, df.res) {
  test <- sum(eig)
  p <- length(eig)
  m <- 0.5 * (abs(p - q) - 1)
  n <- 0.5 * (df.res - p - 1)
  s <- min(p, q)
  tmp1 <- 2 * m + s + 1
  tmp2 <- 2 * (s * n + 1)
  c(test, (tmp2 * test)/s/s/tmp1, s * tmp1, tmp2)
}

Wilks <- function (eig, q, df.res) {
  test <- prod(1/(1 + eig))
  p <- length(eig)
  tmp1 <- df.res - 0.5 * (p - q + 1)
  tmp2 <- (p * q - 2)/4
  tmp3 <- p^2 + q^2 - 5
  tmp3 <- if (tmp3 > 0) 
    sqrt(((p * q)^2 - 4)/tmp3)
  else 1
  c(test, ((test^(-1/tmp3) - 1) * (tmp1 * tmp3 - 2 * tmp2))/p/q, 
    p * q, tmp1 * tmp3 - 2 * tmp2)
}

Pillai <- function (eig, q, df.res) {
  test <- sum(eig/(1 + eig))
  p <- length(eig)
  s <- min(p, q)
  n <- 0.5 * (df.res - p - 1)
  m <- 0.5 * (abs(p - q) - 1)
  tmp1 <- 2 * m + s + 1
  tmp2 <- 2 * n + s + 1
  c(test, (tmp2/tmp1 * test)/(s - test), s * tmp1, s * tmp2)
}

#' @describeIn mplm Export results as html
#' @inheritParams export
#' @param std If TRUE, a table with standardized estimates is included.
#' @export
export.sc_mplm <- function(object, 
                          caption = NA, 
                          footnote = NA, 
                          filename = NA,
                          nice = TRUE,
                          std = FALSE,
                          decimals = 2,
                          ...) {
  

  if (is.na(caption)) {
    caption <- "Multivariate Piecewise-Regression Model"
      #"Multivariate Piecewise-regression model predicting '", 
      #paste0(attr(object, opt("dv")), collapse = "/"), "'"
    #)
  }
  
  results <- .output_mplm(object, std = std)
  
  out <- results$coef
  out <- rownames_to_first_column(out, "Parameter")
  
  if (nice) out$p <- .nice_p(out$p)
  
  if (is.na(footnote)) footnote <- c(
    paste0(results$fit_string)
  )
  
  if (std) {
    footnote <- c(footnote, "Predictors are standardized")
  }
  
  n_dv <- length(attr(object, opt("dv")))
  
  if (getOption("scan.export.engine") == "gt") {
    spanner <- list("Dependent variables" = 2:(1 + n_dv))
  }
  
  out <- round_numeric(out, decimals)
  table <- .create_table(
    out, 
    caption = caption,
    footnote = footnote,
    spanner = spanner
  )
  
  if (getOption("scan.export.engine") == "kable") {
    spanner <- c(" " = 1, "Dependent variables" = n_dv, " " = 3)
    table <- add_header_above(table, spanner)
  }
    
 
    
  # finish ------------------------------------------------------------------
  
  if (!is.na(filename)) .save_export(table, filename)
  table
  
}

#' Data output: Write single-case data to a .csv-file
#'
#' This function restructures and writes single-case data into a .csv-file.
#'
#' This is a wrapper for the write.table function with predefined parameters.
#' It converts the single-case data file into a data frame and writes it to a
#' .csv-file. Each single-case data set is stacked below each other and an
#' additional column "case" is added to identify the different single-cases.
#'
#' @inheritParams .inheritParams
#' @param filename A character string defining the output file name (e.g.
#'   `"scdf_data.csv"`. If `NULL` (default), the data are written to the
#'   console.
#' @param sep The field separator string. Values within each row of x are
#'   separated by this string.
#' @param dec The string to use for decimal points in numeric or complex
#'   columns: must be a single character.
#' @param \dots Further arguments passed to write.table.
#' @author Juergen Wilbert
#' @return Invisibly returns `NULL`.
#' @seealso [write.table()], [saveRDS()]
#' @family io-functions
#' @keywords io
#' @examples
#' ## write single-case data to a .csv-file
#' filename <- tempfile(fileext = ".csv")
#' write_scdf(exampleAB, filename)
#'
#' ## write multiple cases to a .csv-file with semicolon as field and comma as
#' ## decimal separator
#' write_scdf(Grosche2011, filename, sep = ";", dec = ",")
#'
#' ## read_scdf and write_scdf
#' write_scdf(exampleA1B1A2B2_zvt, filename)
#' dat <- read_scdf(filename, cvar = "case", pvar = "part",
#'                  dvar = "zvt", mvar = "day")
#' res1 <- describe(exampleA1B1A2B2_zvt)$descriptives
#' res2 <- describe(dat)$descriptives
#' all.equal(res1,res2)
#' @export
write_scdf <- function(data, filename = NULL, sep = ",", dec = ".", ...) {
  
  utils::write.table(
    as.data.frame(data), 
    file = filename, 
    row.names = FALSE,
    sep = sep,
    dec = dec,
    ...
  )
  
}

#' @rdname deprecated-functions
#' @export
writeSC <- function(...) {
  write_scdf(...)
}

#' @rdname export
#' @export
export.sc_trend <- function(object, 
                            caption = NA, 
                            footnote = NA, 
                            filename = NA,
                            round = 3,
                            decimals = NULL,
                            ...) {
  
  if (is.na(caption)) caption <- c("Trend analysis")
  if (is.na(footnote)) {
    footnote <- paste(
      "Measurement-times start at", object$first_mt, " for each phase"
    )
  }
  out <- object$trend
  #if (isTRUE(flip)) out <- t(out)
  
  tmp.rownames <- rownames(out)
  rownames(out) <- NULL
  
  for (tmp in names(object$formulas)) {
    tmp.rownames <- gsub(paste0(tmp, "."), "", tmp.rownames)
  }
  out <- cbind(Phase = tmp.rownames, out)
  
  out <- round_numeric(out, round)
  
  row_group <- vector("list", length(object$formulas))
  names(row_group) <- paste0(names(object$formulas), " (", object$formulas,")")
  
  for (i in 1:length(object$formulas)) {
    .start <- 1 + (i - 1) * (length(object$design) + 1)
    row_group[[i]] <- .start : (.start + length(object$design))
  }
  
  table <- .create_table(
    out, 
    caption = caption,
    footnote = footnote,
    row_group = row_group,
    round = round,
    decimals = decimals,
    ...
  )
  
  if (getOption("scan.export.engine") == "kable") {
    for (i in 1:length(object$formulas)) {
      table <- group_rows(
        table, object$formulas[i],
        1 + (i - 1) * (length(object$design) + 1),
        i * (length(object$design) + 1)
        #label_row_css = "text-align: center;"
      )
    }
  }
  # finish ------------------------------------------------------------------
  
  if (!is.na(filename)) .save_export(table, filename)
  table
}

#' Dummy function to inherit global descriptions of parameters
#' 
#' This function is only used to inherit parameter descriptions in the
#' documentation of other functions. It has no other purpose and is not meant to
#' be called directly.
#' 
#' @return No return value.
#' @author Juergen Wilbert
#' @family internal functions
#'
#' @param data A single-case data frame. See [scdf()] to learn about this
#'   format.
#' @param scdf A single-case data frame. See [scdf()] to learn about this
#'   format.
#' @param dvar Character string with the name of the dependent variable.
#'   Defaults to the attributes in the scdf file.
#' @param pvar Character string with the name of the phase variable. Defaults to
#'   the attributes in the scdf file.
#' @param mvar Character string with the name of the measurement time variable.
#'   Defaults to the attributes in the scdf file.
#' @param decreasing If you expect data to be lower in the B phase, set
#'   `decreasing = TRUE`. Default is `decreasing = FALSE`.
#' @param phases A vector of two characters or numbers indicating the two phases
#'   that should be compared. E.g., `phases = c("A","C")` or `phases = c(2,4)`
#'   for comparing the second to the fourth phase. Phases could be combined by
#'   providing a list with two elements. E.g., `phases = list(A = c(1,3), B =
#'   c(2,4))` will compare phases 1 and 3 (as A) against 2 and 4 (as B). Default
#'   is `phases = c(1,2)`.
#' @param model Model used for calculating the dummy parameters (see Huitema &
#'   McKean, 2000). Default is `model = "W"`. Possible values are: `"B&L-B"`,
#'   `"H-M"`, `"W"`, and deprecated `"JW"`.
#' @param contrast Sets contrast_level and contrast_slope. Either "first",
#'   "preceding" or a contrast matrix. If NA contrast is ignored.
#' @param contrast_level Either "first", "preceding" or a contrast matrix. If NA
#'   contrast_level is a copy of contrast.
#' @param contrast_slope Either "first", "preceding" or a contrast matrix. If NA
#'   contrast_level is a copy of contrast.
#' @param trend A logical indicating if a trend parameters is included in the
#'   model.
#' @param level A logical indicating if a level parameters is included in the
#'   model.
#' @param slope A logical indicating if a slope parameters is included in the
#'   model.
#' @param nice If set TRUE (default) output values are rounded and optimized for
#'  publication tables.
#' @param ... Further arguments passed to the function.
#' @keywords internal
.inheritParams <- function(data, scdf, dvar, mvar, pvar, decreasing, 
                           phases, model, contrast, contrast_level, 
                           contrast_slope, trend, level, slope, nice,
                           ...) {
  
} 

#' Standardized mean differences for single-case data
#'
#' The `smd()` function provides various standardized mean effect sizes for
#' single-case data.
#' 
#' It computes 'Cohen's d', 'Hedges' g', 'Hedges' g correction',
#' 'Hedges' g durlak correction', and 'Glass' delta' for each single-case
#' included in an scdf.
#'
#' @inheritParams .inheritParams
#' @details 'sd cohen' is the (unweigted) average of the variance of phase A and
#'   B. 'sd Hedges' is the weighted average of the variance of phase A and B
#'   (with a degrees of freedom correction). 'Hedges' g' is the mean difference
#'   divided by 'sd Hedges'. 'Hedges' g correction' and 'Hedges' g durlak
#'   correction' are two approaches of correcting Hedges' g for small sample
#'   sizes. 'Glass' delta' is the mean difference divided by the standard
#'   deviation of the A-phase. 'Cohens d' is the mean difference divided by 'sd
#'   cohen'.
#' @author Juergen Wilbert
#' @seealso [overlap()], [describe()]
#' @examples
#' smd(exampleAB)
#' @export
smd <- function(data, dvar, pvar, mvar, 
                phases = c(1, 2)) {
  
  # set defaults attributes
  if (missing(dvar)) dvar <- dv(data) else dv(data) <- dvar
  if (missing(pvar)) pvar <- phase(data) else phase(data) <- pvar
  if (missing(mvar)) mvar <- mt(data) else mt(data) <- mvar
  
  data_list <- .prepare_scdf(data)
  
  keep <- recombine_phases(data_list, phases = phases)
  data_list <- keep$data
  
  N <- length(data_list)
  
  case_names <- revise_names(names(data_list), length(data_list))
  
  vars <- c(
    "mA", "mB", "sdA", "sdB", "sd cohen", "sd hedges", "Glass' delta",  
    "Hedges' g", "Hedges' g correction",
    "Hedges' g durlak correction", "Cohen's d"
  )
  df <- as.data.frame(matrix(nrow = N, ncol = length(vars)))
  colnames(df) <- vars
  df <- data.frame(Case = case_names, df, check.names = FALSE)
  
  for(i in 1:N) {
    data <- data_list[i][[1]]
    
    A <- data[data[[pvar]] == "A", dvar]
    B <- data[data[[pvar]] == "B", dvar]
    nA <- sum(!is.na(A))
    nB <- sum(!is.na(B))    
    n <- nA + nB
    mA <- mean(A, na.rm = TRUE)
    mB <- mean(B, na.rm = TRUE)    
    sdA <- sd(A, na.rm = TRUE)
    sdB <- sd(B, na.rm = TRUE) 
    
    df$mA[i] <- mA
    df$mB[i] <- mB
    df$sdA[i] <- sdA
    df$sdB[i] <- sdB
    
    df$"Glass' delta"[i] <- (mB - mA) / sdA
    
    df$"sd hedges"[i] <- sqrt(
      ((nA - 1) * sdA^2 + (nB - 1) * sdB^2) / (nA + nB - 2) 
    )  
    
    df$"Hedges' g"[i] <- (mB - mA) / df$"sd hedges"[i]
    
    corr_hedges <- 1 - (3 / (4 * n - 9))
    df$"Hedges' g correction"[i] <- df$"Hedges' g"[i] * corr_hedges
    
    corr_durlak <- (n - 3) / (n - 2.25) * sqrt((n - 2) / n)
    
    df$"Hedges' g durlak correction"[i] <- df$"Hedges' g"[i] * corr_durlak
    
    df$"sd cohen"[i] <- sqrt((sdA^2 + sdB^2) / 2)
    df$"Cohen's d"[i] <- (mB - mA) / df$"sd cohen"[i]
    
  }
  
  out <- list(
    smd = df, 
    phases.A = keep$phases_A, 
    phases.B = keep$phases_B 
  )
  
  class(out) <- c("sc_smd")
  
  source_attributes <- attributes(data_list)[[opt("scdf")]]
  nm <- opts("phase", "mt", "dv")
  attributes(out)[nm] <- source_attributes[nm]
  
  out
  
}

#' Single-Case Data Analyses
#'
#' A collection of procedures for analysing, visualising, and managing
#' single-case data.
#'
#' @name scan-package
#' @aliases scan-package scan
#' @author Juergen Wilbert \[aut, cre\]
#' @keywords package
#' @import nlme
#' @import stats
#' @import graphics
#' @import knitr
#' @import kableExtra
#' @import methods
#' @import gt
#' @import MCMCglmm
#' @importFrom car Anova
#' @importFrom cli col_blue
#' @importFrom utils install.packages read.csv capture.output combn
#' @keywords internal
"_PACKAGE"

#' @rdname export
#' @export
export.sc_smd <- function(object, caption = NA, footnote = NA, 
                          filename = NA,
                          select = c("Case", "Mean A" = "mA", "Mean B" = "mB", 
                                     "SD A" =  "sdA", "SD B" = "sdB", 
                                     "SD Cohen" =  "sd cohen", 
                                     "SD Hedges" = "sd hedges", "Glass' delta", 
                                     "Hedges' g", "Hedges' g correction", 
                                     "Hedges' g durlak correction", 
                                     "Cohen's d"),
                          round = 2,
                          decimals = 2,
                          flip = FALSE,
                          ...) {
  
  if (is.na(caption)) caption <- c(
    "Standardizes mean differences. ",
    .phases_string(
      object$phases.A, 
      object$phases.B
    )
  )
  
  footnote <- c(
    'SD Cohen = unweigted average of the variance of both phases',
    'SD Hedges = weighted average of the variance of both phases with a degrees of freedom correction',
    "Glass' delta = mean difference divided by the standard deviation of the A-phase",
    "Hedges' g = mean difference divided by SD Hedges",
    "Hedges' g (durlak) correction = approaches for correcting Hedges' g for small sample sizes",
    "Cohens d = mean difference divided by SD Cohen"
  )
  
  caption <- paste0(caption, collapse = "")
  
  out <- object$smd
  
  out <- .select(out, select)
  
  if (isTRUE(flip)) {
    cases <- out$Case
    out[-1] <- round(out[-1], round)
    names_par <- colnames(out)[-1]
    out <- t(out[-1]) |> as.data.frame()
    out <- cbind(Statistic = rownames(out), out)
    colnames(out) <- c("Statistic", cases)
  }
  
  table <- .create_table(
    out, 
    caption = caption,
    footnote = footnote,
    decimals = decimals,
    ...
  )
  
  if (!is.na(filename)) .save_export(table, filename)
  
  table
}

#' Select and combine phases for overlap analyses
#'
#' Useful when working with pipe operators. This function allows to select and
#' combine specific phases from an scdf for overlap analyses. For example, in an
#' ABAB design one might want to combine both A phases and both B phases (e.g.,
#' `A = c(1, 3)`, `B = c(2, 4)`). The resulting scdf can then be used in overlap
#' functions such as [overlap()].
#'
#' @details This function selects and combines the specified phases from the
#'   input scdf and returns the resulting scdf with the selected phases only.
#'   This is particularly useful when working with pipe operators, allowing to
#'   select and combine phases before applying overlap functions. The resulting
#'   scdf contains only the selected and combined phases, with phase labels "A"
#'   and "B". If `phase_names = "auto"`, phase names are generated from the
#'   combination of the names of the recombined phases. For example, combining
#'   phases 1 and 3 will result in the phase name "1_3". This function
#'   simplifies the process of preparing data for overlap analyses by allowing
#'   users to easily select and combine phases in a single step.
#'
#' @family overlap functions
#' @keywords overlap
#' @author Juergen Wilbert
#'
#' @inheritParams .inheritParams
#' @param A Selection of the A phase
#' @param B Selection of the B phase
#' @param phase_names A character vector with names for the resulting phases.
#'   The default `"auto"` generates phase names from the combination of the
#'   names of the recombined phases.
#' @return An scdf with selected phases
#' @examples
#' exampleA1B1A2B2_zvt |>
#'   select_phases(A = c(1, 3), B = c(2, 4)) |>
#'   overlap()
#' @export
select_phases <- function(data, A, B, phase_names = "auto") {
  data <- .prepare_scdf(data)
  keep <- recombine_phases(data, phases = list(A = A, B = B), phase_names = phase_names)
  keep$data
}

#' @rdname export
#' @export
export.sc_nap <- function(object, 
                          caption = NA, 
                          footnote = NA, 
                          filename = NA,
                          select = c("Case", "NAP", "NAP Rescaled", 
                                     "w", "p", "d", "R\u00B2"),
                          round = 2,
                          ...) {
  
  if (is.na(caption)) caption <- c("Nonoverlap of all pairs")
  
  out <- as.data.frame(object$nap) 
  out <- .select(out, select)
  out$p <- .nice_p(out$p)
  
  out <- round_numeric(out, round)
  
  table <- .create_table(
    out, 
    caption = caption,
    footnote = footnote,
    ...
  )
  
  # finish ------------------------------------------------------------------
  
  if (!is.na(filename)) .save_export(table, filename)
  
  table
}

# Functions to get package options
opt <- function(x) {
  out <- .opt[[x]]
  if (is.null(out)) abort("Option ", x, " not defined.")
  out
}

opts <- function(...) {
  out <- .opt[c(...)] |> unlist()
  if (is.null(out)) abort("Options not all defined.")
  out
}


# Basic options -----------------------------------------------------------

.opt <- list(
  scdf         = "scdf",
  dv           = "var.values",
  phase        = "var.phase",
  mt           = "var.mt",
  case_name    = "name",
  info         = "info",
  author       = "author",
  rigorous_class_check = TRUE,
  check_arguments = TRUE,
  names_default = paste0("Case", 1:500),
  function_experimental_warning  = paste(
    "This function is in an experimental state.",
    "The syntax and behaviour will probably change in a future version."
  ),
  function_deprecated_warning = paste(
    "This function is deprecated. It might be dropped without any further",
    "notice in a future update of scan."
  ),
  style = list()
  )

.opt$citation <- {
  x <- citation("scan")
  class(x) <- "list"
  attributes(x[[1]])$textVersion
}

.opt$tip <- list(
    "Visit the free online book on scan at: https://jazznbass.github.io/scan-Book/",
    "For information on citing scan, type citation(\"scan\").",
    "Please cite splot (type 'citation(\"scplot\")' for more information).",
    "Start the graphical user interface with shinyscan().",
    "Start scans GUI by clicking 'Addins -> Launch Shiny-Scan'.",
    "Type 'import_scdf()' to start the import helper.",
    "Click 'Addins -> scdf Import' to start the import helper.",
    "Post a question or improvement idea on scan at https://github.com/jazznbass/scan/discussions",
    "Post issues at https://github.com/jazznbass/scan/issues",
    "Find the online help-pages at https://jazznbass.github.io/scan/",
    "The addon package 'scplot' provides sophisticated graphs for single-case data.",
    "Set 'options(scan.export.engine = \"gt\")' and try out the improved table engine for 'export()'.",
    "Use shinyscan to analyse by providing an scdf as an argument, e.g. 'shinyscan(exampleAB)'.",
    "Try out a different visual theme for shinyscan e.g. 'shinyscan(theme = \"united\")'.",
    "Create APA-style html tables with 'export' (e.g. 'hplm(exampleAB) |> export()'."
  )

.opt$startup_message <- paste0(
  "scan ",utils::packageVersion("scan")," (",utils::packageDate('scan'),")\n"
)

#' Single case data frame constructor
#'
#' `scdf()` is the constructor for objects of class `scdf`. It stores data from
#' single-case studies in a structured format suitable for analysis with the `scan` package.
#'
#' @aliases scdf scdf-class as.scdf
#' @inheritParams .inheritParams
#'
#' @param ... One or more vectors representing measurement variables. See the
#'   *Details* section.
#' @param B_start The first measurement point of phase B (simple coding; only
#'   applicable if the design follows a strict AB pattern).
#' @param phase_design A named vector defining the length and label of each
#'   phase. For example: `phase_design = c(A1 = 10, B1 = 10, A2 = 10, B2 = 10)`.
#' @param phase_starts A named vector defining the label and measurement time of
#'   each phase start. For example: `phase_starts = c(A1 = 1, B1 = 6, A2 = 14,
#'   B2 = 19)`.
#' @param name Optional name for the case.
#'
#' @return Returns a single-case data frame `scdf` suitable for all functions in
#'   the `scan` package.
#'
#' @details If no variable matching the name of the dependent variable is
#'   provided (the default name is `values`, which can be changed via the `dvar`
#'   argument), and the first provided variable is unnamed, that variable will
#'   be interpreted as the dependent variable.
#'
#'   If no measurement-time variable is provided (default name `mt`,
#'   configurable via the `mvar` argument), measurement times are automatically
#'   defined as a sequence `(1, 2, 3, ..., n)`.
#'
#'   If the dependent variable is a **named vector**, the names will be used to
#'   define a phase design. For example, `values = c(A = 2, 3, 5, 4, 3, B = 6,
#'   5, 4, 3)` will be interpreted as an AB phase design with five measurements
#'   in phase A and four in phase B.
#'
#'   If a vector matching the name of the phase variable is provided, it will be
#'   used to define the phase design directly.
#'   Otherwise, the phase design can be defined in three alternative ways:
#'   via the `B_start` argument, via the `phase_starts` argument, or via
#'   the `phase_design` argument.
#'   
#'   If `B_start` is provided, a simple AB phase design is assumed, with phase A
#'   starting at measurement time 1 and phase B starting at the measurement time
#'   indicated by `B_start`.
#'   If `phase_starts` is provided, the phase design is constructed based on the
#'   measurement times indicated in the vector.
#'   If `phase_design` is provided, it is used directly to define the phase
#'   design.
#'   If multiple of these options are provided, the priority order is:
#'   `phase_design`, `phase_starts`, `B_start`, phase variable in
#'   data frame, names of dependent variable.
#'   
#'   If none of these options are provided, an error is raised.
#'   
#'   The function can be used to create single-case data frames for multiple cases
#'   separately, which can then be combined into a list for multiple-case
#'   analyses.
#'   
#'   See also the convenience function \code{\link{transform.scdf}} to add new
#'   variables to an existing `scdf` object.
#'   
#'   See the package vignettes for further examples on how to create and work with
#'   `scdf` objects.
#'   
#'   See the section on *Data structure* in the documentation of the
#'   \code{\link{scan}} package for further details on the `scdf` data structure.
#'   
#' @author Juergen Wilbert
#' @family data manipulation functions
#' @examples
#' ## Scores on a letter naming task were collected on eleven days in a row.
#' ## The intervention started after the fifth measurement,
#' ## so the first B phase measurement was 6 (B_start = 6).
#' klaas <- scdf(
#'   c(5, 7, 8, 5, 7, 12, 16, 18, 15, 14, 19),
#'   B_start = 6, name = "Klaas"
#' )
#' describe(klaas)
#'
#' # Alternative: using named vector
#' klaas <- scdf(
#'   c(A = 5, 7, 8, 5, 7, B = 12, 16, 18, 15, 14, 19),
#'   name = "Klaas"
#' )
#'
#' # Alternative: using phase_design
#' klaas <- scdf(
#'   c(5, 7, 8, 5, 7, 12, 16, 18, 15, 14, 19),
#'   phase_design = c(A = 5, B = 6), name = "Klaas"
#' )
#'
#' # Alternative: using phase_starts
#' klaas <- scdf(
#'   c(5, 7, 8, 5, 7, 12, 16, 18, 15, 14, 19),
#'   phase_starts = c(A = 1, B = 7), name = "Klaas"
#' )
#'
#' ## Unfortunately in a similar study there were no data collected on
#' ## days 3 and 9. Use NA to pass them to the function:
#' emmi <- scdf(c(5, 7, NA, 5, 7, 12, 16, 18, NA, 14, 19),
#'   phase_design = c(A = 5, B = 6), name = "Emmi"
#' )
#' describe(emmi)
#'
#' ## In a MBD over three cases, data were collected eleven days in a row.
#' ## Intervention starting points differ between subjects as they were
#' ## randomly assigned. The three SCDFs are then combined in a list for
#' ## further conjoined analyses.
#' charlotte <- scdf(c(A = 5, 7, 10, 5, 12, B = 7, 10, 18, 15, 14, 19))
#' theresa <- scdf(c(A = 3, 4, 3, 5, B = 7, 4, 7, 9, 8, 10, 12))
#' tonia <- scdf(c(A = 9, 8, 8, 7, 5, 7, B = 6, 14, 15, 12, 16))
#' mbd <- c(charlotte, theresa, tonia)
#' names(mbd) <- c("Charlotte", "Theresa", "Tonia")
#' overlap(mbd)
#'
#' ## In a classroom-based intervention it was not possible to measure outcomes
#' ## every day, but only on schooldays. The sequence of measurements is passed
#' ## to the package by using a vector of measurement times.
#' frida <- scdf(
#'   c(A = 3, 2, 4, 2, 2, 3, 5, 6, B = 8, 10, 8, 12, 14, 13, 12),
#'   mt = c(1, 2, 3, 4, 5, 8, 9, 10, 11, 12, 14, 15, 16, 17, 18)
#' )
#' summary(frida)
#' describe(frida)
#'
#' ## example with two independent variables and four phases
#' jim <- scdf(
#'   zvt = c(47, 58, 76, 63, 71, 59, 64, 69, 72, 77, 76, 73),
#'   d2 = c(131, 134, 141, 141, 140, 140, 138, 140, 141, 140, 138, 140),
#'   phase_design = c(A1 = 3, B1 = 3, A2 = 3, B2 = 3), dvar = "zvt"
#' )
#' overlap(jim, phases = list(c("A1", "A2"), c("B1", "B2")))
#'
#' @export
scdf <- function(...,
                 B_start = NULL,
                 phase_starts = NULL,
                 phase_design = NULL,
                 name = NULL,
                 dvar = "values",
                 pvar = "phase",
                 mvar = "mt") {
  
  # Catch ellipse ----
  
  df <- list(...)
  
  ## set first entry to dvar if unnamed and dvar not within df 
  if (is.null(names(df)[1]) || identical(names(df)[1], "")) {
    if (!(dvar %in% names(df))) names(df)[1] <- dvar
  }
  
  # check arguments and df ----

  check_args(
    has_length(pvar, 1),
    has_length(mvar, 1),
    not("var.values" %in% names(df), "'var.values' is deprecated. Use 'dvar'."),
    not(
      "phase.design" %in% names(df), 
      "'phase.design' is deprecated. Use 'phase_design'."
    ),
    not("B.start" %in% names(df), "'B.start' is deprecated. Use 'B_start'."),
    not(("MT" %in% names(df)) && mvar == "mt", "Rename argument 'MT' to 'mt'."),
    is_true(
      all(dvar %in% names(df)), 
      "Dependent variable ", dvar, " not found in scdf definition."
    ),
    is_true(
      all(lengths(df) == lengths(df)[1]), 
      "All variables must have the same length. \nActual lengths are: ", 
      paste0(names(lengths(df)), " = ", lengths(df), collapse = ", "), "."
    )
  )
  
  # create default mt column ----
  if (!(mvar %in% names(df))) df[[mvar]] <- 1:length(df[[dvar]])
  
  # create phase_design ----
  
  ## from a named vector
  if (!is.null(names(df[[dvar]]))) {
    tmp_names <- names(df[[dvar]])
    tmp <- c(which(tmp_names != ""), length(tmp_names) + 1)
    phase_design <- tmp[-1] - tmp[-length(tmp)]
    names(phase_design) <- tmp_names[which(tmp_names != "")]
  }

  ## from phase variable
  if (!is.null(df[[pvar]])) {
    tmp_phase <- rle(df[[pvar]])
    phase_design <- tmp_phase$lengths
    names(phase_design) <- tmp_phase$values
  }
  
  # from B_start
  if (!is.null(B_start)) {
    phase_design <- phase_starts2phase_design(
      list(A = df[[mvar]][1], B = B_start), 
      df[[mvar]]
    )
  }
  
  # from phase_starts
  if (!is.null(phase_starts)) 
    phase_design <- phase_starts2phase_design(phase_starts, df[[mvar]])
  
  if (is.null(phase_design)) {
    abort("Phase design not defined correctly!")
  }
  
  df[[pvar]] <- factor(
    rep(names(phase_design), phase_design),
    levels = unique(names(phase_design))
  )
  
  # return ----
  if (length(df[[pvar]]) != length(df[[dvar]])) {
    abort("The phase argument suggests ", length(df[[pvar]]), " measures but ",
      length(df[[dvar]]), " measurements are available.")
  }
  data <- list(as.data.frame(df))
  attributes(data) <- .default_attributes()
  dv(data) <- dvar
  phase(data) <- pvar
  mt(data) <- mvar
  if (!is.null(name)) names(data) <- name

  data
}

phase_starts2phase_design <- function(starts, mt) {
  ids <- lapply(starts, function(x) which(x == mt))
  check <- lapply(ids, function(x) {
    if (length(x) == 0) 
      abort("phase_starts not defined correctly. ", 
           "Measurement time does not exist.")
  })

  if (ids[1] != 1) {
    abort("phase_starts not defined correctly. ", 
         "First phase must start at the first measurement time which is ", 
         mt[1], ".")
  }
    
  
  phase_design <- list()
  for (i in 2:length(ids)) {
    phase_design[[names(starts)[i - 1]]] <- ids[[i]] - ids[[i - 1]]
  }
  phase_design[[names(starts)[length(starts)]]] <- length(mt) - sum(unlist(phase_design))
  phase_design
}

#' Plot random distribution
#' 
#' This function takes the return of the rand_test function and creates a 
#' histogram with the distribution of the rand sample statistics.
#'
#' @param object Object returned from the rand_test() function
#' @param xlab Label for the x-axis.
#' @param ylab Label for the y-axis.
#' @param type 'hist' or 'xy'.
#' @param title Plot title.
#' @param text_observed Text for marking the number of observed statistic.
#' @param color Bar color.
#' @param ... Further arguments passed to the plot function.
#' @export
plot_rand <- function(object, 
                      type = "hist",
                      xlab = NULL, 
                      ylab = NULL, 
                      title = NULL, 
                      text_observed = "observed", 
                      color = "lightgrey", 
                      ...) {
  
  if (type == "xy") {
    if (is.null(ylab)) ylab <- object$statistic
    if (is.null(xlab)) xlab <- "Start phase B"
    plot(
      object$distribution_startpoints[[1]], 
      object$distribution, 
      xlim = c(min(object$distribution_startpoints), max(object$distribution_startpoints)),
      ylab = ylab, xlab = xlab, main = title,
      xaxt = "n"
    )
    axis(1, at = object$distribution_startpoints[[1]])
    points(object$n1 + 1, object$observed.statistic, col = "blue", pch = 19)
    x <- which(object$distribution > object$observed.statistic)
    y <- object$distribution[x]
    points(x+1, y, col = "red", pch = 19)
    
    if (object$testdirection == "greater") 
      abline(h = max(object$distribution), col = "red")
    else
      abline(h = min(object$distribution), col = "red")
  }
  
  
  if (type == "hist") {
    if (is.null(ylab)) ylab <- "Frequency"
    if (is.null(title)) title <- "Random distribution"
    h <- hist(object$distribution, plot = FALSE)
    lab <- paste0(round(h$counts / length(object$distribution) * 100, 0), "%")
    xlim <- c(min(h$breaks, na.rm = TRUE), max(h$breaks, na.rm = TRUE))
    ylim <- round(max(h$counts * 1.2))
    if (object$observed.statistic < xlim[1]) xlim[1] <- object$observed.statistic
    if (object$observed.statistic > xlim[2]) xlim[2] <- object$observed.statistic
    
    if (is.null(xlab)) xlab <- object$statistic
    hist(
      object$distribution,
      xlab = xlab,
      labels = lab,
      xlim = xlim,
      ylim = c(0, ylim),
      ylab = ylab,
      main = title,
      col = color
    )
    abline(v = object$observed.statistic, lty = 2, lwd = 2, col = "grey")
    if (object$p.value < 0.5) pos <- 2 else pos <- 4
    text(object$observed.statistic, ylim, text_observed, pos = pos)
  }

}

# color styles ------------------------------------------------------------

.opt$style$default <- list(
  frame = "black",las = 1, mai = c(0.6, 0.58, 0.2, 0.2), 
  bty = "o", text.ABlag = NULL, pch = 17, font = "sans", 
  fill = FALSE, fill.bg = FALSE, grid = FALSE,  
  lwd = 2, lwd.seperators = 1.5, lwd.grid = 1,
  lty = "solid", lty.seperators = "dashed", lty.grid = "dotted",
  
  ylab.orientation = 0,
  
  cex = 1, cex.axis = 0.8, cex.text = 1, cex.lab = 1,
  cex.xlab = 1, cex.ylab = 1, cex.xaxis = 0.8, cex.yaxis = 0.8,
  cex.casenames = 1, cex.phasenames = 1, cex.dots = 0.8,
  
  col.lines = "black", col.dots = "black", col.seperators = "black", 
  col.fill = "grey75", col.fill.bg = "grey95", col.bg = "white", col = "black", 
  col.text = "black", col.casenames = "black", col.phasenames = "black",
  col.xlab = "black", col.ylab = "black", col.ridge = NULL,
  
  align.main = "center", align.caption = "left",
  margin.main = 0, margin.caption = 0.02, wrap.caption = NULL, wrap.main = NULL,
  parse.main = FALSE, parse.caption = FALSE,
  seperators.extent = "full",
  
  annotations = NULL, 
  
  names = list(side = 3, line = -1, adj = 0, col = "black", cex = 1)
)

.opt$style$yaxis <- list(
  ylab.orientation = 1, mai = c(0.6, 0.85, 0.2, 0.2), cex.lab = 0.8,
  cex.text = 0.8
)

.opt$style$tiny <- list(
  cex.text = 0.5, cex = 0.5, cex.lab = 0.5, 
  cex.casenames = 0.5, cex.phasenames = 0.5, cex.dots = 0.5,
  lwd = 0.7, lwd.seperators = 0.7, lwd.grid = 0.7,
  mai = c(0.3,0.3,0.1,0.05),
  names = list(side = 3, line = -1, adj = 0, col = "black", cex = 0.5)
)

.opt$style$small <- list(
  cex.text = 0.75, cex = 0.75, cex.lab = 0.75, cex.dots = 0.75,
  cex.casenames = 0.75, cex.phasenames = 0.75,
  lwd = 0.85, lwd.seperators = 0.85, lwd.grid = 0.85,
  mai = c(0.5,0.5,0.15,0.1),
  names = list(side = 3, line = -1, adj = 0, col = "black", cex = 0.75)
)

.opt$style$big <- list(
  cex.text = 1.25, cex = 1.25, cex.lab = 1.25, 
  cex.casenames = 1.25, cex.phasenames = 1.25, cex.dots = 1.25,
  lwd = 1.5, lwd.seperators = 1.5, lwd.grid = 1.5,
  mai = c(0.8,1,0.2,0.2),
  names = list(side = 3, line = -1, adj = 0, col = "black", cex = 0.75)
)


.opt$style$chart <- list(
  col.ridge = "grey50",
  fill.bg = TRUE, col.fill.bg = "grey98", fill = TRUE, col.fill = "grey50", 
  annotations = list(cex = 0.6, col = "black", offset = 0.4, round = 1, pos = 3),
  pch = 19, 
  frame = NA, grid = TRUE, col.grid = "grey75", lwd = 0.7, cex.text = 0.8, 
  cex.lab = 0.8, 
  cex.casenames = 0.8, cex.phasenames = 0.8
)


.opt$style$ridge <- list(
  col.ridge = "grey50",
  fill = "grey50", fill.bg = TRUE, col.fill.bg = "grey95", pch = 20
)

.opt$style$annotate <- list(
  annotations = list(cex = 0.6, col = "black", offset = 0.4, round = 1, pos = 3), pch = 19
)

.opt$style$grid <- list(
  frame = NA, grid = TRUE, col.grid = "lightblue", fill.bg = TRUE, 
  col.fill.bg = "grey95", lwd = 0.7, pch = 19, cex.axis = 0.8
)  

.opt$style$grid2 <- list(
  col.ridge = "white",
  fill = "white", grid = TRUE, col.grid = "lightgreen", frame = "black", 
  fill.bg = TRUE, col.fill.bg = "grey95", lwd = 0.7, pch = 1, cex.axis = 0.8
)  

.opt$style$dark <- list(
  fill.bg = TRUE, col.fill.bg = "black", bty = "o", col.lines = "gold", 
  col.bg = "grey10", col.dots = "red", col.seperators = "white", col = "white", 
  col.text = "white", col.casenames = "white", col.phasenames = "white")

.opt$style$nodots <- list(
  col.ridge = "grey95",
  type = "l", col.dots = "", fill = TRUE, col.fill = "grey95", grid = TRUE, 
  col.grid = "grey80", fill.bg = TRUE, col.fill.bg = "grey99")

.opt$style$sienna <- list(
  grid = TRUE, col.grid = "orange", pch = 18, col.lines = "grey85", 
  col.dots = "seagreen4", lwd = 2, col.bg = "seashell", fill.bg = "moccasin", 
  col.text = "sienna4", col = "darkolivegreen", col.seperators = "sienna4", 
  col.casenames = "sienna4", col.phasenames = "sienna4",
  cex.text = 0.8, cex.lab = 0.8, cex.axis = 0.7, 
  cex.casenames = 0.8, cex.phasenames = 0.8,
  frame = "darkseagreen", 
  font = "serif")

.opt$style$phase_color <- list(
  fill.bg = TRUE, col.fill.bg = c("aliceblue", "mistyrose1", "honeydew")
)

.opt$style$phase_shade <- list(
  fill.bg = TRUE,
  col.fill.bg = c("grey94", "grey99", "grey90")
)

#' Set and get scdf attributes
#'
#' @param x Variable
#' @param var Attribute
#'
#' @return Attribute value
#' @keywords internal
#' @export
scdf_attr <- function(x, var = NULL) {
  if (is.null(var)) return(attr(x, opt("scdf")))
  out <- attr(x, opt("scdf"))
  out[[var]]
}

#' @rdname scdf_attr
#' @param value set value
#' @export
"scdf_attr<-" <- function(x, var, value) {
  
  if (missing(var)) {
    attr(x, opt("scdf")) <- value
    return(x)
  }
  scdf_attr <- attr(x, opt("scdf"))
  if (is.null(scdf_attr)) scdf_attr <- list()
  
  scdf_attr[[var]] <- value
  attr(x, opt("scdf")) <- scdf_attr
  x
}

#' @rdname scdf_attr
#' @param scdf An [scdf] object.
dv <- function(scdf) scdf_attr(scdf, opt("dv"))
"dv<-" <- function(x, value) {
  scdf_attr(x, opt("dv")) <- value
  x
}


#' @rdname scdf_attr
mt <- function(scdf) scdf_attr(scdf, opt("mt"))
"mt<-" <- function(x, value) {
  scdf_attr(x, opt("mt")) <- value
  x
}

#' @rdname scdf_attr
phase <- function(scdf) scdf_attr(scdf, opt("phase"))
"phase<-" <- function(x, value) {
  scdf_attr(x, opt("phase")) <- value
  x
}

#' @rdname export
#' @export
export.sc_pet <- function(object, 
                          caption = NA, 
                          footnote = NA, 
                          filename = NA,
                          round = 1,
                          ...) {
  
  if (is.na(caption)) caption <- c("Percent Exceeding the trend")
  
  if (is.na(footnote)) {
    if (object$decreasing) {
      footnote <- c(
        "Assumed decreasing values in the B-phase",
        "Binomial test alternative hypothesis: true probability < 50%",
        "Single-sided test"
      )
    } else {
      footnote <- c(
        "Assumed increasing values in the B-phase",
        "Binomial test alternative hypothesis: true probability > 50%",
        "Single-sided test"
      )
    }
  }
  
  object$PET$binom.p <- .nice_p(object$PET$binom.p)
  names(object$PET)[4] <- "p (binomial test)"
  
  new <- paste0(
    "Percentage ", if (object$decreasing) "< lower " else "> upper ", 
    object$ci * 100, " CI"
  )
  names(object$PET)[3] <- new
  object$PET <- round_numeric(object$PET, round)
  
  table <- .create_table(
    object$PET, 
    caption = caption,
    footnote = footnote,
    ...
  )
  
  # finish ------------------------------------------------------------------
  
  if (!is.na(filename)) .save_export(table, filename)
  table
}

#' Piecewise linear model / piecewise regression
#'
#' The `plm` function computes a piecewise regression model (see Huitema &
#' McKean, 2000) for one case. The function automatically creates the default
#' model which includes trend, level, and slope effects. The model can be
#' changed with `trend`, `level`, and `slope` arguments or by providing a custom
#' formula.
#'
#' The function uses the `glm` function of the stats package or the `gls`
#' function of the nlme package (for auto-regression models).
#' 
#' @inheritParams .inheritParams
#' @order 1
#' @param AR Maximal lag of autoregression. Modelled based on the
#'   Autoregressive-Moving Average (ARMA) function.  When AR is set, the family
#'   argument must be set to `family = "gaussian"`.
#' @param family Set the distribution family. Defaults to a gaussian
#'   distribution. See the `family` function for more details.
#' @param formula Defaults to the standard piecewise regression model. The
#'   parameter phase followed by the phase name (e.g., phaseB) indicates the
#'   level effect of the corresponding phase. The parameter 'inter' followed by
#'   the phase name (e.g., interB) adresses the slope effect based on the method
#'   provide in the model argument (e.g., "B&L-B"). The formula can be changed
#'   for example to include further variables into the regression model.
#' @param update An easier way to change the regression formula (e.g., `. ~ . +
#'   newvariable`).
#' @param na.action Defines how to deal with missing values.
#' @param r_squared Logical. If TRUE, delta r_squares will be calculated for
#'   each predictor.
#' @param var_trials Name of the variable containing the number of trials (only
#'   for binomial regressions). If a single integer is provided this is
#'   considered to be a the constant number of trials across all measurements.
#' @param dvar_percentage Only for binomial distribution. If set TRUE, the
#'   dependent variable is assumed to represent proportions `[0,1]`. Otherwise
#'   dvar is assumed to represent counts.
#' @param ... Further arguments passed to the [stats::glm()] or [nlme::gls()]
#'   function.
#' @return An object of class `sc_plm`.
#' \item{formula}{plm formula. Uselful if you want to use the update or
#'   formula argument and you don't know the names of the parameters.}
#' \item{model}{Character string from function call (see `Arguments`
#'   above).}
#' \item{F.test}{F-test values of modelfit.} \item{r.squares}{Explained variance
#' R squared for each model parameter.}
#' \item{ar}{Autoregression lag from function call (see `Arguments`
#'   above).}
#' \item{family}{Distribution family from function call
#'   (see `Arguments` above).}
#'   \item{full.model}{Full regression model list from the gls or glm function.}
#' @author Juergen Wilbert
#' @family regression functions
#' @references Beretvas, S., & Chung, H. (2008). An evaluation of modified
#'   R2-change effect size indices for single-subject experimental designs.
#'   *Evidence-Based Communication Assessment and Intervention, 2*,
#'   120-128.
#'
#'   Huitema, B. E., & McKean, J. W. (2000). Design specification issues in
#'   time-series intervention models. *Educational and Psychological
#'   Measurement, 60*, 38-58.
#' @examples
#'
#' ## Compute a piecewise regression model for a random single-case
#' set.seed(123)
#' AB <- design(
#'   phase_design = list(A = 10, B = 20),
#'   level = list(A = 0, B = 1), slope = list(A = 0, B = 0.05),
#'   trend = 0.05
#' )
#' dat <- random_scdf(design = AB)
#' plm(dat, AR = 3)
#'
#' ## Another example with a more complex design
#' A1B1A2B2 <- design(
#'   phase_design = list(A1 = 15, B1 = 20, A2 = 15, B2 = 20),
#'   level = list(A1 = 0, B1 = 1, A2 = -1, B2 = 1),
#'   slope = list(A1 = 0, B1 = 0.0, A2 = 0, B2 = 0.0),
#'   trend = 0.0)
#' dat <- random_scdf(design = A1B1A2B2, seed = 123)
#' plm(dat, contrast = "preceding")
#'
#' ## no slope effects were found. Therefore, you might want to the drop slope
#' ## estimation:
#' plm(dat, slope = FALSE, contrast = "preceding")
#'
#' ## and now drop the trend estimation as well
#' plm(dat, slope = FALSE, trend = FALSE, contrast = "preceding")
#'
#' ## A poisson regression
#' example_A24 |>
#'   plm(family = "poisson")
#'
#' ## A binomial regression (frequencies as dependent variable)
#' plm(exampleAB_score$Christiano, family = "binomial", var_trials = "trials")
#'
#' ## A binomial regression (percentage as dependent variable)
#' exampleAB_score$Christiano |>
#'   transform(percentage = values/trials) |>
#'   set_dvar("percentage") |>
#'   plm(family = "binomial", var_trials = "trials", dvar_percentage = TRUE)
#' @export
plm <- function(data, dvar, pvar, mvar, 
                AR = 0, 
                model = c("W", "H-M", "B&L-B", "JW"),
                family = "gaussian", 
                trend = TRUE, 
                level = TRUE, 
                slope = TRUE,
                contrast = c("first", "preceding"),
                contrast_level = c(NA, "first", "preceding"),
                contrast_slope = c(NA, "first", "preceding"),
                formula = NULL, 
                update = NULL, 
                na.action = na.omit,
                r_squared = TRUE,
                var_trials = NULL,
                dvar_percentage = FALSE,
                ...) {
  
  check_args(
    has_length(data, 1, 
               "plm can not be applied to more than one case (use hplm)."),
    not(family != "gaussian" && AR != 0, 
        "family is not 'gaussian' but AR is set."),
    not(family == "binomial" && is.null(var_trials),
        "family is 'binomial' but 'var_trials' is not defined."),
    by_call(model),
    by_call(contrast_level),
    by_call(contrast_slope),
    by_call(contrast),
    at_least(AR, 0)
  )
  
  # set defaults attributes
  if (missing(dvar)) dvar <- dv(data) else dv(data) <- dvar
  if (missing(pvar)) pvar <- phase(data) else phase(data) <- pvar
  if (missing(mvar)) mvar <- mt(data) else mt(data) <- mvar

  data <- .prepare_scdf(data)

  if (model == "JW") {
    contrast_level <- "preceding"
    contrast_slope <- "preceding"
    model <- "B&L-B"
  }
  
  if (is.na(contrast_level)) contrast_level <- contrast
  if (is.na(contrast_slope)) contrast_slope <- contrast
  
  if (family != "gaussian") r_squared <- FALSE
  
  original_attr <- attributes(data)[[opt("scdf")]]
  
  regmodel <- if (AR == 0) "glm" else "gls"
  
  # formula definition ------------------------------------------------------
  
  tmp_model <- .add_dummy_variables(
    data = data, 
    model = model, 
    contrast_level = contrast_level, 
    contrast_slope = contrast_slope
  )

  data  <- tmp_model$data[[1]]
  
  if(is.null(formula)) {
    formula <- .create_fixed_formula(
      dvar, mvar, slope, level, trend, 
      tmp_model$var_phase, tmp_model$var_inter
    ) 
  } 
  
  if(!is.null(update)) formula <- update(formula, update)
  
  term_labels <- attr(terms(formula), "term.labels")
  predictors  <- if (length(term_labels)) term_labels else character(0)
  
  #predictors <- as.character(formula[3])
  #predictors <- unlist(strsplit(predictors, "\\+"))
  #predictors <- trimws(predictors)
  #if(!is.na(match("1", predictors)))
  #  predictors <- predictors[-match("1", predictors)]
 
  formula_full <- formula
  formulas_restricted  <- sapply(
    predictors, function(x) update(formula, formula(paste0(".~. - ", x)))
  )
  
  # glm models --------------------------------------------------------------
  
  if(regmodel == "glm") {
    
    if (family != "binomial") {
      full <- glm(
        formula_full, data = data, family = family, na.action = na.action, ...
      )
    }
    
    if (family == "binomial") {
      if (is.character(var_trials)) {
        trials <- data[[var_trials]]
      } else {
        trials <- rep(var_trials, length(data[[dvar]]))
      }
      if (!dvar_percentage) data[[dvar]] <- data[[dvar]] / trials
      full <- glm(
        formula_full, data = data, family = family, na.action = na.action, 
        weights = trials, ...
      )
    }
    
    if (r_squared) {
      restricted.models <- lapply(
        formulas_restricted, 
        function(x) 
          glm(x, data = data, family = family, na.action = na.action, ...)
      )
    }
  }
  
  if(regmodel == "gls") {
    full <- gls(
      formula_full, data = data, correlation = corARMA(p = AR), 
      method = "ML", na.action = na.action
    )
    
    if (r_squared) {
      restricted.models <- lapply(
        formulas_restricted, 
        function(x) 
          gls(model = x, data = data, correlation = corARMA(p = AR), 
              method = "ML", na.action = na.action
          )
      )
    }
    
  }
  
  # F and R-Squared ---------------------------------------------------------
  
  if (family == "gaussian") {
    
    if (regmodel == "gls") {
      df_residuals <- full$dims$N - full$dims$p
      df_intercept <- if ("(Intercept)" %in% names(coef(full))) 1 else 0
      y <- fitted(full) + residuals(full)#model.response(model.frame(full))
    }
   
    if (regmodel == "glm") {
      df_residuals <- full$df.residual
      df_intercept <- if (attr(full$terms, "intercept")) 1 else 0
      y <- model.response(model.frame(full))
    }
   
    residuals <- as.numeric(resid(full))
    n <- length(residuals)
    #df_effect <- n - 1 - df_residuals
    df_model <- length(coef(full))
    df_effect <- df_model - df_intercept

    qse <- sum(residuals^2)
    
    if (df_intercept == 1) {
      qst <- sum((y - mean(y))^2)
    } else {
      qst <- sum(y^2)
    }
    
    mqsa <- (qst - qse) / df_effect
    mqse <- qse / df_residuals
    f_value <- mqsa / mqse
   
    if (!is.infinite(f_value)) {
      p <- pf(f_value, df_effect, df_residuals, lower.tail = FALSE)
    } else {
      p <- NULL
    }
    
    if (df_intercept == 1) {
      total_variance <- qst / (n - 1) # identical to var(y)
      mse <- var(residuals)
    } else {
      total_variance <- qst / n
      mse <- sum(residuals^2) / n
    }
    
    r2 <- 1 - mse / total_variance
    r2_adj <- 1 - (1 - r2) * ((n - df_intercept) / df_residuals)
    
    f_test <- c(
      F = f_value, 
      df1 = df_effect, 
      df2 = df_residuals, 
      p = p, 
      R2 = r2, 
      R2.adj = r2_adj
    )

    if (r_squared) {
      r_squares <- lapply(restricted.models, function(x) 
        r2 - (1 - (var(resid(x), na.rm = TRUE) / total_variance))
      )
      r_squares <- unlist(r_squares)
    } else {
      r_squares <- NA
    }
    
 
  }
  
  if (family != "gaussian") {
    f_test <- NA
    r_squares <- NA
  }
  
  
  
  # output ------------------------------------------------------------------
  
  out <- list(
    formula = formula_full, 
    model = model, 
    contrast = list(level = contrast_level, slope = contrast_slope),
    F.test = f_test, 
    r.squares = r_squares, 
    ar = AR, 
    family = family, 
    var_trials = var_trials,
    dvar_percentage = dvar_percentage,
    full.model = full, 
    data = data
  )
  
  class(out) <- c("sc_plm")
  
  nm <- opts("phase", "mt", "dv")
  attributes(out)[nm] <- original_attr[nm]
  
  out
}

test_that("plm", {
  
  result <- plm(exampleAB$Johanna)
  expect_identical(round(sum(coef(result)), 5), 94.64233) 
  expect_s3_class(result, "sc_plm")
  
  result <- plm(
    exampleAB_score$Christiano,
    family = "binomial",
    var_trials = "trials"
  )
  
  expect_identical(round(sum(coef(result)), 3), 3.565)
  
  result_2 <- plm(
    exampleAB_score$Christiano,
    family = "binomial",
    var_trials = 20
  )

  expect_identical(coef(result), coef(result_2))
  
})

test_that("main", {

  ex_scdf <- exampleA1B1A2B2
  
  l2 <- data.frame(
    case = c("Pawel", "Moritz", "Jannis"),
    gender = c(0,1,2),
    age = c(30, 40, 50)
  )
  
  object <- as.data.frame(ex_scdf)
  
  expect_s3_class(object, "data.frame")
  expect_equal(names(object), c("case", names(ex_scdf[[1]])))
  expect_equal(sum(unlist(lapply(ex_scdf, nrow))), nrow(object))
  
  object <- as.data.frame(ex_scdf, l2 = l2)
  expect_equal(names(object), c("case", names(ex_scdf[[1]]), names(l2)[-1]))

})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scdf-functions.R
\name{$.scdf}
\alias{$.scdf}
\alias{[.scdf}
\title{Select an scdf case by name}
\usage{
\method{$}{scdf}(x, i)

\method{[}{scdf}(x, i)
}
\arguments{
\item{x}{A scdf object.}

\item{i}{A case name from x. If i is not a name of x, a warning is issued.}
}
\value{
A scdf object containing only the selected case.
}
\description{
Selects a single case from a scdf object by its name.
}
\author{
Juergen Wilbert
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/select_phases.R
\name{select_phases}
\alias{select_phases}
\title{Select and combine phases for overlap analyses}
\usage{
select_phases(data, A, B, phase_names = "auto")
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{A}{Selection of the A phase}

\item{B}{Selection of the B phase}

\item{phase_names}{A character vector with names for the resulting phases.
The default \code{"auto"} generates phase names from the combination of the
names of the recombined phases.}
}
\value{
An scdf with selected phases
}
\description{
Useful when working with pipe operators. This function allows to select and
combine specific phases from an scdf for overlap analyses. For example, in an
ABAB design one might want to combine both A phases and both B phases (e.g.,
\code{A = c(1, 3)}, \code{B = c(2, 4)}). The resulting scdf can then be used in overlap
functions such as \code{\link[=overlap]{overlap()}}.
}
\details{
This function selects and combines the specified phases from the
input scdf and returns the resulting scdf with the selected phases only.
This is particularly useful when working with pipe operators, allowing to
select and combine phases before applying overlap functions. The resulting
scdf contains only the selected and combined phases, with phase labels "A"
and "B". If \code{phase_names = "auto"}, phase names are generated from the
combination of the names of the recombined phases. For example, combining
phases 1 and 3 will result in the phase name "1_3". This function
simplifies the process of preparing data for overlap analyses by allowing
users to easily select and combine phases in a single step.
}
\examples{
exampleA1B1A2B2_zvt |>
  select_phases(A = c(1, 3), B = c(2, 4)) |>
  overlap()
}
\seealso{
Other overlap functions: 
\code{\link{ird}()},
\code{\link{nap}()},
\code{\link{overlap}()},
\code{\link{pand}()},
\code{\link{pem}()},
\code{\link{pet}()},
\code{\link{pnd}()},
\code{\link{print.sc_cdc}()},
\code{\link{tau_u}()}
}
\author{
Juergen Wilbert
}
\concept{overlap functions}
\keyword{overlap}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fill_missing.R
\name{fill_missing}
\alias{fill_missing}
\title{Replacing missing measurement points in single-case data}
\usage{
fill_missing(data, dvar, mvar, pvar, interpolate_na = TRUE)
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{mvar}{Character string with the name of the measurement time variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{interpolate_na}{If set \code{TRUE}, \code{NA} values in the measurement time variable are
also interpolated. Default is \code{TRUE}.}
}
\value{
A single-case data frame with interpolated missing data points.
}
\description{
The \code{fillmissing()} function replaces missing measurements in single-case
data. It linearly interpolates missing data points between two existing
measurements for all variables except the measurement time and phase. The
measurement time variable is filled with the missing time points. The phase
variable is copied from the previous measurement time point. If mt values are
missing (\code{NA}), they are also interpolated if \code{interpolate_na = TRUE}.
}
\details{
The \code{fill_missing()} function is designed to handle single-case data
with missing measurement points. It performs linear interpolation to estimate
the missing values based on the existing data points. The function iterates
through each single-case in the provided single-case data frame (scdf) and
identifies gaps in the measurement time variable. For each gap, it calculates
the step size for linear interpolation and fills in the missing values for
all target variables (i.e., all variables except the measurement time and
phase). The interpolated data points are then added to the single-case data
frame, and the final result is sorted by measurement time. This function is
particularly useful for preparing single-case data for further analysis, such
as calculating overlap indices or conducting randomization tests, where
continuous measurement times are required. It ensures that the data is
complete by filling in the missing measurement points in a systematic manner.
}
\examples{

## In his study, Grosche (2011) could not realize measurements each
## single week for all participants. During the course of 100 weeks,
## about 20 measurements per person at different times were administered.

## Fill missing values in a single-case dataset with discontinuous
## measurement times
Grosche2011filled <- fill_missing(Grosche2011)
study <- c(Grosche2011[2], Grosche2011filled[2])
names(study) <- c("Original", "Filled")
study

## An example with multiple interpolated variables

rolf_n <- exampleAB_add
rolf_n[[1]] <- rolf_n[[1]][-c(3,7,8),]
rolf_f <- fill_missing(rolf_n)
study1 <- c("original" = exampleAB_add, "interpolated" = rolf_f)
study1

## Example with missing NAs in measurement time
Maggie2 <- random_scdf(design(level = list(0,1)), seed = 123)
Maggie2_n <- Maggie2
Maggie2_n[[1]][c(5,12:14,20), "mt"] <- NA
Maggie2_f <- fill_missing(Maggie2_n)
study2 <- c("original" = Maggie2, "interpolated" = Maggie2_f)
study2

}
\seealso{
Other data manipulation functions: 
\code{\link{add_l2}()},
\code{\link{as.data.frame.scdf}()},
\code{\link{as_scdf}()},
\code{\link{batch_apply}()},
\code{\link{moving_median}()},
\code{\link{print.sc_outlier}()},
\code{\link{ranks}()},
\code{\link{rescale}()},
\code{\link{scdf}()},
\code{\link{select_cases}()},
\code{\link{set_vars}()},
\code{\link{shift}()},
\code{\link{smooth_cases}()},
\code{\link{standardize}()},
\code{\link{truncate_phase}()}
}
\author{
Juergen Wilbert
}
\concept{data manipulation functions}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scan-package.R
\docType{package}
\name{scan-package}
\alias{scan-package}
\alias{scan}
\title{Single-Case Data Analyses}
\description{
A collection of procedures for analysing, visualising, and managing
single-case data.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/jazznbass/scan/}
  \item \url{https://jazznbass.github.io/scan-Book/}
  \item \url{https://jazznbass.github.io/scan/}
  \item Report bugs at \url{https://github.com/jazznbass/scan/issues}
}

}
\author{
Juergen Wilbert [aut, cre]
}
\keyword{internal}
\keyword{package}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.scdf.R
\name{summary.scdf}
\alias{summary.scdf}
\title{Summary function for an scdf object}
\usage{
\method{summary}{scdf}(object, all_cases = FALSE, ...)
}
\arguments{
\item{object}{scdf}

\item{all_cases}{IF TRUE, more that 10 cases are summarized}

\item{...}{not in use}
}
\description{
Provides a summary of an \code{scdf} object, including the number of cases,
measurements per case, and design information.
}
\details{
The summary includes:
\itemize{
\item Total number of cases in the \code{scdf}.
\item A table listing each case with the number of measurements and design.
\item Variable names with annotations for phase, measurement-time, and dependent variable.
\item Additional information and author details if available.
}
}
\author{
Juergen Wilbert
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bplm.R, R/print-export-bplm.R
\name{bplm}
\alias{bplm}
\alias{print.sc_bplm}
\alias{export.sc_bplm}
\title{Bayesian Piecewise Linear Model (bplm)}
\usage{
bplm(
  data,
  dvar,
  pvar,
  mvar,
  model = c("W", "H-M", "B&L-B"),
  contrast_level = c("first", "preceding"),
  contrast_slope = c("first", "preceding"),
  trend = TRUE,
  level = TRUE,
  slope = TRUE,
  random_trend = FALSE,
  random_level = FALSE,
  random_slope = FALSE,
  fixed = NULL,
  random = NULL,
  update_fixed = NULL,
  ...
)

\method{print}{sc_bplm}(x, digits = 3, ...)

\method{export}{sc_bplm}(
  object,
  caption = NA,
  footnote = NA,
  filename = NA,
  round = 2,
  nice = TRUE,
  ...
)
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{mvar}{Character string with the name of the measurement time variable.
Defaults to the attributes in the scdf file.}

\item{model}{Model used for calculating the dummy parameters (see Huitema &
McKean, 2000). Default is \code{model = "W"}. Possible values are: \code{"B&L-B"},
\code{"H-M"}, \code{"W"}, and deprecated \code{"JW"}.}

\item{contrast_level}{Either "first", "preceding" or a contrast matrix. If NA
contrast_level is a copy of contrast.}

\item{contrast_slope}{Either "first", "preceding" or a contrast matrix. If NA
contrast_level is a copy of contrast.}

\item{trend}{A logical indicating if a trend parameters is included in the
model.}

\item{level}{A logical indicating if a level parameters is included in the
model.}

\item{slope}{A logical indicating if a slope parameters is included in the
model.}

\item{random_trend}{If TRUE, includes a random trend effect.}

\item{random_level}{If TRUE, includes a random level effect.}

\item{random_slope}{If TRUE, includes a random slope effect.}

\item{fixed}{A formula that overwrites the automatically created fixed part
of the regression model that defaults to the standard piecewise regression
model. The parameter phase followed by the phase name (e.g., phaseB)
indicates the level effect of the corresponding phase. The parameter
'inter' followed by the phase name (e.g., interB) adresses the slope effect
based on the method provide in the model argument (e.g., "B&L-B"). The
formula can be changed for example to include further L1 or L2 variables
into the regression model.}

\item{random}{A formula that overwrites the automatically created random part
of the regression model.}

\item{update_fixed}{An easier way to change the fixed model part (e.g., \code{. ~ . + newvariable}).}

\item{...}{Further arguments passed to the \code{\link[MCMCglmm:MCMCglmm]{MCMCglmm::MCMCglmm()}} function.}

\item{x}{An object returned by \code{\link[=bplm]{bplm()}}}

\item{digits}{The minimum number of significant digits to be use.
If set to "auto" (default), values are predefined.}

\item{object}{An scdf or an object exported from a scan function.}

\item{caption}{Character string with table caption. If left NA (default) a
caption will be created based on the exported object.}

\item{footnote}{Character string with table footnote. If left NA (default) a
footnote will be created based on the exported object.}

\item{filename}{String containing the file name. If a filename is given the
output will be written to that file.}

\item{round}{Integer passed to the digits argument used to round values.}

\item{nice}{If set TRUE (default) output values are rounded and optimized for
publication tables.}
}
\value{
An object of class \code{sc_bplm} with element:
\itemize{
\item \code{model} - List containing information about the applied model.
\item \code{N} - Number of single-cases.
\item \code{formula} - A list containing the fixed and the random formulas of the hplm model.
\item \code{mcmglmm} -  Object of class MCMglmm.
\item \code{contrast} -  List with contrast definitions.
}
}
\description{
Computes a bayesian (hierarchical) piecewise linear model based on a Markov
chain Monte Carlo sampler. The function automatically creates the fixed and
random part of the regression model.
}
\details{
The function uses the \code{\link[MCMCglmm:MCMCglmm]{MCMCglmm::MCMCglmm()}} function to fit the model. The
default model includes fixed trend, level, and slope effects as well as a
random intercept for each single-case. The fixed part of the model can be
changed by providing a custom formula to the \code{fixed} argument or by using the
\code{update_fixed} argument. The random part of the model can be changed by
providing a custom formula to the \code{random} argument or by setting the
\code{random_trend}, \code{random_level}, or \code{random_slope} arguments to TRUE.
}
\section{Functions}{
\itemize{
\item \code{print(sc_bplm)}: Print results

\item \code{export(sc_bplm)}: Export results as html table (see \code{\link[=export]{export()}})

}}
\examples{
# plm regression
bplm(example_A24)

# Multilevel plm regression with random intercept
bplm(exampleAB_50, nitt = 5000)

# Adding a random slope
bplm(exampleAB_50, random_level = TRUE, nitt = 5000)

# Custom fixed formula
bplm(exampleAB_50, update_fixed = values ~ -1 + mt + phaseB +
  interB, nitt = 5000)
}
\seealso{
Other regression functions: 
\code{\link{fetch}()},
\code{\link{hplm}()},
\code{\link{mplm}()},
\code{\link{plm}()},
\code{\link{print.sc_ac}()},
\code{\link{print.sc_bctau}()},
\code{\link{trend}()}
}
\author{
Juergen Wilbert
}
\concept{regression functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipe.R
\name{\%>\%}
\alias{\%>\%}
\title{Pipe}
\description{
Several functions in scan are designed to work with pipes \verb{\\\%>\\\%}. This
pipe function is directly imported from magrittr.
}
\details{
Since R 4.1 a pipe operator |> is included in base R. This pipe
operator can be used along with scan perfectly fine.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print-export-cdc.R, R/cdc.R
\name{print.sc_cdc}
\alias{print.sc_cdc}
\alias{export.sc_cdc}
\alias{cdc}
\title{Conservative Dual-Criterion Method}
\usage{
\method{print}{sc_cdc}(x, nice = TRUE, ...)

\method{export}{sc_cdc}(object, caption = NA, footnote = NA, filename = NA, nice = TRUE, ...)

cdc(
  data,
  dvar,
  pvar,
  mvar,
  decreasing = FALSE,
  trend_method = c("OLS", "bisplit", "trisplit"),
  conservative = 0.25,
  phases = c(1, 2)
)
}
\arguments{
\item{x}{Object}

\item{nice}{If set TRUE (default) output values are rounded and optimized for
publication tables.}

\item{...}{Further parameters passed to the print function}

\item{object}{An scdf or an object exported from a scan function.}

\item{caption}{Character string with table caption. If left NA (default) a
caption will be created based on the exported object.}

\item{footnote}{Character string with table footnote. If left NA (default) a
footnote will be created based on the exported object.}

\item{filename}{String containing the file name. If a filename is given the
output will be written to that file.}

\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{mvar}{Character string with the name of the measurement time variable.
Defaults to the attributes in the scdf file.}

\item{decreasing}{If you expect data to be lower in the B phase, set
\code{decreasing = TRUE}. Default is \code{decreasing = FALSE}.}

\item{trend_method}{Method used to calculate the trend line. Default is
\code{trend_method = "OLS"}. Possible values are: \code{"OLS"}, \code{"bisplit"}, and
\code{"trisplit"}. \code{"bisplit"}, and \code{"trisplit"} should only be used for cases
with at least five data-points in both relevant phases.}

\item{conservative}{The CDC method adjusts the original mean and trend lines
by adding (expected increase) or subtracting (expected decrease) an
additional .25 SD before evaluating phase B data. Default is the CDC method
with \code{conservative = .25}. To apply the Dual-Criterion (DC) method, set
\code{conservative = 0}.}

\item{phases}{A vector of two characters or numbers indicating the two phases
that should be compared. E.g., \code{phases = c("A","C")} or \code{phases = c(2,4)}
for comparing the second to the fourth phase. Phases could be combined by
providing a list with two elements. E.g., \code{phases = list(A = c(1,3), B = c(2,4))} will compare phases 1 and 3 (as A) against 2 and 4 (as B). Default
is \code{phases = c(1,2)}.}
}
\value{
\tabular{ll}{
    \tab  \cr
   \code{cdc} \tab CDC Evaluation based on a p-value below .05. \cr
   \code{cdc_exc} \tab Number of phase B datapoints indicating expected change. \cr
   \code{cdc_nb} \tab Number of phase B datapoints. \cr
   \code{cdc_p} \tab P value of Binomial Test. \cr
   \code{cdc_all} \tab Overall CDC Evaluation based on all instances/cases of a Multiple Baseline Design. \cr
   \code{N} \tab Number of cases. \cr
   \code{decreasing} \tab Logical argument from function call (see Arguments above). \cr
   \code{conservative} \tab Numeric argument from function call (see Arguments above). \cr
   \code{case_names} \tab Assigned name of single-case. \cr
   \code{phases} \tab - \cr
}
}
\description{
The \code{cdc()} function applies the Conservative Dual-Criterion Method (Fisher,
Kelley, & Lomas, 2003) to scdf objects. It compares phase B data points to
both phase A mean and trend (OLS, bi-split, tri-split) with an additional
increase/decrease of .25 SD. A binomial test against a 50/50 distribution is
computed and p-values below .05 are labelled "systematic change".
}
\section{Functions}{
\itemize{
\item \code{print(sc_cdc)}: Print results

\item \code{export(sc_cdc)}: Export html results

}}
\examples{

## Apply the CDC method (standard OLS line)
design <- design(n = 1, slope = 0.2)
dat <- random_scdf(design, seed = 42)
cdc(dat)

## Apply the CDC with Koenig's bi-split and an expected decrease in phase B.
cdc(exampleAB_decreasing, decreasing = TRUE, trend_method = "bisplit")

## Apply the CDC with Tukey's tri-split, comparing the first and fourth phase
cdc(exampleABAB, trend_method = "trisplit", phases = c(1,4))

## Apply the Dual-Criterion (DC) method (i.e., mean and trend without
##shifting).
cdc(
 exampleAB_decreasing,
 decreasing = TRUE,
 trend_method = "bisplit",
 conservative = 0
)


}
\references{
Fisher, W. W., Kelley, M. E., & Lomas, J. E. (2003). Visual Aids
and Structured Criteria for Improving Visual Inspection and Interpretation
of Single-Case Designs. \emph{Journal of Applied Behavior Analysis, 36},
387-406. https://doi.org/10.1901/jaba.2003.36-387
}
\seealso{
Other overlap functions: 
\code{\link{ird}()},
\code{\link{nap}()},
\code{\link{overlap}()},
\code{\link{pand}()},
\code{\link{pem}()},
\code{\link{pet}()},
\code{\link{pnd}()},
\code{\link{select_phases}()},
\code{\link{tau_u}()}
}
\author{
Timo Lueke
}
\concept{overlap functions}
\keyword{overlap}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/message-system.R
\name{notify}
\alias{notify}
\title{Throws a message}
\usage{
notify(..., type = "!", detail = 1, warning = FALSE)
}
\arguments{
\item{...}{The message. This can be a character string or a
combination of character strings and variables.}

\item{detail}{The level of detail for the message. Messages with a detail
level higher than the current detail level set in the options will not be
shown. The default is 1.}

\item{warning}{If TRUE, the message will be shown as a warning. If FALSE
(default), the message will be shown as an informational message.}
}
\description{
Messages can be shown with different levels
of detail and as warnings. This is a convenience wrapper for internal use.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/design.R
\name{design}
\alias{design}
\title{Generate a single-case design matrix for multiple random single-cases}
\usage{
design(
  n = 1,
  phase_design = list(A = 5, B = 15),
  trend = 0,
  level = list(0),
  slope = list(0),
  start_value = 50,
  s = 10,
  rtt = 0.8,
  extreme_prop = list(0),
  extreme_range = c(-4, -3),
  missing_prop = 0,
  distribution = c("normal", "gaussian", "poisson", "binomial"),
  random_start_value = FALSE,
  n_trials = NULL,
  mt = NULL,
  B_start = NULL,
  m,
  phase.design,
  MT,
  B.start,
  extreme.p,
  extreme.d,
  missing.p
)
}
\arguments{
\item{n}{Number of cases to be designed (Default is \code{n = 1}).}

\item{phase_design, phase.design}{A list defining the length and label of each
phase. E.g., \code{phase_design = list(A1 = 10, B1 = 10, A2 = 10, B2 = 10)}. Use
vectors if you want to define different values for each case \verb{phase_design = list(A = c(10, 15), B = c(10, 15)}.}

\item{trend}{Defines the effect size of a trend added incrementally to each
measurement across the whole data-set. To assign different trends to
several single-cases, use a vector of values (e.g. \code{trend = c(.1, .3, .5)}). If the number of cases exceeds the length of the vector, values are
recycled. When using a 'gaussian' distribution, the \code{trend} parameters
indicate effect size
\emph{d} changes.  When using a binomial or poisson distribution, \code{trend}
indicates an increase in points / counts per measurement.}

\item{level}{A list that defines the level increase (effect size \emph{d}) at the
beginning of each phase relative to the previous phase (e.g. \code{list(A = 0, B = 1)}). The first element must be zero as the first phase of a single-case
has no level effect (if you have one less list element than the number of
phases, scan will add a leading element with 0 values). Use vectors to
define variable level effects for each case (e.g. \code{list(A = c(0, 0), B = c(1, 2))}). When using a 'gaussian' distribution, the \code{level} parameters
indicate effect size \emph{d} changes. When using a binomial or poisson
distribution, \code{level} indicates an increase in points / counts with the
onset of each phase.}

\item{slope}{A list that defines the increase per measurement for each phase
compared to the previous phase. \code{slope = list(A = 0, B = .1)} generates an
incremental increase of 0.1 per measurement starting at the B phase. The
first list element must be zero as the first phase of a single-case has no
slope effect (if you have one less list element than the number of phases,
scan will add a leading element with 0 values). Use vectors to define
variable slope effects for each case (e.g. \code{list(A = c(0, 0), B = c(0.1, 0.2))}). If the number of cases exceeds the length of the vector, values
are recycled. When using a 'gaussian' distribution, the \code{slope} parameters
indicate effect size \emph{d} changes per measurement. When using a binomial or
poisson distribution, \code{slope} indicates an increase in points / counts per
measurement.}

\item{start_value, m}{Starting value at the first measurement. Default is
\code{50}. When \code{distribution = "poission"} the start_value represents
frequency. When \code{distribution = "binomial"} start_value must range between
0 and 1 and they represent the probability of on event. To assign different
start values to several single-cases, use a vector of values (e.g. \code{c(50, 42, 56)}). If the number of cases exceeds the length of the vector, values
are recycled. The \code{m} argument is deprecated.}

\item{s}{Standard deviation used to calculate absolute values from level,
slope, trend effects and to calculate and error distribution from the \code{rtt}
values. Set to \code{10} by default. To assign different variances to several
single-cases, use a vector of values (e.g. \code{s = c(5, 10, 15)}). If the
number of cases exceeds the length of the vector, values are recycled. if
the distribution is 'poisson' or 'binomial' s is not applied.}

\item{rtt}{Reliability of the underlying simulated measurements. Set \code{rtt = .8} by default. To assign different reliabilities to several single-cases,
use a vector of values (e.g. \code{rtt = c(.6, .7, .8)}). If the number of cases
exceeds the length of the vector, values are repeated. \code{rtt} has no effect
when you're using binomial or poisson distributions.}

\item{extreme_prop, extreme.p}{Probability of extreme values. \code{extreme.p = .05} gives a five percent probability of an extreme value. A vector of
values assigns different probabilities to multiple cases. If the number of
cases exceeds the length of the vector, values are recycled.}

\item{extreme_range, extreme.d}{Range for extreme values. \code{extreme_range = c(-7,-6)} uses extreme values within a range of -7 and -6 . In case of a
binomial or poisson distribution, \code{extreme_range} indicates frequencies. In
case of a gaussian (or normal) distribution it indicates effect size d.
Caution: the first value must be smaller than the second, otherwise the
procedure will fail.}

\item{missing_prop, missing.p}{Portion of missing values. \code{missing_prop = 0.1}
creates 10\\% of all values as missing). A vector of values assigns
different probabilities to multiple cases. If the number of cases exceeds
the length of the vector, values are repeated.}

\item{distribution}{Distribution of the criteria varible. Default is
\code{"normal"}. Possible values are \code{"normal"}, \code{"binomial"}, and \code{"poisson"}.}

\item{random_start_value}{If TRUE, the start_values are randomized based on
the distribution.}

\item{n_trials}{If \code{distribution} (see below) is \code{"binomial"}, \code{n_trials} is
the number of trials/observations/items. E.g., if you simulate accuracy
data with 10 items per measurement, set \code{n_trials = 10}. To assign
different n_trials to several single-cases, use a vector of values (e.g.
\code{n_trials = c(10, 15, 20)}). If the number of cases exceeds the length of
the vector, values are recycled.}

\item{mt, MT}{Number of measurements (in each study). Default is \code{mt = 20}.}

\item{B_start, B.start}{Phase B starting point. The default setting \code{B_start = 6} would assign the first five scores (of each case) to phase A, and all
following scores to phase B. To assign different starting points for a set
of multiple single-cases, use a vector of starting values (e.g., \code{B_start = c(6, 7, 8)}). If the number of cases exceeds the length of the vector,
values will be recycled.}
}
\value{
An object of class sc_design.
}
\description{
Generates a parameter list used for generating multiple random single-cases.
This is used within the \code{random_scdf} function and the \code{power_test} function
and for other Monte-Carlo tasks.
}
\examples{
 ## Create random single-case data and inspect it
 design <- design(
   n = 3, rtt = 0.75, slope = 0.1, extreme_prop = 0.1,
   missing_prop = 0.1
 )
 dat <- random_scdf(design, round = 1, random.names = TRUE, seed = 123)
 describe(dat)

 ## And now have a look at poisson-distributed data
 design <- design(
   n = 3, B_start = c(6, 10, 14), mt = c(12, 20, 22), start_value = 10,
   distribution = "poisson", level = -5, missing_prop = 0.1
 )
 dat <- random_scdf(design, seed = 1234)
 pand(dat, decreasing = TRUE)
}
\author{
Juergen Wibert
}
\concept{mc functions}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/export.R, R/export.sc_desc.R,
%   R/export.sc_nap.R, R/export.sc_overlap.R, R/export.sc_pem.R,
%   R/export.sc_pet.R, R/export.sc_pnd.R, R/export.sc_power.R,
%   R/export.sc_smd.R, R/export.sc_trend.R, R/export.scdf.R,
%   R/export.scdf_summary.R
\name{export}
\alias{export}
\alias{export.sc_desc}
\alias{export.sc_nap}
\alias{export.sc_overlap}
\alias{export.sc_pem}
\alias{export.sc_pet}
\alias{export.sc_pnd}
\alias{export.sc_power}
\alias{export.sc_smd}
\alias{export.sc_trend}
\alias{export.scdf}
\alias{export.scdf_summary}
\title{Export scan objects to html or latex}
\usage{
export(object, ...)

\method{export}{sc_desc}(
  object,
  caption = NA,
  footnote = NA,
  filename = NA,
  flip = FALSE,
  round = 2,
  ...
)

\method{export}{sc_nap}(
  object,
  caption = NA,
  footnote = NA,
  filename = NA,
  select = c("Case", "NAP", "NAP Rescaled", "w", "p", "d", "R²"),
  round = 2,
  ...
)

\method{export}{sc_overlap}(
  object,
  caption = NA,
  footnote = NULL,
  filename = NA,
  round = 2,
  decimals = 2,
  flip = FALSE,
  ...
)

\method{export}{sc_pem}(object, caption = NA, footnote = NA, filename = NA, round = 2, ...)

\method{export}{sc_pet}(object, caption = NA, footnote = NA, filename = NA, round = 1, ...)

\method{export}{sc_pnd}(
  object,
  caption = NA,
  footnote = NA,
  filename = NA,
  select = c("Case", "PND", "Total", "Exceeds"),
  round = 2,
  ...
)

\method{export}{sc_power}(object, caption = NA, footnote = NA, filename = NA, round = 3, ...)

\method{export}{sc_smd}(
  object,
  caption = NA,
  footnote = NA,
  filename = NA,
  select = c("Case", `Mean A` = "mA", `Mean B` = "mB", `SD A` = "sdA", `SD B` = "sdB",
    `SD Cohen` = "sd cohen", `SD Hedges` = "sd hedges", "Glass' delta", "Hedges' g",
    "Hedges' g correction", "Hedges' g durlak correction", "Cohen's d"),
  round = 2,
  decimals = 2,
  flip = FALSE,
  ...
)

\method{export}{sc_trend}(
  object,
  caption = NA,
  footnote = NA,
  filename = NA,
  round = 3,
  decimals = NULL,
  ...
)

\method{export}{scdf}(
  object,
  summary = FALSE,
  caption = NA,
  footnote = NA,
  filename = NA,
  cols,
  round = 2,
  ...
)

\method{export}{scdf_summary}(object, caption = NA, footnote = NA, filename = NA, round = 2, ...)
}
\arguments{
\item{object}{An scdf or an object exported from a scan function.}

\item{...}{Further Arguments passed to internal functions.}

\item{caption}{Character string with table caption. If left NA (default) a
caption will be created based on the exported object.}

\item{footnote}{Character string with table footnote. If left NA (default) a
footnote will be created based on the exported object.}

\item{filename}{String containing the file name. If a filename is given the
output will be written to that file.}

\item{flip}{If TRUE, some objects are exported with rows and columns flipped.}

\item{round}{Integer passed to the digits argument used to round values.}

\item{select}{A character vector containing the names of the variables to be
included. If the vector is named, the variables will be renamed
accordingly.}

\item{decimals}{Decimal places that are reported.}

\item{summary}{If TRUE, exports the summary of an \code{scdf}.}

\item{cols}{Defines which columns are included when exporting an scdf. It is
either a vector of variable names or the string "main" will select the
central variables.}
}
\value{
Returns or displays a specially formatted html (or latex) file.
}
\description{
Export creates html files of tables or displays them directly in the viewer
pane of rstudio. When applied in rmarkdown/quarto, tables can also be created
for pdf/latex output.
}
\details{
The function uses either the kableExtra or the gt package to create the tables.
The default engine can be set via the option \code{scan.export.engine}. Additional
options for kable and kable_styling can be set via the options
\code{scan.export.kable} and \code{scan.export.kable_styling}.
The default options can be viewed and modified via
\code{options("scan.export.kable")} and
\code{options("scan.export.kable_styling")}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ird.R
\name{ird}
\alias{ird}
\alias{print.sc_ird}
\alias{export.sc_ird}
\title{IRD - Improvement rate difference}
\usage{
ird(data, dvar, pvar, decreasing = FALSE, phases = c(1, 2))

\method{print}{sc_ird}(x, digits = 3, ...)

\method{export}{sc_ird}(object, caption = NA, footnote = NA, filename = NA, round = 3, ...)
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{decreasing}{If you expect data to be lower in the B phase, set
\code{decreasing = TRUE}. Default is \code{decreasing = FALSE}.}

\item{phases}{A vector of two characters or numbers indicating the two phases
that should be compared. E.g., \code{phases = c("A","C")} or \code{phases = c(2,4)}
for comparing the second to the fourth phase. Phases could be combined by
providing a list with two elements. E.g., \code{phases = list(A = c(1,3), B = c(2,4))} will compare phases 1 and 3 (as A) against 2 and 4 (as B). Default
is \code{phases = c(1,2)}.}

\item{x}{An object returned by \code{\link[=ird]{ird()}}}

\item{digits}{The minimum number of significant digits to be use.}

\item{...}{Further arguments passed to the function.}

\item{object}{An scdf or an object exported from a scan function.}

\item{caption}{Character string with table caption. If left NA (default) a
caption will be created based on the exported object.}

\item{footnote}{Character string with table footnote. If left NA (default) a
footnote will be created based on the exported object.}

\item{filename}{String containing the file name. If a filename is given the
output will be written to that file.}

\item{round}{Integer passed to the digits argument used to round values.}
}
\description{
\code{ird()} calculates the robust improvement rate difference as proposed by
Parker and colleagues (2011).
}
\details{
The adaptation of the improvement rate difference for single-case phase
comparisons was developed by Parker and colleagues (2009). A variation called
robust improvement rate difference was proposed by Parker and colleagues in
2011. This function calculates the robust improvement rate difference. It
follows the formula suggested by Pustejovsky (2019). For a multiple case
design, ird is based on the overall improvement rate of all cases which is
the average of the irds for each case.
}
\section{Functions}{
\itemize{
\item \code{print(sc_ird)}: Print results

\item \code{export(sc_ird)}: Export results to html

}}
\references{
Parker, R. I., Vannest, K. J., & Brown, L. (2009). The
improvement rate difference for single-case research. Exceptional Children,
75(2), 135-150.

Parker, R. I., Vannest, K. J., & Davis, J. L. (2011). Effect Size in
Single-Case Research: A Review of Nine Nonoverlap Techniques. Behavior
Modification, 35(4), 303-322. https://doi.org/10.1177/0145445511399147

Pustejovsky, J. E. (2019). Procedural sensitivities of effect sizes for
single-case designs with directly observed behavioral outcome measures.
\emph{Psychological Methods}, \emph{24(2)}, 217-235.
https://doi.org/10.1037/met0000179
}
\seealso{
Other overlap functions: 
\code{\link{nap}()},
\code{\link{overlap}()},
\code{\link{pand}()},
\code{\link{pem}()},
\code{\link{pet}()},
\code{\link{pnd}()},
\code{\link{print.sc_cdc}()},
\code{\link{select_phases}()},
\code{\link{tau_u}()}
}
\concept{overlap functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/power_test.R
\name{power_test}
\alias{power_test}
\title{Empirical power analysis for single-case data}
\usage{
power_test(
  design,
  method = c("plm_level", "rand", "tauU"),
  effect = "level",
  n_sim = 100,
  design_is_one_study = TRUE,
  alpha_test = TRUE,
  power_test = TRUE,
  binom_test = FALSE,
  binom_test_alpha = FALSE,
  binom_test_power = FALSE,
  binom_test_correct = FALSE,
  ci = FALSE,
  alpha_level = 0.05
)
}
\arguments{
\item{design}{An object returned from the \code{design} function.}

\item{method}{A (named) list that defines the methods the power analysis is
based on. Each element can contain a function (that takes an scdf file and
returns a p value) or a character string (the name of predefined
functions). default \code{method = list("plm_level", "rand", "tauU")} computes a
power analysis based on \code{\link[=tau_u]{tau_u()}}, \code{\link[=rand_test]{rand_test()}} and \code{\link[=plm]{plm()}} analyses.
(Further predefined functions are: "plm_slope", "plm_poisson_level",
"plm_poisson_slope", "hplm_level", "hplm_slope", "base_tau".}

\item{effect}{Either "level" or "slope". The respective effect of the
provided design is set to 0 when computing the alpha-error proportion.}

\item{n_sim}{Number of sample studies created for the the Monte-Carlo study.
Default is \code{n = 100}. Ignored if design_is_one_study = FALSE.}

\item{design_is_one_study}{If TRUE, the design is assumed to define all cases
of one study that is repeatedly randomly created \code{n_sim} times. If FALSE,
the design is assumed to contain all cases from which a random sample is
generated. This is useful for very specific complex simulation studies.
In this case, n_sim indicates how many random samples are drawn from the
provided design. Default is TRUE.}

\item{alpha_test}{Logical. If TRUE, alpha error is calculated.}

\item{power_test}{Logical. If TRUE, power is calculated.}

\item{binom_test}{Shortcut. When set TRUE, binom_test_power is set to 0.80,
binom_test_alpha is set to 0.05, and binom_test_correct is set to 0.875.}

\item{binom_test_alpha}{Either FALSE or a value. If a value is provided, a
binomial test is calculated testing if the alpha error proportion is less
than the provided value.}

\item{binom_test_power}{Either FALSE or a value. If a value is provided, a
binomial test is calculated testing if the power is greater than the
provided value.}

\item{binom_test_correct}{Either FALSE or a value. If a value is provided, a
binomial test is calculated testing if the correct proportion is greater
than the provided value.}

\item{ci}{Either FALSE or a value. If a value is provided, confidence
intervals at the provided level are calculated for power, alpha error, and
correct proportions.}

\item{alpha_level}{Alpha level used to calculate the proportion of
significant tests. Default is \code{alpha_level = 0.05}.}
}
\value{
A data frame with the power, alpha error, and correct proportions for
each provided method. If binomial tests are requested, p values for these
tests are also provided. If confidence intervals are requested, these
are also provided.
}
\description{
Conducts a Monte-Carlo study on the test-power and alpha-error probability of
a statistical function.
}
\details{
Based on a \code{\link[=design]{design()}} object, a large number of single-cases are generated
and re-analysed with a provided statistical function. The proportion of
significant analyses is the test power. In a second step, a specified effect
of the design object is set to 0 and again single-cases are generated and
re-analysed. The proportion of significant analyses is the alpha error
probability.
}
\examples{

## Assume you want to conduct a single-case study with 15 measurements
## (phases: A = 6 and B = 9) using a highly reliable test and
## an expected level effect of d = 1.4.
## A (strong) trend effect is trend = 0.05. What is the power?
## (Note: n_sims is set to 10. Set n_sims to 1000 for a serious calculation.)
design <- design(
  n = 1, phase_design = list(A = 6, B = 9),
  rtt = 0.8, level = 1.4, trend = 0.05
)
power_test(design, n_sim = 10)

## Would you achieve higher power by setting up a MBD with three cases?
design <- design(
  n = 3, phase_design = list(A = 6, B = 9),
  rtt = 0.8, level = 1.4, trend = 0.05
)
power_test(design, n_sim=10, method=list("hplm_level", "rand", "tauU_meta"))
}
\seealso{
\code{\link[=random_scdf]{random_scdf()}}, \code{\link[=design]{design()}}
}
\author{
Juergen Wilbert
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rescale.R
\name{rescale}
\alias{rescale}
\title{Rescales values of an scdf}
\usage{
rescale(data, ..., m = 0, sd = 1)
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{...}{Names of variables to be standardized. If none are given, all numeric
variables are standardized.}

\item{m}{The target mean. If set NULL, it is not changed.}

\item{sd}{The target standard deviation. If set NULL, it is not changed.}
}
\value{
An scdf with the scaled values.
}
\description{
This function scales the measured values of an scdf file. It allows for mean
centering and standardization across all cases included in an scdf.
}
\examples{

## Standardize a multiple case scdf and compute an hplm
exampleAB_50 |>
  rescale(values, mt) |>
  hplm()
}
\seealso{
Other data manipulation functions: 
\code{\link{add_l2}()},
\code{\link{as.data.frame.scdf}()},
\code{\link{as_scdf}()},
\code{\link{batch_apply}()},
\code{\link{fill_missing}()},
\code{\link{moving_median}()},
\code{\link{print.sc_outlier}()},
\code{\link{ranks}()},
\code{\link{scdf}()},
\code{\link{select_cases}()},
\code{\link{set_vars}()},
\code{\link{shift}()},
\code{\link{smooth_cases}()},
\code{\link{standardize}()},
\code{\link{truncate_phase}()}
}
\author{
Juergen Wilbert
}
\concept{data manipulation functions}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hplm.R, R/print-export-hplm.R,
%   R/coefficents.sc_plm.R
\name{hplm}
\alias{hplm}
\alias{print.sc_hplm}
\alias{export.sc_hplm}
\alias{coef.sc_hplm}
\title{Hierarchical piecewise linear model / piecewise regression for multiple cases}
\usage{
hplm(
  data,
  dvar,
  pvar,
  mvar,
  model = c("W", "H-M", "B&L-B", "JW"),
  contrast = c("first", "preceding"),
  contrast_level = NA,
  contrast_slope = NA,
  method = c("ML", "REML"),
  control = list(opt = "optim"),
  random.slopes = FALSE,
  lr.test = FALSE,
  ICC = TRUE,
  trend = TRUE,
  level = TRUE,
  slope = TRUE,
  random_trend = FALSE,
  random_level = FALSE,
  random_slope = FALSE,
  fixed = NULL,
  random = NULL,
  ar = 0,
  unequal_variances = FALSE,
  update.fixed = NULL,
  data.l2 = NULL,
  ...
)

\method{print}{sc_hplm}(x, digits = 3, bcsmd = FALSE, casewise = FALSE, ...)

\method{export}{sc_hplm}(
  object,
  caption = NA,
  footnote = NA,
  filename = NA,
  round = 2,
  nice = TRUE,
  casewise = FALSE,
  ...
)

\method{coef}{sc_hplm}(object, casewise = FALSE, ...)
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{mvar}{Character string with the name of the measurement time variable.
Defaults to the attributes in the scdf file.}

\item{model}{Model used for calculating the dummy parameters (see Huitema &
McKean, 2000). Default is \code{model = "W"}. Possible values are: \code{"B&L-B"},
\code{"H-M"}, \code{"W"}, and deprecated \code{"JW"}.}

\item{contrast}{Sets contrast_level and contrast_slope. Either "first",
"preceding" or a contrast matrix. If NA contrast is ignored.}

\item{contrast_level}{Either "first", "preceding" or a contrast matrix. If NA
contrast_level is a copy of contrast.}

\item{contrast_slope}{Either "first", "preceding" or a contrast matrix. If NA
contrast_level is a copy of contrast.}

\item{method}{Method used to fit your model. Pass \code{"REML"} to maximize the
restricted log-likelihood or \code{"ML"} for maximized log-likelihood. Default
is \code{"ML"}.}

\item{control}{A list of settings for the estimation algorithm, replacing the
default values passed to the function \code{lmeControl} of the \code{nlme} package.}

\item{random.slopes}{If \code{random.slopes = TRUE} random slope effects of the
level, trend, and treatment parameter are estimated.}

\item{lr.test}{If set TRUE likelihood ratio tests are calculated comparing
model with vs. without random slope parameters.}

\item{ICC}{If \code{ICC = TRUE} an intraclass-correlation is estimated.}

\item{trend}{A logical indicating if a trend parameters is included in the
model.}

\item{level}{A logical indicating if a level parameters is included in the
model.}

\item{slope}{A logical indicating if a slope parameters is included in the
model.}

\item{random_trend}{If TRUE, includes a random trend trend effect.}

\item{random_level}{If TRUE, includes a random level trend effect.}

\item{random_slope}{If TRUE, includes a random slope trend effect.}

\item{fixed}{Defaults to the fixed part of the standard piecewise regression
model. The parameter phase followed by the phase name (e.g., phaseB)
indicates the level effect of the corresponding phase. The parameter
'inter' followed by the phase name (e.g., interB) adresses the slope effect
based on the method provide in the model argument (e.g., "B&L-B"). The
formula can be changed for example to include further L1 or L2 variables
into the regression model.}

\item{random}{The random part of the model. Defaults to a random intercept
model. The formula can be changed to include random slope effects for
level, trend, and treatment effects.}

\item{ar}{Maximal lag of autoregression. Modelled based on the
Autoregressive-Moving Average (ARMA) function.}

\item{unequal_variances}{Logical. If set TRUE, estimations are weighted by
phase variances.}

\item{update.fixed}{An easier way to change the fixed model part (e.g., \code{. ~ . + newvariable}).}

\item{data.l2}{A data frame providing additional variables at Level 2. The
scdf File has to have names for all cases and the Level 2 data frame has to
have a column named 'cases' with the names of the cases the Level 2
variables belong to.}

\item{...}{Further arguments passed to the lme function.}

\item{x}{An object returned by \code{\link[=hplm]{hplm()}}}

\item{digits}{The minimum number of significant digits to be use.
If set to "auto" (default), values are predefined.}

\item{bcsmd}{If TRUE, reports between-case standardized mean differences.}

\item{casewise}{Returns the estimations for each case separately}

\item{object}{An scdf or an object exported from a scan function.}

\item{caption}{Character string with table caption. If left NA (default) a
caption will be created based on the exported object.}

\item{footnote}{Character string with table footnote. If left NA (default) a
footnote will be created based on the exported object.}

\item{filename}{String containing the file name. If a filename is given the
output will be written to that file.}

\item{round}{Integer passed to the digits argument used to round values.}

\item{nice}{If set TRUE (default) output values are rounded and optimized for
publication tables.}
}
\value{
An object of class \code{sc_hplm}.
\itemize{
\item \code{model} | List containing infromation about
\item \code{N} | Number of single-cases.
\item \code{formula} | A list containing the fixed and the random formulas of the hplm model.
\item \code{hplm} | Object of class lme contaning the multilevel model.
\item \code{model.0} | Object of class lme containing the zero model.
\item \code{ICC} | List containing intraclass correlation and test parameters.
\item \code{model.without} | Object of class gls containing the fixed effect model.
\item \code{contrast} | List with contrast definitions.
}
}
\description{
The \code{\link[=hplm]{hplm()}} function computes a hierarchical piecewise regression model. It
extends the standard piecewise regression model to multiple cases by
estimating fixed and random effects. The function uses the \code{lme} function of
the \code{nlme} package to fit linear mixed-effects models. The model can include
random intercepts and random slopes for level, trend, and treatment effects.
Additionally, it allows for the inclusion of autoregressive structures and
unequal variances across phases. The function also provides options for
likelihood ratio tests to compare models with and without random slope
effects, as well as the calculation of intraclass correlations (ICC) to
assess the proportion of variance attributable to between-case differences.
This function is particularly useful for analyzing data from multiple
single-case experimental designs (SCEDs) where observations are nested within
cases.
}
\section{Functions}{
\itemize{
\item \code{print(sc_hplm)}: Print results

\item \code{export(sc_hplm)}: Export results as html table (see \code{\link[=export]{export()}})

\item \code{coef(sc_hplm)}: Extract model coefficients

}}
\section{Model specification}{
 The fixed effects part of the model can be
specified using the \code{fixed} argument, while the random effects part can be
specified using the \code{random} argument. If not provided, default formulas
based on the specified model type (e.g., "B&L-B") are created. The function
also allows for the inclusion of autoregressive structures through the \code{ar}
argument and unequal variances across phases through the
\code{unequal_variances} argument.
}

\section{Random slopes}{
 By setting the \code{random.slopes} argument to TRUE, the
model will include random slope effects for level, trend, and treatment
effects. This allows for individual differences in how cases respond to
these effects.
}

\section{Likelihood ratio tests}{
 If the \code{lr.test} argument is set to TRUE,
the function will perform likelihood ratio tests to compare models with and
without random slope effects. This helps to determine whether including
random slopes significantly improves model fit.
}

\section{Intraclass correlation}{
 If the \code{ICC} argument is set to TRUE, the
function will calculate the intraclass correlation coefficient (ICC) to
assess the proportion of variance attributable to between-case differences.
This provides insight into the degree of similarity among observations
within the same case.
}

\examples{

## Compute hplm model on a MBD over fifty cases (restricted log-likelihood)
hplm(exampleAB_50, method = "REML", random.slopes = FALSE)

## Analyzing with additional L2 variables
Leidig2018 |>
  add_l2(Leidig2018_l2) |>
  hplm(update.fixed = .~. + gender + migration + ITRF_TOTAL*phaseB,
       slope = FALSE, random.slopes = FALSE, lr.test = FALSE
  )

}
\seealso{
Other regression functions: 
\code{\link{bplm}()},
\code{\link{fetch}()},
\code{\link{mplm}()},
\code{\link{plm}()},
\code{\link{print.sc_ac}()},
\code{\link{print.sc_bctau}()},
\code{\link{trend}()}
}
\author{
Juergen Wilbert
}
\concept{regression functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/batch_apply.R
\name{batch_apply}
\alias{batch_apply}
\title{Apply a function to each element in an scdf.}
\usage{
batch_apply(scdf, fn, simplify = FALSE)
}
\arguments{
\item{scdf}{A list of inputs to apply the function to.}

\item{fn}{The function to apply to each element. Use a \code{.} as a
placeholder for the scdf (e.g. \code{describe(.)}).}

\item{simplify}{If simplify is TRUE and \code{fn} returns a vector of values,
\code{batch_apply} will return a data frame case names.}
}
\value{
A list of the output of each function call.
}
\description{
This function applies a given function to each case of a multiple case scdf,
returning a list of the output of each function call.
}
\details{
If \code{simplify} is TRUE and the function returns a vector of values, the output
is combined into a data frame with case names as additional columns.

This is particularly useful for applying statistical models or summary
statistics to each case in an scdf.
}
\examples{
batch_apply(exampleAB, coef(plm(.)))

}
\seealso{
Other data manipulation functions: 
\code{\link{add_l2}()},
\code{\link{as.data.frame.scdf}()},
\code{\link{as_scdf}()},
\code{\link{fill_missing}()},
\code{\link{moving_median}()},
\code{\link{print.sc_outlier}()},
\code{\link{ranks}()},
\code{\link{rescale}()},
\code{\link{scdf}()},
\code{\link{select_cases}()},
\code{\link{set_vars}()},
\code{\link{shift}()},
\code{\link{smooth_cases}()},
\code{\link{standardize}()},
\code{\link{truncate_phase}()}
}
\author{
Juergen Wilbert
}
\concept{data manipulation functions}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprec-truncate_phase.R
\name{truncate_phase}
\alias{truncate_phase}
\title{Truncate single-case data}
\usage{
truncate_phase(
  data,
  dvar,
  pvar,
  truncate = list(A = c(0, 0), B = c(0, 0)),
  na = TRUE
)
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{truncate}{A list with a vector of two (beginning and end) values for
each phase defining the number of data points to be deleted. For lists of
single-case data frames, the truncation is adapted to the length of each
phase for each single case.}

\item{na}{If FALSE, the truncated measurement times are deleted. If TRUE, NAs
are set for the dependent variable.}
}
\value{
A truncated data frame (for each single-case).
}
\description{
#' \emph{This function is superseded by the more versatile
\code{\link{transform.scdf}} function.}
This function truncates data points at the beginning and / or end of each
phase in each case.
}
\examples{

## Truncate the first two data points of both phases and compare the two 
## data sets
study <- c(
  "Original" = byHeart2011[1],
  "Selected" = truncate_phase(
    byHeart2011[1], truncate = list(A = c(2, 0), B = c(2, 0))
  )
)
plot(study)
}
\seealso{
Other data manipulation functions: 
\code{\link{add_l2}()},
\code{\link{as.data.frame.scdf}()},
\code{\link{as_scdf}()},
\code{\link{batch_apply}()},
\code{\link{fill_missing}()},
\code{\link{moving_median}()},
\code{\link{print.sc_outlier}()},
\code{\link{ranks}()},
\code{\link{rescale}()},
\code{\link{scdf}()},
\code{\link{select_cases}()},
\code{\link{set_vars}()},
\code{\link{shift}()},
\code{\link{smooth_cases}()},
\code{\link{standardize}()}
}
\author{
Juergen Wilbert
}
\concept{data manipulation functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/inherit_roxygen_descriptions.R
\name{.inheritParams}
\alias{.inheritParams}
\title{Dummy function to inherit global descriptions of parameters}
\usage{
.inheritParams(
  data,
  scdf,
  dvar,
  mvar,
  pvar,
  decreasing,
  phases,
  model,
  contrast,
  contrast_level,
  contrast_slope,
  trend,
  level,
  slope,
  nice,
  ...
)
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{scdf}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{mvar}{Character string with the name of the measurement time variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{decreasing}{If you expect data to be lower in the B phase, set
\code{decreasing = TRUE}. Default is \code{decreasing = FALSE}.}

\item{phases}{A vector of two characters or numbers indicating the two phases
that should be compared. E.g., \code{phases = c("A","C")} or \code{phases = c(2,4)}
for comparing the second to the fourth phase. Phases could be combined by
providing a list with two elements. E.g., \code{phases = list(A = c(1,3), B = c(2,4))} will compare phases 1 and 3 (as A) against 2 and 4 (as B). Default
is \code{phases = c(1,2)}.}

\item{model}{Model used for calculating the dummy parameters (see Huitema &
McKean, 2000). Default is \code{model = "W"}. Possible values are: \code{"B&L-B"},
\code{"H-M"}, \code{"W"}, and deprecated \code{"JW"}.}

\item{contrast}{Sets contrast_level and contrast_slope. Either "first",
"preceding" or a contrast matrix. If NA contrast is ignored.}

\item{contrast_level}{Either "first", "preceding" or a contrast matrix. If NA
contrast_level is a copy of contrast.}

\item{contrast_slope}{Either "first", "preceding" or a contrast matrix. If NA
contrast_level is a copy of contrast.}

\item{trend}{A logical indicating if a trend parameters is included in the
model.}

\item{level}{A logical indicating if a level parameters is included in the
model.}

\item{slope}{A logical indicating if a slope parameters is included in the
model.}

\item{nice}{If set TRUE (default) output values are rounded and optimized for
publication tables.}

\item{...}{Further arguments passed to the function.}
}
\value{
No return value.
}
\description{
This function is only used to inherit parameter descriptions in the
documentation of other functions. It has no other purpose and is not meant to
be called directly.
}
\author{
Juergen Wilbert
}
\concept{internal functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert.R
\name{convert}
\alias{convert}
\title{Convert scdf to R code}
\usage{
convert(
  scdf,
  file = "",
  study_name = "study",
  case_name = "case",
  inline = FALSE,
  indent = 2,
  silent = FALSE
)
}
\arguments{
\item{scdf}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{file}{A filename for exporting the syntax. If an empty string (default),
the syntax is not written to a file.}

\item{study_name}{Character string. Name of the study object.}

\item{case_name}{Character string. Name of the scdf objects.}

\item{inline}{If TRUE, phase definition is in an online version.}

\item{indent}{Integer. Indentation.}

\item{silent}{If TRUE, syntax is not printed to the console.}
}
\value{
Returns a string (invisible).
}
\description{
Converts an scdf object into R code that can be used to recreate the object.
}
\details{
This function is useful for exporting scdf objects for sharing or
documentation purposes. The generated R code can be sourced to recreate the
original scdf object.
}
\examples{
filename <- tempfile()
convert(exampleABC, file = filename)
source(filename)
all.equal(study, exampleABC)
unlink(filename)
}
\seealso{
Other io-functions: 
\code{\link{read_scdf}()},
\code{\link{write_scdf}()}
}
\author{
Juergen Wilbert
}
\concept{io-functions}
\keyword{io}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transform-helper-functions.R,
%   R/transform.scdf.R
\name{moving_median}
\alias{moving_median}
\alias{moving_mean}
\alias{local_regression}
\alias{set_na_at}
\alias{center_at}
\alias{first_of}
\alias{across_cases}
\alias{all_cases}
\alias{rowwise}
\alias{transform.scdf}
\title{Transform every single case of a single case data frame}
\usage{
moving_median(x, lag = 1)

moving_mean(x, lag = 1)

local_regression(x, mt = 1:length(x), f = 0.2)

set_na_at(x, first_of, positions = 0)

center_at(x, at = TRUE, shift = 0, part = 0)

first_of(x, positions = 0)

across_cases(...)

all_cases(...)

rowwise(...)

\method{transform}{scdf}(`_data`, ...)
}
\arguments{
\item{x}{A logical vector.}

\item{lag}{Number of values surrounding a value to calculate the average}

\item{mt}{A vector with measurement times.}

\item{f}{the proportion of surrounding data influencing each data point.}

\item{first_of}{A logical vector}

\item{positions}{A numeric vector with relative positions to the first
appearance of a TRUE value in x.}

\item{at}{A logical vector. E.g. \code{phase == "A"}. The first TRUE value of that
vector is the target position for centring. By default, this is the first
position.}

\item{shift}{A value indicating a shift in measurement times for centring.
E.g. \code{shift = 4} will centre four measurement-times after the position
defined by the \code{at} and \code{part} arguments.}

\item{part}{A numeric value between 0 and 1. \code{0} refers to the first \code{TRUE}
in the \code{at} vector, \code{1} to the last, and \code{0.5} to the midpoint of the
sequence of \code{TRUE} values. E.g. if you want to centre at the middle of
phase A, set \verb{at = phase == A, part = 0.5}. Note: decimals are rounded to
integers.}

\item{...}{Expressions.}

\item{_data}{An scdf.}
}
\value{
An scdf.
}
\description{
Takes an scdf and applies transformations to each individual case. This is
useful to calculate or modify new variables.
}
\details{
This function is a method of the generic \code{transform()} function.
Unlike the generic version, expressions are evaluated \strong{serially}:
the result of one expression is used as the basis for subsequent computations.

Several helper functions can be used inside expressions:
\itemize{
\item \strong{\code{n()}}: returns the number of measurements in a case.
\item \strong{\code{all_cases()}}: extracts the values of a variable across all cases.
Takes an expression as argument.
For example:
\itemize{
\item \code{mean(all_cases(values))} calculates the mean of \code{values} across all cases.
\item \code{mean(all_cases(values[phase == "A"]))} calculates the mean of all values where \code{phase == "A"}.
}
\item \strong{\code{rowwise()}}: applies a calculation separately to each row.
Example: \code{rowwise(sum(values, mt, na.rm = TRUE))}.
\item \strong{\code{across_cases()}}: creates new variables or replaces existing ones across all cases.
Example: \code{across_cases(values_ranked = rank(values, na.last = "keep"))}
ranks the \code{values} variable across all cases and creates a new variable \code{values_ranked}.
}
}
\examples{
## Creates a single-case with frequency distributions. The proportion and
## percentage of the frequencies are calculated with transform:
design <- design(
 n = 3,
 level = 5,
 distribution = "binomial",
 n_trials = 20,
 start_value = 0.5
)
study <- random_scdf(design)
transform(study, proportion = values/trials, percentage = proportion * 100)

## Z standardize the dependent variable and add two new variables:
exampleAB |>
  transform(
    values = scale(values),
    mean_values = mean(values),
    sd_values = sd(values)
  )

## Use `all` to calculate global variables.
exampleAB |>
  transform(
    values_center_case = values - mean(values[phase == "A"]),
    values_center_global = values - mean(all(values[phase == "A"])),
    value_dif = values_center_case - values_center_global
  )

## Use `across_cases` to calculate or replace a variable with values from
## all cases. E.g., standardize the dependent variable:
exampleABC |>
  transform(
    across_cases(values = scale(values))
  )

## Rank transform the values based on all cases vs. within each case:
exampleABC |>
  transform(
    across_cases(values_across = rank(values, na.last="keep")),
    value_within = rank(values, na.last="keep")
  )

## Three helper functions to smooth the data
Huber2014$Berta |>
transform(
  "compliance (moving median)" = moving_median(compliance),
  "compliance (moving mean)" = moving_mean(compliance),
  "compliance (local regression)" = local_regression(compliance, mt)
)

## Function first_of() helps to set NAs for specific phases.
## E.g., you want to replace the first two values of phase A and the first
## value of phase B and its preceding value.

byHeart2011 |>
  transform(
    values = set_na_at(values, phase == "A", 0:1),
    values = set_na_at(values, phase == "B", -1:0)
  )
}
\seealso{
Other data manipulation functions: 
\code{\link{add_l2}()},
\code{\link{as.data.frame.scdf}()},
\code{\link{as_scdf}()},
\code{\link{batch_apply}()},
\code{\link{fill_missing}()},
\code{\link{print.sc_outlier}()},
\code{\link{ranks}()},
\code{\link{rescale}()},
\code{\link{scdf}()},
\code{\link{select_cases}()},
\code{\link{set_vars}()},
\code{\link{shift}()},
\code{\link{smooth_cases}()},
\code{\link{standardize}()},
\code{\link{truncate_phase}()}
}
\author{
Juergen Wilbert
}
\concept{data manipulation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.export-bctau.R, R/corrected_tau.R
\name{print.sc_bctau}
\alias{print.sc_bctau}
\alias{export.sc_bctau}
\alias{corrected_tau}
\title{Baseline corrected tau}
\usage{
\method{print}{sc_bctau}(x, nice = TRUE, digits = "auto", ...)

\method{export}{sc_bctau}(
  object,
  caption = NA,
  footnote = NA,
  filename = NA,
  nice = TRUE,
  round = 2,
  ...
)

corrected_tau(
  data,
  dvar,
  pvar,
  mvar,
  phases = c(1, 2),
  alpha = 0.05,
  continuity = FALSE,
  tau_method = c("b", "a")
)
}
\arguments{
\item{x}{An object returned by \code{\link[=corrected_tau]{corrected_tau()}}}

\item{nice}{If set TRUE (default) output values are rounded and optimized for
publication tables.}

\item{digits}{The minimum number of significant digits to be use.
If set to "auto" (default), values are predefined.}

\item{...}{Further parameters passed to the print function}

\item{object}{An scdf or an object exported from a scan function.}

\item{caption}{Character string with table caption. If left NA (default) a
caption will be created based on the exported object.}

\item{footnote}{Character string with table footnote. If left NA (default) a
footnote will be created based on the exported object.}

\item{filename}{String containing the file name. If a filename is given the
output will be written to that file.}

\item{round}{Integer passed to the digits argument used to round values.}

\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{mvar}{Character string with the name of the measurement time variable.
Defaults to the attributes in the scdf file.}

\item{phases}{A vector of two characters or numbers indicating the two phases
that should be compared. E.g., \code{phases = c("A","C")} or \code{phases = c(2,4)}
for comparing the second to the fourth phase. Phases could be combined by
providing a list with two elements. E.g., \code{phases = list(A = c(1,3), B = c(2,4))} will compare phases 1 and 3 (as A) against 2 and 4 (as B). Default
is \code{phases = c(1,2)}.}

\item{alpha}{Sets the p-value at and below which a baseline correction is
applied.}

\item{continuity}{If TRUE applies a continuity correction for calculating p}

\item{tau_method}{Character with values "a" or "b" (default) indicating
whether Kendall Tau A or Kendall Tau B is applied.}
}
\description{
Kendall's tau correlation for the dependent variable and the phase variable
is calculated after correcting for a baseline trend.
}
\details{
This method has been proposed by Tarlow (2016). The baseline data
are checked for a significant autocorrelation (based on Kendall's Tau). If
so, a non-parametric Theil-Sen regression is applied for the baseline data
where the dependent values are regressed on the measurement time. The
resulting slope information is then used to predict data of the B-phase.
The dependent variable is now corrected for this baseline trend and the
residuals of the Theil-Sen regression are taken for further calculations.
Finally, Kendall's tau is calculated for the dependent variable and the
dichotomous phase variable. The function here provides two extensions to
this procedure: The more accurate continuity correction is applied when
\code{continuity = TRUE}.
}
\section{Functions}{
\itemize{
\item \code{print(sc_bctau)}: Print results

\item \code{export(sc_bctau)}: Export results as html

}}
\examples{
dat <- scdf(c(A = 33,25,17,25,14,13,15, B = 15,16,16,5,7,9,6,5,3,3,8,11,7))
corrected_tau(dat)
}
\references{
Tarlow, K. R. (2016). An Improved Rank Correlation Effect Size
Statistic for Single-Case Designs: Baseline Corrected Tau. \emph{Behavior
Modification, 41(4)}, 427–467. https://doi.org/10.1177/0145445516676750
}
\seealso{
Other regression functions: 
\code{\link{bplm}()},
\code{\link{fetch}()},
\code{\link{hplm}()},
\code{\link{mplm}()},
\code{\link{plm}()},
\code{\link{print.sc_ac}()},
\code{\link{trend}()}
}
\concept{regression functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/anova.R
\name{anova.sc_plm}
\alias{anova.sc_plm}
\alias{anova.sc_hplm}
\alias{anova.sc_mplm}
\title{ANOVA Table for Piecewise Linear Models}
\usage{
\method{anova}{sc_plm}(object, ...)

\method{anova}{sc_hplm}(object, ...)

\method{anova}{sc_mplm}(object, ...)
}
\arguments{
\item{object}{An object containing the results returned by \code{plm()},
\code{hplm()}, or \code{mplm()}.}

\item{...}{additional objects for model comparison.}
}
\value{
An object of class \code{anova} containing the results of the model
comparison.
}
\description{
Model comparison for piecewise regression models fitted with \code{plm()},
\code{hplm()}, or \code{mplm()} using likelihood ratio tests.
}
\details{
The function performs likelihood ratio tests to compare nested piecewise
regression models. It extracts the underlying model from the sc_plm, sc_hplm,
or sc_mplm object and passes them to the generic anova() function
for model comparison.
}
\examples{
## For glm models with family = "gaussian"
mod1 <- plm(exampleAB$Johanna, level = FALSE, slope = FALSE)
mod2 <- plm(exampleAB$Johanna)
anova(mod1, mod2)
## For glm models with family = "poisson"
mod0 <- plm(example_A24, formula = injuries ~ 1, family = "poisson")
mod1 <- plm(example_A24, trend = FALSE, family = "poisson")
anova(mod0, mod1, mod2)
## For glm with family = "binomial"
mod0 <- plm(
  exampleAB_score$Christiano, 
  formula = values ~ 1, 
  family = "binomial", 
  var_trials = "trials"
)
mod1 <- plm(
  exampleAB_score$Christiano, 
  trend = FALSE, 
  family = "binomial", 
  var_trials = "trials"
)
anova(mod0, mod1)
## For multilevel models:
mod0 <- hplm(Leidig2018, trend = FALSE, slope = FALSE, level = FALSE)
mod1 <- hplm(Leidig2018, trend = FALSE)
mod2 <- hplm(Leidig2018)
anova(mod0, mod1, mod2)
## For mplm
mod0 <- mplm(
  Leidig2018$`1a1`, 
  update = . ~  1, dvar = c("academic_engagement", "disruptive_behavior")
)
mod1 <- mplm(
  Leidig2018$`1a1`, 
  trend = FALSE, 
  dvar = c("academic_engagement", "disruptive_behavior")
)
mod2 <- mplm(
  Leidig2018$`1a1`, 
  dvar = c("academic_engagement", "disruptive_behavior")
)

anova(mod0, mod1, mod2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print-export-ac.R, R/autocorr.R
\name{print.sc_ac}
\alias{print.sc_ac}
\alias{export.sc_ac}
\alias{autocorr}
\title{Autocorrelation within and across phases}
\usage{
\method{print}{sc_ac}(x, digits = "auto", ...)

\method{export}{sc_ac}(object, caption = NA, footnote = NA, filename = NA, round = 3, ...)

autocorr(data, dvar, pvar, mvar, lag_max = 3, lag.max, ...)
}
\arguments{
\item{x}{An object returned by \code{\link[=autocorr]{autocorr()}}}

\item{digits}{The minimum number of significant digits to be use.
If set to "auto" (default), values are predefined.}

\item{...}{Further arguments passed to the \code{\link[=acf]{acf()}} function.}

\item{object}{An scdf or an object exported from a scan function.}

\item{caption}{Character string with table caption. If left NA (default) a
caption will be created based on the exported object.}

\item{footnote}{Character string with table footnote. If left NA (default) a
footnote will be created based on the exported object.}

\item{filename}{String containing the file name. If a filename is given the
output will be written to that file.}

\item{round}{Integer passed to the digits argument used to round values.}

\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{mvar}{Character string with the name of the measurement time variable.
Defaults to the attributes in the scdf file.}

\item{lag_max, lag.max}{The lag up to which autocorrelations will be computed.}
}
\value{
A data frame containing separate autocorrelations for each phase and
for all phases (for each single-case). If \code{lag_max} exceeds the length
of a phase minus one, NA is returned for this cell.
}
\description{
The autocorr function calculates autocorrelations within each phase and
across all phases.
}
\details{
Autocorrelations are computed using the \code{\link[=acf]{acf()}} function from the stats
package. For each single-case in the scdf object, a data frame is returned
containing the autocorrelations for each phase and for all phases up to the
specified lag.
}
\section{Functions}{
\itemize{
\item \code{print(sc_ac)}: Print method

\item \code{export(sc_ac)}: Export results to html

}}
\examples{
## Compute autocorrelations for a list of four single-cases up to lag 2.
autocorr(Huber2014, lag_max = 2)
}
\seealso{
\code{\link[=acf]{acf()}}

Other regression functions: 
\code{\link{bplm}()},
\code{\link{fetch}()},
\code{\link{hplm}()},
\code{\link{mplm}()},
\code{\link{plm}()},
\code{\link{print.sc_bctau}()},
\code{\link{trend}()}
}
\author{
Juergen Wilbert
}
\concept{Autocorrelation}
\concept{Serial correlation}
\concept{regression functions}
\keyword{regression}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pnd.R
\name{pnd}
\alias{pnd}
\title{Percentage of non-overlapping data (PND)}
\usage{
pnd(data, dvar, pvar, decreasing = FALSE, phases = c(1, 2))
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{decreasing}{If you expect data to be lower in the B phase, set
\code{decreasing = TRUE}. Default is \code{decreasing = FALSE}.}

\item{phases}{A vector of two characters or numbers indicating the two phases
that should be compared. E.g., \code{phases = c("A","C")} or \code{phases = c(2,4)}
for comparing the second to the fourth phase. Phases could be combined by
providing a list with two elements. E.g., \code{phases = list(A = c(1,3), B = c(2,4))} will compare phases 1 and 3 (as A) against 2 and 4 (as B). Default
is \code{phases = c(1,2)}.}
}
\value{
\item{PND}{Percentage of non-overlapping data.}
}
\description{
This function returns the percentage of non-overlapping data.  Due to its
error-proneness the PND should not be used, but \code{\link{nap}} or
\code{\link{pand}} instead (see Parker & Vannest, 2009).
}
\details{
PND is calculated by determining the number of data points in phase B that
exceed the highest data point in phase A (or are lower than the lowest data
point in phase A for decreasing data) divided by the total number of data
points in phase B. This value is then multiplied by 100 to get a percentage
value.
}
\examples{

## Calculate the PND for multiple single-case data
pnd(GruenkeWilbert2014)

}
\seealso{
Other overlap functions: 
\code{\link{ird}()},
\code{\link{nap}()},
\code{\link{overlap}()},
\code{\link{pand}()},
\code{\link{pem}()},
\code{\link{pet}()},
\code{\link{print.sc_cdc}()},
\code{\link{select_phases}()},
\code{\link{tau_u}()}
}
\author{
Juergen Wilbert
}
\concept{overlap functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fetch.R
\name{fetch}
\alias{fetch}
\alias{fetch.sc_plm}
\alias{fetch.sc_hplm}
\alias{fetch.sc_bplm}
\alias{fetch.sc_mplm}
\title{Fetches elements from scan objects}
\usage{
fetch(object, what, ...)
}
\arguments{
\item{object}{Object returned from a scan function.}

\item{what}{Element/part to be extracted. Currently, only "model" is
supported to extract the fitted regression model.}

\item{...}{Further parameters passed to the function.}
}
\value{
An object of the respective regression model class.
}
\description{
The fetch function is a getter function for scan objects
returned from regression functions such as \code{plm()}, \code{hplm()}, \code{bplm()}, and
\code{mplm()}. It allows users to extract specific elements from these objects,
such as the fitted model.
}
\examples{
# plm regression
model1 <- plm(example_A24)
fetch(model1, what = "model") |> summary()
# Multilevel plm regression
model2 <- hplm(exampleAB_50)
fetch(model2, what = "model") |> summary()
# Bayesian plm regression
model3 <- bplm(exampleAB_50, nitt = 5000)
fetch(model3, what = "model") |> summary()
}
\seealso{
Other regression functions: 
\code{\link{bplm}()},
\code{\link{hplm}()},
\code{\link{mplm}()},
\code{\link{plm}()},
\code{\link{print.sc_ac}()},
\code{\link{print.sc_bctau}()},
\code{\link{trend}()}
}
\author{
Juergen Wilbert
}
\concept{regression functions}
\keyword{regression}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print-export-outlier.R, R/outlier.R
\name{print.sc_outlier}
\alias{print.sc_outlier}
\alias{export.sc_outlier}
\alias{outlier}
\title{Handling outliers in single-case data}
\usage{
\method{print}{sc_outlier}(x, digits = "auto", ...)

\method{export}{sc_outlier}(object, caption = NA, footnote = NA, filename = NA, ...)

outlier(
  data,
  dvar,
  pvar,
  mvar,
  method = c("MAD", "Cook", "SD", "CI"),
  criteria = 3.5
)
}
\arguments{
\item{x}{An object returned by \code{\link[=outlier]{outlier()}}}

\item{digits}{The minimum number of significant digits to be use.
If set to "auto" (default), values are predefined.}

\item{...}{Further parameters passed to the print function}

\item{object}{An scdf or an object exported from a scan function.}

\item{caption}{Character string with table caption. If left NA (default) a
caption will be created based on the exported object.}

\item{footnote}{Character string with table footnote. If left NA (default) a
footnote will be created based on the exported object.}

\item{filename}{String containing the file name. If a filename is given the
output will be written to that file.}

\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{mvar}{Character string with the name of the measurement time variable.
Defaults to the attributes in the scdf file.}

\item{method}{Specifies the method for outlier identification. Set \code{method = "MAD"} for mean average deiviation, \code{method = "SD"} for standard
deviations, \code{method = "CI"} for confidence intervals, \code{method = "Cook"} for
Cook's Distance based on the Piecewise Linear Regression Model.}

\item{criteria}{Specifies the criteria for outlier identification. Based on
the \code{method} setting.}
}
\value{
\tabular{ll}{
    \tab  \cr
   \code{data} \tab A single-case data frame with substituted outliers. \cr
   \code{dropped.n} \tab A list with the number of dropped data points for each single-case. \cr
   \code{dropped.mt} \tab A list with the measurement-times of dropped data points for each single-case (values are based on the mt variable of each single-case data frame). \cr
   \code{sd.matrix} \tab A list with a matrix for each case with values for the upper and lower boundaries based on the standard deviation. \cr
   \code{ci.matrix} \tab A list with a matrix for each single-case with values for the upper and lower boundaries based on the confidence interval. \cr
   \code{cook} \tab A list of Cook's Distances for each measurement of each single-case. \cr
   \code{criteria} \tab Criteria used for outlier analysis. \cr
   \code{N} \tab Number of single-cases. \cr
   \code{case.names} \tab Case identifier. \cr
}
}
\description{
Identifies and drops outliers within a single-case data frame (scdf).
Outliers can be identified based on mean average deviation (MAD),
standard deviation (SD), confidence intervals (CI), or Cook's Distance
from a Piecewise Linear Regression Model.
}
\details{
For \code{method = "SD"}, \code{criteria = 2} would refer t0 two standard
deviations. For \code{method = "MAD"}, \code{criteria = 3.5} would refer to 3.5 times
the mean average deviation. For \code{method = "CI"}, \code{criteria = 0.99} would
refer to a 99 percent confidence interval. For \code{method = "cook"}, \code{criteria = "4/n"} would refer to a Cook's Distance greater than 4/n.
}
\section{Functions}{
\itemize{
\item \code{print(sc_outlier)}: Print results

\item \code{export(sc_outlier)}: Export html results

}}
\examples{

## Identify outliers using 1.5 standard deviations as criterion
susanne <- random_scdf(level = 1.0)
res_outlier <- outlier(susanne, method = "SD", criteria = 1.5)
res_outlier

## Identify outliers in the original data from Grosche (2011)
## using Cook's Distance greater than 4/n as criterion
res_outlier <- outlier(Grosche2011, method = "Cook", criteria = "4/n")
res_outlier

}
\seealso{
Other data manipulation functions: 
\code{\link{add_l2}()},
\code{\link{as.data.frame.scdf}()},
\code{\link{as_scdf}()},
\code{\link{batch_apply}()},
\code{\link{fill_missing}()},
\code{\link{moving_median}()},
\code{\link{ranks}()},
\code{\link{rescale}()},
\code{\link{scdf}()},
\code{\link{select_cases}()},
\code{\link{set_vars}()},
\code{\link{shift}()},
\code{\link{smooth_cases}()},
\code{\link{standardize}()},
\code{\link{truncate_phase}()}
}
\author{
Juergen Wilbert
}
\concept{data manipulation functions}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.sc.R, R/print.sc_desc.R,
%   R/print.sc_design.R, R/print.sc_nap.R, R/print.sc_overlap.R,
%   R/print.sc_pem.R, R/print.sc_pet.R, R/print.sc_pnd.R, R/print.sc_power.R,
%   R/print.sc_rci.R, R/print.sc_smd.R, R/print.sc_trend.R
\name{print.sc}
\alias{print.sc}
\alias{print.sc_desc}
\alias{print.sc_design}
\alias{print.sc_nap}
\alias{print.sc_overlap}
\alias{print.sc_pem}
\alias{print.sc_pet}
\alias{print.sc_pnd}
\alias{print.sc_power}
\alias{print.sc_rci}
\alias{print.sc_smd}
\alias{print.sc_trend}
\title{Print methods for scan objects}
\usage{
\method{print}{sc_desc}(x, digits = "auto", ...)

\method{print}{sc_design}(x, ...)

\method{print}{sc_nap}(x, digits = "auto", nice = TRUE, complete = FALSE, ...)

\method{print}{sc_overlap}(x, digits = "auto", ...)

\method{print}{sc_pem}(x, ...)

\method{print}{sc_pet}(x, digits = 3, ...)

\method{print}{sc_pnd}(x, ...)

\method{print}{sc_power}(x, duration = FALSE, digits = 1, ...)

\method{print}{sc_rci}(x, digits = 3, ...)

\method{print}{sc_smd}(x, digits = "auto", ...)

\method{print}{sc_trend}(x, digits = 3, ...)
}
\arguments{
\item{x}{Object}

\item{digits}{The minimum number of significant digits to be use.
If set to "auto" (default), values are predefined.}

\item{...}{Further parameters passed to the print function}

\item{nice}{If set TRUE (default) output values are rounded and optimized for
publication tables.}

\item{duration}{If TRUE the duration for computation is printed.}
}
\description{
Print methods for scan objects
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scdf-functions.R
\name{combine}
\alias{combine}
\alias{c.scdf}
\title{Combine single-case data frames into one scdf}
\usage{
combine(..., dvar = NULL, pvar = NULL, mvar = NULL, info = NULL, author = NULL)

\method{c}{scdf}(...)
}
\arguments{
\item{...}{scdf objects to be combined.}

\item{dvar}{Character string. Name of the dependent variable. Defaults to the
dependent variable of the first case provided.}

\item{pvar}{Character string. Name of the phase variable. Defaults to the
phase variable of the first case provided.}

\item{mvar}{Character string. Name of the measurement-time variable. Defaults
to the measurement-time variable of the first case provided.}

\item{info}{additional information on the scdf file.}

\item{author}{author of the data.}
}
\value{
A scdf. If not set differently, the attributes of this scdf are
copied from the first scdf provided (i.e the first argument of the
function).
}
\description{
Combines several single-case data frames (scdf) into one scdf object.
}
\author{
Juergen Wilbert
}
\keyword{data}
\keyword{manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprec-shift.R
\name{shift}
\alias{shift}
\title{Shift values in a single-case data file}
\usage{
shift(data, value, var)
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{value}{Number by which to shift the values}

\item{var}{Character string with the name of the target variable.
Defaults to the measurement time variable.}
}
\value{
A scdf with shifted data
}
\description{
\emph{This function has been superseded by the much more versatile
\code{\link{transform.scdf}} function.}
Shifting the values might be helpful in cases where the measurement time
is given as a time variable (see example below).
}
\examples{
### Shift the measurement time for a better estimation of the intercept
ex <- shift(example_A24, value = -1996)
plm(ex)

# Please use transform instead:
example_A24 |>
  transform(year = year - 1996) |>
  plm()
}
\seealso{
Other data manipulation functions: 
\code{\link{add_l2}()},
\code{\link{as.data.frame.scdf}()},
\code{\link{as_scdf}()},
\code{\link{batch_apply}()},
\code{\link{fill_missing}()},
\code{\link{moving_median}()},
\code{\link{print.sc_outlier}()},
\code{\link{ranks}()},
\code{\link{rescale}()},
\code{\link{scdf}()},
\code{\link{select_cases}()},
\code{\link{set_vars}()},
\code{\link{smooth_cases}()},
\code{\link{standardize}()},
\code{\link{truncate_phase}()}
}
\concept{data manipulation functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/random_scdf.R
\name{random_scdf}
\alias{random_scdf}
\title{Single-case data generator}
\usage{
random_scdf(design = NULL, round = NA, random_names = FALSE, seed = NULL, ...)
}
\arguments{
\item{design}{A design matrix which is created by \code{design} and specifies
all parameters. If \code{design} is \code{NULL} (default), the design
parameters have to be specified via the \code{...} argument. If a numeric
value is provided instead of a design matrix, it is interpreted as the
number of cases \code{n}.}

\item{round}{Rounds the scores to the defined decimal. To round to the second
decimal, set \code{round = 2}. Default is \code{NA} (no rounding).}

\item{random_names}{Is \code{FALSE} by default. If set \code{random_names =
  TRUE} cases are assigned random first names. If set \code{"neutral", "male"
  or "female"} only gender neutral, male, or female names are chosen. The
names are drawn from the 2,000 most popular names for newborns in 2012 in
the U.S. (1,000 male and 1,000 female names).}

\item{seed}{A seed number for the random generator. If \code{NULL} (default),
no seed}

\item{...}{arguments that are directly passed to the \code{design} function
for a more concise coding.}
}
\value{
A single-case data frame. See \code{\link{scdf}} to learn about this
format.
}
\description{
The \code{random_scdf} function generates random single-case data frames for
monte-carlo studies and demonstration purposes. \code{design} is used to set
up a design matrix with all parameters needed for the \code{random_scdf}
function.
}
\details{
The generated data can be normally distributed, Poisson-distributed, or
binomially distributed. The default is normally distributed data.
}
\examples{

## Create random single-case data and inspect it
design <- design(
  n = 3, rtt = 0.75, slope = 0.1, extreme_prop = 0.1,
  missing_prop = 0.1
)
dat <- random_scdf(design, round = 1, random_names = TRUE, seed = 123)
describe(dat)

## And now have a look at poisson-distributed data
design <- design(
  n = 3, B_start = c(6, 10, 14), mt = c(12, 20, 22), start_value = 10,
  distribution = "poisson", level = -5, missing_prop = 0.1
)
dat <- random_scdf(design, seed = 1234)
pand(dat, decreasing = TRUE)
}
\author{
Juergen Wibert
}
\concept{mc fucntions}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprec-smooth_cases.R
\name{smooth_cases}
\alias{smooth_cases}
\title{Smoothing single-case data}
\usage{
smooth_cases(data, dvar, mvar, method = "mean", intensity = NULL, FUN = NULL)
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{mvar}{Character string with the name of the measurement time variable.
Defaults to the attributes in the scdf file.}

\item{method, FUN}{Method determining the smoothed scores. Default \code{method =
"median"} is a moving median function. Further possible values are:
\code{"mean"} and a non-parametric \code{"regression"}.}

\item{intensity}{For \code{method = "median"} and \code{"mean"} it
is the lag used for computing the average. Default is \code{intensity = 1}.
In case of \code{method = "regression"} it is the proportion of
surrounding data influencing each data point, which is \code{intensity =
0.2} by default.}
}
\value{
Returns a data frame (for each single-case) with smoothed data
points. See \code{\link{scdf}} to learn about the format of these data
frames.
}
\description{
\emph{This function is superseded by the more versatile
\code{\link{transform.scdf}} function.}
The \code{smooth_cases} function provides procedures to smooth single-case
data (i.e., to eliminate noise). A moving average function (mean- or
median-based) replaces each data point by the average of the surrounding
data points step-by-step. With a local regression function, each data point
is regressed by its surrounding data points.
}
\details{
\code{moving_median}, \code{moving_mean}, and \code{local_regression} are
helper function for \code{\link{transform.scdf}} returning the smoothed
values of a numeric vector.
}
\examples{

## Use the three different smoothing functions and compare the results
study <- c(
  "Original" = Huber2014$Berta,
  "Moving median" = smooth_cases(Huber2014$Berta, method = "median"),
  "Moving mean" = smooth_cases(Huber2014$Berta, method = "mean"),
  "Local regression" = smooth_cases(Huber2014$Berta, method = "regression")
)
plot(study)

Huber2014$Berta |>
transform(
  "compliance (moving median)" = moving_median(compliance),
  "compliance (moving mean)" = moving_mean(compliance),
  "compliance (local regression)" = local_regression(compliance, mt)
)

}
\seealso{
Other data manipulation functions: 
\code{\link{add_l2}()},
\code{\link{as.data.frame.scdf}()},
\code{\link{as_scdf}()},
\code{\link{batch_apply}()},
\code{\link{fill_missing}()},
\code{\link{moving_median}()},
\code{\link{print.sc_outlier}()},
\code{\link{ranks}()},
\code{\link{rescale}()},
\code{\link{scdf}()},
\code{\link{select_cases}()},
\code{\link{set_vars}()},
\code{\link{shift}()},
\code{\link{standardize}()},
\code{\link{truncate_phase}()}
}
\author{
Juergen Wilbert
}
\concept{data manipulation functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mplm.R, R/print-export-mplm.R
\name{mplm}
\alias{mplm}
\alias{print.sc_mplm}
\alias{export.sc_mplm}
\title{Multivariate Piecewise linear model / piecewise regression}
\usage{
mplm(
  data,
  dvar,
  mvar,
  pvar,
  model = c("W", "H-M", "B&L-B", "JW"),
  contrast = c("first", "preceding"),
  contrast_level = c(NA, "first", "preceding"),
  contrast_slope = c(NA, "first", "preceding"),
  trend = TRUE,
  level = TRUE,
  slope = TRUE,
  formula = NULL,
  update = NULL,
  na.action = na.omit,
  ...
)

\method{print}{sc_mplm}(x, digits = "auto", std = FALSE, ...)

\method{export}{sc_mplm}(
  object,
  caption = NA,
  footnote = NA,
  filename = NA,
  nice = TRUE,
  std = FALSE,
  decimals = 2,
  ...
)
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{mvar}{Character string with the name of the measurement time variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{model}{Model used for calculating the dummy parameters (see Huitema &
McKean, 2000). Default is \code{model = "W"}. Possible values are: \code{"B&L-B"},
\code{"H-M"}, \code{"W"}, and deprecated \code{"JW"}.}

\item{contrast}{Sets contrast_level and contrast_slope. Either "first",
"preceding" or a contrast matrix. If NA contrast is ignored.}

\item{contrast_level}{Either "first", "preceding" or a contrast matrix. If NA
contrast_level is a copy of contrast.}

\item{contrast_slope}{Either "first", "preceding" or a contrast matrix. If NA
contrast_level is a copy of contrast.}

\item{trend}{A logical indicating if a trend parameters is included in the
model.}

\item{level}{A logical indicating if a level parameters is included in the
model.}

\item{slope}{A logical indicating if a slope parameters is included in the
model.}

\item{formula}{Defaults to the standard piecewise regression model. The
parameter phase followed by the phase name (e.g., \code{phaseB}) indicates the
level effect of the corresponding phase. The parameter 'inter' followed by
the phase name (e.g., \code{interB}) adresses the slope effect based on the
method provide in the model argument (e.g., \code{"B&L-B"}). The formula can be
changed for example to include further variables into the regression model.}

\item{update}{An easier way to change the regression formula (e.g., \code{. ~ . + newvariable}).}

\item{na.action}{Defines how to deal with missing values.}

\item{...}{Further arguments passed to the \code{\link[=lm]{lm()}} function.}

\item{x}{Object returned from \code{\link[=mplm]{mplm()}}.}

\item{digits}{The minimum number of significant digits to be use.
If set to "auto" (default), values are predefined.}

\item{std}{If TRUE, a table with standardized estimates is included.}

\item{object}{An scdf or an object exported from a scan function.}

\item{caption}{Character string with table caption. If left NA (default) a
caption will be created based on the exported object.}

\item{footnote}{Character string with table footnote. If left NA (default) a
footnote will be created based on the exported object.}

\item{filename}{String containing the file name. If a filename is given the
output will be written to that file.}

\item{nice}{If set TRUE (default) output values are rounded and optimized for
publication tables.}

\item{decimals}{Decimal places that are reported.}
}
\value{
\tabular{ll}{
    \tab  \cr
   \code{model} \tab Character string from function call (see arguments above). \cr
   \code{contrast} \tab List with contrast definitions. \cr
   \code{full.model} \tab Full regression model list. \cr
   \code{formula} \tab Formula of the mplm model. \cr
}
}
\description{
The \code{\link[=mplm]{mplm()}} function computes a multivariate piecewise regression model. The
function automatically creates the regression formula based on the provided
data and the selected options. The default model includes trend, level, and
slope effects for each dependent variable. The regression formula can be
changed by providing a custom formula to the \code{update} argument.
}
\details{
The function currently only supports single-case data (i.e., one case per
dataset). For multilevel piecewise regression models, please use the \link{hplm}
function.
}
\section{Functions}{
\itemize{
\item \code{print(sc_mplm)}: Print results

\item \code{export(sc_mplm)}: Export results as html

}}
\examples{
res <- mplm(Leidig2018$`1a1`,
  dvar = c("academic_engagement", "disruptive_behavior")
)
print(res)
## also report standardized coefficients:
print(res, std = TRUE)
}
\seealso{
Other regression functions: 
\code{\link{bplm}()},
\code{\link{fetch}()},
\code{\link{hplm}()},
\code{\link{plm}()},
\code{\link{print.sc_ac}()},
\code{\link{print.sc_bctau}()},
\code{\link{trend}()}
}
\author{
Juergen Wilbert
}
\concept{regression functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pem.R
\name{pem}
\alias{pem}
\title{Percent exceeding the median (PEM)}
\usage{
pem(
  data,
  dvar,
  pvar,
  decreasing = FALSE,
  binom.test = TRUE,
  chi.test = FALSE,
  FUN = median,
  phases = c(1, 2),
  ...
)
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{decreasing}{If you expect data to be lower in the B phase, set
\code{decreasing = TRUE}. Default is \code{decreasing = FALSE}.}

\item{binom.test}{Computes a binomial test for a 50/50 distribution. Default
is \code{binom.test = TRUE}.}

\item{chi.test}{Computes a Chi-square test. The default setting
\code{chi.test = FALSE} skips the Chi-square test.}

\item{FUN}{Data points are compared with the phase A median. Use this
argument to implement alternative measures of central tendency. Default is
\code{FUN = median}.}

\item{phases}{A vector of two characters or numbers indicating the two phases
that should be compared. E.g., \code{phases = c("A","C")} or \code{phases = c(2,4)}
for comparing the second to the fourth phase. Phases could be combined by
providing a list with two elements. E.g., \code{phases = list(A = c(1,3), B = c(2,4))} will compare phases 1 and 3 (as A) against 2 and 4 (as B). Default
is \code{phases = c(1,2)}.}

\item{\dots}{Additional arguments for the \code{FUN} parameter (e.g.
\code{FUN = mean, trim = 0.1} will use the 10 percent trimmed arithmetic
mean instead of the median for comparisons). The function must take a vector
of numeric values and the \code{na.rm} argument and return a numeric value.}
}
\description{
The \code{pem} function returns the percentage of phase B data exceeding the
phase A median.  Additionally, a chi square test against a 50/50
distribution is computed.  Different measures of central tendency can be
addressed for alternative analyses.
}
\details{
The Percent Exceeding the Median is calculated as
the percentage of data points in phase B that exceed the median of phase A.
If the \code{decreasing} argument is set to \code{TRUE},
the percentage of data points in phase B that are below the median of
phase A is calculated.  The PEM is expressed as a percentage ranging
from 0 to 100.  Higher values indicate a greater degree of
improvement from phase A to phase B.
}
\examples{

## Calculate the PEM including the Binomial and Chi-square tests for a single-case
dat <- random_scdf(5, level = 0.5)
pem(dat, chi.test = TRUE)

}
\seealso{
Other overlap functions: 
\code{\link{ird}()},
\code{\link{nap}()},
\code{\link{overlap}()},
\code{\link{pand}()},
\code{\link{pet}()},
\code{\link{pnd}()},
\code{\link{print.sc_cdc}()},
\code{\link{select_phases}()},
\code{\link{tau_u}()}
}
\author{
Juergen Wilbert
}
\concept{overlap functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plm.R, R/print-export-plm.R
\name{plm}
\alias{plm}
\alias{print.sc_plm}
\alias{export.sc_plm}
\title{Piecewise linear model / piecewise regression}
\usage{
plm(
  data,
  dvar,
  pvar,
  mvar,
  AR = 0,
  model = c("W", "H-M", "B&L-B", "JW"),
  family = "gaussian",
  trend = TRUE,
  level = TRUE,
  slope = TRUE,
  contrast = c("first", "preceding"),
  contrast_level = c(NA, "first", "preceding"),
  contrast_slope = c(NA, "first", "preceding"),
  formula = NULL,
  update = NULL,
  na.action = na.omit,
  r_squared = TRUE,
  var_trials = NULL,
  dvar_percentage = FALSE,
  ...
)

\method{print}{sc_plm}(
  x,
  lag_max = 3,
  ci = 0.95,
  q = FALSE,
  r_squared = getOption("scan.rsquared"),
  ...
)

\method{export}{sc_plm}(
  object,
  caption = NA,
  footnote = NA,
  filename = NA,
  nice = TRUE,
  ci = 0.95,
  q = FALSE,
  round = 2,
  r_squared = getOption("scan.rsquared"),
  ...
)
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{mvar}{Character string with the name of the measurement time variable.
Defaults to the attributes in the scdf file.}

\item{AR}{Maximal lag of autoregression. Modelled based on the
Autoregressive-Moving Average (ARMA) function.  When AR is set, the family
argument must be set to \code{family = "gaussian"}.}

\item{model}{Model used for calculating the dummy parameters (see Huitema &
McKean, 2000). Default is \code{model = "W"}. Possible values are: \code{"B&L-B"},
\code{"H-M"}, \code{"W"}, and deprecated \code{"JW"}.}

\item{family}{Set the distribution family. Defaults to a gaussian
distribution. See the \code{family} function for more details.}

\item{trend}{A logical indicating if a trend parameters is included in the
model.}

\item{level}{A logical indicating if a level parameters is included in the
model.}

\item{slope}{A logical indicating if a slope parameters is included in the
model.}

\item{contrast}{Sets contrast_level and contrast_slope. Either "first",
"preceding" or a contrast matrix. If NA contrast is ignored.}

\item{contrast_level}{Either "first", "preceding" or a contrast matrix. If NA
contrast_level is a copy of contrast.}

\item{contrast_slope}{Either "first", "preceding" or a contrast matrix. If NA
contrast_level is a copy of contrast.}

\item{formula}{Defaults to the standard piecewise regression model. The
parameter phase followed by the phase name (e.g., phaseB) indicates the
level effect of the corresponding phase. The parameter 'inter' followed by
the phase name (e.g., interB) adresses the slope effect based on the method
provide in the model argument (e.g., "B&L-B"). The formula can be changed
for example to include further variables into the regression model.}

\item{update}{An easier way to change the regression formula (e.g., \code{. ~ . + newvariable}).}

\item{na.action}{Defines how to deal with missing values.}

\item{r_squared}{Either "delta", "partial", or "none".}

\item{var_trials}{Name of the variable containing the number of trials (only
for binomial regressions). If a single integer is provided this is
considered to be a the constant number of trials across all measurements.}

\item{dvar_percentage}{Only for binomial distribution. If set TRUE, the
dependent variable is assumed to represent proportions \verb{[0,1]}. Otherwise
dvar is assumed to represent counts.}

\item{...}{Further arguments passed to the \code{\link[stats:glm]{stats::glm()}} or \code{\link[nlme:gls]{nlme::gls()}}
function.}

\item{x}{Object}

\item{lag_max}{Maximum lag to be reported for autocorrelation of residuals.
Default is \code{3}. Set \code{FALSE} for no report of autocorrelations.}

\item{ci}{Print confidence intervals. Either FALSE, TRUE or a number
between 0 and 1 (0.90 for a 90\% intervals).}

\item{q}{Logical. If set \code{TRUE}, Yule's Q is reported.}

\item{object}{An scdf or an object exported from a scan function.}

\item{caption}{Character string with table caption. If left NA (default) a
caption will be created based on the exported object.}

\item{footnote}{Character string with table footnote. If left NA (default) a
footnote will be created based on the exported object.}

\item{filename}{String containing the file name. If a filename is given the
output will be written to that file.}

\item{nice}{If set TRUE (default) output values are rounded and optimized for
publication tables.}

\item{round}{Integer passed to the digits argument used to round values.}
}
\value{
An object of class \code{sc_plm}.
\item{formula}{plm formula. Uselful if you want to use the update or
formula argument and you don't know the names of the parameters.}
\item{model}{Character string from function call (see \code{Arguments}
above).}
\item{F.test}{F-test values of modelfit.} \item{r.squares}{Explained variance
R squared for each model parameter.}
\item{ar}{Autoregression lag from function call (see \code{Arguments}
above).}
\item{family}{Distribution family from function call
(see \code{Arguments} above).}
\item{full.model}{Full regression model list from the gls or glm function.}
}
\description{
The \code{plm} function computes a piecewise regression model (see Huitema &
McKean, 2000) for one case. The function automatically creates the default
model which includes trend, level, and slope effects. The model can be
changed with \code{trend}, \code{level}, and \code{slope} arguments or by providing a custom
formula.
}
\details{
The function uses the \code{glm} function of the stats package or the \code{gls}
function of the nlme package (for auto-regression models).
}
\section{Functions}{
\itemize{
\item \code{print(sc_plm)}: Print

\item \code{export(sc_plm)}: Export results as html table (see \code{\link[=export]{export()}})

}}
\examples{

## Compute a piecewise regression model for a random single-case
set.seed(123)
AB <- design(
  phase_design = list(A = 10, B = 20),
  level = list(A = 0, B = 1), slope = list(A = 0, B = 0.05),
  trend = 0.05
)
dat <- random_scdf(design = AB)
plm(dat, AR = 3)

## Another example with a more complex design
A1B1A2B2 <- design(
  phase_design = list(A1 = 15, B1 = 20, A2 = 15, B2 = 20),
  level = list(A1 = 0, B1 = 1, A2 = -1, B2 = 1),
  slope = list(A1 = 0, B1 = 0.0, A2 = 0, B2 = 0.0),
  trend = 0.0)
dat <- random_scdf(design = A1B1A2B2, seed = 123)
plm(dat, contrast = "preceding")

## no slope effects were found. Therefore, you might want to the drop slope
## estimation:
plm(dat, slope = FALSE, contrast = "preceding")

## and now drop the trend estimation as well
plm(dat, slope = FALSE, trend = FALSE, contrast = "preceding")

## A poisson regression
example_A24 |>
  plm(family = "poisson")

## A binomial regression (frequencies as dependent variable)
plm(exampleAB_score$Christiano, family = "binomial", var_trials = "trials")

## A binomial regression (percentage as dependent variable)
exampleAB_score$Christiano |>
  transform(percentage = values/trials) |>
  set_dvar("percentage") |>
  plm(family = "binomial", var_trials = "trials", dvar_percentage = TRUE)
## Print
plm(exampleAB$Johanna) |> 
  print(ci = 0.9, r_squared = c("delta", "partial"))
## Export
plm(exampleAB$Johanna) |> export()

}
\references{
Beretvas, S., & Chung, H. (2008). An evaluation of modified
R2-change effect size indices for single-subject experimental designs.
\emph{Evidence-Based Communication Assessment and Intervention, 2},
120-128.

Huitema, B. E., & McKean, J. W. (2000). Design specification issues in
time-series intervention models. \emph{Educational and Psychological
Measurement, 60}, 38-58.
}
\seealso{
Other regression functions: 
\code{\link{bplm}()},
\code{\link{fetch}()},
\code{\link{hplm}()},
\code{\link{mplm}()},
\code{\link{print.sc_ac}()},
\code{\link{print.sc_bctau}()},
\code{\link{trend}()}
}
\author{
Juergen Wilbert
}
\concept{regression functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pet.R
\name{pet}
\alias{pet}
\title{Percent exceeding the trend (PET)}
\usage{
pet(data, dvar, pvar, mvar, ci = 0.95, decreasing = FALSE, phases = c(1, 2))
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{mvar}{Character string with the name of the measurement time variable.
Defaults to the attributes in the scdf file.}

\item{ci}{Width of the confidence interval. Default is \code{ci = 0.95}.}

\item{decreasing}{If you expect data to be lower in the B phase, set
\code{decreasing = TRUE}. Default is \code{decreasing = FALSE}.}

\item{phases}{A vector of two characters or numbers indicating the two phases
that should be compared. E.g., \code{phases = c("A","C")} or \code{phases = c(2,4)}
for comparing the second to the fourth phase. Phases could be combined by
providing a list with two elements. E.g., \code{phases = list(A = c(1,3), B = c(2,4))} will compare phases 1 and 3 (as A) against 2 and 4 (as B). Default
is \code{phases = c(1,2)}.}
}
\value{
\tabular{ll}{
    \tab  \cr
   \code{PET} \tab Percent exceeding the trend. \cr
   \code{ci} \tab Width of confidence interval. \cr
   \code{decreasing} \tab Logical argument from function call (see Arguments above). \cr
}
}
\description{
The \code{pet} function returns the percentage of Phase B data points that exceed
the prediction based on the Phase A trend. A binomial test against a 50/50
distribution is calculated. It also calculates the percentage of Phase B data
points that exceed the upper (or lower) 95 percent confidence interval of the
predicted progression.
}
\details{
The PET is calculated by first fitting a linear model to the Phase A data
to estimate the trend. Then, for each data point in Phase B, it is
determined whether it exceeds the predicted value from the Phase A trend.
The PET is the percentage of Phase B data points that exceed this predicted
value. Additionally, a binomial test is performed to assess whether the
observed PET is significantly greater than what would be expected by chance
(i.e., 50\%).
}
\examples{

## Calculate the PET and use a 99\%-CI for the additional calculation
# create random example data
design <- design(n = 5, slope = 0.2)
dat <- random_scdf(design, seed = 23)
pet(dat, ci = .99)

}
\seealso{
Other overlap functions: 
\code{\link{ird}()},
\code{\link{nap}()},
\code{\link{overlap}()},
\code{\link{pand}()},
\code{\link{pem}()},
\code{\link{pnd}()},
\code{\link{print.sc_cdc}()},
\code{\link{select_phases}()},
\code{\link{tau_u}()}
}
\author{
Juergen Wilbert
}
\concept{overlap functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/private_check.R
\name{check_args}
\alias{check_args}
\title{Check function arguments}
\usage{
check_args(...)
}
\value{
This function does not return a value. If any of the tests fail,
it raises an error with a descriptive message.
}
\description{
This function checks function arguments based on a set of predefined tests.
It is intended for internal use within other functions to validate their
arguments.
#' @param ... A set of expressions defining the tests to be performed on the
function arguments. Each expression should use one of the predefined test
functions available within the scope of \code{check_args()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/between_smd.R, R/print-export-bcsmd.R
\name{between_smd}
\alias{between_smd}
\alias{print.sc_bcsmd}
\alias{export.sc_bcsmd}
\title{Between-Case Standardized Mean Difference}
\usage{
between_smd(
  data,
  method = c("REML", "MCMCglmm"),
  ci = 0.95,
  include_residuals = TRUE,
  ...
)

\method{print}{sc_bcsmd}(x, digits = 2, ...)

\method{export}{sc_bcsmd}(
  object,
  caption = NA,
  footnote = NA,
  filename = NA,
  digits = 2,
  round = 2,
  ...
)
}
\arguments{
\item{data}{Either an scdf or an object returned from the \code{\link[=hplm]{hplm()}} or
\code{\link[=bplm]{bplm()}} function.}

\item{method}{Either \code{"REML"} or "\verb{MCMglmm"}. This indicated which
statistical method is applied to calculate the model.}

\item{ci}{A numeric between 0 and 1 setting the width of the confidence
interval (when method is REML) or the credible interval (when method is
MCMCglmm). The default is \code{0.95} for a 95-percent interval.}

\item{include_residuals}{Logical. See details.}

\item{...}{Further arguments passed to the \code{\link[=hplm]{hplm()}} or \code{\link[=bplm]{bplm()}}function.}

\item{x}{An object returned by \code{baseline_smd()}.}

\item{digits}{The minimum number of significant digits to be use.
If set to "auto" (default), values are predefined.}

\item{object}{An scdf or an object exported from a scan function.}

\item{caption}{Character string with table caption. If left NA (default) a
caption will be created based on the exported object.}

\item{footnote}{Character string with table footnote. If left NA (default) a
footnote will be created based on the exported object.}

\item{filename}{String containing the file name. If a filename is given the
output will be written to that file.}

\item{round}{Integer passed to the digits argument used to round values.}
}
\value{
An object of class sc_bcsmd. It is a list containing the following
elements:
\itemize{
\item \code{models}: A list of data frames containing the BC-SMD results for each
model calculated.
\item \code{ci}: The width of the confidence/credible interval.
\item \code{method}: The method used for model estimation.
}
}
\description{
Calculates a standardized mean difference from a multilevel model as
described in Pustejovsky et al. (2014)
}
\details{
The BC-SMD is calculate as \verb{BC-SMD = Phase difference / sqrt(residual + random_intercept)}. This is most closely related to Cohen's \emph{d}. If you want
to have the most exact estimation based on the between case variance, you
have to exclude the residual variance by setting the argument
\code{include_residuals = FALSE} you get \verb{BC-SMD = Phase difference / sqrt(random_intercept)}. The 'base' model only includes the phase level as a
predictor like originally proposed by Hedges et al. Whereas the 'Full plm'
model includes the trend and the phase slope as additional predictors.
}
\section{Functions}{
\itemize{
\item \code{print(sc_bcsmd)}: Print results

\item \code{export(sc_bcsmd)}: export results

}}
\examples{
## Create a example scdf:
des <- design(
  n = 150,
  phase_design = list(A1 = 10, B1 = 10, A2 = 10, B2 = 10, C = 10),
  level = list(B1 = 1, A2 = 0, B2 = 1, C = 1),
  rtt = 0.7,
  random_start_value = TRUE
)
study <- random_scdf(des)

## Standard BC-SMD return:
between_smd(study)

## Specify the model and provide an hplm object:
model <- hplm(study, contrast_level = "preceding", slope = FALSE,  trend = FALSE)
between_smd(model)

## excluding the residuals gives a more accurate estimation:
between_smd(model, include_residuals = FALSE)

}
\references{
Pustejovsky, J. E., Hedges, L. V., & Shadish, W. R. (2014).
Design-Comparable Effect Sizes in Multiple Baseline Designs: A General
Modeling Framework. Journal of Educational and Behavioral Statistics,
39(5), 368–393. https://doi.org/10.3102/1076998614547577
}
\author{
Juergen Wilbert
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.scdf.R
\name{plot.scdf}
\alias{plot.scdf}
\alias{plotSC}
\title{(Deprecated) Plot single-case data}
\usage{
\method{plot}{scdf}(...)

plotSC(
  data,
  dvar,
  pvar,
  mvar,
  ylim = NULL,
  xlim = NULL,
  xinc = 1,
  lines = NULL,
  marks = NULL,
  phase.names = NULL,
  xlab = NULL,
  ylab = NULL,
  main = "",
  case.names = NULL,
  style = getOption("scan.plot.style"),
  ...
)
}
\arguments{
\item{...}{Further arguments passed to the plot command.}

\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{mvar}{Character string with the name of the measurement time variable.
Defaults to the attributes in the scdf file.}

\item{ylim}{Lower and upper limits of the y-axis (e.g., \code{ylim = c(0,
  20)} sets the y-axis to a scale from 0 to 20). With multiple single-cases
you can use \code{ylim = c(0, NA)} to scale the y-axis from 0 to the
maximum of each case. \code{ylim} is not set by default, which makes
\code{scan} set a proper scale based on the given data.}

\item{xlim}{Lower and upper limits of the x-axis (e.g., \code{xlim = c(0,
  20)} sets the x-axis to a scale from 0 to 20). With multiple single-cases
you can use \code{ylim = c(0, NA)} to scale the x-axis from 0 to the
maximum of each case. \code{xlim} is not set by default, which makes
\code{scan} set a proper scale based on the given data.}

\item{xinc}{An integer. Increment of the x-axis. 1 :each mt value will be
printed, 2 : every other value, 3 : every third values etc.}

\item{lines}{A list defining one or multiple lines or curves to be plotted.
The argument is passed as a list (e.g., \code{list(type = "median")}). Some
of the procedures can be refined with an additional argument (e.g.,
\code{lines = list(type = "mean", trim = 0.2)} adds a 20\\% trimmed mean
line. For multiple lines, provide a list element for each line (e.g.,
\code{list( list(type = "median", col = "red"), list(type = "trend", col =
  "blue"))}.
Possible lines are: \itemize{
\item\code{"median"} Separate lines for phase A and B medians.
\item\code{"mean"} Separate lines for phase A and B means. By default it is
10\\%-trimmed. Other trims can be set, using a trim parameter (e.g.,
\code{lines = list(type = "mean", trim = 0.2)} draws a 20\\%-trimmed mean line).
\item\code{"trend"} Separate lines for phase A and B trends.
\item\code{"trendA"} OLS trend line for phase A, extrapolated throughout phase B.
\item\code{"trendA_bisplit"} Split middle (bi-split) trend line for phase A, extrapolated throughout phase B.
\item\code{"trendA_trisplit"} Tukey tri-split trend line for phase A, extrapolated throughout phase B.
\item\code{"maxA/minA"} Line at the level of the highest or lowest phase A score.
\item\code{"medianA"} Line at the phase A median score.  \item\code{"meanA"}
Line at the phase A 10\\%-trimmed mean score. Apply a different trim, by
using the additional argument (e.g., \code{lines = list(type = "meanA", trim = 0.2)}).
\item\code{"plm"} Regression lines for piecewise linear regression model.
\item\code{"plm.ar"} Regression lines for piecewise autoregression model.
The lag is specified like this: \code{lines = list(type = "plm.ar", ar = 2)}. Default lag is set to 2.
\item\code{"movingMean"} Draws a moving mean curve, with a specified lag:
\code{lines = list(type = "movingMean", lag = 2)}. Default is a lag 1 curve.
\item\code{"movingMedian"} Draws a moving median curve, with a specified
lag: \code{lines = list(type = "movingMedian", lag = 3)}. Default is a lag 1 curve.
\item\code{"loreg"} Draws a non-parametric local regression line. The
proportion of data influencing each data point can be specified using
\code{lines = list(type = "loreg"m f = 0.66)}. The default is 0.5.  \item\code{"lty"}
Use this argument to define the line type. Examples are: \code{"solid"},
\code{"dashed"}, \code{"dotted"}.  \item\code{"lwd"} Use this argument to
define the line's thickness, e.g., \code{lwd = 4}.  \item\code{"col"} Use
this argument to define the line's color, e.g., \code{col = "red"}.  }}

\item{marks}{A list of parameters defining markings of certain data points.
\itemize{ \item\code{"positions"} A vector or a list of vectors indicating
measurement-times to be highlighted. In case of a vector, the marked
measurement-times are the same for all plotted cases. In case of a list of
vectors, marks are set differently for each case. The list must have the
same length as there are cases in the data file.  \item\code{"col"} Color of
the marks.  \item\code{"cex"} Size of the marks.  } Use for example
\code{marks = list(positions = c(1, 8, 15), col = "red", cex = 3)} to make
the MTs one, eight and 18 appear big and red.}

\item{phase.names}{By default phases are labeled based on the levels of the
phase variable.
Use this argument to specify different labels: \code{phase.names = c("Baseline",
"Intervention")}.}

\item{xlab}{The label of the x-axis. Default is \code{xlab = "Measurement
time"}.}

\item{ylab}{The labels of the y-axis. Default is \code{ylab = "Score"}.}

\item{main}{Main title of the plot.}

\item{case.names}{Case names. If not provided, names are taken from the scdf.
Set \code{case.names = ""} if you don't like to include case names.}

\item{style}{Either a character with the name of a pre-implemented style or a
style object. See \code{\link{style_plot}} to learn about this format.}
}
\value{
Returns a plot of one or multiple single-cases.
}
\description{
This function provides a plot of a single-case or multiple single-cases.
}
\examples{

## Request the default plot of the data from Borckhardt (2014)
plot(Borckardt2014)

## Plot the three cases from Grosche (2011) and visualize the phase A trend
plot(Grosche2011, style = "grid", lines = "trendA")

## Request the local regression line for Georg from that data set and customize the plot
plot(Grosche2011$Georg, style = "sienna", ylim = c(0,NA),
       xlab = "Training session", ylab = "Words per minute",
       phase.names = c("Baseline", "Intervention"), xinc = 5,
       lines = list(type = "loreg", f = 0.2, lty = "solid", col = "black", lwd = 3))

## Plot a random MBD over three cases and mark interesting MTs
dat <- random_scdf(design = design(3))
plot(dat, marks = list(positions = list(c(2,4,5),c(1,2,3),c(7,8,9)), col = "blue",
       cex = 1.4), style = c("grid", "annotate", "tiny"))

}
\author{
Juergen Wilbert
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pand.R, R/print.sc_pand.R
\name{pand}
\alias{pand}
\alias{print.sc_pand}
\alias{export.sc_pand}
\title{Percentage of all non-overlapping data}
\usage{
pand(
  data,
  dvar,
  pvar,
  decreasing = FALSE,
  phases = c(1, 2),
  method = c("sort", "minimum")
)

\method{print}{sc_pand}(x, ...)

\method{export}{sc_pand}(object, caption = NA, footnote = NA, filename = NA, round = 1, ...)
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{decreasing}{If you expect data to be lower in the B phase, set
\code{decreasing = TRUE}. Default is \code{decreasing = FALSE}.}

\item{phases}{A vector of two characters or numbers indicating the two phases
that should be compared. E.g., \code{phases = c("A","C")} or \code{phases = c(2,4)}
for comparing the second to the fourth phase. Phases could be combined by
providing a list with two elements. E.g., \code{phases = list(A = c(1,3), B = c(2,4))} will compare phases 1 and 3 (as A) against 2 and 4 (as B). Default
is \code{phases = c(1,2)}.}

\item{method}{Either \code{"sort"}" or \code{"minimum"}. See details.}

\item{x}{An object returned by \code{\link[=pand]{pand()}}}

\item{...}{Further arguments passed to the function.}

\item{object}{An scdf or an object exported from a scan function.}

\item{caption}{Character string with table caption. If left NA (default) a
caption will be created based on the exported object.}

\item{footnote}{Character string with table footnote. If left NA (default) a
footnote will be created based on the exported object.}

\item{filename}{String containing the file name. If a filename is given the
output will be written to that file.}

\item{round}{Integer passed to the digits argument used to round values.}
}
\value{
\tabular{ll}{
    \tab  \cr
   \code{pand} \tab Percentage of all non-overlapping data. \cr
   \code{method} \tab Calculation method. \cr
   \code{phi} \tab Effect size Phi based on expected and observed values. \cr
   \code{perc_overlap} \tab Percentage of overlapping data points. \cr
   \code{overlaps} \tab Number of overlapping data points. \cr
   \code{n} \tab Number of data points. \cr
   \code{N} \tab Number of cases. \cr
   \code{n_a} \tab Number of data points in phase A. \cr
   \code{n_b} \tab Number of data points in phase B. \cr
   \code{matrix} \tab 2x2 frequency matrix of phase A and B comparisons. \cr
   \code{matrix_counts} \tab 2x2 counts matrix of phase A and B comparisons. \cr
   \code{chi_test} \tab A Chi-squared analysis of expected and observed data (chisq.test()). \cr
   \code{fisher_test} \tab A Fisher exact test analysis of expected and observed data (fisher.test()). \cr
}
}
\description{
The \code{pand()} function calculates the percentage of all non-overlapping data
(PAND; Parker, Hagan-Burke, & Vannest, 2007), an index to quantify a level
increase (or decrease) in performance after the onset of an intervention.
}
\details{
PAND was proposed by Parker, Hagan-Burke, and Vannest in 2007. The authors
emphasize that PAND is designed for application in a multiple case design with
a substantial number of measurements, technically at least 20 to 25, but
preferably 60 or more. PAND is defined as 100\% minus the percentage of data
points that need to be removed from either phase in order to ensure nonoverlap
between the phases. Several approaches have been suggested to calculate PAND,
leading to potentially different outcomes. In their 2007 paper, Parker and
colleagues present an algorithm for computing PAND. The algorithm involves
sorting the scores of a time series, including the associated phases, and
comparing the resulting phase order with the original phase order using a
contingency table. To account for ties, the algorithm includes a randomization
process where ties are randomly assigned to one of the two phases.
Consequently, executing the algorithm multiple times could yield different
results. It is important to note that this algorithm does not produce the same
results as the PAND definition provided earlier in the same paper. However, it
offers the advantage of allowing the calculation of an effect size measure
\code{phi}, and the application of statistical tests for frequency distributions.
Pustejovsky (2019) presented a mathematical formulation of Parker's original
definition for comparing two phases of a single case: \deqn{PAND =
\frac{1}{m+n}max\{(i+j)I(y^A_{i}<y^B_{n+1-j}\}} This formulation provides
accurate results for PAND, but the original definition has the drawback of an
unknown distribution under the null hypothesis, making a statistical test
difficult. The \code{pand()} function enables the calculation of PAND using both
methods. The first approach (\code{method = "sort"}) follows the algorithm
described above, with the exclusion of randomization before sorting to avoid
ambiguity. It calculates a phi measure and provides the results of a
chi-squared test and a Fisher exact test. The second approach (\code{method = "minimum"}) applies the aforementioned formula. The code of this function is
based on the code of the \code{SingleCaseES} package (function \code{calc_PAND}). For a
multiple case design, overlaps are calculated for each case, summed, and then
divided by the total number of measurements. No statistical test is conducted
for this method.
}
\section{Functions}{
\itemize{
\item \code{print(sc_pand)}: Print results

\item \code{export(sc_pand)}: Export results as html table (see \code{\link[=export]{export()}})

}}
\examples{
## REplication of the Parker et al. 2007 example
pand(Parker2007)

## Calculate the PAND with an expected decrease of phase B scores
cubs <- scdf(c(20,22,24,17,21,13,10,9,20,9,18), B_start = 5)
pand(cubs, decreasing = TRUE)

}
\references{
Parker, R. I., Hagan-Burke, S., & Vannest, K. (2007). Percentage
of All Non-Overlapping Data (PAND): An Alternative to PND. \emph{The Journal of
Special Education, 40}, 194-204.

Parker, R. I., & Vannest, K. (2009). An Improved Effect Size for Single-Case
Research: Nonoverlap of All Pairs. \emph{Behavior Therapy, 40}, 357-367.

Pustejovsky, J. E. (2019). Procedural sensitivities of effect sizes for
single-case designs with directly observed behavioral outcome measures.
\emph{Psychological Methods}, \emph{24(2)}, 217-235.
https://doi.org/10.1037/met0000179

Pustejovsky JE, Chen M, Swan DM (2023). SingleCaseES: A Calculator for
Single-Case Effect Sizes. R package version 0.7.1.9999,
https://jepusto.github.io/SingleCaseES/.
}
\seealso{
Other overlap functions: 
\code{\link{ird}()},
\code{\link{nap}()},
\code{\link{overlap}()},
\code{\link{pem}()},
\code{\link{pet}()},
\code{\link{pnd}()},
\code{\link{print.sc_cdc}()},
\code{\link{select_phases}()},
\code{\link{tau_u}()}
}
\author{
Juergen Wilbert
}
\concept{overlap functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rci.R
\name{rci}
\alias{rci}
\title{Reliable change index}
\usage{
rci(data, dvar, pvar, rel, ci = 0.95, graph = FALSE, phases = c(1, 2))
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{rel}{Reliability of the measure, used to compute the standard error.}

\item{ci}{Width of confidence interval as a decimal. Default is \code{ci = 0.95}
applying a 95 percent confidence interval.}

\item{graph}{If set \code{TRUE}, a box plot of phase A and B scores is displayed.
\code{graph = FALSE} by default.}

\item{phases}{A vector of two characters or numbers indicating the two phases
that should be compared. E.g., \code{phases = c("A","C")} or \code{phases = c(2,4)}
for comparing the second to the fourth phase. Phases could be combined by
providing a list with two elements. E.g., \code{phases = list(A = c(1,3), B = c(2,4))} will compare phases 1 and 3 (as A) against 2 and 4 (as B). Default
is \code{phases = c(1,2)}.}
}
\description{
The \code{rci()} function computes indices of reliable change (Wise, 2004)
and corresponding descriptive statistics.
}
\details{
The reliable change index (RCI) indicates whether the change in scores from
phase A to phase B is statistically reliable, that is, whether the change is
larger than would be expected due to measurement error alone. The RCI can be
calculated using different methods (Jacobsen & Truax, 1991; Christensen &
Mendoza, 1986). The most common method is that of Jacobsen and Truax (1991)
which uses the standard error of measurement in phase A to
calculate the RCI. Christensen and Mendoza (1986) proposed an alteration of
the RCI which uses the standard error of the difference between phase A and
phase B means. Both methods are implemented in the \code{rci()} function.
}
\examples{

## Report the RCIs of the first case from the byHeart data and include a graph
rci(byHeart2011[1], graph = TRUE, rel = 0.8)

}
\references{
Christensen, L., & Mendoza, J. L. (1986). A method of assessing
change in a single subject: An alteration of the RC index. \emph{Behavior
Therapy, 17}, 305-308.

Jacobson, N. S., & Truax, P. (1991). Clinical Significance: A statistical
approach to defining meaningful change in psychotherapy research.
\emph{Journal of Consulting and Clinical Psychology, 59}, 12-19.

Wise, E. A. (2004). Methods for analyzing psychotherapy outcomes: A review
of clinical significance, reliable change, and recommendations for future
directions. \emph{Journal of Personality Assessment, 82}, 50 - 59.
}
\author{
Juergen Wilbert
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tau_u.R, R/print.sc_tauu.R, R/export.sc_tauu.R
\name{tau_u}
\alias{tau_u}
\alias{print.sc_tauu}
\alias{export.sc_tauu}
\title{Tau-U for single-case data}
\usage{
tau_u(
  data,
  dvar,
  pvar,
  method = c("complete", "parker", "tarlow"),
  phases = c(1, 2),
  meta_analyses = TRUE,
  ci = 0.95,
  ci_method = c("z", "tau", "s"),
  meta_weight_method = c("z", "tau"),
  tau_method = c("b", "a"),
  continuity_correction = FALSE
)

\method{print}{sc_tauu}(
  x,
  complete = FALSE,
  digits = "auto",
  select = c("Tau", "CI lower", "CI upper", "SD_S", "Z", "p"),
  nice_p = TRUE,
  ...
)

\method{export}{sc_tauu}(
  object,
  caption = NA,
  footnote = NA,
  filename = NA,
  select = "auto",
  meta = FALSE,
  round = 3,
  decimals = 3,
  ...
)
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{method}{\code{"complete"} (default), \code{"parker"} or \code{"tarlow"}. The
\code{"parker"} calculates the number of possible pairs as described in Parker
et al. (2011) which might lead to tau-U values greater than 1. \code{"tarlow"}
follows an online calculator and R code developed by Tarlow (2017).}

\item{phases}{A vector of two characters or numbers indicating the two phases
that should be compared. E.g., \code{phases = c("A","C")} or \code{phases = c(2,4)}
for comparing the second to the fourth phase. Phases could be combined by
providing a list with two elements. E.g., \code{phases = list(A = c(1,3), B = c(2,4))} will compare phases 1 and 3 (as A) against 2 and 4 (as B). Default
is \code{phases = c(1,2)}.}

\item{meta_analyses}{If TRUE, a meta analysis is conducted.}

\item{ci}{Confidence intervals level. If NULL or NA, no confidence intervals
are calculated.}

\item{ci_method}{String to specify the method for calculating the standard
error of tau. Either "tau", "z", or "s" (not recommended).}

\item{meta_weight_method}{String to specify the method for calculating the
weights of the studies. Either "tau" or "z".}

\item{tau_method}{Character with values "a" or "b" (default) indicating
whether Kendall Tau A or Kendall Tau B is applied. Ignored for methods
'tarlow' and 'parker'.}

\item{continuity_correction}{If TRUE, a continuity correction is applied for
calculating p-values of correlations (here: S will be reduced by one before
calculating Z). Ignored for methods 'tarlow' and 'parker'.}

\item{x}{Object returned from \code{\link[=tau_u]{tau_u()}}.}

\item{complete}{Print all parameters.}

\item{digits}{The minimum number of significant digits to be use.
If set to "auto" (default), values are predefined.}

\item{select}{Character vector with name of variables to be included. When
the vector is named, variables are renamed appropriately.}

\item{nice_p}{If TRUE, p-values are printed in publication friendly form.}

\item{...}{Further arguments passed to the function.}

\item{object}{An scdf or an object exported from a scan function.}

\item{caption}{Character string with table caption. If left NA (default) a
caption will be created based on the exported object.}

\item{footnote}{Character string with table footnote. If left NA (default) a
footnote will be created based on the exported object.}

\item{filename}{String containing the file name. If a filename is given the
output will be written to that file.}

\item{meta}{If TRUE, the results of the meta analysis will be exported. If
FALSE, each single-case is exported.}

\item{round}{Integer passed to the digits argument used to round values.}

\item{decimals}{Decimal places that are reported.}
}
\value{
\item{table}{A data frame containing statistics from the Tau-U
family, including: Pairs, positive and negative comparisons, S, and Tau}
\item{matrix}{The matrix of comparisons used for calculating the
statistics.} \item{tau_u}{Tau-U value.}
}
\description{
This function calculates indices of the Tau-U family as proposed by Parker et
al. (2011a). It allows to calculate Tau-U values for single cases as well
as overall Tau-U values across several single cases by applying a meta
analysis.
}
\details{
Tau-U is an inconsistently operationalized construct. Parker et al.
(2011b) describe a method which may result in Tau-U outside the [-1;1]
interval. A different implementation of the method (provided at
http://www.singlecaseresearch.org/calculators/tau-u) uses tau-b (instead of
tau-a as in the original formulation by Parker). Bossart et. al (2018)
describe inconsistencies in the results from this implementation as well.
Another problems lies in the calculation in overall Tau-U values from
several single cases. The function presented here applies a meta-analysis
to gain the overall values. Each tau value is weighted by the inverse of
the variance (ie. the tau standard error). The confidence intervals for
single cases are calculated by Fisher-Z transforming tau, calculating the
confidence intervals, and inverse transform them back to tau (see Long &
Cliff, 1997).
}
\section{Functions}{
\itemize{
\item \code{print(sc_tauu)}: Print results

\item \code{export(sc_tauu)}: Export results as html table

}}
\examples{

tau_u(Grosche2011$Eva)

## Replicate  tau-U calculation from Parker et al. (2011)
bob <- scdf(c(A = 2, 3, 5, 3, B = 4, 5, 5, 7, 6), name = "Bob")
res <- tau_u(bob, method = "parker")
print(res, complete = TRUE)

## Request tau-U for all single-cases from the Grosche2011 data set
tau_u(Grosche2011)
}
\references{
Brossart, D. F., Laird, V. C., & Armstrong, T. W. (2018).
Interpreting Kendall’s Tau and Tau-U for single-case experimental designs.
\emph{Cogent Psychology, 5(1)}, 1–26.
https://doi.org/10.1080/23311908.2018.1518687.

Long, J. D., & Cliff, N. (1997). Confidence intervals for Kendall’s tau.
\emph{British Journal of Mathematical and Statistical Psychology}, 50(1),
31–41. https://doi.org/10.1111/j.2044-8317.1997.tb01100.x

Parker, R. I., Vannest, K. J., & Davis, J. L. (2011a). Effect Size in
Single-Case Research: A Review of Nine Nonoverlap Techniques.
\emph{Behavior Modification}, 35(4), 303–322. https://doi.org/10/dsdfs4

Parker, R. I., Vannest, K. J., Davis, J. L., & Sauber, S. B. (2011b).
Combining Nonoverlap and Trend for Single-Case Research: Tau-U.
\emph{Behavior Therapy, 42}(2), 284–299.
https://doi.org/10.1016/j.beth.2010.08.006

Tarlow, K. R. (2017, March). Tau-U for single-case research (R code).
Retrieved from http://ktarlow.com/stats/
}
\seealso{
Other overlap functions: 
\code{\link{ird}()},
\code{\link{nap}()},
\code{\link{overlap}()},
\code{\link{pand}()},
\code{\link{pem}()},
\code{\link{pet}()},
\code{\link{pnd}()},
\code{\link{print.sc_cdc}()},
\code{\link{select_phases}()}
}
\author{
Juergen Wilbert
}
\concept{overlap functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/message-system.R
\name{abort}
\alias{abort}
\title{Throws an error}
\usage{
abort(..., class = NULL, call = FALSE)
}
\arguments{
\item{...}{The message. This can be a character string or a
combination of character strings and variables.}

\item{class}{An optional class or vector of classes to add to the error.}

\item{call}{If TRUE, the call will be included in the error message. Default
is FALSE.}
}
\description{
This is a convenience wrapper for internal use.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plm_contrasts.R
\name{add_dummy_variables}
\alias{add_dummy_variables}
\title{Add Dummy Variables for Piecewise Linear Models}
\usage{
add_dummy_variables(
  scdf,
  model = c("W", "H-M", "B&L-B"),
  contrast_level = c("first", "preceding"),
  contrast_slope = c("first", "preceding")
)
}
\arguments{
\item{scdf}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{model}{Model used for calculating the dummy parameters (see Huitema &
McKean, 2000). Default is \code{model = "W"}. Possible values are: \code{"B&L-B"},
\code{"H-M"}, \code{"W"}, and deprecated \code{"JW"}.}

\item{contrast_level}{Either "first", "preceding" or a contrast matrix. If NA
contrast_level is a copy of contrast.}

\item{contrast_slope}{Either "first", "preceding" or a contrast matrix. If NA
contrast_level is a copy of contrast.}
}
\description{
Adds dummy variables to an scdf for calculating piecewise linear models.
}
\details{
This function creates dummy variables for phase levels and phase slopes
according to the specified piecewise regression model. It supports different
contrast coding schemes for both level and slope contrasts.
}
\examples{
add_dummy_variables(
 scdf = exampleABC, 
 model = "W", 
 contrast_level = "first", 
 contrast_slope = "first"
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/describe.R
\name{describe}
\alias{describe}
\title{Descriptive statistics for single-case data}
\usage{
describe(data, dvar, pvar, mvar)
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{mvar}{Character string with the name of the measurement time variable.
Defaults to the attributes in the scdf file.}
}
\value{
A list containing a data frame of descriptive statistics
(descriptives); the cse design (design); the number of cases (N).
}
\description{
The \code{describe()} function provides common descriptive statistics for
single-case data.
}
\details{
It computes the number of measurements, number of missing values, mean,
median, standard deviation, median average deviation, minimum, maximum, and
trend (slope of dependent variable regressed on measurement-time) for each
phase of each single-case included in an scdf.

n = number of measurements; mis = number of missing vaues; m = mean;
md = median; sd = standard deviation; mad = median average deviation; min =
minimum; max = maximum; trend = weight of depended variable regressed on
time (values ~ mt).
}
\examples{

## Descriptive statistics for a study of three single-cases
describe(Grosche2011)

## Descriptives of a three phase design
describe(exampleABC)

## Write descriptive statistics to .csv-file
study <- describe(Waddell2011)
write.csv(study$descriptives, file = tempfile())
}
\seealso{
\code{\link[=overlap]{overlap()}}, \code{\link[=plot.scdf]{plot.scdf()}}
}
\author{
Juergen Wilbert
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as.data.frame.scdf.R
\name{as.data.frame.scdf}
\alias{as.data.frame.scdf}
\title{Creating a long format data frame from several single-case data frames
(scdf).}
\usage{
\method{as.data.frame}{scdf}(x, ..., l2 = NULL, id = "case")
}
\arguments{
\item{x}{An scdf object.}

\item{...}{Not implemented.}

\item{l2}{A data frame providing additional variables at Level 2. The scdf
has to have names for all cases and the Level 2 data frame has to have a
column with corresponding case names.}

\item{id}{Variable name of the Level 2 data frame that contains the case
names. Defaults to "case".}
}
\value{
Returns one data frame with data of all single-cases structured by
the case variable.
}
\description{
The as.data.frame function transposes an scdf into one long data frame.
Additionally, a data frame can be merged that includes level 2 data of the
subjects. This might be helpful to prepare data to be used with other
packages than scan.
}
\examples{

## Convert the list of three single-case data frames from Grosche (2011)
### into one long data frame
Grosche2011
Grosche2011_long <- as.data.frame(Grosche2011)
Grosche2011_long

## Combine an scdf with data for l2
Leidig2018_long <- as.data.frame(Leidig2018, l2 = Leidig2018_l2)
names(Leidig2018_long)
summary(Leidig2018_long)

}
\seealso{
Other data manipulation functions: 
\code{\link{add_l2}()},
\code{\link{as_scdf}()},
\code{\link{batch_apply}()},
\code{\link{fill_missing}()},
\code{\link{moving_median}()},
\code{\link{print.sc_outlier}()},
\code{\link{ranks}()},
\code{\link{rescale}()},
\code{\link{scdf}()},
\code{\link{select_cases}()},
\code{\link{set_vars}()},
\code{\link{shift}()},
\code{\link{smooth_cases}()},
\code{\link{standardize}()},
\code{\link{truncate_phase}()}
}
\author{
Juergen Wilbert
}
\concept{data manipulation functions}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprec-standardize.R
\name{standardize}
\alias{standardize}
\title{Standardize values of an scdf file}
\usage{
standardize(
  data,
  var,
  center = TRUE,
  scale = FALSE,
  m = 0,
  sd = 1,
  grand = TRUE
)
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{var}{A character string or a vector of character strings with variable
names that should be scaled.}

\item{center}{If set TRUE, data are mean centered.}

\item{scale}{If set TRUE, the standard deviation is set.}

\item{m}{The target mean for centering.}

\item{sd}{The target standard deviation for scaling}

\item{grand}{If set TRUE, scaling is based on the mean and standard deviation
of all values across all single-cases within the scdf.}
}
\value{
An scdf with the scaled values.
}
\description{
\emph{This function is superseded by the much more versatile
\code{\link{transform.scdf}} function (see example below).}
This function scales the measured values of an scdf file. It allows for mean
centering and standardization based on each single-case data set or a
scaling across all cases included in an scdf.
}
\examples{

## Standardize a multiple case scdf and compute an hplm
exampleAB_50 |>
  standardize("values", center = TRUE, scale = TRUE) |>
  hplm()

## The more versatile transform function supersedes standardize:
exampleAB_50 |>
  transform(values = (values - mean(all(values))) / sd(all(values))) |>
  hplm()
}
\seealso{
Other data manipulation functions: 
\code{\link{add_l2}()},
\code{\link{as.data.frame.scdf}()},
\code{\link{as_scdf}()},
\code{\link{batch_apply}()},
\code{\link{fill_missing}()},
\code{\link{moving_median}()},
\code{\link{print.sc_outlier}()},
\code{\link{ranks}()},
\code{\link{rescale}()},
\code{\link{scdf}()},
\code{\link{select_cases}()},
\code{\link{set_vars}()},
\code{\link{shift}()},
\code{\link{smooth_cases}()},
\code{\link{truncate_phase}()}
}
\author{
Juergen Wilbert
}
\concept{data manipulation functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scdf-functions.R
\name{is.scdf}
\alias{is.scdf}
\title{Test for scdf objects}
\usage{
is.scdf(x)
}
\arguments{
\item{x}{An object to be tested.}
}
\value{
Returns TRUE or FALSE depending on whether its argument is of scdf type or not.
}
\description{
Tests for objects of type "scdf".
}
\author{
Juergen Wilbert
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_vars.R
\name{set_vars}
\alias{set_vars}
\alias{set_dvar}
\alias{set_mvar}
\alias{set_pvar}
\title{Set analysis variables in an scdf object}
\usage{
set_vars(data, dvar, mvar, pvar)

set_dvar(data, dvar)

set_mvar(data, mvar)

set_pvar(data, pvar)
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string. Name of the dependent variable.}

\item{mvar}{Character string. Name of the measurement-time variable.}

\item{pvar}{Character string. Name of the phase variable.}
}
\value{
An \code{scdf} object with updated variable settings.
}
\description{
This function allows to set or change the dependent variable,
measurement-time variable, and phase variable in an scdf object.
}
\examples{
exampleAB_add |>
  set_dvar("depression") |>
  describe()
}
\seealso{
Other data manipulation functions: 
\code{\link{add_l2}()},
\code{\link{as.data.frame.scdf}()},
\code{\link{as_scdf}()},
\code{\link{batch_apply}()},
\code{\link{fill_missing}()},
\code{\link{moving_median}()},
\code{\link{print.sc_outlier}()},
\code{\link{ranks}()},
\code{\link{rescale}()},
\code{\link{scdf}()},
\code{\link{select_cases}()},
\code{\link{shift}()},
\code{\link{smooth_cases}()},
\code{\link{standardize}()},
\code{\link{truncate_phase}()}
}
\author{
Juergen Wilbert
}
\concept{data manipulation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write_scdf.R
\name{write_scdf}
\alias{write_scdf}
\title{Data output: Write single-case data to a .csv-file}
\usage{
write_scdf(data, filename = NULL, sep = ",", dec = ".", ...)
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{filename}{A character string defining the output file name (e.g.
\code{"scdf_data.csv"}. If \code{NULL} (default), the data are written to the
console.}

\item{sep}{The field separator string. Values within each row of x are
separated by this string.}

\item{dec}{The string to use for decimal points in numeric or complex
columns: must be a single character.}

\item{\dots}{Further arguments passed to write.table.}
}
\value{
Invisibly returns \code{NULL}.
}
\description{
This function restructures and writes single-case data into a .csv-file.
}
\details{
This is a wrapper for the write.table function with predefined parameters.
It converts the single-case data file into a data frame and writes it to a
.csv-file. Each single-case data set is stacked below each other and an
additional column "case" is added to identify the different single-cases.
}
\examples{
## write single-case data to a .csv-file
filename <- tempfile(fileext = ".csv")
write_scdf(exampleAB, filename)

## write multiple cases to a .csv-file with semicolon as field and comma as
## decimal separator
write_scdf(Grosche2011, filename, sep = ";", dec = ",")

## read_scdf and write_scdf
write_scdf(exampleA1B1A2B2_zvt, filename)
dat <- read_scdf(filename, cvar = "case", pvar = "part",
                 dvar = "zvt", mvar = "day")
res1 <- describe(exampleA1B1A2B2_zvt)$descriptives
res2 <- describe(dat)$descriptives
all.equal(res1,res2)
}
\seealso{
\code{\link[=write.table]{write.table()}}, \code{\link[=saveRDS]{saveRDS()}}

Other io-functions: 
\code{\link{convert}()},
\code{\link{read_scdf}()}
}
\author{
Juergen Wilbert
}
\concept{io-functions}
\keyword{io}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_scdf.R
\name{read_scdf}
\alias{read_scdf}
\title{Load single-case data from files}
\usage{
read_scdf(
  file,
  cvar = "case",
  pvar = "phase",
  dvar = "values",
  mvar = "mt",
  sort_cases = FALSE,
  phase_names = NULL,
  type = NA,
  na = c("", "NA"),
  sort.labels = NULL,
  phase.names = NULL,
  ...
)
}
\arguments{
\item{file}{Either a character string defining the file to be loaded (e.g.
\code{"SC_Anita.csv"} (if left empty a dialog box for choosing will be opened)
or a data.frame.}

\item{cvar}{Sets the variable name of the "case" variable. Defaults to
\code{"case"}.}

\item{pvar}{Sets the variable name of the "phase" variable. Defaults to
\code{"phase"}.}

\item{dvar}{Sets the variable name of the "values" variable. Defaults to
\code{"values"}.}

\item{mvar}{Sets the variable name of the "mt" variable. Defaults to \code{"mt"}.}

\item{sort_cases, sort.labels}{If set TRUE, the resulting list is sorted by
label names (alphabetically increasing).}

\item{phase_names, phase.names}{A character vector with phase names. Defaults
to the phase names provided in the phase variable.}

\item{type}{Format of the file to be loaded. Either "csv", "xlsx", "xls",
"excel", "yml" is possible. By default (NA) the type is extracted from the
file extension.}

\item{na}{Character vector of strings to interpret as missing values.}

\item{\dots}{Further arguments passed to the respective read function.}
}
\value{
Returns a single-case data frame. See \code{\link{scdf}} to learn
about the format of these data frames.
}
\description{
Use the \code{read_scdf} function to load single-case data csv, excel, or yaml
files.
}
\details{
The data file must be in a "long" format with one column for case
identifiers, one column for phase identifiers, one column for measurement
time, and one column for the dependent variable values. Each row represents
a single measurement.
}
\examples{

## Read SC-data from a file named "study1.csv" in your working directory
# study1 <- read_scdf("study1.csv")

## Read SC-data from a .csv-file with semicolon as field and comma as decimal separator
# study2 <- read_scdf("study2.csv", sep = ";", dec = ",")

## write_scdf and read_scdf
filename <- file.path(tempdir(), "test.csv")
write_scdf(exampleA1B1A2B2_zvt, filename)
dat <- read_scdf(filename, cvar = "case", pvar = "part", dvar = "zvt", mvar = "day")
res1 <- describe(exampleA1B1A2B2_zvt)$descriptives
res2 <- describe(dat)$descriptives
all.equal(res1,res2)

}
\seealso{
\code{\link[=read.table]{read.table()}}, \code{\link[=readRDS]{readRDS()}}

Other io-functions: 
\code{\link{convert}()},
\code{\link{write_scdf}()}
}
\author{
Juergen Wilbert
}
\concept{io-functions}
\keyword{io}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/style_plot.R
\name{style_plot}
\alias{style_plot}
\title{(Deprecated) Create styles for single-case data plots}
\usage{
style_plot(style = "default", ...)
}
\arguments{
\item{style}{A character string or a vector of character strings with
predefined styles.}

\item{...}{Further arguments passed to the plot command.}
}
\value{
Returns a list to be provided for the style argument of the
\code{\link[=plot.scdf]{plot.scdf()}} function.

\itemize{ \item\code{fill} If set, the area under the line is filled
with the given color (e.g., \code{fill = "tomato"}). Use the standard R
command colors() to get a list of all possible colours. \code{fill} is
empty by default. \item\code{annotations} A list of parameters defining
annotations to each data point. This adds the score of each MT to your
plot.
\itemize{
\item\code{"pos"} Position of the annotations: 1 = below, 2 =
left, 3 = above, 4 = right.
\item\code{"col"} Color of the annotations.
\item\code{"cex"} Size of the annotations.
\item\code{"round"} Rounds the values to the specified decimal.}
\item\code{annotations = list(pos = 3, col =
"brown", round = 1)} adds scores rounded to one decimal above the data point
in brown color to the plot. \item\code{"names"} A list of parameters
defining the depiction of phase names (e.g. \code{names = list(cex = 0.8,
  col = "red", side = 1)}: cex for size, col for color, and side for
position). See \code{\link{mtext}} for more details. \item\code{"lwd"}
Width of the plot line. Default is \code{lwd = 2}. \item\code{"pch"} Point
type. Default is \code{pch = 17} (triangles). Other options are for
example: 16 (filled circles) or "A" (uses the letter A). \item\code{"main"}
Main title of the plot. \item\code{"mai"} Sets the margins of the plot.
\item\code{"bty"} Shape of the frame surrounding the inner plot
\item\code{"fill.bg"} Background color of the plot. If a vector is
provided, these colors will be assigned to phases (each phase name becomes
a color). \item\code{"grid"} Color of a grid. \item\code{"text.ABlag"} Text
displayed between phases. \item\code{"cex.axis"} Size of the axis
annotations \item\code{"las"} Orientation of the axis annotations
\item\code{"col.lines"} Color of the lines \item\code{"col.dots"} Color of
the dots \item\code{"col.seperator"} Color of the phase seperating lines
\item\code{"col.bg"} Color of the outer plot \item\code{"col"} General
color setting for the plot \item\code{"col.text"} Color of all labels of
the plot. }
}
\description{
The \code{style_plot} function is used to create graphical styles for a
single-case plot
}
\details{
\code{style_plot("")} will return a list of predefined styles.
Predefined styles can be combined \code{style_plot(style = c("grid2",
  "tiny"))} where settings of a latter style overwrite settings of the
former. Additional style paramters are set following the style argument and
can be combined with those: \code{style_plot(style = "grid2", fill =
  "grey50", pch = 18)}.
}
\examples{
newstyle <- style_plot(style = "default")
newstyle$text.ABlag <- c("START", "END")
newstyle$col.dots <- ""
newstyle$annotations <- list(cex = 0.6, col = "grey10", offset = 0.4)
newstyle$names <- list(cex = 0.8, col = "blue", side = 1, adj = 1, line = -1, at = 31)
newstyle$fill.bg <- c("grey99", "grey95", "grey90")
plot(exampleABC, style = newstyle, main = "Example Plot")

}
\seealso{
\code{\link[=plot.scdf]{plot.scdf()}}
}
\author{
Juergen Wilbert
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scdf-functions.R
\name{na.omit.scdf}
\alias{na.omit.scdf}
\title{Remove missing values from scdf}
\usage{
\method{na.omit}{scdf}(object, ...)
}
\arguments{
\item{object}{A scdf.}

\item{...}{not implemented yet.}
}
\value{
A scdf object.
}
\description{
Removes any row with a missing value from each single-case data frame in an scdf.
}
\author{
Juergen Wilbert
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coefficents.sc_plm.R
\name{coef.sc_plm}
\alias{coef.sc_plm}
\title{Extract coefficients from plm/hplm objects}
\usage{
\method{coef}{sc_plm}(object, ...)
}
\arguments{
\item{object}{plm or hplm object}

\item{...}{not implemented}
}
\value{
data frame with coefficient table
}
\description{
Extract coefficients from plm/hplm objects
}
\examples{
coefficients(plm(exampleAB$Johanna))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nap.R
\name{nap}
\alias{nap}
\title{Nonoverlap of all Pairs (NAP)}
\usage{
nap(data, dvar, pvar, decreasing = FALSE, phases = c(1, 2))
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{decreasing}{If you expect data to be lower in the B phase, set
\code{decreasing = TRUE}. Default is \code{decreasing = FALSE}.}

\item{phases}{A vector of two characters or numbers indicating the two phases
that should be compared. E.g., \code{phases = c("A","C")} or \code{phases = c(2,4)}
for comparing the second to the fourth phase. Phases could be combined by
providing a list with two elements. E.g., \code{phases = list(A = c(1,3), B = c(2,4))} will compare phases 1 and 3 (as A) against 2 and 4 (as B). Default
is \code{phases = c(1,2)}.}
}
\value{
\tabular{ll}{
    \tab  \cr
   \code{nap} \tab A data frame with NAP and additional values for each case. \cr
   \code{N} \tab Number of cases. \cr
}
}
\description{
The \code{\link[=nap]{nap()}} function calculates the nonoverlap of all pairs (NAP; Parker &
Vannest, 2009).  NAP summarizes the overlap between all pairs of phase A and
phase B data points.  If an increase of phase B scores is expected, a
non-overlapping pair has a higher phase B data point. The NAP equals
\emph{number of pairs showing no overlap / number of pairs} where ties are
counted as half non-overlaps.  Because NAP can take values between 0 and 100
percent where values below 50 percent indicate an inverse effect, an nap
rescaled from -100 to 100 percent where negative
values indicate an inverse effect is also displayed (\eqn{nap_{rescaled} = 2
* nap - 100}).
}
\details{
If a decrease of phase B scores is expected, set the argument
\code{decreasing = TRUE}.
}
\examples{

## Calculate NAP for a study with  lower expected phase B scores
## (e.g. aggressive behavior)
gretchen <- scdf(c(A = 12, 14, 9, 10, B = 10, 6, 4, 5, 3, 4))
nap(gretchen, decreasing = TRUE)

## Request NAP for all cases from the Grosche2011 scdf
nap(Grosche2011)

## Calculate NAP for phase 1 and phase 3 of an ABAB design
nap(exampleABAB, phases = c(1, 3))


}
\references{
Parker, R. I., & Vannest, K. (2009). An improved effect size for
single-case research: Nonoverlap of all pairs. \emph{Behavior Therapy}, \emph{40},
357-367.
}
\seealso{
Other overlap functions: 
\code{\link{ird}()},
\code{\link{overlap}()},
\code{\link{pand}()},
\code{\link{pem}()},
\code{\link{pet}()},
\code{\link{pnd}()},
\code{\link{print.sc_cdc}()},
\code{\link{select_phases}()},
\code{\link{tau_u}()}
}
\author{
Juergen Wilbert
}
\concept{overlap functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_rand.R
\name{plot_rand}
\alias{plot_rand}
\title{Plot random distribution}
\usage{
plot_rand(
  object,
  type = "hist",
  xlab = NULL,
  ylab = NULL,
  title = NULL,
  text_observed = "observed",
  color = "lightgrey",
  ...
)
}
\arguments{
\item{object}{Object returned from the rand_test() function}

\item{type}{'hist' or 'xy'.}

\item{xlab}{Label for the x-axis.}

\item{ylab}{Label for the y-axis.}

\item{title}{Plot title.}

\item{text_observed}{Text for marking the number of observed statistic.}

\item{color}{Bar color.}

\item{...}{Further arguments passed to the plot function.}
}
\description{
This function takes the return of the rand_test function and creates a
histogram with the distribution of the rand sample statistics.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trend.R
\name{trend}
\alias{trend}
\title{Trend analysis for single-cases data}
\usage{
trend(
  data,
  dvar,
  pvar,
  mvar,
  offset = "deprecated",
  first_mt = 0,
  model = NULL
)
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{mvar}{Character string with the name of the measurement time variable.
Defaults to the attributes in the scdf file.}

\item{offset}{(Deprecated. Please use first_mt). An offset for the first
measurement-time of each phase. If \code{offset = 0}, the phase measurement is
handled as MT 1. Default is \code{offset = -1}, setting the first value of MT to
0.}

\item{first_mt}{A numeric setting the value for the first measurement-time.
Default = 0.}

\item{model}{A string or a list of (named) strings each depicting one
regression model. This is a formula expression of the standard R class. The
parameters of the model are \code{values}, \code{mt} and \code{phase}.}
}
\value{
A list of class \code{sc_trend} containing:
\item{trend}{A matrix containing the results (Intercept, B and beta)
of separate regression models for phase A, phase B, and the whole data.}
\item{first_mt}{Numeric argument from function call (see arguments
section).}
}
\description{
The \code{trend()} function provides an overview of linear trends in single case
data. By default, it provides the intercept and slope of a linear and
quadratic regression of measurement time on scores. Models are calculated
separately for each phase and across all phases. For more advanced use, you
can add regression models using the R-specific formula class.
}
\details{
The function computes separate regression models for each phase and
for the whole data. By default two models are computed: a linear model and
a quadratic model. Additionally, custom models can be specified using the
\code{model} argument. The measurement time variable is adjusted such that the
first measurement time point of each phase is set to the value specified in
the \code{first_mt} argument (default = 0). This means that if \code{first_mt = 0},
the first measurement time point of each phase is set to 0, if \code{first_mt = 1}, the first measurement time point of each phase is set to 1, and so on.
This adjustment allows for a more intuitive interpretation of the
regression coefficients, especially the intercept, which then represents
the estimated value at the beginning of each phase.
}
\examples{

## Compute the linear and squared regression for a random single-case
design <- design(slope = 0.5)
matthea <- random_scdf(design)
trend(matthea)

## Besides the linear and squared regression models compute two custom models:
## a) a cubic model, and
## b) the values predicted by the natural logarithm of the
## measurement time.
design <- design(slope = 0.3)
ben <- random_scdf(design)
trend(
  ben,
  model = list("Cubic" = values ~ mt^3, "Log Time" = values ~ log(mt)),
  first_mt = 1 # must be set to 1 because log(0) would be -Inf
)

}
\seealso{
\code{\link[=describe]{describe()}}

Other regression functions: 
\code{\link{bplm}()},
\code{\link{fetch}()},
\code{\link{hplm}()},
\code{\link{mplm}()},
\code{\link{plm}()},
\code{\link{print.sc_ac}()},
\code{\link{print.sc_bctau}()}
}
\author{
Juergen Wilbert
}
\concept{regression functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprec-ranks.R
\name{ranks}
\alias{ranks}
\title{(Deprecated) Rank-transformation of single-case data files}
\usage{
ranks(data, var, grand = TRUE, ...)
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{var}{A string or string vector with the names of the variables to be
ranked.}

\item{grand}{If TRUE, ranks will be calculated across all cases. If FALSE
ranks are calculated within each case.}

\item{...}{Additional parameters passed to the \code{\link{rank}} function.}
}
\value{
An \code{scdf} object where the values of the variable(s) are
replaced with ranks.
}
\description{
\emph{This function is superseded by the more versatile
\code{\link{transform.scdf}} function.}
}
\examples{
# The ranks function is deprecated. Please use transform:
res1 <- ranks(Huber2014, var = "compliance")
res2 <- transform(Huber2014, across_cases(compliance = rank(compliance, na.last="keep")))
identical(res1, res2)

res1 <- ranks(Huber2014, var = "compliance", grand = FALSE)
res2 <- transform(Huber2014, compliance = rank(compliance, na.last="keep"))
identical(res1, res2)
}
\seealso{
Other data manipulation functions: 
\code{\link{add_l2}()},
\code{\link{as.data.frame.scdf}()},
\code{\link{as_scdf}()},
\code{\link{batch_apply}()},
\code{\link{fill_missing}()},
\code{\link{moving_median}()},
\code{\link{print.sc_outlier}()},
\code{\link{rescale}()},
\code{\link{scdf}()},
\code{\link{select_cases}()},
\code{\link{set_vars}()},
\code{\link{shift}()},
\code{\link{smooth_cases}()},
\code{\link{standardize}()},
\code{\link{truncate_phase}()}
}
\author{
Juergen Wilbert
}
\concept{data manipulation functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smd.R
\name{smd}
\alias{smd}
\title{Standardized mean differences for single-case data}
\usage{
smd(data, dvar, pvar, mvar, phases = c(1, 2))
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{mvar}{Character string with the name of the measurement time variable.
Defaults to the attributes in the scdf file.}

\item{phases}{A vector of two characters or numbers indicating the two phases
that should be compared. E.g., \code{phases = c("A","C")} or \code{phases = c(2,4)}
for comparing the second to the fourth phase. Phases could be combined by
providing a list with two elements. E.g., \code{phases = list(A = c(1,3), B = c(2,4))} will compare phases 1 and 3 (as A) against 2 and 4 (as B). Default
is \code{phases = c(1,2)}.}
}
\description{
The \code{smd()} function provides various standardized mean effect sizes for
single-case data.
}
\details{
It computes 'Cohen's d', 'Hedges' g', 'Hedges' g correction',
'Hedges' g durlak correction', and 'Glass' delta' for each single-case
included in an scdf.

'sd cohen' is the (unweigted) average of the variance of phase A and
B. 'sd Hedges' is the weighted average of the variance of phase A and B
(with a degrees of freedom correction). 'Hedges' g' is the mean difference
divided by 'sd Hedges'. 'Hedges' g correction' and 'Hedges' g durlak
correction' are two approaches of correcting Hedges' g for small sample
sizes. 'Glass' delta' is the mean difference divided by the standard
deviation of the A-phase. 'Cohens d' is the mean difference divided by 'sd
cohen'.
}
\examples{
smd(exampleAB)
}
\seealso{
\code{\link[=overlap]{overlap()}}, \code{\link[=describe]{describe()}}
}
\author{
Juergen Wilbert
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scan-Data.R
\docType{data}
\name{example_data_sets}
\alias{example_data_sets}
\alias{Beretvas2008}
\alias{Borckardt2014}
\alias{Huitema2000}
\alias{Waddell2011}
\alias{byHeart2011}
\alias{Grosche2011}
\alias{Grosche2014}
\alias{GruenkeWilbert2014}
\alias{Huber2014}
\alias{example_A24}
\alias{exampleAB_simple}
\alias{exampleAB}
\alias{exampleAB_add}
\alias{exampleAB_decreasing}
\alias{exampleAB_score}
\alias{exampleABC}
\alias{exampleABC_outlier}
\alias{exampleABC_50}
\alias{exampleABAB}
\alias{exampleA1B1A2B2}
\alias{exampleA1B1A2B2_zvt}
\alias{exampleAB_50}
\alias{exampleAB_50.l2}
\alias{exampleABC_150}
\alias{exampleAB_mpd}
\alias{example_stranger}
\alias{example_atd}
\alias{Lenz2013}
\alias{Leidig2018}
\alias{Leidig2018_l2}
\alias{SSDforR2017}
\alias{Parker2011}
\alias{Parker2007}
\alias{Parker2009}
\alias{Parker2009b}
\alias{Parker2011b}
\alias{Tarlow2017}
\title{Single-case example data sets}
\usage{
data(exampleAB, package = "scan")
}
\description{
The \link{scan} package comes with a set of fictitious and authentic
single-case study data. These data sets can be used to practice single-case
data analysis and to reproduce results from the respective publications.
\itemize{
\item \strong{Beretvas2008} — Fictitious single-case intervention study from Beretvas & Chung, 2008.
\item \strong{Borckardt2014} — Fictitious daily pain ratings evaluating a psychological treatment from Borckardt & Nash, 2014.
\item \strong{byHeart2011} — Multiple-baseline (11 cases) flash card vocabulary learning (Wilbert, unpublished).
\item \strong{example_A24} — Number of injuries on a German autobahn before and after implementation of a speedlimit (130km/h) (Ministerium fuer Infrastruktur und Landesplanung. Land Brandenburg).
\item \strong{exampleA1B1A2B2} — Fictitious A1-B1-A2-B2 example dataset.
\item \strong{exampleA1B1A2B2_zvt} — Non-fictitious A1-B1-A2-B2 example with ZVT (intelligence measure) and D2 (concentration measure) scores.
\item \strong{exampleAB} — Fictitious AB example dataset with three cases.
\item \strong{exampleAB_50} — Fictitious AB example dataset (50 cases)s.
\item \strong{exampleAB_50.l2} — Level-2 data for \strong{exampleAB_50}.
\item \strong{exampleAB_add} — Fictitious AB example with added covariates.
\item \strong{exampleAB_decreasing} — Fictitious AB example with an expected decreasing effect in the intervention phase.
\item \strong{exampleAB_mpd} — Fictitious example with different phase structures for each case.
\item \strong{exampleAB_score} — Fictitious AB example with binomial distributed score outcome.
\item \strong{exampleAB_simple} — Simple fictitious AB example with three cases.
\item \strong{exampleABAB} — Fictitious ABAB reversal design example.
\item \strong{exampleABC} — Fictitious ABC example dataset.
\item \strong{exampleABC_150} — Fictitious ABC example (150 cases).
\item \strong{exampleABC_50} — Fictitious ABC example (50 cases).
\item \strong{exampleABC_outlier} — Fictitious ABC example with outlier.
\item \strong{example_atd} — Fictitious AB alternating treatment design.
\item \strong{example_stranger} - Example for screen time of Stranger Things characters.
\item \strong{Grosche2011} — Multiple-baseline (three cases) from a direct-instructive reading intervention (Grosche, 2011).
\item \strong{Grosche2014} — Multiple-baseline (3×3 materials) reading intervention (Grosche, Lueke, & Wilbert, unpublished).
\item \strong{GruenkeWilbert2014} — Multiple-baseline (six cases) from a story mapping reading intervention (Gruenke, Wilbert, & Stegemann-Calder, 2013).
\item \strong{Huber2014} — Multiple-baseline (four cases) with DBR ratings from a behavioral compliance intervention (Huber, unpublished).
\item \strong{Huitema2000} — Fictitious single-case intervention study (Huitema & McKean, 2000).
\item \strong{Lenz2013} — Fictitious example (Lenz, 2013).
\item \strong{Leidig2018} — Multiple-baseline good behavior game intervention (Leidig et al., 2022).
\item \strong{Leidig2018_l2} — Level-2 data for \strong{Leidig2018} (Leidig et al., 2022).
\item \strong{Parker2007} — Example dataset after Parker et al. (2007).
\item \strong{Parker2009} — Example dataset after Parker et al. (2009).
\item \strong{Parker2009b} — Example dataset after Parker & Vannest (2009).
\item \strong{Parker2011} — Example dataset after Parker et al. (2011).
\item \strong{Parker2011b} — Example from Parker, Vannest, & Davis (2011).
\item \strong{SSDforR2017} — Example from the R package SSDforR.
\item \strong{Tarlow2017} — Fictitious single-case intervention study (Tarlow, 2017).
\item \strong{Waddell2011} — Fictitious single-case intervention study (Waddell, Nassar, & Gustafson, 2011).
}
}
\references{
Beretvas, S., & Chung, H. (2008). An evaluation of modified
R2-change effect size indices for single-subject experimental designs.
\emph{Evidence-Based Communication Assessment and Intervention, 2}, 120–128.

Borckardt, J. J., & Nash, M. R. (2014). Simulation modelling analysis for
small sets of single-subject data collected over time. \emph{Neuropsychological
Rehabilitation, 24}, 492–506.

Gruenke, M., Wilbert, J., & Stegemann-Calder, K. (2013). Analyzing the
effects of story mapping on the reading comprehension of children with low
intellectual abilities. \emph{Learning Disabilities: A Contemporary Journal, 11},
51–64.

Grosche, M. (2011). Effekte einer direkt-instruktiven Foerderung der
Lesegenauigkeit. \emph{Empirische Sonderpaedagogik, 3}, 147–161.

Huitema, B. E., & McKean, J. W. (2000). Design specification issues in
time-series intervention models. \emph{Educational and Psychological Measurement,
60}, 38–58.

Lenz, A. S. (2013). Calculating Effect Size in Single-Case Research: A
Comparison of Nonoverlap Methods. \emph{Measurement and Evaluation in Counseling
and Development, 46}(1), 64–73.

Leidig, T., Casale, G., Wilbert, J., Hennemann, T., Volpe, R. J., Briesch,
A., & Grosche, M. (2022). Individual, generalized, and moderated effects of
the good behavior game on at-risk primary school students: A multilevel
multiple baseline study using behavioral progress monitoring. \emph{Frontiers in
Education, 7}. https://www.frontiersin.org/articles/10.3389/feduc.2022.917138

Parker, R. I., Hagan-Burke, S., & Vannest, K. (2007). Percentage of All
Non-Overlapping Data (PAND) An Alternative to PND. The Journal of Special
Education, 40(4), 194-204.

Parker, R. I., Vannest, K. J., & Brown, L. (2009). The improvement rate
difference for single-case research. Exceptional Children, 75(2), 135-150.

Parker, R. I., & Vannest, K. (2009). An improved effect size for single-case
research: Nonoverlap of all pairs. Behavior Therapy, 40(4), 357-367.

Parker, R. I., Vannest, K. J., Davis, J. L., & Sauber, S. B. (2011).
Combining Nonoverlap and Trend for Single-Case Research: Tau-U. \emph{Behavior
Therapy, 42}(2), 284–299. https://doi.org/10.1016/j.beth.2010.08.006

Parker, R. I., Vannest, K. J., & Davis, J. L. (2011). Effect Size in
Single-Case Research: A Review of Nine Nonoverlap Techniques. Behavior
Modification, 35(4), 303-322. https://doi.org/10.1177/0145445511399147

Tarlow, K. R. (2017). An Improved Rank Correlation Effect Size Statistic for
Single-Case Designs: Baseline Corrected Tau. \emph{Behavior Modification, 41}(4),
427–467. https://doi.org/10.1177/0145445516676750

Waddell, D. E., Nassar, S. L., & Gustafson, S. A. (2011). Single-Case Design
in Psychophysiological Research: Part II: Statistical Analytic Approaches.
\emph{Journal of Neurotherapy, 15}, 160–169.
}
\author{
Juergen Wilbert
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scdf.R
\name{scdf}
\alias{scdf}
\alias{scdf-class}
\alias{as.scdf}
\title{Single case data frame constructor}
\usage{
scdf(
  ...,
  B_start = NULL,
  phase_starts = NULL,
  phase_design = NULL,
  name = NULL,
  dvar = "values",
  pvar = "phase",
  mvar = "mt"
)
}
\arguments{
\item{...}{One or more vectors representing measurement variables. See the
\emph{Details} section.}

\item{B_start}{The first measurement point of phase B (simple coding; only
applicable if the design follows a strict AB pattern).}

\item{phase_starts}{A named vector defining the label and measurement time of
each phase start. For example: \code{phase_starts = c(A1 = 1, B1 = 6, A2 = 14, B2 = 19)}.}

\item{phase_design}{A named vector defining the length and label of each
phase. For example: \code{phase_design = c(A1 = 10, B1 = 10, A2 = 10, B2 = 10)}.}

\item{name}{Optional name for the case.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{mvar}{Character string with the name of the measurement time variable.
Defaults to the attributes in the scdf file.}
}
\value{
Returns a single-case data frame \code{scdf} suitable for all functions in
the \code{scan} package.
}
\description{
\code{scdf()} is the constructor for objects of class \code{scdf}. It stores data from
single-case studies in a structured format suitable for analysis with the \code{scan} package.
}
\details{
If no variable matching the name of the dependent variable is
provided (the default name is \code{values}, which can be changed via the \code{dvar}
argument), and the first provided variable is unnamed, that variable will
be interpreted as the dependent variable.

If no measurement-time variable is provided (default name \code{mt},
configurable via the \code{mvar} argument), measurement times are automatically
defined as a sequence \verb{(1, 2, 3, ..., n)}.

If the dependent variable is a \strong{named vector}, the names will be used to
define a phase design. For example, \code{values = c(A = 2, 3, 5, 4, 3, B = 6, 5, 4, 3)} will be interpreted as an AB phase design with five measurements
in phase A and four in phase B.

If a vector matching the name of the phase variable is provided, it will be
used to define the phase design directly.
Otherwise, the phase design can be defined in three alternative ways:
via the \code{B_start} argument, via the \code{phase_starts} argument, or via
the \code{phase_design} argument.

If \code{B_start} is provided, a simple AB phase design is assumed, with phase A
starting at measurement time 1 and phase B starting at the measurement time
indicated by \code{B_start}.
If \code{phase_starts} is provided, the phase design is constructed based on the
measurement times indicated in the vector.
If \code{phase_design} is provided, it is used directly to define the phase
design.
If multiple of these options are provided, the priority order is:
\code{phase_design}, \code{phase_starts}, \code{B_start}, phase variable in
data frame, names of dependent variable.

If none of these options are provided, an error is raised.

The function can be used to create single-case data frames for multiple cases
separately, which can then be combined into a list for multiple-case
analyses.

See also the convenience function \code{\link{transform.scdf}} to add new
variables to an existing \code{scdf} object.

See the package vignettes for further examples on how to create and work with
\code{scdf} objects.

See the section on \emph{Data structure} in the documentation of the
\code{\link{scan}} package for further details on the \code{scdf} data structure.
}
\examples{
## Scores on a letter naming task were collected on eleven days in a row.
## The intervention started after the fifth measurement,
## so the first B phase measurement was 6 (B_start = 6).
klaas <- scdf(
  c(5, 7, 8, 5, 7, 12, 16, 18, 15, 14, 19),
  B_start = 6, name = "Klaas"
)
describe(klaas)

# Alternative: using named vector
klaas <- scdf(
  c(A = 5, 7, 8, 5, 7, B = 12, 16, 18, 15, 14, 19),
  name = "Klaas"
)

# Alternative: using phase_design
klaas <- scdf(
  c(5, 7, 8, 5, 7, 12, 16, 18, 15, 14, 19),
  phase_design = c(A = 5, B = 6), name = "Klaas"
)

# Alternative: using phase_starts
klaas <- scdf(
  c(5, 7, 8, 5, 7, 12, 16, 18, 15, 14, 19),
  phase_starts = c(A = 1, B = 7), name = "Klaas"
)

## Unfortunately in a similar study there were no data collected on
## days 3 and 9. Use NA to pass them to the function:
emmi <- scdf(c(5, 7, NA, 5, 7, 12, 16, 18, NA, 14, 19),
  phase_design = c(A = 5, B = 6), name = "Emmi"
)
describe(emmi)

## In a MBD over three cases, data were collected eleven days in a row.
## Intervention starting points differ between subjects as they were
## randomly assigned. The three SCDFs are then combined in a list for
## further conjoined analyses.
charlotte <- scdf(c(A = 5, 7, 10, 5, 12, B = 7, 10, 18, 15, 14, 19))
theresa <- scdf(c(A = 3, 4, 3, 5, B = 7, 4, 7, 9, 8, 10, 12))
tonia <- scdf(c(A = 9, 8, 8, 7, 5, 7, B = 6, 14, 15, 12, 16))
mbd <- c(charlotte, theresa, tonia)
names(mbd) <- c("Charlotte", "Theresa", "Tonia")
overlap(mbd)

## In a classroom-based intervention it was not possible to measure outcomes
## every day, but only on schooldays. The sequence of measurements is passed
## to the package by using a vector of measurement times.
frida <- scdf(
  c(A = 3, 2, 4, 2, 2, 3, 5, 6, B = 8, 10, 8, 12, 14, 13, 12),
  mt = c(1, 2, 3, 4, 5, 8, 9, 10, 11, 12, 14, 15, 16, 17, 18)
)
summary(frida)
describe(frida)

## example with two independent variables and four phases
jim <- scdf(
  zvt = c(47, 58, 76, 63, 71, 59, 64, 69, 72, 77, 76, 73),
  d2 = c(131, 134, 141, 141, 140, 140, 138, 140, 141, 140, 138, 140),
  phase_design = c(A1 = 3, B1 = 3, A2 = 3, B2 = 3), dvar = "zvt"
)
overlap(jim, phases = list(c("A1", "A2"), c("B1", "B2")))

}
\seealso{
Other data manipulation functions: 
\code{\link{add_l2}()},
\code{\link{as.data.frame.scdf}()},
\code{\link{as_scdf}()},
\code{\link{batch_apply}()},
\code{\link{fill_missing}()},
\code{\link{moving_median}()},
\code{\link{print.sc_outlier}()},
\code{\link{ranks}()},
\code{\link{rescale}()},
\code{\link{select_cases}()},
\code{\link{set_vars}()},
\code{\link{shift}()},
\code{\link{smooth_cases}()},
\code{\link{standardize}()},
\code{\link{truncate_phase}()}
}
\author{
Juergen Wilbert
}
\concept{data manipulation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/importer-gadget.R
\name{import_scdf}
\alias{import_scdf}
\title{Import scdf – RStudio Addin}
\usage{
import_scdf()
}
\description{
Import scdf – RStudio Addin
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/message-system.R
\name{warn}
\alias{warn}
\title{Throws a warning
This is a convenience wrapper for internal use.}
\usage{
warn(...)
}
\arguments{
\item{...}{The message. This can be a character string or a
combination of character strings and variables.}
}
\description{
Throws a warning
This is a convenience wrapper for internal use.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/overlap.R
\name{overlap}
\alias{overlap}
\title{Overlap indices for single-case data}
\usage{
overlap(data, dvar, pvar, mvar, decreasing = FALSE, phases = c(1, 2))
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{mvar}{Character string with the name of the measurement time variable.
Defaults to the attributes in the scdf file.}

\item{decreasing}{If you expect data to be lower in the B phase, set
\code{decreasing = TRUE}. Default is \code{decreasing = FALSE}.}

\item{phases}{A vector of two characters or numbers indicating the two phases
that should be compared. E.g., \code{phases = c("A","C")} or \code{phases = c(2,4)}
for comparing the second to the fourth phase. Phases could be combined by
providing a list with two elements. E.g., \code{phases = list(A = c(1,3), B = c(2,4))} will compare phases 1 and 3 (as A) against 2 and 4 (as B). Default
is \code{phases = c(1,2)}.}
}
\value{
\tabular{ll}{
    \tab  \cr
   \code{overlap} \tab A data frame consisting of the following indices for each single-case for all cases: PND, PEM, PET, NAP, PAND, IRD, Tau-U (A vs. B - Trend A), Diff_mean, Diff_trend, SMD, Hedges-g. \cr
   \code{phases.A} \tab Selection for A phase. \cr
   \code{phases.B} \tab Selection for B phase. \cr
   \code{design} \tab Phase design. \cr
}
}
\description{
The \code{overlap} function provides the most common overlap indices for
single-case data and some additional statistics.
}
\details{
It computes PND, PEM, PET, NAP, PAND, IRD, Tau-U, mean difference,
difference in trend, SMD, and Hedges-g for each single-case included in an
scdf.

See corresponding functions of PND, PEM, PET, NAP, PAND for
calculation. Tau_U(A) reports "A vs. B - Trend A" whereas Tau_U(BA) reports
"A vs. B + Trend B - Trend A". Base_Tau is baseline corrected tau
(correction applied when autocorrelation in phase A is significant).
Diff_mean is the mean difference. Diff_trend is the difference in the
regression estimation of the dependent variable on measurement-time (\code{x ~ mt}) for each phase. SMD is the mean difference divided by the standard
deviation of phase A. Hedges_g is the mean difference divided by the pooled
standard deviation: \eqn{\sqrt{ (n_A - 1)sd_A^2 + (n_B - 1)sd_B^2 \over n_A
  + n_B - 2 }} with a hedges correction applied: \eqn{Hedges_g * (1 -
  \frac{3}{4n - 9} ) )}.
}
\examples{

## Display overlap indices for one single-case
overlap(Huitema2000, decreasing = TRUE)

## Display overlap indices for six single-cases
overlap(GruenkeWilbert2014)

## Combining phases for analyszing designs with more than two phases
overlap(exampleA1B1A2B2, phases = list(c("A1","A2"), c("B1","B2")))

}
\seealso{
\code{\link[=pnd]{pnd()}}, \code{\link[=pem]{pem()}}, \code{\link[=pet]{pet()}}, \code{\link[=nap]{nap()}}, \code{\link[=pand]{pand()}}, \code{\link[=ird]{ird()}}, \code{\link[=tau_u]{tau_u()}}, \code{\link[=corrected_tau]{corrected_tau()}}

Other overlap functions: 
\code{\link{ird}()},
\code{\link{nap}()},
\code{\link{pand}()},
\code{\link{pem}()},
\code{\link{pet}()},
\code{\link{pnd}()},
\code{\link{print.sc_cdc}()},
\code{\link{select_phases}()},
\code{\link{tau_u}()}
}
\author{
Juergen Wilbert
}
\concept{overlap functions}
\keyword{overlap}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shinyscan.R
\name{shinyscan}
\alias{shinyscan}
\title{A Shiny app for scan}
\usage{
shinyscan(
  scdf = NULL,
  quiet = TRUE,
  browser = c("external", "viewer"),
  theme = "cerulean",
  ...
)
}
\arguments{
\item{scdf}{If you provide an \emph{scdf} here, it will be loaded at startup.}

\item{quiet}{If TRUE (default) does not report shiny messages in the console.}

\item{browser}{c("external","viewer")}

\item{theme}{Bootstrap 5 theme. Default is "cerulean".}

\item{\dots}{Further arguments passed to the \code{shiny::runApp()} function.}
}
\value{
This function launches a Shiny application.
}
\description{
Run a Shiny app with most of the scan functions.
}
\details{
This function launches a shiny application.
You need to have \code{scplot} and \code{shiny} installed.
These packages are suggested but not necessarily installed along with scan.
\code{shinyscan()} will ask to install missing packages.
}
\author{
Juergen Wilbert
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print-export-rand-test.R, R/rand_test.R
\name{print.sc_rand}
\alias{print.sc_rand}
\alias{export.sc_rand}
\alias{rand_test}
\title{Randomization Tests for single-case data}
\usage{
\method{print}{sc_rand}(x, ...)

\method{export}{sc_rand}(object, caption = NA, footnote = NA, filename = NA, ...)

rand_test(
  data,
  dvar,
  pvar,
  statistic = c("Mean B-A", "Mean A-B", "Median B-A", "Median A-B", "Mean |A-B|",
    "Median |A-B|", "SMD hedges", "SMD glass", "W-test", "T-test", "NAP",
    "NAP decreasing", "Slope B-A", "Slope A-B"),
  statistic_function = NULL,
  number = 500,
  complete = FALSE,
  limit = 5,
  startpoints = NA,
  exclude.equal = FALSE,
  phases = c(1, 2),
  graph = FALSE,
  output = NULL,
  seed = NULL
)
}
\arguments{
\item{x}{An object returned by \code{\link[=rand_test]{rand_test()}}}

\item{...}{Further parameters passed to the print function}

\item{object}{An scdf or an object exported from a scan function.}

\item{caption}{Character string with table caption. If left NA (default) a
caption will be created based on the exported object.}

\item{footnote}{Character string with table footnote. If left NA (default) a
footnote will be created based on the exported object.}

\item{filename}{String containing the file name. If a filename is given the
output will be written to that file.}

\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{statistic}{Defines the statistic on which the comparison of phases A
and B is based on. Default setting is \code{statistic = "Mean B-A"}. See
details.}

\item{statistic_function}{A list with a user defined function to calculate
the statistic. When set, overwrites the \code{statistic} argument. See details.}

\item{number}{Sample size of the randomization distribution. The exactness of
the p-value can not exceed \eqn{1/number} (i.e., \code{number = 100} results in
p-values with an exactness of one percent). Default is \code{number = 500}. For
faster processing use \code{number = 100}. For more precise p-values set \code{number = 1000}).}

\item{complete}{If TRUE, the distribution is based on a complete permutation
of all possible starting combinations. This setting overwrites the number
Argument. The default setting is FALSE.}

\item{limit}{Minimal number of data points per phase in the sample. The first
number refers to the A-phase and the second to the B-phase (e.g., \code{limit = c(5,3)}). If only one number is given, this number is applied to both
phases. Default is \code{limit = 5}.}

\item{startpoints}{Alternative to the \code{limit}-parameter \code{startpoints} exactly
defines the possible start points of phase B (e.g., \code{startpoints = 4:9}
restricts the phase B start points to measurements 4 to 9. \code{startpoints}
overruns the \code{limit}-parameter. If a list of start points is provided, each
single-case can have different start points (e.g., \code{startpoints = list(4:9, 5:10)} for a two-case design). Default is \code{startpoints = NA}, which uses
the \code{limit}-parameter to define possible start points.}

\item{exclude.equal}{If set to \code{exclude.equal = FALSE}, which is the
default, random distribution values equal to the observed distribution are
counted as null-hypothesis conform. That is, they decrease the probability
of rejecting the null-hypothesis (increase the p-value).
\code{exclude.equal} should be set to \code{TRUE} if you analyse one single-case
design (not a multiple baseline data set) to reach a sufficient power. But
be aware, that it increases the chance of an alpha-error. If
\code{exclude.equal = "auto" } is used, the function automatically sets
\code{exclude.equal = TRUE} for single-case designs and to \code{FALSE} for
multiple-case designs.}

\item{phases}{A vector of two characters or numbers indicating the two phases
that should be compared. E.g., \code{phases = c("A","C")} or \code{phases = c(2,4)}
for comparing the second to the fourth phase. Phases could be combined by
providing a list with two elements. E.g., \code{phases = list(A = c(1,3), B = c(2,4))} will compare phases 1 and 3 (as A) against 2 and 4 (as B). Default
is \code{phases = c(1,2)}.}

\item{graph}{If \code{graph = TRUE}, a histogram of the resulting distribution is
plotted. It is \code{FALSE} by default. \emph{Note: use the more versatile
\code{\link[=plot_rand]{plot_rand()}} function instead.}}

\item{output}{(deprecated and not implemented)}

\item{seed}{A seed number for the random generator. Default is \code{NULL}.}
}
\value{
An object of class \code{sc_rand}. It is a list containing the following
elements:
\item{statistic}{Character string from function call (see
\code{Arguments} above).} \item{N}{Number of single-cases.}
\item{n1}{Number of data points in phase A.} \item{n2}{Number of data
points in phase B.} \item{limit}{Numeric from function call (see
\code{Arguments} above).}
\item{startpoints}{A vector defining the start points passed from the
function call (see \code{Arguments} above).} \item{p.value}{P-value of the
randomization test for the given data.} \item{number}{Sample size of
randomization distribution from function call (see \code{Arguments} above).}
\item{complete}{Logical argument from function call (see \code{Arguments}
above).} \item{observed.statistic}{Test statistic observed for the given
single-case data. (see \code{statistic} in the \code{Arguments} above.)}
\item{Z}{Z-value of observed test statistic.} \item{p.z.single}{Probability
of z-value.} \item{distribution}{Test statistic distribution from randomized
data sets.} \item{possible.combinations}{Number of possible combinations
under the given restrictions.} \item{auto.corrected.number}{\code{TRUE}
indicates that a corrected number of combinations was used. This happens, if
the number of possible combinations (under the given restrictions) undercuts
the requested \code{number} of combinations.} \item{ecxlude.equal}{see argument
above}
\item{testdirection}{Indicates the direction of the test ("greater" or
"less").}
}
\description{
The \code{rand_test} function computes a randomization test for single or multiple
baseline single-case data.  The function is based on an algorithm from the
\code{SCRT} package (Bulte & Onghena, 2009, 2012), but rewritten and extended for
the use in AB designs.
}
\section{Functions}{
\itemize{
\item \code{print(sc_rand)}: Print results

\item \code{export(sc_rand)}: Export html results

}}
\section{Details}{
\subsection{Predefinded statisic}{

Use the \code{statistic} argument to choose a predefnied statistic. The
following comparisons are possible: \itemize{ \item\verb{Mean A-B}: Uses
the difference between the mean of phase A and the mean of phase B. This is
appropriate if a decrease of scores was expected for phase B.
\item\verb{Mean B-A}: Uses the difference between the mean of phase B and
the mean of phase A. This is appropriate if an increase of scores was
expected for phase B.  \item\verb{Mean |A-B|}: Uses the absolute value of
the difference between the means of phases A and B.  \item\verb{Median A-B}: The same as \verb{Mean A-B}, but based on the median.
\item\verb{Median B-A}: The same as \verb{Mean B-A}, but based on the
median.  \item\verb{SMD hedges / SMD glass}: Standardizes mean difference of B-A
as Hedges's g or Glass' delta. \item\code{NAP}: Non-overlap of all pairs.
\item\code{W-test}: Wilcoxon-test statistic W.\item\code{T-test}: T-test statistic t.}
}

\subsection{Create own statistic function}{

Use the \code{statistic_function} argument to proved your own function in a list.
This list must have an element named \code{statistic} with a function that takes
two arguments \code{a} and \code{b} and returns a single numeric value.  E.g.
\verb{list(statistic = function(a, b) mean(a) - mean(b)}. A second element of the
list is named \code{aggregate} which takes a function with one numeric argument
that returns a numeric argument. This function is used to aggregate the
values of a multiple case design. If you do not provide this element, it uses
the default \code{function(x) sum(x)/length(x)}. The third optional argument is
\code{name} which provides a name for your user function.
}
}

\examples{

## Compute a randomization test on the first case of the byHeart2011 data and include a graph
rand_test(byHeart2011[1], statistic = "Median B-A", graph = TRUE, seed = 123)

## Compute a randomization test on the Grosche2011 data using complete permutation
rand_test(Grosche2011, statistic = "Median B-A", complete = TRUE, limit = 4, seed = 123)

}
\references{
Bulte, I., & Onghena, P. (2009). Randomization tests for
multiple-baseline designs: An extension of the SCRT-R package.
\emph{Behavior Research Methods, 41}, 477-485.

Bulte, I., & Onghena, P. (2012). \emph{SCRT: Single-Case Randomization Tests}.
Available from: \url{https://CRAN.R-project.org/package=SCRT}
}
\author{
Juergen Wilbert
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimate_design.R
\name{estimate_design}
\alias{estimate_design}
\title{Estimate single-case design}
\usage{
estimate_design(
  data,
  dvar,
  pvar,
  mvar,
  s = NULL,
  rtt = NULL,
  overall_effects = FALSE,
  overall_rtt = TRUE,
  model = "JW",
  ...
)
}
\arguments{
\item{data}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{dvar}{Character string with the name of the dependent variable.
Defaults to the attributes in the scdf file.}

\item{pvar}{Character string with the name of the phase variable. Defaults to
the attributes in the scdf file.}

\item{mvar}{Character string with the name of the measurement time variable.
Defaults to the attributes in the scdf file.}

\item{s}{The standard deviation depicting the between case variance of the
overall performance. If more than two single-cases are included in the
scdf, the variance is estimated if s is set to NULL. If s is provided, this
value is used.}

\item{rtt}{The reliability of the measurements. The reliability is estimated
when rtt = NULL. If rtt is provided, this value is used for all single-cases.}

\item{overall_effects}{If TRUE, trend, level, and slope effect estimations
will be identical for each case. If FALSE, effects are estimated for each
case separately. Default is FALSE.}

\item{overall_rtt}{Ignored when \code{rtt} is set. If TRUE, rtt estimations will
be based on all cases and identical for each case. If FALSE rtt is
estimated for each case separately. Default is TRUE.}

\item{model}{Model used for calculating the dummy parameters (see Huitema &
McKean, 2000). Default is \code{model = "W"}. Possible values are: \code{"B&L-B"},
\code{"H-M"}, \code{"W"}, and deprecated \code{"JW"}.}

\item{...}{Further arguments passed to the plm function used for parameter
estimation. For example, the model argument can be used to select
different piecewise regression models. See plm for details.}
}
\value{
A list of parameters for each single-case. Parameters include name,
length, and starting measurement time of each phase, trend, level, and
slope effects for each phase, start value, standard deviation, and
reliability for each case. This list can be used as input for the
random_scdf function to create new random scdf files based on the estimated
parameters.
}
\description{
This functions takes an scdf and extracts design parameters. The resulting
object can be used to randomly create new scdf files with the same underlying
parameters. This is useful for Monte-Carlo studies and bootstrapping
procedures.
}
\details{
The function uses the plm function to estimate parameters for each
single-case. If more than two single-cases are included in the scdf,
the between case variance depicting the overall performance (s) is
estimated unless s is provided. The reliability of the measurements (rtt)
is estimated for each case unless rtt is provided. If overall_rtt is set
to TRUE, rtt estimations will be based on all cases and identical for each
case. If overall_effects is set to TRUE, trend, level, and slope
effect estimations will be identical for each case.

The resulting design object can be used as input for the random_scdf
function to create new random scdf files based on the estimated parameters.
This allows to create bootstrap samples or Monte-Carlo datasets based on
the characteristics of an existing dataset.
}
\examples{
# create a random scdf with predefined parameters
set.seed(1234)
design <- design(
  n = 10, trend = -0.02,
  level = list(0, 1), rtt = 0.8,
  s = 1
)
scdf<- random_scdf(design)

# Estimate the parameters based on the scdf and create a new random scdf
# based on these estimations
design_est <- estimate_design(scdf, rtt = 0.8)
scdf_est <- random_scdf(design_est)

# Analyze both datasets with an hplm model. See how similar the estimations
# are:
hplm(scdf, slope = FALSE)
hplm(scdf_est, slope = FALSE)

# Also similar results for pand and randomization tests:
pand(scdf)
pand(scdf_est)
rand_test(scdf)
rand_test(scdf_est)
}
\author{
Juergen Wilbert
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R, R/read_scdf.R, R/write_scdf.R
\name{deprecated-functions}
\alias{deprecated-functions}
\alias{tauUSC}
\alias{power_testSC}
\alias{fillmissingSC}
\alias{overlapSC}
\alias{randSC}
\alias{rand.test}
\alias{rciSC}
\alias{rSC}
\alias{readSC.excel}
\alias{readSC}
\alias{writeSC}
\title{List of old deprecated function names}
\usage{
tauUSC(...)

power_testSC(...)

fillmissingSC(...)

overlapSC(...)

randSC(...)

rand.test(...)

rciSC(...)

rSC(...)

readSC.excel(...)

readSC(...)

writeSC(...)
}
\arguments{
\item{...}{Arguments passed through to the new function.}
}
\description{
This is a list with functions names that have been replaced by new names
which are in line with R syntax guidelines. The old function names still
work. They are wrappers that call the new function.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_l2.R
\name{add_l2}
\alias{add_l2}
\title{Add level-2 data to an scdf}
\usage{
add_l2(scdf, data_l2, cvar = "case")
}
\arguments{
\item{scdf}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{data_l2}{A level 2 dataset.}

\item{cvar}{Character string with the name of the "case" variable in the L2
dataset (default is 'case').}
}
\value{
An scdf with added level-2 variables.
}
\description{
Merges variables with corresponding case names from a data.frame with an scdf.
}
\details{
This function is mostly used in combination with the \code{\link[=hplm]{hplm()}} function.
It adds level-2 variables to each single-case data frame in an scdf based on
matching case names.
}
\examples{
## Example with the default case variable name 'case'
Leidig2018 |> add_l2(Leidig2018_l2)
## Example with a different case variable name in the L2 data
Leidig2018_l2_renamed <- Leidig2018_l2
names(Leidig2018_l2_renamed)[2] <- "subject"
Leidig2018 |> add_l2(Leidig2018_l2_renamed, cvar = "subject")
}
\seealso{
\code{\link[=hplm]{hplm()}}

Other data manipulation functions: 
\code{\link{as.data.frame.scdf}()},
\code{\link{as_scdf}()},
\code{\link{batch_apply}()},
\code{\link{fill_missing}()},
\code{\link{moving_median}()},
\code{\link{print.sc_outlier}()},
\code{\link{ranks}()},
\code{\link{rescale}()},
\code{\link{scdf}()},
\code{\link{select_cases}()},
\code{\link{set_vars}()},
\code{\link{shift}()},
\code{\link{smooth_cases}()},
\code{\link{standardize}()},
\code{\link{truncate_phase}()}
}
\author{
Juergen Wilbert
}
\concept{data manipulation functions}
\keyword{transform}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/subset.scdf.R
\name{subset.scdf}
\alias{subset.scdf}
\title{Subset cases, rows, and variables of an scdf}
\usage{
\method{subset}{scdf}(x, subset, select, cases, ...)
}
\arguments{
\item{x}{An scdf object.}

\item{subset}{Logical expression indicating rows to keep: missing values are
taken as false.}

\item{select}{Expression, indicating columns to select from an scdf.}

\item{cases}{Expression, indicating cases to keep from an scdf.}

\item{...}{not implemented}
}
\value{
An scdf.
}
\description{
This function is mainly used to filter rows by a logical expression. It has
also arguments to filter variables and cases.
}
\details{
For subsetting rows, a logical expression can be provided in the \code{subset}
argument. Missing values are treated as FALSE. The expression is evaluated
within each single-case data frame of the scdf.
}
\examples{
exampleAB |>
  subset((values < 60 & phase == "A") | (values >= 60 & phase == "B"))
subset(exampleAB_add, select = c(-cigarrets, -depression))
subset(exampleAB, cases = c(Karolina, Johanna))
subset(exampleA1B1A2B2, phase \%in\% c("A1", "B2"), cases = Pawel:Moritz)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.scdf.R
\name{print.scdf}
\alias{print.scdf}
\title{Print an scdf}
\usage{
\method{print}{scdf}(
  x,
  cases = getOption("scan.print.cases"),
  rows = getOption("scan.print.rows"),
  cols = getOption("scan.print.cols"),
  long = getOption("scan.print.long"),
  digits = getOption("scan.print.digits"),
  ...
)
}
\arguments{
\item{x}{An scdf object}

\item{cases}{Number of cases to be printed. "fit" fits the number to the current screen width.}

\item{rows}{Number of rows to be printed.}

\item{cols}{Columns to be printed. "Main" only prints the dependent, measurement-time and phase variable.}

\item{long}{Logical. If TRUE cases are printed in one by a time.}

\item{digits}{Number of digits.}

\item{...}{Further arguments passed to the print function.}
}
\description{
Print an scdf
}
\details{
Print options for scdf objects could be set globally:
option(scan.print.cases = "all"),
option(scan.print.rows = 10),
option(scan.print.cols = "main"),
option(scan.print.long = TRUE),
option(scan.print.digits = 0),
option(scan.print.scdf.name = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_scdf.R
\name{as_scdf}
\alias{as_scdf}
\title{as_scdf}
\usage{
as_scdf(
  object,
  cvar = "case",
  pvar = "phase",
  dvar = "values",
  mvar = "mt",
  phase_names = NULL,
  sort_cases = FALSE
)
}
\arguments{
\item{object}{A data.frame}

\item{cvar}{Sets the "case" variable. Defaults to \code{case}.}

\item{pvar}{Sets the "phase" variable. Defaults to \code{phase}.}

\item{dvar}{Sets the "values" variable. Defaults to \code{values}.}

\item{mvar}{Sets the variable name of the "mt" variable. Defaults to \code{mt}.}

\item{phase_names}{A character vector with phase names. Defaults to the phase
names provided in the phase variable.}

\item{sort_cases}{If set TRUE, the resulting list is sorted by label names
(alphabetically increasing).}
}
\value{
An scdf.
}
\description{
Converts a data frame to an scdf object.
}
\seealso{
Other data manipulation functions: 
\code{\link{add_l2}()},
\code{\link{as.data.frame.scdf}()},
\code{\link{batch_apply}()},
\code{\link{fill_missing}()},
\code{\link{moving_median}()},
\code{\link{print.sc_outlier}()},
\code{\link{ranks}()},
\code{\link{rescale}()},
\code{\link{scdf}()},
\code{\link{select_cases}()},
\code{\link{set_vars}()},
\code{\link{shift}()},
\code{\link{smooth_cases}()},
\code{\link{standardize}()},
\code{\link{truncate_phase}()}
}
\concept{data manipulation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sample_names.R
\name{sample_names}
\alias{sample_names}
\title{Samples random names}
\usage{
sample_names(n = 1, type = "neutral", seed = NULL)
}
\arguments{
\item{n}{Number of names to sample.}

\item{type}{"neutral", "male", "female", or "mixed" type of names to sample.}

\item{seed}{A seed for the random number generator. If provided, the sampling will be reproducible.}
}
\value{
A character vector with random names.
}
\description{
Generates random names for cases. Names are sampled from a predefined list
of neutral, male, female, or mixed names.
}
\details{
This function is useful for anonymizing case names in datasets or for generating
random identifiers.
}
\examples{
sample_names(3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/select_cases.R
\name{select_cases}
\alias{select_cases}
\title{Select a subset of cases from an scdf}
\usage{
select_cases(scdf, ...)
}
\arguments{
\item{scdf}{A single-case data frame. See \code{\link[=scdf]{scdf()}} to learn about this
format.}

\item{...}{Selection criteria. Either numeric, objectnames, or as characters.}
}
\value{
An scdf with a subset of cases.
}
\description{
This function allows to select a subset of cases from an scdf by
specifying either the case names or their numeric indices. Negative
selection is also supported.
}
\examples{
select_cases(exampleAB, Johanna, Karolina)
select_cases(exampleAB, c(Johanna, Karolina))
select_cases(exampleAB, 1,2)
select_cases(exampleAB, 1:2)
select_cases(exampleAB, -Johanna)
select_cases(exampleAB, -c(Johanna, Karolina))
v <- c("Moritz", "Jannis")
select_cases(exampleA1B1A2B2, v)
}
\seealso{
Other data manipulation functions: 
\code{\link{add_l2}()},
\code{\link{as.data.frame.scdf}()},
\code{\link{as_scdf}()},
\code{\link{batch_apply}()},
\code{\link{fill_missing}()},
\code{\link{moving_median}()},
\code{\link{print.sc_outlier}()},
\code{\link{ranks}()},
\code{\link{rescale}()},
\code{\link{scdf}()},
\code{\link{set_vars}()},
\code{\link{shift}()},
\code{\link{smooth_cases}()},
\code{\link{standardize}()},
\code{\link{truncate_phase}()}
}
\author{
Juergen Wilbert
}
\concept{data manipulation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scdf_attr.R
\name{scdf_attr}
\alias{scdf_attr}
\alias{scdf_attr<-}
\alias{dv}
\alias{mt}
\alias{phase}
\title{Set and get scdf attributes}
\usage{
scdf_attr(x, var = NULL)

scdf_attr(x, var) <- value

dv(scdf)

mt(scdf)

phase(scdf)
}
\arguments{
\item{x}{Variable}

\item{var}{Attribute}

\item{value}{set value}

\item{scdf}{An \link{scdf} object.}
}
\value{
Attribute value
}
\description{
Set and get scdf attributes
}
\keyword{internal}
